Package: googlesheets4
Title: Access Google Sheets using the Sheets API V4
Version: 1.1.2
Authors@R: c(
    person("Jennifer", "Bryan", , "jenny@posit.co", role = c("cre", "aut"),
           comment = c(ORCID = "0000-0002-6983-2759")),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: Interact with Google Sheets through the Sheets API v4
    <https://developers.google.com/sheets/api>. "API" is an acronym for
    "application programming interface"; the Sheets API allows users to
    interact with Google Sheets programmatically, instead of via a web
    browser. The "v4" refers to the fact that the Sheets API is currently
    at version 4. This package can read and write both the metadata and
    the cell data in a Sheet.
License: MIT + file LICENSE
URL: https://googlesheets4.tidyverse.org,
        https://github.com/tidyverse/googlesheets4
BugReports: https://github.com/tidyverse/googlesheets4/issues
Depends: R (>= 3.6)
Imports: cellranger, cli (>= 3.0.0), curl, gargle (>= 1.6.0), glue (>=
        1.3.0), googledrive (>= 2.1.0), httr, ids, lifecycle, magrittr,
        methods, purrr, rematch2, rlang (>= 1.0.2), tibble (>= 2.1.1),
        utils, vctrs (>= 0.2.3), withr
Suggests: readr, rmarkdown, spelling, testthat (>= 3.1.7)
ByteCompile: true
Config/Needs/website: tidyverse, tidyverse/tidytemplate
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.2.9000
NeedsCompilation: no
Packaged: 2025-09-03 23:00:43 UTC; jenny
Author: Jennifer Bryan [cre, aut] (ORCID:
    <https://orcid.org/0000-0002-6983-2759>),
  Posit Software, PBC [cph, fnd]
Maintainer: Jennifer Bryan <jenny@posit.co>
Repository: CRAN
Date/Publication: 2025-09-03 23:20:02 UTC
Built: R 4.6.0; ; 2026-04-28 22:49:21 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as_CellData,"NULL")
S3method(as_CellData,Date)
S3method(as_CellData,POSIXct)
S3method(as_CellData,character)
S3method(as_CellData,default)
S3method(as_CellData,factor)
S3method(as_CellData,googlesheets4_formula)
S3method(as_CellData,googlesheets4_schema_CellData)
S3method(as_CellData,list)
S3method(as_CellData,logical)
S3method(as_CellData,numeric)
S3method(as_GridCoordinate,default)
S3method(as_GridCoordinate,range_spec)
S3method(as_GridRange,default)
S3method(as_GridRange,range_spec)
S3method(as_NamedRange,default)
S3method(as_NamedRange,range_spec)
S3method(as_Sheet,"NULL")
S3method(as_Sheet,character)
S3method(as_Sheet,data.frame)
S3method(as_Sheet,default)
S3method(as_id,googlesheets4_spreadsheet)
S3method(as_id,sheets_id)
S3method(as_range_spec,"NULL")
S3method(as_range_spec,cell_limits)
S3method(as_range_spec,character)
S3method(as_range_spec,default)
S3method(as_sheets_id,"NULL")
S3method(as_sheets_id,character)
S3method(as_sheets_id,default)
S3method(as_sheets_id,dribble)
S3method(as_sheets_id,drive_id)
S3method(as_sheets_id,googlesheets4_spreadsheet)
S3method(as_sheets_id,sheets_id)
S3method(as_tibble,googlesheets4_schema_GridRange)
S3method(as_tibble,googlesheets4_schema_NamedRange)
S3method(as_tibble,googlesheets4_schema_ProtectedRange)
S3method(as_tibble,googlesheets4_schema_Sheet)
S3method(as_tibble,googlesheets4_schema_SheetProperties)
S3method(ctype,"NULL")
S3method(ctype,SHEETS_CELL)
S3method(ctype,character)
S3method(ctype,default)
S3method(ctype,list)
S3method(format,googlesheets4_schema_CellData)
S3method(format,googlesheets4_spreadsheet)
S3method(format,range_spec)
S3method(format,sheets_id)
S3method(patch,default)
S3method(patch,googlesheets4_schema)
S3method(print,googlesheets4_schema_CellData)
S3method(print,googlesheets4_spreadsheet)
S3method(print,range_spec)
S3method(print,sheets_id)
S3method(vec_cast,character.sheets_id)
S3method(vec_cast,drive_id.sheets_id)
S3method(vec_cast,googlesheets4_formula)
S3method(vec_cast,sheets_id.character)
S3method(vec_cast,sheets_id.drive_id)
S3method(vec_cast,sheets_id.sheets_id)
S3method(vec_cast.character,googlesheets4_formula)
S3method(vec_cast.googlesheets4_formula,character)
S3method(vec_cast.googlesheets4_formula,default)
S3method(vec_cast.googlesheets4_formula,googlesheets4_formula)
S3method(vec_ptype2,character.sheets_id)
S3method(vec_ptype2,drive_id.sheets_id)
S3method(vec_ptype2,googlesheets4_formula)
S3method(vec_ptype2,sheets_id.character)
S3method(vec_ptype2,sheets_id.drive_id)
S3method(vec_ptype2,sheets_id.sheets_id)
S3method(vec_ptype2.character,googlesheets4_formula)
S3method(vec_ptype2.googlesheets4_formula,character)
S3method(vec_ptype2.googlesheets4_formula,default)
S3method(vec_ptype2.googlesheets4_formula,googlesheets4_formula)
S3method(vec_ptype_abbr,googlesheets4_formula)
S3method(vec_ptype_abbr,sheets_id)
export("%>%")
export(anchored)
export(as_sheets_id)
export(cell_cols)
export(cell_limits)
export(cell_rows)
export(gs4_api_key)
export(gs4_auth)
export(gs4_auth_configure)
export(gs4_browse)
export(gs4_create)
export(gs4_deauth)
export(gs4_endpoints)
export(gs4_example)
export(gs4_examples)
export(gs4_find)
export(gs4_fodder)
export(gs4_formula)
export(gs4_get)
export(gs4_has_token)
export(gs4_oauth_app)
export(gs4_oauth_client)
export(gs4_random)
export(gs4_scopes)
export(gs4_token)
export(gs4_user)
export(local_gs4_quiet)
export(range_autofit)
export(range_clear)
export(range_delete)
export(range_flood)
export(range_read)
export(range_read_cells)
export(range_speedread)
export(range_write)
export(read_sheet)
export(request_generate)
export(request_make)
export(sheet_add)
export(sheet_append)
export(sheet_copy)
export(sheet_delete)
export(sheet_names)
export(sheet_properties)
export(sheet_relocate)
export(sheet_rename)
export(sheet_resize)
export(sheet_write)
export(spread_sheet)
export(vec_cast.googlesheets4_formula)
export(vec_ptype2.googlesheets4_formula)
export(with_gs4_quiet)
export(write_sheet)
import(rlang)
import(vctrs)
importFrom(cellranger,anchored)
importFrom(cellranger,cell_cols)
importFrom(cellranger,cell_limits)
importFrom(cellranger,cell_rows)
importFrom(gargle,bulletize)
importFrom(gargle,gargle_map_cli)
importFrom(glue,glue)
importFrom(glue,glue_collapse)
importFrom(glue,glue_data)
importFrom(googledrive,as_id)
importFrom(lifecycle,deprecated)
importFrom(magrittr,"%>%")
importFrom(methods,setOldClass)
importFrom(purrr,compact)
importFrom(purrr,discard)
importFrom(purrr,imap)
importFrom(purrr,keep)
importFrom(purrr,map)
importFrom(purrr,map2)
importFrom(purrr,map_chr)
importFrom(purrr,map_dbl)
importFrom(purrr,map_int)
importFrom(purrr,map_lgl)
importFrom(purrr,modify_if)
importFrom(purrr,pluck)
importFrom(purrr,pmap)
importFrom(purrr,pmap_chr)
importFrom(purrr,transpose)
importFrom(purrr,walk)
importFrom(tibble,as_tibble)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# googlesheets4 <a href="https://googlesheets4.tidyverse.org"><img src="man/figures/logo.png" align="right" height="138" alt = ""/></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/googlesheets4)](https://CRAN.R-project.org/package=googlesheets4)
[![R-CMD-check](https://github.com/tidyverse/googlesheets4/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/tidyverse/googlesheets4/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/tidyverse/googlesheets4/branch/main/graph/badge.svg)](https://app.codecov.io/gh/tidyverse/googlesheets4?branch=main)
<!-- badges: end -->

## Overview

googlesheets4 provides an R interface to [Google
Sheets](https://docs.google.com/spreadsheets/) via the [Sheets API
v4](https://developers.google.com/sheets/api/). It is a reboot of an
earlier package called
[googlesheets](https://github.com/jennybc/googlesheets#readme).

*Why **4**? Why googlesheets**4**? Did I miss googlesheets1 through 3?
No. The idea is to name the package after the corresponding version of
the Sheets API. In hindsight, the original googlesheets should have been
googlesheets**3**.*

## Installation

You can install the released version of googlesheets4 from
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("googlesheets4")
```

And the development version from [GitHub](https://github.com/) with:

``` r
#install.packages("pak")
pak::pak("tidyverse/googlesheets4")
```

## Cheatsheet

You can see how to read data with googlesheets4 in the **data import
cheatsheet**, which also covers similar functionality in the related
packages readr and readxl.

<a href="https://github.com/rstudio/cheatsheets/blob/main/data-import.pdf"><img src="https://raw.githubusercontent.com/rstudio/cheatsheets/main/pngs/thumbnails/data-import-cheatsheet-thumbs.png" width="630" height="252" alt="thumbnail of data import cheatsheet"/></a>

## Auth

googlesheets4 will, by default, help you interact with Sheets as an
authenticated Google user. If you don’t plan to write Sheets or to read
private Sheets, use `gs4_deauth()` to indicate there is no need for a
token. See the article [googlesheets4
auth](https://googlesheets4.tidyverse.org/articles/articles/auth.html)
for more.

For this overview, we’ve logged into Google as a specific user in a
hidden chunk.

## Attach googlesheets4

``` r
library(googlesheets4)
```

## Read

The main “read” function of the googlesheets4 package goes by two names,
because we want it to make sense in two contexts:

- `read_sheet()` evokes other table-reading functions, like
  `readr::read_csv()` and `readxl::read_excel()`. The `sheet` in this
  case refers to a Google (spread)Sheet.
- `range_read()` is the right name according to the [naming
  convention](https://googlesheets4.tidyverse.org/articles/articles/function-class-names.html)
  used throughout the googlesheets4 package.

`read_sheet()` and `range_read()` are synonyms and you can use either
one. Here we’ll use `read_sheet()`.

googlesheets4 is [pipe-friendly](https://r4ds.had.co.nz/pipes.html) (and
reexports `%>%`), but works just fine without the pipe.

Read from

- a URL
- a Sheet ID
- a
  [`dribble`](https://googledrive.tidyverse.org/reference/dribble.html)
  produced by the googledrive package, which can lookup by file name

These all achieve the same thing:

``` r
# URL
read_sheet("https://docs.google.com/spreadsheets/d/1U6Cf_qEOhiR9AZqTqS3mbMF3zt2db48ZP5v3rkrAEJY/edit#gid=780868077")
#> ✔ Reading from "gapminder".
#> ✔ Range 'Africa'.
#> # A tibble: 624 × 6
#>   country continent  year lifeExp      pop gdpPercap
#>   <chr>   <chr>     <dbl>   <dbl>    <dbl>     <dbl>
#> 1 Algeria Africa     1952    43.1  9279525     2449.
#> 2 Algeria Africa     1957    45.7 10270856     3014.
#> 3 Algeria Africa     1962    48.3 11000948     2551.
#> 4 Algeria Africa     1967    51.4 12760499     3247.
#> 5 Algeria Africa     1972    54.5 14760787     4183.
#> # ℹ 619 more rows

# Sheet ID
read_sheet("1U6Cf_qEOhiR9AZqTqS3mbMF3zt2db48ZP5v3rkrAEJY")
#> ✔ Reading from "gapminder".
#> ✔ Range 'Africa'.
#> # A tibble: 624 × 6
#>   country continent  year lifeExp      pop gdpPercap
#>   <chr>   <chr>     <dbl>   <dbl>    <dbl>     <dbl>
#> 1 Algeria Africa     1952    43.1  9279525     2449.
#> 2 Algeria Africa     1957    45.7 10270856     3014.
#> 3 Algeria Africa     1962    48.3 11000948     2551.
#> 4 Algeria Africa     1967    51.4 12760499     3247.
#> 5 Algeria Africa     1972    54.5 14760787     4183.
#> # ℹ 619 more rows

# a googledrive "dribble"
googledrive::drive_get("gapminder") %>% 
  read_sheet()
#> ✔ The input `path` resolved to exactly 1 file.
#> ✔ Reading from "gapminder".
#> ✔ Range 'Africa'.
#> # A tibble: 624 × 6
#>   country continent  year lifeExp      pop gdpPercap
#>   <chr>   <chr>     <dbl>   <dbl>    <dbl>     <dbl>
#> 1 Algeria Africa     1952    43.1  9279525     2449.
#> 2 Algeria Africa     1957    45.7 10270856     3014.
#> 3 Algeria Africa     1962    48.3 11000948     2551.
#> 4 Algeria Africa     1967    51.4 12760499     3247.
#> 5 Algeria Africa     1972    54.5 14760787     4183.
#> # ℹ 619 more rows
```

*Note: the only reason we can read a sheet named “gapminder” (the last
example) is because the account we’re logged in as has a Sheet named
“gapminder”.*

See the article [Find and Identify
Sheets](https://googlesheets4.tidyverse.org/articles/articles/find-identify-sheets.html)
for more about specifying the Sheet you want to address. See the article
[Read
Sheets](https://googlesheets4.tidyverse.org/articles/articles/find-identify-sheets.html)
for more about reading from specific sheets or ranges, setting column
type, and getting low-level cell data.

## Write

`gs4_create()` creates a brand new Google Sheet and can optionally send
some initial data.

``` r
(ss <- gs4_create("fluffy-bunny", sheets = list(flowers = head(iris))))
#> ✔ Creating new Sheet: "fluffy-bunny".
#> 
#> ── <googlesheets4_spreadsheet> ─────────────────────────────────────────────────
#> Spreadsheet name: "fluffy-bunny"                              
#>               ID: 1enILX4tYJeFEJ1RL8MsGgDRjb0NHTdm3ZD92R2RMWYI
#>           Locale: en_US                                       
#>        Time zone: Etc/GMT                                     
#>      # of sheets: 1                                           
#> 
#> ── <sheets> ────────────────────────────────────────────────────────────────────
#> (Sheet name): (Nominal extent in rows x columns)
#>    'flowers': 7 x 5
```

`sheet_write()` (over)writes a whole data frame into a (work)sheet
within a (spread)Sheet.

``` r
head(mtcars) %>% 
  sheet_write(ss, sheet = "autos")
#> ✔ Writing to "fluffy-bunny".
#> ✔ Writing to sheet 'autos'.
ss
#> 
#> ── <googlesheets4_spreadsheet> ─────────────────────────────────────────────────
#> Spreadsheet name: "fluffy-bunny"                              
#>               ID: 1enILX4tYJeFEJ1RL8MsGgDRjb0NHTdm3ZD92R2RMWYI
#>           Locale: en_US                                       
#>        Time zone: Etc/GMT                                     
#>      # of sheets: 2                                           
#> 
#> ── <sheets> ────────────────────────────────────────────────────────────────────
#> (Sheet name): (Nominal extent in rows x columns)
#>    'flowers': 7 x 5
#>      'autos': 7 x 11
```

`sheet_append()`, `range_write()`, `range_flood()`, and `range_clear()`
are more specialized writing functions. See the article [Write
Sheets](https://googlesheets4.tidyverse.org/articles/articles/write-sheets.html)
for more about writing to Sheets.

## Where to learn more

[Get
started](https://googlesheets4.tidyverse.org/articles/googlesheets4.html)
is a more extensive general introduction to googlesheets4.

Browse the [articles
index](https://googlesheets4.tidyverse.org/articles/index.html) to find
articles that cover various topics in more depth.

See the [function
index](https://googlesheets4.tidyverse.org/reference/index.html) for an
organized, exhaustive listing.

## Contributing

If you’d like to contribute to the development of googlesheets4, please
read [these
guidelines](https://googlesheets4.tidyverse.org/CONTRIBUTING.html).

Please note that the googlesheets4 project is released with a
[Contributor Code of
Conduct](https://googlesheets4.tidyverse.org/CODE_OF_CONDUCT.html). By
contributing to this project, you agree to abide by its terms.

## Privacy

[Privacy policy](https://www.tidyverse.org/google_privacy_policy)

## Context

googlesheets4 draws on and complements / emulates other packages in the
tidyverse:

- [googlesheets](https://cran.r-project.org/package=googlesheets) is the
  package that googlesheets4 replaces. Main improvements in
  googlesheets4: (1) wraps the current, most modern Sheets API; (2)
  leaves all “whole file” operations to googledrive; and (3) uses shared
  infrastructure for auth and more, from the gargle package. The v3 API
  wrapped by googlesheets is deprecated. [Starting in April/May
  2020](https://workspace.google.com/blog/product-announcements/migrate-your-apps-use-latest-sheets-api),
  features will gradually be disabled and it’s anticipated the API will
  fully shutdown in September 2020. At that point, the original
  googlesheets package must be retired.
- [googledrive](https://googledrive.tidyverse.org) provides a
  fully-featured interface to the Google Drive API. Any “whole file”
  operations can be accomplished with googledrive: upload or download or
  update a spreadsheet, copy, rename, move, change permission, delete,
  etc. googledrive supports Team Drives.
- [readxl](https://readxl.tidyverse.org) is the tidyverse package for
  reading Excel files (xls or xlsx) into an R data frame. googlesheets4
  takes cues from parts of the readxl interface, especially around
  specifying which cells to read.
- [readr](https://readr.tidyverse.org) is the tidyverse package for
  reading delimited files (e.g., csv or tsv) into an R data frame.
  googlesheets4 takes cues from readr with respect to column type
  specification.

---
output: github_document
---

<!-- README.md is generated from README.Rmd. Please edit that file -->

```{r setup, include = FALSE}
can_decrypt <- gargle:::secret_can_decrypt("googlesheets4")
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  error = TRUE,
  fig.path = "man/figures/README-",
  out.width = "100%",
  purl = can_decrypt,
  eval = can_decrypt
)
options(tibble.print_min = 5L, tibble.print_max = 5L)
```

```{r eval = !can_decrypt, echo = FALSE, comment = NA}
message("No token available. Code chunks will not be evaluated.")
```

```{r readme-auth, include = FALSE}
googlesheets4:::gs4_auth_docs()
```

# googlesheets4 <a href='https:/googlesheets4.tidyverse.org'><img src='man/figures/logo.png' align="right" height="138.5" /></a>

<!-- badges: start -->
[![CRAN status](https://www.r-pkg.org/badges/version/googlesheets4)](https://CRAN.R-project.org/package=googlesheets4)
[![R-CMD-check](https://github.com/tidyverse/googlesheets4/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/tidyverse/googlesheets4/actions/workflows/R-CMD-check.yaml)
[![Codecov test coverage](https://codecov.io/gh/tidyverse/googlesheets4/branch/main/graph/badge.svg)](https://app.codecov.io/gh/tidyverse/googlesheets4?branch=main)
<!-- badges: end -->

## Overview

googlesheets4 provides an R interface to [Google Sheets](https://docs.google.com/spreadsheets/) via the [Sheets API v4](https://developers.google.com/sheets/api/). It is a reboot of an earlier package called [googlesheets](https://github.com/jennybc/googlesheets#readme).

*Why **4**? Why googlesheets**4**? Did I miss googlesheets1 through 3? No. The idea is to name the package after the corresponding version of the Sheets API. In hindsight, the original googlesheets should have been googlesheets**3**.*

## Installation

You can install the released version of googlesheets4 from [CRAN](https://CRAN.R-project.org) with:

```{r, eval = FALSE}
install.packages("googlesheets4")
```

And the development version from [GitHub](https://github.com/) with:

```{r, eval = FALSE}
# install.packages("devtools")
devtools::install_github("tidyverse/googlesheets4")
```

## Cheatsheet

You can see how to read data with googlesheets4 in the **data import cheatsheet**, which also covers similar functionality in the related packages readr and readxl.

<a href="https://github.com/rstudio/cheatsheets/blob/master/data-import.pdf"><img src="https://raw.githubusercontent.com/rstudio/cheatsheets/master/pngs/thumbnails/data-import-cheatsheet-thumbs.png" width="630" height="252"/></a>

## Auth

googlesheets4 will, by default, help you interact with Sheets as an authenticated Google user. If you don't plan to write Sheets or to read private Sheets, use `gs4_deauth()` to indicate there is no need for a token. See the article [googlesheets4 auth](https://googlesheets4.tidyverse.org/articles/articles/auth.html) for more.

For this overview, we've logged into Google as a specific user in a hidden chunk.

## Attach googlesheets4

```{r}
library(googlesheets4)
```

## Read

The main "read" function of the googlesheets4 package goes by two names, because we want it to make sense in two contexts:

* `read_sheet()` evokes other table-reading functions, like
 `readr::read_csv()` and `readxl::read_excel()`. The `sheet` in this case
  refers to a Google (spread)Sheet.
* `range_read()` is the right name according to the
  [naming convention](https://googlesheets4.tidyverse.org/articles/articles/function-class-names.html)
  used throughout the googlesheets4 package.
  
`read_sheet()` and `range_read()` are synonyms and you can use either one. Here we'll use `read_sheet()`.

googlesheets4 is [pipe-friendly](https://r4ds.had.co.nz/pipes.html) (and reexports `%>%`), but works just fine without the pipe.

Read from

  * a URL
  * a Sheet ID
  * a [`dribble`](https://googledrive.tidyverse.org/reference/dribble.html)
    produced by the googledrive package, which can lookup by file name

These all achieve the same thing:

```{r}
# URL
read_sheet("https://docs.google.com/spreadsheets/d/1U6Cf_qEOhiR9AZqTqS3mbMF3zt2db48ZP5v3rkrAEJY/edit#gid=780868077")

# Sheet ID
read_sheet("1U6Cf_qEOhiR9AZqTqS3mbMF3zt2db48ZP5v3rkrAEJY")

# a googledrive "dribble"
googledrive::drive_get("gapminder") %>% 
  read_sheet()
```

*Note: the only reason we can read a sheet named "gapminder" (the last example) is because the account we're logged in as has a Sheet named "gapminder".*

See the article [Find and Identify Sheets](https://googlesheets4.tidyverse.org/articles/articles/find-identify-sheets.html) for more about specifying the Sheet you want to address. See the article [Read Sheets](https://googlesheets4.tidyverse.org/articles/articles/find-identify-sheets.html) for more about reading from specific sheets or ranges, setting column type, and getting low-level cell data.

## Write

`gs4_create()` creates a brand new Google Sheet and can optionally send some initial data.

```{r}
(ss <- gs4_create("fluffy-bunny", sheets = list(flowers = head(iris))))
```

`sheet_write()` (over)writes a whole data frame into a (work)sheet within a (spread)Sheet.

```{r}
head(mtcars) %>% 
  sheet_write(ss, sheet = "autos")
ss
```

`sheet_append()`, `range_write()`, `range_flood()`, and `range_clear()` are more
specialized writing functions. See the article [Write Sheets](https://googlesheets4.tidyverse.org/articles/articles/write-sheets.html) for more about writing to Sheets.

```{r include = FALSE}
googledrive::drive_trash(ss)
```

## Where to learn more

[Get started](https://googlesheets4.tidyverse.org/articles/googlesheets4.html) is a more extensive general introduction to googlesheets4.

Browse the [articles index](https://googlesheets4.tidyverse.org/articles/index.html) to find articles that cover various topics in more depth.

See the [function index](https://googlesheets4.tidyverse.org/reference/index.html) for an organized, exhaustive listing.

## Contributing

If you'd like to contribute to the development of googlesheets4, please read [these guidelines](https://googlesheets4.tidyverse.org/CONTRIBUTING.html).

Please note that the googlesheets4 project is released with a [Contributor Code of Conduct](https://googlesheets4.tidyverse.org/CODE_OF_CONDUCT.html). By contributing to this project, you agree to abide by its terms.

## Privacy

[Privacy policy](https://www.tidyverse.org/google_privacy_policy)

## Context

googlesheets4 draws on and complements / emulates other packages in the tidyverse:

  * [googlesheets](https://cran.r-project.org/package=googlesheets) is the package that googlesheets4 replaces. Main improvements in googlesheets4: (1) wraps the current, most modern Sheets API; (2) leaves all "whole file"
  operations to googledrive; and (3) uses shared infrastructure for auth and more, from the gargle package. The v3 API wrapped by googlesheets is deprecated. [Starting in April/May 2020](https://cloud.google.com/blog/products/g-suite/migrate-your-apps-use-latest-sheets-api),
  features will gradually be disabled and it's anticipated the API will fully
  shutdown in September 2020. At that point, the original googlesheets
  package must be retired.
  * [googledrive](https://googledrive.tidyverse.org) provides a fully-featured interface to the Google Drive API. Any "whole file" operations can be accomplished with googledrive: upload or download or update a spreadsheet, copy, rename, move, change permission, delete, etc. googledrive supports Team Drives.
  * [readxl](https://readxl.tidyverse.org) is the tidyverse package for reading Excel files (xls or xlsx) into an R data frame. googlesheets4 takes cues from parts of the readxl interface, especially around specifying which cells to read.
  * [readr](https://readr.tidyverse.org) is the tidyverse package for reading delimited files (e.g., csv or tsv) into an R data frame. googlesheets4 takes cues from readr with respect to column type specification.

# https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/other#GridRange
#
# All indexes are zero-based. Indexes are half open, e.g the start index is
# inclusive and the end index is exclusive -- [startIndex, endIndex). Missing
# indexes indicate the range is unbounded on that side.

#' @export
as_tibble.googlesheets4_schema_GridRange <- function(x, ...) {
  tibble::tibble(
    # if there is only 1 sheet, sheetId might not be sent!
    # https://github.com/tidyverse/googlesheets4/issues/29
    # don't be shocked if this is NA
    sheet_id = glean_int(x, "sheetId"),
    # API sends zero-based row and column
    #   => we add one
    # API indices are half-open, i.e. [start, end)
    #   => we substract one from end_[row|column]
    # net effect
    #   => we add one to start_[row|column] but not to end_[row|column]
    start_row = glean_int(x, "startRowIndex") + 1L,
    end_row = glean_int(x, "endRowIndex"),
    start_column = glean_int(x, "startColumnIndex") + 1L,
    end_column = glean_int(x, "endColumnIndex")
  )
}

as_GridRange <- function(x, ...) {
  UseMethod("as_GridRange")
}

#' @export
as_GridRange.default <- function(x, ...) {
  abort_unsupported_conversion(x, to = "GridRange")
}

#' @export
as_GridRange.range_spec <- function(x, ...) {
  if (!is.null(x$named_range)) {
    gs4_abort("This function does not accept a named range as {.arg range}.")
  }
  s <- lookup_sheet(x$sheet_name, sheets_df = x$sheets_df)
  out <- new("GridRange", sheetId = s$id)

  if (is.null(x$cell_limits)) {
    if (is.null(x$cell_range)) {
      return(out)
    }
    x$cell_limits <- limits_from_range(x$cell_range)
  }

  cl <- list(
    startRowIndex = x$cell_limits$ul[1] - 1,
    endRowIndex = x$cell_limits$lr[1],
    startColumnIndex = x$cell_limits$ul[2] - 1,
    endColumnIndex = x$cell_limits$lr[2]
  )
  cl <- discard(cl, is.na)
  patch(out, !!!cl)
}

test_that("we can make a GridRange from a range_spec", {
  sheets_df <- tibble::tibble(name = "abc", id = 123)

  # test cases are taken from examples given for GridRange schema
  # https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/other#GridRange
  spec <- new_range_spec(sheet_name = "abc", sheets_df = sheets_df)
  out <- as_GridRange(spec)
  expect_equal(out$sheetId, 123)

  spec <- new_range_spec(
    sheet_name = "abc",
    cell_range = "A3:B4",
    sheets_df = sheets_df
  )
  out <- as_GridRange(spec)
  expect_equal(out$sheetId, 123)
  expect_equal(out$startRowIndex, 2)
  expect_equal(out$endRowIndex, 4)
  expect_equal(out$startColumnIndex, 0)
  expect_equal(out$endColumnIndex, 2)

  spec <- new_range_spec(
    sheet_name = "abc",
    cell_range = "A5:B",
    sheets_df = sheets_df
  )
  out <- as_GridRange(spec)
  expect_equal(out$sheetId, 123)
  expect_equal(out$startRowIndex, 4)
  expect_null(out$endRowIndex)
  expect_equal(out$startColumnIndex, 0)
  expect_equal(out$endColumnIndex, 2)

  spec <- new_range_spec(
    sheet_name = "abc",
    cell_range = "A:B",
    sheets_df = sheets_df
  )
  out <- as_GridRange(spec)
  expect_equal(out$sheetId, 123)
  expect_null(out$startRowIndex)
  expect_null(out$endRowIndex)
  expect_equal(out$startColumnIndex, 0)
  expect_equal(out$endColumnIndex, 2)

  spec <- new_range_spec(
    sheet_name = "abc",
    cell_range = "A1:A1",
    sheets_df = sheets_df
  )
  out <- as_GridRange(spec)
  expect_equal(out$sheetId, 123)
  expect_equal(out$startRowIndex, 0)
  expect_equal(out$endRowIndex, 1)
  expect_equal(out$startColumnIndex, 0)
  expect_equal(out$endColumnIndex, 1)

  spec1 <- new_range_spec(
    sheet_name = "abc",
    cell_range = "C3:C3",
    sheets_df = sheets_df
  )
  spec2 <- new_range_spec(
    sheet_name = "abc",
    cell_range = "C3",
    sheets_df = sheets_df
  )
  expect_equal(as_GridRange(spec1), as_GridRange(spec2))
})

test_that("we refuse to make a GridRange from a named_range", {
  spec <- new_range_spec(named_range = "thingy")
  expect_error(as_GridRange(spec), "does not accept a named range")
})

#' `sheets_id` class
#'
#' @description

#' `sheets_id` is an S3 class that marks a string as a Google Sheet's id, which
#' the Sheets API docs refer to as `spreadsheetId`.
#'
#' Any object of class `sheets_id` also has the [`drive_id`][googledrive::as_id]
#' class, which is used by [googledrive] for the same purpose. This means you
#' can provide a `sheets_id` to [googledrive] functions, in order to do anything
#' with your Sheet that has nothing to do with it being a spreadsheet. Examples:
#' change the Sheet's name, parent folder, or permissions. Read more about using
#' [googlesheets4] and [googledrive] together in `vignette("drive-and-sheets")`.
#' Note that a `sheets_id` object is intended to hold **just one** id, while the
#' parent class `drive_id` can be used for multiple ids.
#'

#' `as_sheets_id()` is a generic function that converts various inputs into an
#' instance of `sheets_id`. See more below.
#'

#' When you print a `sheets_id`, we attempt to reveal the Sheet's current
#' metadata, via [gs4_get()]. This can fail for a variety of reasons (e.g. if
#' you're offline), but the input `sheets_id` is always revealed and returned,
#' invisibly.

#' @section `as_sheets_id()`:
#'
#' These inputs can be converted to a `sheets_id`:
#'   * Spreadsheet id, "a string containing letters, numbers, and some special
#'   characters", typically 44 characters long, in our experience. Example:
#'   `1qpyC0XzvTcKT6EISywvqESX3A0MwQoFDE8p-Bll4hps`.
#'   * A URL, from which we can excavate a spreadsheet or file id. Example:
#'     `"https://docs.google.com/spreadsheets/d/1BzfL0kZUz1TsI5zxJF1WNF01IxvC67FbOJUiiGMZ_mQ/edit#gid=1150108545"`.
#'   * A one-row [`dribble`][googledrive::dribble], a "Drive tibble" used by the
#'     [googledrive] package. In general, a `dribble` can represent several
#'     files, one row per file. Since googlesheets4 is not vectorized over
#'     spreadsheets, we are only prepared to accept a one-row `dribble`.
#'     - [`googledrive::drive_get("YOUR_SHEET_NAME")`][googledrive::drive_get()]
#'     is a great way to look up a Sheet via its name.
#'     - [`gs4_find("YOUR_SHEET_NAME")`][gs4_find()] is another good way
#'     to get your hands on a Sheet.
#'   * Spreadsheet meta data, as returned by, e.g., [gs4_get()]. Literally,
#'     this is an object of class `googlesheets4_spreadsheet`.
#'

#' @name sheets_id
#' @seealso [googledrive::as_id]

#' @param x Something that contains a Google Sheet id: an id string, a
#'   [`drive_id`][googledrive::as_id], a URL, a one-row
#'   [`dribble`][googledrive::dribble], or a `googlesheets4_spreadsheet`.
#' @param ... Other arguments passed down to methods. (Not used.)

#' @examplesIf gs4_has_token()
#' mini_gap_id <- gs4_example("mini-gap")
#' class(mini_gap_id)
#' mini_gap_id
#'
#' as_sheets_id("abc")
NULL

# constructor and validator ----
new_sheets_id <- function(x = character()) {
  vec_assert(x, character())
  new_vctr(x, class = c("sheets_id", "drive_id"), inherit_base_type = TRUE)
}

validate_sheets_id <- function(x) {
  if (length(x) > 1) {
    gs4_abort(c(
      "A {.cls sheets_id} object can't have length greater than 1.",
      x = "Actual input has length {length(x)}."
    ))
  }
  validate_drive_id(x)
}

new_drive_id <- function(x = character()) {
  utils::getFromNamespace("new_drive_id", "googledrive")(x)
}
validate_drive_id <- function(x) {
  utils::getFromNamespace("validate_drive_id", "googledrive")(x)
}

# vctrs methods ----

# sheets_id is intended to hold ONE id, so I want:
# c(sheets_id, sheets_id) = drive_id
# I'm willing to accept that this is not quite right / necessary if one or both
# inputs has length 1
#' @export
vec_ptype2.sheets_id.sheets_id <- function(x, y, ...) new_drive_id()

#' @export
vec_ptype2.sheets_id.character <- function(x, y, ...) character()
#' @export
vec_ptype2.character.sheets_id <- function(x, y, ...) character()

#' @export
vec_ptype2.sheets_id.drive_id <- function(x, y, ...) new_drive_id()
#' @export
vec_ptype2.drive_id.sheets_id <- function(x, y, ...) new_drive_id()

#' @export
vec_cast.sheets_id.sheets_id <- function(x, to, ...) x

#' @export
vec_cast.sheets_id.character <- function(x, to, ...) {
  validate_sheets_id(new_sheets_id(x))
}
#' @export
vec_cast.character.sheets_id <- function(x, to, ...) vec_data(x)

#' @export
vec_cast.sheets_id.drive_id <- function(x, to, ...) {
  validate_sheets_id(new_sheets_id(vec_data(x)))
}
#' @export
vec_cast.drive_id.sheets_id <- function(x, to, ...) as_id(vec_data(x))

#' @export
vec_ptype_abbr.sheets_id <- function(x, ...) "sht_id"

# googledrive ----

#' @export
as_id.sheets_id <- function(x, ...) as_id(vec_data(x))
#' @export
as_id.googlesheets4_spreadsheet <- function(x, ...) as_id(x$spreadsheet_id)

# user-facing  ----

#' @export
#' @rdname sheets_id
as_sheets_id <- function(x, ...) UseMethod("as_sheets_id")

#' @export
as_sheets_id.NULL <- function(x, ...) {
  abort_unsupported_conversion(x, to = "sheets_id")
}

#' @export
as_sheets_id.default <- function(x, ...) {
  abort_unsupported_conversion(x, to = "sheets_id")
}

#' @export
as_sheets_id.sheets_id <- function(x, ...) x

#' @export
as_sheets_id.drive_id <- function(x, ...) {
  validate_sheets_id(new_sheets_id(vec_data(x)))
}

#' @export
as_sheets_id.dribble <- function(x, ...) {
  if (nrow(x) != 1) {
    gs4_abort(c(
      "{.cls dribble} input must have exactly 1 row.",
      x = "Actual input has {nrow(x)} rows."
    ))
  }
  # not worrying about whether we are authed as same user with Sheets and Drive
  # revealing the MIME type is local to the dribble, so this makes no API calls
  mime_type <- googledrive::drive_reveal(x, "mime_type")[["mime_type"]]
  target <- "application/vnd.google-apps.spreadsheet"
  if (!identical(mime_type, target)) {
    gs4_abort(c(
      "{.cls dribble} input must refer to a Google Sheet, i.e. a file with \\
       MIME type {.field {target}}.",
      i = "File name: {.s_sheet {x$name}}",
      i = "File id: {.field {x$id}}",
      x = "MIME TYPE: {.field {mime_type}}"
    ))
  }
  as_sheets_id(x$id)
}

#' @export
as_sheets_id.character <- function(x, ...) {
  # we're leaning on as_id() for URL detection and processing
  id <- as_id(x)
  validate_sheets_id(new_sheets_id(vec_data(id)))
}

#' @export
as_sheets_id.googlesheets4_spreadsheet <- function(x, ...) {
  validate_sheets_id(new_sheets_id(x$spreadsheet_id))
}

#' @export
print.sheets_id <- function(x, ...) {
  cli::cat_line(format(x))
  invisible(x)
}

#' @export
format.sheets_id <- function(x, ...) {
  meta <- tryCatch(
    gs4_get(x),
    # seen with a failed request
    gargle_error_request_failed = function(e) e,
    # seen when we can't get a token but auth is active
    googlesheets4_error = function(e) e
  )

  if (inherits(meta, "googlesheets4_spreadsheet")) {
    return(format(meta))
  }

  # meta is an error, i.e. gs4_get() failed
  out <- new_googlesheets4_spreadsheet(list(spreadsheetId = x))
  c(
    format(out),
    "",
    "Unable to get metadata for this Sheet. Error details:",
    meta$message,
    cli::cli_format_method(
      cli::cli_bullets(meta$body)
    )
  )
}

test_that("can coerce simple strings and drive_id's to sheets_id", {
  expect_s3_class(as_sheets_id("123"), "sheets_id")
  expect_identical(as_sheets_id(as_sheets_id("123")), as_sheets_id("123"))
  expect_identical(as_sheets_id(as_id("123")), as_sheets_id("123"))
})

test_that("string with invalid character is rejected", {
  expect_snapshot(as_sheets_id("abc&123"), error = TRUE)
})

test_that("invalid inputs are caught", {
  expect_error(as_sheets_id(NULL))
  expect_error(as_sheets_id(1))
  expect_snapshot(as_sheets_id(letters[1:2]), error = TRUE)
})

test_that("id can be dug out of a URL", {
  expect_identical(
    as_sheets_id("https://docs.google.com/spreadsheets/d/abc123/"),
    as_sheets_id("abc123")
  )
  expect_identical(
    as_sheets_id("https://docs.google.com/spreadsheets/d/abc123/edit#gid=123"),
    as_sheets_id("abc123")
  )
})

test_that("invalid URL is interpreted as filepath, results in NA", {
  out <- as_sheets_id("https://www.r-project.org")
  expect_equal(vec_data(NA), NA)
})

# how I created the reference dribble, which represents two files:
#   * one Google Sheet
#   * one non-Google Sheet
# dat <- googledrive::drive_examples_remote()
# dat <- dat[dat$name %in% c("chicken.txt", "chicken_sheet"), ]
# saveRDS(dat, file = test_path("ref/dribble.rds"), version = 2)

test_that("multi-row dribble is rejected", {
  d <- readRDS(test_path("ref/dribble.rds"))
  expect_snapshot(as_sheets_id(d), error = TRUE)
})

test_that("dribble with non-Sheet file is rejected", {
  d <- readRDS(test_path("ref/dribble.rds"))
  d <- googledrive::drive_reveal(d, what = "mime_type")
  d <- d[d$mime_type == "text/plain", ]
  expect_snapshot(as_sheets_id(d), error = TRUE)
})

test_that("dribble with one Sheet can be coerced", {
  d <- readRDS(test_path("ref/dribble.rds"))
  d <- googledrive::drive_reveal(d, what = "mime_type")
  d <- d[d$mime_type == "application/vnd.google-apps.spreadsheet", ]
  expect_s3_class(as_sheets_id(d), "sheets_id")
})

test_that("a googlesheets4_spreadsheet can be coerced", {
  x <- new("Spreadsheet", spreadsheetId = "123")
  out <- as_sheets_id(new_googlesheets4_spreadsheet(x))
  expect_s3_class(out, "sheets_id")
  expect_identical(out, as_sheets_id("123"))
})

test_that("as_id.googlesheets4_spreadsheet works", {
  x <- new_googlesheets4_spreadsheet(list(spreadsheetId = "123"))
  expect_identical(as_id(x), as_id("123"))
})

## sheets_id print method ----

test_that("sheets_id print method reveals metadata", {
  skip_if_offline()
  skip_if_no_token()

  expect_snapshot(print(gs4_example("gapminder")))
})

test_that("sheets_id print method doesn't error for nonexistent ID", {
  skip_if_offline()
  skip_if_no_token()

  expect_no_error(format(as_sheets_id("12345")))
  expect_snapshot(as_sheets_id("12345"))
})

test_that("can print public sheets_id if deauth'd", {
  skip_if_offline()
  skip_on_cran()

  local_deauth()
  expect_snapshot(print(gs4_example("mini-gap")))
})

test_that("sheets_id print does not error for lack of cred", {
  skip_if_offline()
  skip_on_cran()

  local_deauth()
  local_interactive(FALSE)
  withr::local_options(list(gargle_oauth_cache = FALSE))

  # typical initial state: auth_active, but no token yet
  .auth$clear_cred()
  .auth$set_auth_active(TRUE)

  expect_no_error(format(gs4_example("mini-gap")))
  expect_snapshot(print(gs4_example("mini-gap")))
})

## low-level helpers ----
test_that("new_sheets_id() handles 0-length input and NA", {
  expect_no_error(
    out <- new_sheets_id(character())
  )
  expect_length(out, 0)
  expect_s3_class(out, "sheets_id")

  expect_no_error(
    out <- new_sheets_id(NA_character_)
  )
  expect_true(is.na(out))
  expect_s3_class(out, "sheets_id")
})

test_that("combining 2 sheets_id yields drive_id", {
  id1 <- as_sheets_id("abc")
  id2 <- as_sheets_id("def")
  expect_s3_class(c(id1, id2), class(new_drive_id()))
})

# ---- nm_fun ----
me_ <- nm_fun("TEST-sheet_delete")

# ---- tests ----
test_that("sheet_delete() rejects invalid `sheet`", {
  expect_error(
    sheet_delete(as_sheets_id("123"), sheet = TRUE),
    "must be either"
  )
})

test_that("sheet_delete() works", {
  skip_if_offline()
  skip_if_no_token()

  ss <- local_ss(me_())

  sheet_add(ss, c("alpha", "beta", "gamma", "delta"))

  expect_no_error(
    sheet_delete(ss, 1)
  )
  expect_no_error(
    sheet_delete(ss, "gamma")
  )
  expect_no_error(
    sheet_delete(ss, list("alpha", 2))
  )

  sheets_df <- sheet_properties(ss)

  expect_identical(sheets_df$name, "delta")
})

#' @export
as_tibble.googlesheets4_schema_Sheet <- function(x, ...) {
  out <- as_tibble(new("SheetProperties", !!!x$properties))
  # TODO: come back to deal with `data`
  tibble::add_column(out, data = list(NULL))
}

as_Sheet <- function(x, ...) {
  UseMethod("as_Sheet")
}

#' @export
as_Sheet.default <- function(x, ...) {
  abort_unsupported_conversion(x, to = "Sheet")
}

#' @export
as_Sheet.NULL <- function(x, ...) {
  return(new(id = "Sheet", properties = NULL))
}

#' @export
as_Sheet.character <- function(x, ...) {
  check_length_one(x)
  new(
    "Sheet",
    properties = new(id = "SheetProperties", title = x),
    ...
  )
}

#' @export
as_Sheet.data.frame <- function(x, ...) {
  # do first, so that gridProperties derived from x overwrite anything passed
  # via `...`
  sp <- new("SheetProperties", ...)

  sp <- patch(
    sp,
    gridProperties = new(
      "GridProperties",
      rowCount = nrow(x) + 1, # make room for column names
      columnCount = ncol(x),
    )
  )

  new(
    "Sheet",
    properties = sp,
    data = list(
      # an array of instances of GridData
      list(
        rowData = as_RowData(x) # an array of instances of RowData
      )
    )
  )
}

#' Read a Sheet into a data frame
#'
#' @description
#' This is the main "read" function of the googlesheets4 package. It goes by two
#' names, because we want it to make sense in two contexts:
#' * `read_sheet()` evokes other table-reading functions, like
#'   `readr::read_csv()` and `readxl::read_excel()`. The `sheet` in this case
#'    refers to a Google (spread)Sheet.
#' * `range_read()` is the right name according to the naming convention used
#'   throughout the googlesheets4 package.
#'
#' `read_sheet()` and `range_read()` are synonyms and you can use either one.
#'
#' @section Column Specification:
#'
#'   Column types must be specified in a single string of readr-style short
#'   codes, e.g. "cci?l" means "character, character, integer, guess, logical".
#'   This is not where googlesheets4's col spec will end up, but it gets the
#'   ball rolling in a way that is consistent with readr and doesn't reinvent
#'   any wheels.
#'
#'   Shortcodes for column types:

#'   * `_` or `-`: Skip. Data in a skipped column is still requested from the
#'   API (the high-level functions in this package are rectangle-oriented), but
#'   is not parsed into the data frame output.
#'   * `?`: Guess. A type is guessed for each cell and then a consensus type is
#'   selected for the column. If no atomic type is suitable for all cells, a
#'   list-column is created, in which each cell is converted to an R object of
#'   "best" type. If no column types are specified, i.e. `col_types = NULL`,
#'   all types are guessed.
#'   * `l`: Logical.
#'   * `i`: Integer. This type is never guessed from the data, because Sheets
#'   have no formal cell type for integers.
#'   * `d` or `n`: Numeric, in the sense of "double".
#'   * `D`: Date. This type is never guessed from the data, because date cells
#'   are just serial datetimes that bear a "date" format.
#'   * `t`: Time of day. This type is never guessed from the data, because time
#'   cells are just serial datetimes that bear a "time" format. *Not implemented
#'   yet; returns POSIXct.*
#'   * `T`: Datetime, specifically POSIXct.
#'   * `c`: Character.
#'   * `C`: Cell. This type is unique to googlesheets4. This returns raw cell
#'   data, as an R list, which consists of everything sent by the Sheets API for
#'   that cell. Has S3 type of `"CELL_SOMETHING"` and `"SHEETS_CELL"`. Mostly
#'   useful internally, but exposed for those who want direct access to, e.g.,
#'   formulas and formats.
#'   * `L`: List, as in "list-column". Each cell is a length-1 atomic vector of
#'   its discovered type.
#'   * *Still to come*: duration (code will be `:`) and factor (code will be
#'   `f`).
#'
#' @inheritParams range_read_cells
#' @param col_names `TRUE` to use the first row as column names, `FALSE` to get
#'   default names, or a character vector to provide column names directly. If
#'   user provides `col_types`, `col_names` can have one entry per column or one
#'   entry per unskipped column.
#' @param col_types Column types. Either `NULL` to guess all from the
#'   spreadsheet or a string of readr-style shortcodes, with one character or
#'   code per column. If exactly one `col_type` is specified, it is recycled.
#'   See Column Specification for more.
#' @param na Character vector of strings to interpret as missing values. By
#'   default, blank cells are treated as missing data.
#' @param trim_ws Logical. Should leading and trailing whitespace be trimmed
#'   from cell contents?
#' @param guess_max Maximum number of data rows to use for guessing column
#'   types.
#' @param .name_repair Handling of column names. By default, googlesheets4
#'   ensures column names are not empty and are unique. There is full support
#'   for `.name_repair` as documented in [tibble::tibble()].
#'
#' @return A [tibble][tibble::tibble-package]
#' @export
#'
#' @examplesIf gs4_has_token()
#' ss <- gs4_example("deaths")
#' read_sheet(ss, range = "A5:F15")
#' read_sheet(ss, range = "other!A5:F15", col_types = "ccilDD")
#' read_sheet(ss, range = "arts_data", col_types = "ccilDD")
#'
#' read_sheet(gs4_example("mini-gap"))
#' read_sheet(
#'   gs4_example("mini-gap"),
#'   sheet = "Europe",
#'   range = "A:D",
#'   col_types = "ccid"
#' )
range_read <- function(
  ss,
  sheet = NULL,
  range = NULL,
  col_names = TRUE,
  col_types = NULL,
  na = "",
  trim_ws = TRUE,
  skip = 0,
  n_max = Inf,
  guess_max = min(1000, n_max),
  .name_repair = "unique"
) {
  # check these first, so we don't download cells in vain
  col_spec <- standardise_col_spec(col_names, col_types, call = current_env())
  check_character(na)
  check_bool(trim_ws)
  check_non_negative_integer(guess_max)

  # range spec params are checked inside get_cells():
  # ss, sheet, range, skip, n_max
  df <- get_cells(
    ss = ss,
    sheet = sheet,
    range = range,
    col_names_in_sheet = isTRUE(col_spec$col_names),
    skip = skip,
    n_max = n_max
  )

  spread_sheet_impl_(
    df,
    col_spec = col_spec,
    na = na,
    trim_ws = trim_ws,
    guess_max = guess_max,
    .name_repair = .name_repair
  )
}

#' @rdname range_read
#' @export
read_sheet <- range_read

#' Spread a data frame of cells into spreadsheet shape
#'
#' Reshapes a data frame of cells (presumably the output of
#' [range_read_cells()]) into another data frame, i.e., puts it back into the
#' shape of the source spreadsheet. This function exists primarily for internal
#' use and for testing. The flagship function [range_read()], a.k.a.
#' [read_sheet()], is what most users are looking for. It is basically
#' [range_read_cells()] + `spread_sheet()`.
#'
#' @inheritParams range_read
#' @param df A data frame with one row per (nonempty) cell, integer variables
#'   `row` and `column` (probably referring to location within the spreadsheet),
#'   and a list-column `cell` of `SHEET_CELL` objects.
#'
#' @return A tibble in the shape of the original spreadsheet, but enforcing
#'   user's wishes regarding column names, column types, `NA` strings, and
#'   whitespace trimming.
#' @export
#'
#' @examplesIf gs4_has_token()
#' df <- gs4_example("mini-gap") %>%
#'   range_read_cells()
#' spread_sheet(df)
#'
#' # ^^ gets same result as ...
#' read_sheet(gs4_example("mini-gap"))
spread_sheet <- function(
  df,
  col_names = TRUE,
  col_types = NULL,
  na = "",
  trim_ws = TRUE,
  guess_max = min(1000, max(df$row)),
  .name_repair = "unique"
) {
  col_spec <- standardise_col_spec(col_names, col_types, call = current_env())
  check_character(na)
  check_bool(trim_ws)
  check_non_negative_integer(guess_max)

  spread_sheet_impl_(
    df,
    col_spec = col_spec,
    na = na,
    trim_ws = trim_ws,
    guess_max = guess_max,
    .name_repair = .name_repair
  )
}

spread_sheet_impl_ <- function(
  df,
  col_spec = list(
    col_names = TRUE,
    col_types = NULL
  ),
  na = "",
  trim_ws = TRUE,
  guess_max = min(1000, max(df$row)),
  .name_repair = "unique",
  call = caller_env()
) {
  if (nrow(df) == 0) {
    return(tibble::tibble())
  }
  col_names <- col_spec$col_names
  ctypes <- col_spec$ctypes
  col_names_in_sheet <- isTRUE(col_names)

  # absolute spreadsheet coordinates no longer relevant
  # update row, col to refer to location in output data frame
  # row 0 holds cells designated as column names
  df$row <- df$row - min(df$row) + !col_names_in_sheet
  nr <- max(df$row)
  df$col <- df$col - min(df$col) + 1

  if (is.logical(col_names)) {
    # if col_names is logical, this is first chance to check/set length of
    # ctypes, using the cell data
    ctypes <- rep_ctypes(
      max(df$col),
      ctypes,
      "column{?s} found in sheet",
      call = call
    )
  }

  # drop cells in skipped cols, update df$col and ctypes
  skipped_col <- ctypes == "COL_SKIP"
  if (any(skipped_col)) {
    df <- df[!df$col %in% which(skipped_col), ]
    df$col <- match(df$col, sort(unique(df$col)))
    ctypes <- ctypes[!skipped_col]
  }
  nc <- max(df$col)

  # if column names were provided explicitly, we need to check that length
  # of col_names (and, therefore, ctypes) == nc
  if (is.character(col_names) && length(col_names) != nc) {
    gs4_abort(
      c(
        "Length of {.arg col_names} is not compatible with the data:",
        "*" = "{.arg col_names} has length {length(col_names)}.",
        "x" = "But data has {nc} un-skipped column{?s}."
      ),
      call = call
    )
  }

  df$cell <- apply_ctype(df$cell, na = na, trim_ws = trim_ws)

  if (is.logical(col_names)) {
    col_names <- character(length = nc)
  }
  if (col_names_in_sheet) {
    this <- df$row == 0
    col_names[df$col[this]] <-
      as_character(df$cell[this], na = na, trim_ws = trim_ws)
    df <- df[!this, ]
  }

  df_split <- map(seq_len(nc), ~ df[df$col == .x, ])

  out_scratch <- map2(
    df_split,
    ctypes,
    make_column,
    na = na,
    trim_ws = trim_ws,
    nr = nr,
    guess_max = guess_max
  ) %>%
    set_names(col_names) %>%
    discard(is.null)

  as_tibble(out_scratch, .name_repair = .name_repair)
}

## helpers ---------------------------------------------------------------------

standardise_col_spec <- function(col_names, col_types, call = caller_env()) {
  check_col_names(col_names, call = call)
  ctypes <- standardise_ctypes(col_types, call = call)
  if (is.character(col_names)) {
    ctypes <- rep_ctypes(
      length(col_names),
      ctypes,
      "column name{?s}",
      call = call
    )
    col_names <- filter_col_names(col_names, ctypes)
    # if column names were provided explicitly, this is now true
    # length(col_names) == length(ctypes[ctypes != "COL_SKIP"])
  }
  list(col_names = col_names, ctypes = ctypes)
}

check_col_names <- function(col_names, call = caller_env()) {
  if (is.logical(col_names)) {
    return(check_bool(col_names, call = call))
  }
  check_character(col_names, call = call)
  check_has_length(col_names, call = call)
}

# input:  a string of readr-style shortcodes or NULL
# output: a vector of col types of length >= 1
standardise_ctypes <- function(col_types, call = caller_env()) {
  col_types <- col_types %||% "?"
  check_string(col_types, call = call)

  if (identical(col_types, "")) {
    gs4_abort(
      "
      {.arg col_types}, when provided, must be a string that contains at \\
      least one readr-style shortcode.",
      call = call
    )
  }

  accepted_codes <- keep(names(.ctypes), nzchar)

  col_types_split <- strsplit(col_types, split = "")[[1]]
  ok <- col_types_split %in% accepted_codes
  if (!all(ok)) {
    gs4_abort(
      c(
        "{.arg col_types} must be a string of readr-style shortcodes. \\
         Unrecognized code{?s}{cli::qty(sum(!ok))}:",
        bulletize(gargle_map_cli(col_types_split[!ok]), bullet = "x")
      ),
      call = call
    )
  }
  ctypes <- ctype(col_types_split)
  if (all(ctypes == "COL_SKIP")) {
    gs4_abort(
      "{.arg col_types} can't request that all columns be skipped.",
      call = call
    )
  }
  ctypes
}

# makes sure there are n ctypes or n ctypes that are not COL_SKIP
rep_ctypes <- function(n, ctypes, comparator = "n", call = caller_env()) {
  if (length(ctypes) == n) {
    return(ctypes)
  }
  n_col_types <- sum(ctypes != "COL_SKIP")
  if (n_col_types == n) {
    return(ctypes)
  }
  if (length(ctypes) == 1) {
    return(rep_len(ctypes, length.out = n))
  }
  # must pre-pluralize the comparator, e.g.
  # column{?s} found in sheet
  # column name{?s}
  comparator <- cli::pluralize(sprintf("{cli::qty(n)}%s{?s}", comparator))
  gs4_abort(
    c(
      "Length of {.arg col_types} is not compatible with {comparator}:",
      x = "{length(ctypes)} column type{?s} specified.",
      x = "{n_col_types} un-skipped column type{?s} specified.",
      x = "But there {cli::qty(n)}{?is/are} {n} {comparator}."
    ),
    call = call
  )
}

# removes col_names for skipped columns
# rep_ctypes() is called before and ensures that col_names and ctypes are
# conformable (hence the non-user facing stopifnot())
filter_col_names <- function(col_names, ctypes) {
  stopifnot(length(col_names) <= length(ctypes))
  col_names[ctypes != "COL_SKIP"]
}

# ---- nm_fun ----
me_ <- nm_fun("TEST-sheet_write")

# ---- tests ----
test_that("sheet_write() writes what it should", {
  skip_if_offline()
  skip_if_no_token()

  dat <- range_read(
    test_sheet("googlesheets4-col-types"),
    sheet = "lots-of-types",
    col_types = "lccinDT" # TODO: revisit when 'f' means factor
  )
  dat$factor <- factor(dat$factor)

  ss <- local_ss(me_("datetimes"))
  sheet_write(dat, ss)
  x <- range_read(ss, sheet = "dat", col_types = "C")

  # the main interesting bit to test is whether we successfully sent
  # correct value for the date and datetime, with a sane (= ISO 8601) format
  expect_equal(
    purrr::pluck(x, "date", 1, "formattedValue"),
    format(dat$date[1])
  )
  expect_equal(
    purrr::pluck(x, "date", 1, "effectiveFormat", "numberFormat", "type"),
    "DATE"
  )
  expect_equal(
    purrr::pluck(x, "date", 1, "effectiveFormat", "numberFormat", "pattern"),
    "yyyy-mm-dd"
  )

  expect_equal(
    purrr::pluck(x, "datetime", 1, "formattedValue"),
    format(dat$datetime[1])
  )
  expect_equal(
    purrr::pluck(x, "datetime", 1, "effectiveFormat", "numberFormat", "type"),
    "DATE_TIME"
  )
  expect_equal(
    purrr::pluck(
      x,
      "datetime",
      1,
      "effectiveFormat",
      "numberFormat",
      "pattern"
    ),
    "yyyy-mm-dd hh:mm:ss"
  )
})

test_that("sheet_write() can figure out (work)sheet name", {
  skip_if_offline()
  skip_if_no_token()

  foofy <- data.frame(x = 1:3, y = letters[1:3])

  ss <- local_ss(me_("sheetnames"))

  # get (work)sheet name from data frame's name
  sheet_write(foofy, ss)
  expect_equal(tail(sheet_names(ss), 1), "foofy")

  # we don't clobber existing (work)sheet if name was inferred
  sheet_write(foofy, ss)
  expect_equal(tail(sheet_names(ss), 1), "Sheet2")

  # we do write into existing (work)sheet if name is explicitly given
  sheet_write(foofy, ss, sheet = "foofy")
  expect_setequal(sheet_names(ss), c("Sheet1", "Sheet2", "foofy"))

  # we do write into existing (work)sheet if position is explicitly given
  sheet_write(foofy, ss, sheet = 2)
  expect_setequal(sheet_names(ss), c("Sheet1", "Sheet2", "foofy"))
})

test_that("read_sheet() works and discovers reasonable types", {
  skip_if_offline()
  skip_if_no_token()

  dat <- range_read(
    test_sheet("googlesheets4-col-types"),
    sheet = "lots-of-types"
  )
  expect_type(dat$logical, "logical")
  expect_type(dat$character, "character")
  expect_type(dat$factor, "character")
  expect_type(dat$integer, "double")
  expect_type(dat$double, "double")
  expect_s3_class(dat$date, "POSIXct")
  expect_s3_class(dat$datetime, "POSIXct")
})

test_that("read_sheet() enacts user-specified coltypes", {
  skip_if_offline()
  skip_if_no_token()

  dat <- range_read(
    test_sheet("googlesheets4-col-types"),
    sheet = "lots-of-types",
    col_types = "lccinDT"
  )
  expect_type(dat$logical, "logical")
  expect_type(dat$character, "character")
  expect_type(dat$factor, "character") # TODO: revisit when 'f' means factor
  expect_type(dat$integer, "integer")
  expect_type(dat$double, "double")
  expect_s3_class(dat$date, "Date")
  expect_s3_class(dat$datetime, "POSIXct")
})

test_that("read_sheet() can skip columns", {
  skip_if_offline()
  skip_if_no_token()

  dat <- range_read(
    test_sheet("googlesheets4-col-types"),
    sheet = "lots-of-types",
    col_types = "?-_-_-?"
  )
  expect_equal(ncol(dat), 2)
  expect_type(dat$logical, "logical")
  expect_s3_class(dat$datetime, "POSIXct")
})

# https://github.com/tidyverse/googlesheets4/issues/73
# https://github.com/tidyverse/googlesheets4/issues/174
test_that("read_sheet() honors `na`", {
  skip_if_offline()
  skip_if_no_token()

  # default behaviour
  dat <- read_sheet(
    test_sheet("googlesheets4-col-types"),
    sheet = "NAs"
  )
  expect_true(all(map_lgl(dat, is.character)))
  expect_false(is.na(dat$...NA[2]))
  expect_true(is.na(dat$space[2]))
  expect_true(is.na(dat$empty_string[2]))
  expect_true(is.na(dat$truly_empty[2]))

  # can explicit whitespace survive?
  dat <- read_sheet(
    test_sheet("googlesheets4-col-types"),
    sheet = "NAs",
    trim_ws = FALSE
  )
  expect_equal(dat$space[2], " ")

  # can we request empty string instead of NA?
  dat <- read_sheet(
    test_sheet("googlesheets4-col-types"),
    sheet = "NAs",
    na = character()
  )
  expect_equal(dat$space[2], "")
  expect_equal(dat$empty_string[2], "")
  expect_equal(dat$truly_empty[2], "")

  # explicit whitespace and empty-string-for-NA
  dat <- read_sheet(
    test_sheet("googlesheets4-col-types"),
    sheet = "NAs",
    na = character(),
    trim_ws = FALSE
  )
  expect_equal(dat$space[2], " ")

  # more NA strings
  dat <- read_sheet(
    test_sheet("googlesheets4-col-types"),
    sheet = "NAs",
    na = c("", "NA", "Missing")
  )
  expect_true(is.na(dat$...Missing[2]))
  expect_true(is.na(dat$...NA[2]))
  expect_true(is.na(dat$space[2]))
  expect_true(is.na(dat$empty_string[2]))
  expect_true(is.na(dat$truly_empty[2]))

  # column name that is NA string
  dat <- read_sheet(
    test_sheet("googlesheets4-col-types"),
    sheet = "NAs",
    na = "complete",
    .name_repair = ~ vec_as_names(.x, repair = "unique", quiet = TRUE)
  )
  expect_match(rev(names(dat))[1], "^...")

  # how NA strings interact with column typing
  dat <- read_sheet(
    test_sheet("googlesheets4-col-types"),
    sheet = "NAs",
    na = c("one", "three")
  )
  expect_true(is.character(dat$...Missing))
  expect_true(is.character(dat$...NA))
  expect_true(is.character(dat$space))
  expect_true(is.character(dat$complete))
  expect_true(is.logical(dat$empty_string))
  expect_true(is.logical(dat$truly_empty))
})

# helpers to check arguments ----
test_that("col_names must be logical or character and have length", {
  wrapper_fun <- function(...) check_col_names(...)
  expect_snapshot(wrapper_fun(1:3), error = TRUE)
  expect_snapshot(wrapper_fun(factor("a")), error = TRUE)
  expect_snapshot(wrapper_fun(character()), error = TRUE)
})

test_that("logical col_names must be TRUE or FALSE", {
  wrapper_fun <- function(...) check_col_names(...)
  expect_snapshot(wrapper_fun(NA), error = TRUE)
  expect_snapshot(wrapper_fun(c(TRUE, FALSE)), error = TRUE)
  expect_identical(check_col_names(TRUE), TRUE)
  expect_identical(check_col_names(FALSE), FALSE)
})

test_that("standardise_ctypes() turns NULL col_types into 'COL_GUESS'", {
  expect_equal(standardise_ctypes(NULL), c("?" = "COL_GUESS"))
})

test_that("standardise_ctypes() errors for only 'COL_SKIP'", {
  errmsg <- "can't request that all columns be skipped"
  expect_error(standardise_ctypes("-"), errmsg)
  expect_error(standardise_ctypes("-_"), errmsg)
})

test_that("standardise_ctypes() understands and requires readr shortcodes", {
  good <- "_-lidnDtTcCL?"
  expect_equal(
    standardise_ctypes(good),
    c(
      `_` = "COL_SKIP",
      `-` = "COL_SKIP",
      l = "CELL_LOGICAL",
      i = "CELL_INTEGER",
      d = "CELL_NUMERIC",
      n = "CELL_NUMERIC",
      D = "CELL_DATE",
      t = "CELL_TIME",
      T = "CELL_DATETIME",
      c = "CELL_TEXT",
      C = "COL_CELL",
      L = "COL_LIST",
      `?` = "COL_GUESS"
    )
  )
  expect_error(standardise_ctypes("abe"), "Unrecognized")
  expect_error(standardise_ctypes("f:"), "Unrecognized")
  expect_error(standardise_ctypes(""), "at least one")
})

test_that("col_types of right length are tolerated", {
  expect_identical(rep_ctypes(1, ctypes = "a"), "a")
  expect_identical(rep_ctypes(2, ctypes = c("a", "b")), c("a", "b"))
  expect_identical(
    rep_ctypes(2, ctypes = c("a", "b", "COL_SKIP")),
    c("a", "b", "COL_SKIP")
  )
})

test_that("a single col_types is repeated to requested length", {
  expect_identical(rep_ctypes(2, ctypes = "a"), c("a", "a"))
})

test_that("col_types with length > 1 and != n throw error", {
  expect_error(rep_ctypes(1, ctypes = rep("a", 2)), "not compatible")
  expect_error(rep_ctypes(3, ctypes = rep("a", 2)), "not compatible")
})

test_that("filter_col_names() removes entries for skipped columns", {
  expect_identical(filter_col_names(letters[1:2], letters[3:4]), letters[1:2])
  expect_identical(
    filter_col_names(letters[1:3], ctypes = c("a", "COL_SKIP", "c")),
    letters[c(1, 3)]
  )
})

# ---- nm_fun ----
me_ <- nm_fun("TEST-sheet_append")

# ---- tests ----
test_that("sheet_append() works", {
  skip_if_offline()
  skip_if_no_token()

  dat <- tibble::tibble(x = as.numeric(1:10), y = LETTERS[1:10])
  ss <- local_ss(me_(), sheets = list(test = dat[0, ]))

  sheet_append(ss, dat[1, ], sheet = "test")
  out <- range_read(ss, sheet = "test")
  expect_equal(out, dat[1, ])

  sheet_append(ss, dat[2:10, ], sheet = "test")
  out <- range_read(ss, sheet = "test")
  expect_equal(out, dat)
})

#' Delete one or more (work)sheets
#'
#' Deletes one or more (work)sheets from a (spread)Sheet.
#'
#' @eval param_ss()
#' @eval param_sheet(
#'   action = "delete",
#'   "You can pass a vector to delete multiple sheets at once or even a list,",
#'   "if you need to mix names and positions."
#' )
#'
#' @return The input `ss`, as an instance of [`sheets_id`]
#' @export
#' @family worksheet functions
#' @seealso Makes an `DeleteSheetsRequest`:
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#DeleteSheetRequest>
#'
#' @examplesIf gs4_has_token()
#' ss <- gs4_create("delete-sheets-from-me")
#' sheet_add(ss, c("alpha", "beta", "gamma", "delta"))
#'
#' # get an overview of the sheets
#' sheet_properties(ss)
#'
#' # delete sheets
#' sheet_delete(ss, 1)
#' sheet_delete(ss, "gamma")
#' sheet_delete(ss, list("alpha", 2))
#'
#' # get an overview of the sheets
#' sheet_properties(ss)
#'
#' # clean up
#' gs4_find("delete-sheets-from-me") %>%
#'   googledrive::drive_trash()
sheet_delete <- function(ss, sheet) {
  ssid <- as_sheets_id(ss)
  walk(sheet, ~ check_sheet(.x, arg = "sheet"))

  # retrieve spreadsheet metadata ----------------------------------------------
  x <- gs4_get(ssid)

  # capture sheet ids ----------------------------------------------------------
  s <- map(
    sheet,
    ~ lookup_sheet(.x, sheets_df = x$sheets, call = quote(sheet_delete()))
  )
  sheet_names <- map_chr(s, "name")
  n <- length(sheet_names)
  gs4_bullets(c(
    v = "Deleting {n} sheet{?s} from {.s_sheet {x$name}}:",
    bulletize(gargle_map_cli(sheet_names, template = "{.field <<x>>}"))
  ))

  sid <- map(s, "id")
  requests <- map(sid, ~ list(deleteSheet = list(sheetId = .x)))

  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = requests
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)
  invisible(ssid)
}

#' Delete cells
#'
#' Deletes a range of cells and shifts other cells into the deleted area. There
#' are several related tasks that are implemented by other functions:
#'   * To clear cells of their value and/or format, use [range_clear()].
#'   * To delete an entire (work)sheet, use [sheet_delete()].
#'   * To change the dimensions of a (work)sheet, use [sheet_resize()].
#'
#' @eval param_ss()
#' @eval param_sheet(
#'   action = "delete",
#'   "Ignored if the sheet is specified via `range`. If neither argument",
#'   "specifies the sheet, defaults to the first visible sheet."
#' )
#' @param range Cells to delete. There are a couple differences between `range`
#'   here and how it works in other functions (e.g. [range_read()]):
#'   * `range` must be specified.
#'   * `range` must not be a named range.
#'   * `range` must not be the name of a (work) sheet. Instead, use
#'     [sheet_delete()] to delete an entire sheet.
#'  Row-only and column-only ranges are especially relevant, such as "2:6" or
#'  "D". Remember you can also use the helpers in [`cell-specification`],
#'  such as `cell_cols(4:6)`, or `cell_rows(5)`.
#' @param shift Must be one of "up" or "left", if specified. Required if `range`
#'   is NOT a rows-only or column-only range (in which case, we can figure it
#'   out for you). Determines whether the deleted area is filled by shifting
#'   surrounding cells up or to the left.
#'
#' @template ss-return
#' @export
#' @family write functions
#' @seealso Makes a `DeleteRangeRequest`:
#' * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#DeleteRangeRequest>
#'
#' @examplesIf gs4_has_token()
#' # create a data frame to use as initial data
#' df <- gs4_fodder(10)
#'
#' # create Sheet
#' ss <- gs4_create("range-delete-example", sheets = list(df))
#'
#' # delete some rows
#' range_delete(ss, range = "2:4")
#'
#' # delete a column
#' range_delete(ss, range = "C")
#'
#' # delete a rectangle and specify how to shift remaining cells
#' range_delete(ss, range = "B3:F4", shift = "left")
#'
#' # clean up
#' gs4_find("range-delete-example") %>%
#'   googledrive::drive_trash()
range_delete <- function(ss, sheet = NULL, range, shift = NULL) {
  ssid <- as_sheets_id(ss)
  maybe_sheet(sheet)
  check_range(range)
  stopifnot(!is.null(range))
  if (is.null(shift)) {
    shift_dimension <- NULL
  } else {
    shift <- match.arg(shift, c("up", "left"))
    shift_dimension <- switch(shift, up = "ROWS", left = "COLUMNS")
  }

  x <- gs4_get(ssid)
  gs4_bullets(c(v = "Editing {.s_sheet {x$name}}."))

  # determine (work)sheet and range --------------------------------------------
  range_spec <- as_range_spec(
    range,
    sheet = sheet,
    sheets_df = x$sheets,
    nr_df = x$named_ranges
  )
  if (is.null(range_spec$cell_range) && is.null(range_spec$cell_limits)) {
    gs4_abort("{.fun range_delete} requires a cell range.")
  }
  range_spec$sheet_name <- range_spec$sheet_name %||%
    first_visible_name(x$sheets)
  # as_GridRange() throws an error for a named range
  grid_range <- as_GridRange(range_spec)
  gs4_bullets(c(
    v = "Deleting cells in sheet {.w_sheet {range_spec$sheet_name}}."
  ))

  # form batch update request --------------------------------------------------
  shift_dimension <- shift_dimension %||% determine_shift(grid_range)
  if (is.null(shift_dimension)) {
    gs4_abort(c(
      "The {.arg shift} direction must be specified for this {.arg range}.",
      "It can't be automatically determined."
    ))
  }

  # form batch update request --------------------------------------------------
  delete_req <- list(
    deleteRange = new(
      "DeleteRangeRequest",
      range = grid_range,
      shiftDimension = shift_dimension
    )
  )

  # do it ----------------------------------------------------------------------
  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = list(delete_req)
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)

  invisible(ssid)
}

determine_shift <- function(gr, call = caller_env()) {
  stopifnot(inherits(gr, "googlesheets4_schema_GridRange"))
  bounded_on_bottom <- !is.null(gr$endRowIndex) && notNA(gr$endRowIndex)
  bounded_on_right <- !is.null(gr$endColumnIndex) && notNA(gr$endColumnIndex)

  if (bounded_on_bottom && bounded_on_right) {
    # user must specify shift
    return(NULL)
  }

  if (bounded_on_bottom) {
    # and not bounded_on_right
    return("ROWS")
  }

  if (bounded_on_right) {
    # and not bounded_on_bottom
    return("COLUMNS")
  }

  gs4_abort(
    c(
      "{.arg range} must be bounded on the bottom and/or on the right.",
      i = "Use {.fun sheet_delete} or {.fun sheet_resize} to delete or \\
           resize a (work)sheet."
    ),
    call = call
  )
}

# ---- nm_fun ----
me_ <- nm_fun("TEST-range_delete")

# ---- tests ----
test_that("range_delete() works", {
  skip_if_offline()
  skip_if_no_token()

  df <- gs4_fodder(4)
  ss <- local_ss(me_(), sheets = list(df))

  range_delete(ss, range = "2:3")
  range_delete(ss, range = "B")
  range_delete(ss, range = "B2", shift = "left")

  out <- range_read(ss)

  expect_match(sub("[A-Z](\\d)", "\\1", out[2, ]), "5")
  expect_equal(names(out), c("A", "C", "D"))
  expect_equal(out[[1, 2]], "D4")
})

# helpers ----

test_that("determine_shift() 'works' for ranges where user input is required", {
  sheets_df <- tibble::tibble(name = "Sheet1", index = 0)

  # these are all essentially true rectangles and the user will have to tell
  # us how to shift cells into the deleted region
  bounded_bottom_and_right <- list(
    cell_limits(c(NA, NA), c(3, 5)),
    cell_limits(c(1, NA), c(3, 5)),
    cell_limits(c(NA, 3), c(3, 5)),
    cell_limits(c(1, 3), c(3, 5))
  )

  out <- purrr::map(
    bounded_bottom_and_right,
    ~ determine_shift(as_GridRange(as_range_spec(.x, sheets_df = sheets_df)))
  )
  purrr::map(out, expect_null)
})

test_that("determine_shift() detects ranges where we shift ROWS up", {
  sheets_df <- tibble::tibble(name = "Sheet1", index = 0)

  # these are bounded on the bottom, but not the on the right
  bounded_bottom <- list(
    cell_limits(c(NA, NA), c(3, NA)),
    cell_limits(c(1, NA), c(3, NA)),
    cell_limits(c(NA, 3), c(3, NA)),
    cell_limits(c(1, 3), c(3, NA))
  )

  out <- purrr::map_chr(
    bounded_bottom,
    ~ determine_shift(as_GridRange(as_range_spec(.x, sheets_df = sheets_df)))
  )
  expect_match(out, "ROWS")
})

test_that("determine_shift() detects ranges where we shift COLUMNS left", {
  sheets_df <- tibble::tibble(name = "Sheet1", index = 0)

  # these are bounded on the bottom, but not the on the right
  bounded_right <- list(
    cell_limits(c(NA, NA), c(NA, 5)),
    cell_limits(c(1, NA), c(NA, 5)),
    cell_limits(c(NA, 3), c(NA, 5)),
    cell_limits(c(1, 3), c(NA, 5))
  )

  out <- purrr::map_chr(
    bounded_right,
    ~ determine_shift(as_GridRange(as_range_spec(.x, sheets_df = sheets_df)))
  )
  expect_match(out, "COLUMNS")
})

test_that("determine_shift() detects ranges where we must error", {
  sheets_df <- tibble::tibble(name = "Sheet1", index = 0)

  # these are not bounded at on either the bottom or the right
  not_bounded <- list(
    cell_limits(c(NA, NA), c(NA, NA)),
    cell_limits(c(1, NA), c(NA, NA)),
    cell_limits(c(NA, 3), c(NA, NA)),
    cell_limits(c(1, 3), c(NA, NA))
  )

  expect_bad_range <- function(x) {
    grid_range <- as_GridRange(as_range_spec(x, sheets_df = sheets_df))
    expect_error(determine_shift(grid_range), "must be bounded")
  }

  purrr::walk(not_bounded, expect_bad_range)
})

.onLoad <- function(libname, pkgname) {
  # .auth is created in R/gs4_auth.R
  # this is to insure we get an instance of gargle's AuthState using the
  # current, locally installed version of gargle
  utils::assignInMyNamespace(
    ".auth",
    gargle::init_AuthState(package = "googlesheets4", auth_active = TRUE)
  )

  if (identical(Sys.getenv("IN_PKGDOWN"), "true")) {
    tryCatch(
      gs4_auth_docs(),
      googlesheets4_auth_internal_error = function(e) NULL
    )
  }

  invisible()
}

#' List Sheets endpoints
#'
#' Returns a list of selected Sheets API v4 endpoints, as stored inside the
#' googlesheets4 package. The names of this list (or the `id` sub-elements) are
#' the nicknames that can be used to specify an endpoint in
#' [request_generate()]. For each endpoint, we store its nickname or `id`, the
#' associated HTTP `method`, the `path`, and details about the parameters. This
#' list is derived programmatically from the Sheets API v4 Discovery
#' Document (`https://www.googleapis.com/discovery/v1/apis/sheets/v4/rest`).
#'
#' @param i The name(s) or integer index(ices) of the endpoints to return.
#'   Optional. By default, the entire list is returned.
#'
#' @return A list containing some or all of the subset of the Sheets API v4
#'   endpoints that are used internally by googlesheets4.
#' @export
#'
#' @examples
#' str(gs4_endpoints(), max.level = 2)
#' gs4_endpoints("sheets.spreadsheets.values.get")
#' gs4_endpoints(4)
gs4_endpoints <- function(i = NULL) {
  if (is.null(i)) {
    i <- seq_along(.endpoints)
  }
  .endpoints[i]
}

test_that("endpoints can be retrieved en masse", {
  endpoints <- gs4_endpoints()
  expect_true(length(endpoints) >= 14)
  expect_match(names(endpoints), "^sheets\\.spreadsheets\\.")
})

test_that("a single endpoint can be retrieved", {
  nm <- "sheets.spreadsheets.values.batchClear"
  endpoint <- gs4_endpoints(nm)[[1]]
  expect_true(
    all(c("id", "path", "parameters", "scopes") %in% names(endpoint))
  )
})

# hack-y implementation of typed pluck with an NA default
glean_lgl <- function(.x, ..., .default = NA) {
  map_lgl(list(.x), ..., .default = .default)
}

glean_chr <- function(.x, ..., .default = NA) {
  map_chr(list(.x), ..., .default = .default)
}

glean_int <- function(.x, ..., .default = NA) {
  map_int(list(.x), ..., .default = .default)
}

test_that("glean_lgl() works", {
  expect_identical(glean_lgl(list(a = TRUE), "a"), TRUE)
  expect_identical(glean_lgl(list(b = TRUE), "a"), NA)
  expect_identical(glean_lgl(list(), "a"), NA)
  expect_identical(glean_lgl(list(b = TRUE), "a", .default = FALSE), FALSE)
  expect_error(glean_lgl(list(a = "a"), "a"), "Can't coerce")
})

test_that("glean_chr() works", {
  expect_identical(glean_chr(list(a = "hi"), "a"), "hi")
  expect_identical(glean_chr(list(b = "bye"), "a"), NA_character_)
  expect_identical(glean_chr(list(), "a"), NA_character_)
  expect_identical(glean_chr(list(b = "bye"), "a", .default = "huh"), "huh")
})

test_that("glean_int() works", {
  expect_identical(glean_int(list(a = 1L), "a"), 1L)
  expect_identical(glean_int(list(b = 1L), "a"), NA_integer_)
  expect_identical(glean_int(list(), "a"), NA_integer_)
  expect_identical(glean_int(list(b = 1L), "a", .default = 2L), 2L)
  expect_error(glean_int(list(a = "a"), "a"), "Can't coerce")
})

#' Create useful spreadsheet filler
#'
#' Creates a data frame that is useful for filling a spreadsheet, when you just
#' need a sheet to experiment with. The data frame has `n` rows and `m` columns
#' with these properties:
#' * Column names match what Sheets displays: "A", "B", "C", and so on.
#' * Inner cell values reflect the coordinates where each value will land in
#'   the sheet, in A1-notation. So the first row is "B2", "C2", and so on.
#' Note that this `n`-row data frame will occupy `n + 1` rows in the sheet,
#' because the column names occupy the first row.
#'
#' @param n Number of rows.
#' @param m Number of columns.
#'
#' @return A data frame of character vectors.
#' @export
#'
#' @examples
#' gs4_fodder()
#' gs4_fodder(5, 3)
gs4_fodder <- function(n = 10, m = n) {
  columns <- LETTERS[seq_len(m)]
  names(columns) <- columns
  f <- function(number, letter) paste0(letter, number)
  as.data.frame(
    outer(seq_len(n) + 1, columns, f),
    stringsAsFactors = FALSE
  )
}

test_that("gs4_fodder() works", {
  dat <- gs4_fodder(3, 5)
  expect_named(dat, LETTERS[1:5])
  ltrs <- rep(LETTERS[1:5], each = 3)
  nbrs <- rep(1:3, 5) + 1
  expect_equal(
    as.vector(as.matrix(dat)),
    paste0(ltrs, nbrs)
  )
})

#' Protect a cell range
#'
#' @description
#' *Note: not yet exported, still very alpha. Usage still requires using
#' low-level helpers. This documentation is for ME.*
#'
#' `range_add_protection()` protects a range of cells against editing.
#'
#' @eval param_ss()
#' @eval param_sheet()
#' @param range Cells to protect. This `range` argument works very much like
#'   `range` in, for example, [range_read()]). Specific things to note:
#'   You can omit `range` to protect a whole sheet and `range` can be a named
#'   range.
#' @param ... Optional arguments used when constructing the `ProtectedRange`
#'   object. Use this is you want to set `description`, `warningOnly`,
#'   `unprotectedRanges`, or `editors`. For advanced use.
#'
#' @template ss-return
#' @seealso Makes an `AddProtectedRangeRequest`:
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#addprotectedrangerequest>
#'
#' Documentation on the `ProtectedRange` object:
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#protectedrange>
#'
#' @keywords internal
#' @noRd
#'
#' @examplesIf gs4_has_token()
#' # create a data frame to use as initial data
#' dat <- gs4_fodder(3)
#'
#' # create Sheet, add a couple more sheets
#' ss <- gs4_create("range-add-protection-example", sheets = dat)
#' sheet_write(head(chickwts), ss, sheet = "chickwts")
#' sheet_write(head(mtcars), ss, sheet = "mtcars")
#' sheet_write(ToothGrowth, ss, sheet = "ToothGrowth")
#'
#' # add myself and get it open in the browser
#' gs4_share(ss, type = "user", emailAddress = "jenny@rstudio.com", role = "writer")
#' gs4_browse(ss)
#'
#' # protect a whole sheet
#' ss %>%
#'   range_add_protection(sheet = "dat", description = "whole sheet")
#'
#' # create a named range, then protect it
#' ss %>%
#'   range_add_named("feed", sheet = "chickwts", range = "B:B") %>%
#'   range_add_protection(range = "feed", description = "named range")
#'
#' # protect an arbitrary rectangle and add an editor
#' ss %>%
#'   range_add_protection(
#'     range = "mtcars!1:1",
#'     description = "single row",
#'     editors = new("Editors", users = "jenny@rstudio.com")
#'   )
#'
#' # check in on the protected ranges we've created
#' ss_info <- gs4_get(ss)
#' ss_info$protected_ranges
#'
#' # protect a sheet EXCEPT certain columns that can be edited
#' unprotect_this <- as_range_spec(
#'   "C:C",
#'   sheet = "ToothGrowth",
#'   sheets_df = ss_info$sheets, nr_df = ss_info$named_ranges
#' )
#' unprotect_range <- as_GridRange(unprotect_this)
#' ss %>%
#'   range_add_protection(
#'     sheet = "ToothGrowth",
#'     description = "sheet MINUS some cols",
#'     unprotectedRanges = unprotect_range
#'   )
#'
#' # look at the editors for our protected ranges
#' ss_info <- gs4_get(ss)
#' ss_info$protected_ranges
#' ss_info$protected_ranges$editors
#'
#' # add an editor to a protected range
#' id <- ss_info$protected_ranges$protected_range_id[[1]]
#' range_update_protection(
#'   ss,
#'   protectedRangeId = id,
#'   editors = new("Editors", users = "jenny@rstudio.com")
#' )
#'
#' # confirm the editor change happened
#' ss_info <- gs4_get(ss)
#' ss_info$protected_ranges$editors
#'
#' # delete protections from a range
#' id <- ss_info$protected_ranges$protected_range_id[[3]]
#' range_delete_protection(ss, id = id)
#'
#' # confirm the deletion happened
#' ss_info <- gs4_get(ss)
#' ss_info$protected_ranges
#'
#' # clean up
#' gs4_find("range-add-protection-example") %>%
#'   googledrive::drive_trash()
range_add_protection <- function(ss, sheet = NULL, range = NULL, ...) {
  ssid <- as_sheets_id(ss)
  maybe_sheet(sheet)
  check_range(range)

  x <- gs4_get(ssid)
  gs4_bullets(c(v = "Editing {.s_sheet {x$name}}."))

  # determine range ------------------------------------------------------------
  range_spec <- as_range_spec(
    range,
    sheet = sheet,
    sheets_df = x$sheets,
    nr_df = x$named_ranges
  )
  if (is.null(range_spec$named_range)) {
    range_spec$sheet_name <- range_spec$sheet_name %||%
      first_visible_name(x$sheets)
    gs4_bullets(c(
      v = "Protecting cells on sheet: {.w_sheet {range_spec$sheet_name}}."
    ))
  } else {
    gs4_bullets(c(
      v = "Protecting named range: {.range {range_spec$named_range}}."
    ))
  }

  # form batch update request --------------------------------------------------
  prot_req <- list(
    addProtectedRange = new(
      "AddProtectedRangeRequest",
      protectedRange = new_ProtectedRange(range_spec, ...)
    )
  )

  # do it ----------------------------------------------------------------------
  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = list(prot_req)
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)

  invisible(ssid)
}

# helpers ----
new_ProtectedRange <- function(range_spec, ...) {
  if (is.null(range_spec$named_range)) {
    out <- new("ProtectedRange", range = as_GridRange(range_spec))
  } else {
    out <- new(
      "ProtectedRange",
      namedRangeId = vlookup(
        range_spec$named_range,
        range_spec$nr_df,
        "name",
        "id"
      )
    )
  }
  out <- patch(out, editors = new("Editors", domainUsersCanEdit = FALSE))
  patch(out, ...)
}

# even less polished one-offs used during development
range_update_protection <- function(ss, ...) {
  ssid <- as_sheets_id(ss)

  x <- gs4_get(ssid)
  gs4_bullets(c(v = "Editing {.s_sheet {x$name}}."))

  # form batch update request --------------------------------------------------
  protected_range <- new("ProtectedRange", ...)
  mask <- gargle::field_mask(protected_range)
  # I have no idea why this is necessary, but it's the only way I've been able
  # to updated editors
  mask <- sub("editors.users", "editors", mask)
  prot_req <- list(
    updateProtectedRange = new(
      "UpdateProtectedRangeRequest",
      protectedRange = protected_range,
      fields = mask
    )
  )

  # do it ----------------------------------------------------------------------
  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = list(prot_req)
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)

  invisible(ssid)
}

range_delete_protection <- function(ss, id) {
  ssid <- as_sheets_id(ss)

  x <- gs4_get(ssid)
  gs4_bullets(c(v = "Editing {.s_sheet {x$name}}."))

  # form batch update request --------------------------------------------------
  prot_req <- list(
    deleteProtectedRange = new(
      "DeleteProtectedRangeRequest",
      protectedRangeId = id
    )
  )

  # do it ----------------------------------------------------------------------
  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = list(prot_req)
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)

  invisible(ssid)
}

#' Make a Google Sheets API request
#'
#' Low-level function to execute a Sheets API request. Most users should,
#' instead, use higher-level wrappers that facilitate common tasks, such as
#' reading or writing worksheets or cell ranges. The functions here are intended
#' for internal use and for programming around the Sheets API.
#'
#' `make_request()` is a very thin wrapper around [gargle::request_retry()],
#' only adding the googlesheets4 user agent. Typically the input has been
#' created with [request_generate()] or [gargle::request_build()] and the output
#' is processed with `process_response()`.
#'
#' [gargle::request_retry()] retries requests that error with `429
#' RESOURCE_EXHAUSTED`. Its basic scheme is exponential backoff, with one tweak
#' that is very specific to the Sheets API, which has documented [usage
#' limits](https://developers.google.com/sheets/api/limits):
#'
#' "a limit of 500 requests per 100 seconds per project and 100 requests per 100
#' seconds per user"
#'
#' Note that the "project" here means everyone using googlesheets4 who hasn't
#' configured their own OAuth client. This is potentially a lot of users, all
#' acting independently.
#'
#' If you hit the "100 requests per 100 seconds per **user**" limit (which
#' really does mean YOU), the first wait time is a bit more than 100 seconds,
#' then we revert to exponential backoff.
#'
#' If you experience lots of retries, especially with 100 second delays, it
#' means your use of googlesheets4 is more than casual and **it's time for you
#' to get your own OAuth client or use a service account token**. This is explained
#' in the gargle vignette `vignette("get-api-credentials", package = "gargle")`.
#'
#' @param x List. Holds the components for an HTTP request, presumably created
#'   with [request_generate()] or [gargle::request_build()]. Must contain a
#'   `method` and `url`. If present, `body` and `token` are used.
#' @param ... Optional arguments passed through to the HTTP method.
#' @param encode If the body is a named list, how should it be encoded? This has
#'   the same meaning as `encode` in all the [httr::VERB()]s, such as
#'   [httr::POST()]. Note, however, that we default to `encode = "json"`, which
#'   is what you want most of the time when calling the Sheets API. The httr
#'   default is `"multipart"`. Other acceptable values are `"form"` and `"raw"`.
#'
#' @return Object of class `response` from [httr].
#' @export
#' @family low-level API functions
request_make <- function(x, ..., encode = "json") {
  gargle::request_retry(
    x,
    ...,
    encode = encode,
    user_agent = gs4_user_agent()
  )
}

gs4_user_agent <- function() {
  httr::user_agent(paste0(
    "googlesheets4/",
    utils::packageVersion("googlesheets4"),
    " ",
    "(GPN:RStudio; )",
    " ",
    "gargle/",
    utils::packageVersion("gargle"),
    " ",
    "httr/",
    utils::packageVersion("httr")
  ))
}

#' Auto-fit columns or rows to the data
#'
#' Applies automatic resizing to either columns or rows of a (work)sheet. The
#' width or height of targeted columns or rows, respectively, is determined
#' from the current cell contents. This only affects the appearance of a sheet
#' in the browser and doesn't affect its values or dimensions in any way.
#'
#' @eval param_ss()
#' @eval param_sheet(
#'   action = "modify",
#'   "Ignored if the sheet is specified via `range`. If neither argument",
#'   "specifies the sheet, defaults to the first visible sheet."
#' )
#' @param range Which columns or rows to resize. Optional. If you want to resize
#'   all columns or all rows, use `dimension` instead. All the usual `range`
#'   specifications are accepted, but the targeted range must specify only
#'   columns (e.g. "B:F") or only rows (e.g. "2:7").
#' @param dimension Ignored if `range` is given. If consulted, `dimension` must
#'   be either `"columns"` (the default) or `"rows"`. This is the simplest way
#'   to request auto-resize for all columns or all rows.
#'
#' @template ss-return
#' @export
#' @family formatting functions
#' @seealso Makes an `AutoResizeDimensionsRequest`:
#' * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#autoresizedimensionsrequest>
#'
#' @examplesIf gs4_has_token()
#' dat <- tibble::tibble(
#'   fruit = c("date", "lime", "pear", "plum")
#' )
#'
#' ss <- gs4_create("range-autofit-demo", sheets = dat)
#' ss
#'
#' # open in the browser
#' gs4_browse(ss)
#'
#' # shrink column A to fit the short fruit names
#' range_autofit(ss)
#' # in the browser, notice how the column width shrank
#'
#' # send some longer fruit names
#' dat2 <- tibble::tibble(
#'   fruit = c("cucumber", "honeydew")
#' )
#' ss %>% sheet_append(dat2)
#' # in the browser, see that column A is now too narrow to show the data
#'
#' range_autofit(ss)
#' # in the browser, see the column A reveals all the data now
#'
#' # clean up
#' gs4_find("range-autofit-demo") %>%
#'   googledrive::drive_trash()
range_autofit <- function(
  ss,
  sheet = NULL,
  range = NULL,
  dimension = c("columns", "rows")
) {
  ssid <- as_sheets_id(ss)
  maybe_sheet(sheet)
  check_range(range)

  x <- gs4_get(ssid)

  # determine targeted sheet ---------------------------------------------------
  range_spec <- as_range_spec(
    range,
    sheet = sheet,
    sheets_df = x$sheets,
    nr_df = x$named_ranges
  )
  range_spec$sheet_name <- range_spec$sheet_name %||%
    first_visible_name(x$sheets)
  s <- lookup_sheet(range_spec$sheet_name, sheets_df = x$sheets)

  # form request ---------------------------------------------------------------
  if (is.null(range)) {
    dimension <- match.arg(dimension)
    resize_req <- list(bureq_auto_resize_dimensions(
      sheetId = s$id,
      dimension = toupper(dimension)
    ))
  } else {
    resize_req <- prepare_auto_resize_request(s$id, range_spec)
  }
  resize_dim <- pluck(
    resize_req,
    1,
    "autoResizeDimensions",
    "dimensions",
    "dimension"
  )

  gs4_bullets(c(
    v = "Editing {.s_sheet {x$name}}.",
    v = "Resizing one or more {tolower(resize_dim)} in \\
         {.w_sheet {range_spec$sheet_name}}."
  ))

  # do it ----------------------------------------------------------------------
  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = resize_req
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)

  invisible(ssid)
}

force_cell_limits <- function(x) {
  if (!is.null(x$cell_limits)) {
    return(x)
  }

  if (is.null(x$cell_range)) {
    x$cell_limits <- cell_limits()
  } else {
    x$cell_limits <- limits_from_range(x$cell_range)
  }
  x
}

check_only_one_dimension <- function(x, call = caller_env()) {
  limits <- x$cell_limits

  if (is.na(limits$ul[1]) && is.na(limits$lr[1])) {
    return(invisible(x))
  }
  if (is.na(limits$ul[2]) && is.na(limits$lr[2])) {
    return(invisible(x))
  }

  gs4_abort(
    "The {.arg range} must target only columns or only rows, but not both.",
    call = call
  )
}

determine_dimension <- function(x) {
  limits <- x$cell_limits

  if (notNA(limits$ul[1]) || notNA(limits$lr[1])) {
    "ROWS"
  } else {
    "COLUMNS"
  }
}

prepare_auto_resize_request <- function(sheet_id, range_spec) {
  range_spec <- force_cell_limits(range_spec)
  check_only_one_dimension(range_spec)
  dimension <- determine_dimension(range_spec)

  element <- if (dimension == "ROWS") 1L else 2L

  list(bureq_auto_resize_dimensions(
    sheetId = sheet_id,
    dimension = dimension,
    start = pluck(range_spec, "cell_limits", "ul", element),
    end = pluck(range_spec, "cell_limits", "lr", element)
  ))
}

# ---- nm_fun ----
me_ <- nm_fun("TEST-range_autofit")

# ---- tests ----
test_that("range_autofit() works", {
  skip_if_offline()
  skip_if_no_token()

  dat <- tibble::tribble(
        ~x,     ~y,     ~z,     ~a,     ~b,     ~c,
    "abcd", "efgh", "ijkl", "mnop", "qrst", "uvwx"
  )
  ss <- local_ss(me_(), sheets = list(dat = dat))
  ssid <- as_sheets_id(ss)

  range_autofit(ss)
  before <- gs4_get_impl_(
    ssid,
    fields = "sheets.data.columnMetadata.pixelSize"
  )

  dat2 <- purrr::modify(dat, ~ paste0(.x, "_", .x))
  dat4 <- purrr::modify(dat2, ~ paste0(.x, "_", .x))
  sheet_append(ss, dat4)
  range_autofit(ss)

  after <- gs4_get_impl_(
    ssid,
    fields = "sheets.data.columnMetadata.pixelSize"
  )

  before <- pluck(before, "sheets", 1, "data", 1, "columnMetadata")
  after <- pluck(after, "sheets", 1, "data", 1, "columnMetadata")
  expect_true(all(unlist(before) < unlist(after)))
})

# ---- helpers ----
test_that("A1-style ranges can be turned into a request", {
  req <- prepare_auto_resize_request(123, as_range_spec("D:H"))
  req <- pluck(req, 1, "autoResizeDimensions", "dimensions")
  expect_equal(req$dimension, "COLUMNS")
  expect_equal(req$startIndex, cellranger::letter_to_num("D") - 1)
  expect_equal(req$endIndex, cellranger::letter_to_num("H"))

  req <- prepare_auto_resize_request(123, as_range_spec("3:7"))
  req <- pluck(req, 1, "autoResizeDimensions", "dimensions")
  expect_equal(req$dimension, "ROWS")
  expect_equal(req$startIndex, 3 - 1)
  expect_equal(req$endIndex, 7)
})

test_that("cell_limits can be turned into a request", {
  req <- prepare_auto_resize_request(
    123,
    as_range_spec(cell_limits())
  )
  req <- pluck(req, 1, "autoResizeDimensions", "dimensions")
  expect_equal(req$dimension, "COLUMNS")
  expect_null(req$startIndex)
  expect_null(req$endIndex)

  req <- prepare_auto_resize_request(
    123,
    as_range_spec(cell_cols(c(3, NA)))
  )
  req <- pluck(req, 1, "autoResizeDimensions", "dimensions")
  expect_equal(req$dimension, "COLUMNS")
  expect_equal(req$startIndex, 3 - 1)
  expect_null(req$endIndex)

  req <- prepare_auto_resize_request(
    123,
    as_range_spec(cell_cols(c(NA, 5)))
  )
  req <- pluck(req, 1, "autoResizeDimensions", "dimensions")
  expect_equal(req$dimension, "COLUMNS")
  expect_equal(req$endIndex, 5)
})

test_that("an invalid range is rejected", {
  expect_error(
    prepare_auto_resize_request(123, as_range_spec("D3:H")),
    "only columns or only rows"
  )
})

#' @param range A cell range to read from. If `NULL`, all non-empty cells are
#'   read. Otherwise specify `range` as described in [Sheets A1
#'   notation](https://developers.google.com/sheets/api/guides/concepts#a1_notation)
#'   or using the helpers documented in [cell-specification]. Sheets uses
#'   fairly standard spreadsheet range notation, although a bit different from
#'   Excel. Examples of valid ranges: `"Sheet1!A1:B2"`, `"Sheet1!A:A"`,
#'   `"Sheet1!1:2"`, `"Sheet1!A5:A"`, `"A1:B2"`, `"Sheet1"`. Interpreted
#'   strictly, even if the range forces the inclusion of leading, trailing, or
#'   embedded empty rows or columns. Takes precedence over `skip`, `n_max` and
#'   `sheet`. Note `range` can be a named range, like `"sales_data"`, without
#'   any cell reference.

# Why not `new("CellData", ...)`? It seems excessive to store the schema as
# an attribute for each cell. Possibly a premature concern.
new_CellData <- function(...) {
  # explicit 'list' class is a bit icky but it makes jsonlite happy
  structure(
    list2(...),
    class = c(
      "googlesheets4_schema_CellData",
      "googlesheets4_schema",
      "list"
    )
  )
}

# Use this instead of `new_CellData()` when (light) validation makes sense.
CellData <- function(...) {
  dots <- list2(...)
  stopifnot(is_dictionaryish(dots))
  check_against_schema(dots, id = "CellData")
  new_CellData(...)
}

is_CellData <- function(x) inherits(x, "googlesheets4_schema_CellData")

as_CellData <- function(x, .na = NULL) {
  UseMethod("as_CellData")
}

#' @export
as_CellData.default <- function(x, .na = NULL) {
  abort_unsupported_conversion(x, to = "CellData")
}

# I want to centralize what we send for NA, even though -- for now, at
# least -- I have not exposed this in user-facing functions. You could imagine
# generalizing to allow user to request we send #N/A.
# More about #N/A:
# https://support.google.com/docs/answer/3093359?hl=en
# However, again, for now, we leave NA as NA and let jsonlite do its usual
# thing, which is to encode as JSON `null`.
empty_cell <- function(.na = NULL) {
  if (is.null(.na)) {
    new_CellData(userEnteredValue = NA)
  } else {
    CellData(!!!.na)
  }
}

# Note that this always returns a **list** of instances of
# googlesheets4_schema_CellData
# of the same length as x.
cell_data <- function(x, val_type, .na = NULL) {
  force(val_type)
  f <- function(y) {
    new_CellData(userEnteredValue = list2(!!val_type := y))
  }
  out <- map(x, f)
  out[is.na(x)] <- list(empty_cell(.na = .na))
  out
}

#' @export
as_CellData.NULL <- function(x, .na = NULL) {
  list(empty_cell(.na))
}

#' @export
as_CellData.googlesheets4_schema_CellData <- function(x, .na = NULL) {
  x
}

#' @export
as_CellData.logical <- function(x, .na = NULL) {
  cell_data(x, val_type = "boolValue", .na = .na)
}

#' @export
as_CellData.character <- function(x, .na = NULL) {
  cell_data(x, val_type = "stringValue", .na = .na)
}

#' @export
as_CellData.factor <- function(x, .na = NULL) {
  as_CellData(as.character(x), .na = .na)
}

#' @export
as_CellData.numeric <- function(x, .na = NULL) {
  cell_data(x, val_type = "numberValue", .na = .na)
}

#' @export
as_CellData.list <- function(x, .na = NULL) {
  out <- map(x, as_CellData, .na = .na)
  # awkwardness possibly solved by using vctrs to create an S3 class for
  # CellData ... but not pursuing at this time
  needs_flatten <- !map_lgl(x, is_CellData)
  out[needs_flatten] <- purrr::flatten(out[needs_flatten])
  out
}

#' @export
as_CellData.googlesheets4_formula <- function(x, .na = NULL) {
  cell_data(vec_data(x), val_type = "formulaValue", .na = .na)
}

add_format <- function(x, fmt) {
  x[["userEnteredFormat"]] <- list(numberFormat = list2(!!!fmt))
  x
}

#' @export
as_CellData.Date <- function(x, .na = NULL) {
  # 25569 = DATEVALUE("1970-01-01), i.e. Unix epoch as a serial date, when the
  # date origin is December 30th 1899
  x <- unclass(x) + 25569
  x <- cell_data(x, val_type = "numberValue", .na = .na)
  map(x, add_format, fmt = list(type = "DATE", pattern = "yyyy-mm-dd"))
}

#' @export
as_CellData.POSIXct <- function(x, .na = NULL) {
  # 86400 = 60 * 60 * 24 = number of seconds in a day
  x <- (unclass(x) / 86400) + 25569
  x <- cell_data(x, val_type = "numberValue", .na = .na)
  map(
    x,
    add_format,
    # I decided that going with R's default format was more important than
    # a militant stance re: ISO 8601
    # the space (vs. a 'T') between date and time is "blessed" in RFC 3339
    # https://tools.ietf.org/html/rfc3339#section-5.6
    fmt = list(type = "DATE_TIME", pattern = "yyyy-mm-dd hh:mm:ss")
  )
}

# Currently (overly) focused on userEnteredValue, because I am thinking about
# writing. But with a reading focus, one would want to see effectiveValue.
#' @export
format.googlesheets4_schema_CellData <- function(x, ...) {
  # TODO: convey something about userEnteredFormat?
  user_entered_value <- pluck(x, "userEnteredValue")
  if (is.null(user_entered_value) || is.na(user_entered_value)) {
    return("--no userEnteredValue --")
  }
  nm <- pluck(user_entered_value, names)
  fval <- format(user_entered_value)
  as.character(glue("{nm}: {fval}"))
}

#' @export
print.googlesheets4_schema_CellData <- function(x, ...) {
  header <- as.character(glue("<CellData>"))
  cat(c(header, format(x)), sep = "\n")
  invisible(x)
}

## helpers ---------------------------------------------------------------------
expect_empty_cell <- function(object) {
  expect_equal(object[["userEnteredValue"]], NA)
}

expect_cell_value <- function(object, nm, val) {
  expect_equal(
    object[["userEnteredValue"]],
    list2(!!nm := val)
  )
}

expect_cell_format <- function(object, fmt) {
  expect_equal(object[["userEnteredFormat"]], fmt)
}

## tests -----------------------------------------------------------------------
test_that("expect_empty_cell() is synced with empty_cell()", {
  expect_empty_cell(empty_cell())
})

test_that("as_CellData() treats NULL as empty cell", {
  expect_empty_cell(as_CellData(NULL)[[1]])
})

test_that("as_CellData() works for logical", {
  out <- as_CellData(c(TRUE, NA, FALSE, NA))
  expect_cell_value(out[[1]], "boolValue", TRUE)
  expect_empty_cell(out[[2]])
  expect_cell_value(out[[3]], "boolValue", FALSE)
  expect_empty_cell(out[[4]])
})

test_that("as_CellData() works for character and factor", {
  out <- as_CellData(c("a", NA, "c", NA))
  expect_cell_value(out[[1]], "stringValue", "a")
  expect_empty_cell(out[[2]])
  expect_cell_value(out[[3]], "stringValue", "c")
  expect_empty_cell(out[[4]])

  out <- as_CellData(factor(c("a", NA, "c", NA)))
  expect_cell_value(out[[1]], "stringValue", "a")
  expect_empty_cell(out[[2]])
  expect_cell_value(out[[3]], "stringValue", "c")
  expect_empty_cell(out[[4]])
})

test_that("as_CellData() works for integer or double", {
  out <- as_CellData(c(1L, NA, 3L, NA))
  expect_cell_value(out[[1]], "numberValue", 1L)
  expect_empty_cell(out[[2]])
  expect_cell_value(out[[3]], "numberValue", 3L)
  expect_empty_cell(out[[4]])

  out <- as_CellData(c(1.5, NA, 3.5, NA))
  expect_cell_value(out[[1]], "numberValue", 1.5)
  expect_empty_cell(out[[2]])
  expect_cell_value(out[[3]], "numberValue", 3.5)
  expect_empty_cell(out[[4]])
})

test_that("as_CellData() works for googlesheets4_schema_CellData", {
  out <- as_CellData("a")[[1]]
  expect_identical(out, as_CellData(out))

  out <- as_CellData(list("a", TRUE, 1.5))
  expect_identical(out, as_CellData(out))
})

test_that("as_CellData() works for Date", {
  input <- as.Date(c("2003-06-06", NA, "1982-12-05"))
  naked_input <- unclass(input)
  out <- as_CellData(input)
  # 25569 = DATEVALUE("1970-01-01), i.e. Unix epoch as a serial date, when the
  # date origin is December 30th 1899
  expect_cell_value(out[[1]], "numberValue", naked_input[[1]] + 25569)
  expect_empty_cell(out[[2]])
  expect_cell_value(out[[3]], "numberValue", naked_input[[3]] + 25569)

  fmt <- list(numberFormat = list(type = "DATE", pattern = "yyyy-mm-dd"))
  expect_cell_format(out[[1]], fmt)
  expect_cell_format(out[[2]], fmt)
  expect_cell_format(out[[3]], fmt)
})

test_that("as_CellData() works for POSIXct", {
  input <- as.POSIXct(c("1978-05-31 04:24:32", NA, "2006-07-19 23:27:37"))
  naked_input <- unclass(input)
  attributes(naked_input) <- NULL

  out <- as_CellData(input)
  # 86400 = 60 * 60 * 24 = number of seconds in a day
  # 25569 = DATEVALUE("1970-01-01), i.e. Unix epoch as a serial date, when the
  # date origin is December 30th 1899
  expect_cell_value(
    out[[1]],
    "numberValue",
    (naked_input[[1]] / 86400) + 25569
  )
  expect_empty_cell(out[[2]])
  expect_cell_value(
    out[[3]],
    "numberValue",
    (naked_input[[3]] / 86400) + 25569
  )

  fmt <- list(
    numberFormat = list(
      type = "DATE_TIME",
      pattern = "yyyy-mm-dd hh:mm:ss"
    )
  )
  expect_cell_format(out[[1]], fmt)
  expect_cell_format(out[[2]], fmt)
  expect_cell_format(out[[3]], fmt)
})

test_that("as_CellData() works for list", {
  input <- list(TRUE, NA, "a", 1.5, factor("a"), 4L)
  out <- as_CellData(input)
  expect_cell_value(out[[1]], "boolValue", TRUE)
  expect_empty_cell(out[[2]])
  expect_cell_value(out[[3]], "stringValue", "a")
  expect_cell_value(out[[4]], "numberValue", 1.5)
  expect_cell_value(out[[5]], "stringValue", "a")
  expect_cell_value(out[[6]], "numberValue", 4L)
})

test_that("as_CellData() works for formula", {
  hyperlink <- "=HYPERLINK(\"http://www.google.com/\",\"Google\")"
  image <- "=IMAGE(\"https://www.google.com/images/srpr/logo3w.png\")"
  out <- as_CellData(gs4_formula(c(hyperlink, NA, image)))
  expect_cell_value(out[[1]], "formulaValue", hyperlink)
  expect_empty_cell(out[[2]])
  expect_cell_value(out[[3]], "formulaValue", image)
})

test_that("as_CellData() doesn't add extra nesting to list-cols", {
  expect_identical(
    as_CellData(c("a", "b")),
    as_CellData(list("a", "b"))
  )
})

#' @param n_max Maximum number of data rows to parse into the returned tibble.
#'   Trailing empty rows are automatically skipped, so this is an upper bound on
#'   the number of rows in the result. Ignored if `range` is given. `n_max` is
#'   imposed locally, after reading all non-empty cells, so, if speed is an
#'   issue, it is better to use `range`.

# https://github.com/tidyverse/googlesheets4/issues/73
devtools::load_all() # I assume we're in googlesheets4 source
library(googledrive)
library(tidyverse)

googlesheets4:::gs4_auth_testing()

ss <- test_sheet_create("googlesheets4-col-types")
gs4_browse(ss)

df <- tibble(
  A = list(1, "Missing", 3),
  B = list(1, "NA", 3),
  C = c(1, NA, 3),
  D = 1:3
)
sheet_add(ss, sheet = "NAs")
sheet_write(df, ss, sheet = "NAs")

#' Create a new Sheet
#'
#' @description
#'
#' Creates an entirely new (spread)Sheet (or, in Excel-speak, workbook).
#' Optionally, you can also provide names and/or data for the initial set of
#' (work)sheets. Any initial data provided via `sheets` is styled as a table,
#' as described in [sheet_write()].
#'
#' @seealso
#' Wraps the `spreadsheets.create` endpoint:
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/create>
#'
#' There is an article on writing Sheets:
#'   * <https://googlesheets4.tidyverse.org/articles/articles/write-sheets.html>
#'
#' @param name The name of the new spreadsheet.
#' @param ... Optional spreadsheet properties that can be set through this API
#'   endpoint, such as locale and time zone.
#' @param sheets Optional input for initializing (work)sheets. If unspecified,
#'   the Sheets API automatically creates an empty "Sheet1". You can provide a
#'   vector of sheet names, a data frame, or a (possibly named) list of data
#'   frames. See the examples.
#'
#' @template ss-return
#' @export
#' @family write functions
#'
#' @examplesIf gs4_has_token()
#' gs4_create("gs4-create-demo-1")
#'
#' gs4_create("gs4-create-demo-2", locale = "en_CA")
#'
#' gs4_create(
#'   "gs4-create-demo-3",
#'   locale = "fr_FR",
#'   timeZone = "Europe/Paris"
#' )
#'
#' gs4_create(
#'   "gs4-create-demo-4",
#'   sheets = c("alpha", "beta")
#' )
#'
#' my_data <- data.frame(x = 1)
#' gs4_create(
#'   "gs4-create-demo-5",
#'   sheets = my_data
#' )
#'
#' gs4_create(
#'   "gs4-create-demo-6",
#'   sheets = list(chickwts = head(chickwts), mtcars = head(mtcars))
#' )
#'
#' # Clean up
#' gs4_find("gs4-create-demo") %>%
#'   googledrive::drive_trash()
gs4_create <- function(name = gs4_random(), ..., sheets = NULL) {
  sheets <- enlist_sheets(enquo(sheets))
  sheets_given <- !is.null(sheets)
  data_given <- sheets_given && !is.null(unlist(sheets$value))

  # create the (spread)Sheet ---------------------------------------------------
  gs4_bullets(c(v = "Creating new Sheet: {.s_sheet {name}}."))
  ss_body <- new(
    "Spreadsheet",
    properties = new("SpreadsheetProperties", title = name, ...)
  )
  if (sheets_given) {
    ss_body <- ss_body %>%
      patch(sheets = map(sheets$name, as_Sheet))
  }
  req <- request_generate(
    "sheets.spreadsheets.create",
    params = ss_body
  )
  resp_raw <- request_make(req)
  resp_create <- gargle::response_process(resp_raw)
  ss <- new_googlesheets4_spreadsheet(resp_create)
  ssid <- as_sheets_id(ss)

  if (!data_given) {
    return(invisible(ssid))
  }

  request_populate_sheets <- map2(ss$sheets$id, sheets$value, prepare_df)
  request_populate_sheets <- purrr::flatten(request_populate_sheets)
  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = request_populate_sheets,
      responseIncludeGridData = FALSE
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)

  invisible(ssid)
}

prepare_df <- function(sheet_id, df, skip = 0) {
  # if df is a 0-row data frame, we must send a 1-row data frame of NAs in order
  # to shrink wrap the data and freeze the top row
  # https://github.com/tidyverse/googlesheets4/issues/92
  if (nrow(df) == 0) {
    df <- vec_init(df, n = 1)
  }

  # pack the data --------------------------------------------------------------
  # `start` (or `range`) must be sent, even if `skip = 0`
  start <- new("GridCoordinate", sheetId = sheet_id)
  if (skip > 0) {
    start <- patch(start, rowIndex = skip)
  }
  request_values <- list(
    updateCells = new(
      "UpdateCellsRequest",
      start = start,
      rows = as_RowData(df), # an array of instances of RowData
      fields = "userEnteredValue,userEnteredFormat"
    )
  )

  # set sheet dimensions and freeze top row -------------------------------------
  request_sheet_properties <- bureq_set_grid_properties(
    sheetId = sheet_id,
    nrow = nrow(df) + skip + 1,
    ncol = ncol(df),
    frozenRowCount = skip + 1
  )

  c(
    list(request_sheet_properties),
    list(request_values),
    list(bureq_header_row(sheetId = sheet_id, row = skip + 1))
  )
}

#' Generate a random Sheet name
#'
#' Generates a random name, suitable for a newly created Sheet, using
#' [ids::adjective_animal()].
#'
#' @param n Number of names to generate.
#'
#' @return A character vector.
#' @export
#'
#' @examples
#' gs4_random()
gs4_random <- function(n = 1) {
  ids::adjective_animal(n = n, max_len = 10, style = "kebab")
}

#' Copy a (work)sheet
#'
#' Copies a (work)sheet, within its current (spread)Sheet or to another Sheet.
#'
#' @eval param_ss(pname = "from_ss")
#' @eval param_sheet(
#'   pname = "from_sheet",
#'   action = "copy",
#'   "Defaults to the first visible sheet."
#' )
#' @param to_ss The Sheet to copy *to*. Accepts all the same types of input as
#'   `from_ss`, which is also what this defaults to, if unspecified.
#' @param to_sheet Optional. Name of the new sheet, as a string. If you don't
#'   specify this, Google generates a name, along the lines of "Copy of blah".
#'   Note that sheet names must be unique within a Sheet, so if the automatic
#'   name would violate this, Google also de-duplicates it for you, meaning you
#'   could conceivably end up with "Copy of blah 2". If you have better ideas
#'   about sheet names, specify `to_sheet`.
#' @eval param_before_after("sheet")
#'
#' @return The receiving Sheet, `to_ ss`, as an instance of [`sheets_id`].
#' @export
#' @family worksheet functions
#' @seealso
#' If the copy happens within one Sheet, makes a `DuplicateSheetRequest`:
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#duplicatesheetrequest>
#'
#' If the copy is from one Sheet to another, wraps the
#' `spreadsheets.sheets/copyTo` endpoint:
#' * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets.sheets/copyTo>
#'
#' and possibly makes a subsequent `UpdateSheetPropertiesRequest`:
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#UpdateSheetPropertiesRequest>
#'
#' @examplesIf gs4_has_token()
#' ss_aaa <- gs4_create(
#'   "sheet-copy-demo-aaa",
#'   sheets = list(mtcars = head(mtcars), chickwts = head(chickwts))
#' )
#'
#' # copy 'mtcars' sheet within existing Sheet, accept autogenerated name
#' ss_aaa %>%
#'   sheet_copy()
#'
#' # copy 'mtcars' sheet within existing Sheet
#' # specify new sheet's name and location
#' ss_aaa %>%
#'   sheet_copy(to_sheet = "mtcars-the-sequel", .after = 1)
#'
#' # make a second Sheet
#' ss_bbb <- gs4_create("sheet-copy-demo-bbb")
#'
#' # copy 'chickwts' sheet from first Sheet to second
#' # accept auto-generated name and default location
#' ss_aaa %>%
#'   sheet_copy("chickwts", to_ss = ss_bbb)
#'
#' # copy 'chickwts' sheet from first Sheet to second,
#' # WITH a specific name and into a specific location
#' ss_aaa %>%
#'   sheet_copy(
#'     "chickwts",
#'     to_ss = ss_bbb, to_sheet = "chicks-two", .before = 1
#'   )
#'
#' # clean up
#' gs4_find("sheet-copy-demo") %>%
#'   googledrive::drive_trash()
sheet_copy <- function(
  from_ss,
  from_sheet = NULL,
  to_ss = from_ss,
  to_sheet = NULL,
  .before = NULL,
  .after = NULL
) {
  from_ssid <- as_sheets_id(from_ss)
  to_ssid <- as_sheets_id(to_ss)
  maybe_sheet(from_sheet)

  if (identical(from_ssid, to_ssid)) {
    sheet_copy_internal(
      ssid = from_ssid,
      from_sheet = from_sheet,
      to_sheet = to_sheet,
      .before = .before,
      .after = .after
    )
  } else {
    sheet_copy_external(
      from_ssid = from_ssid,
      from_sheet = from_sheet,
      to_ssid = to_ssid,
      to_sheet = to_sheet,
      .before = .before,
      .after = .after
    )
  }
}

sheet_copy_internal <- function(
  ssid,
  from_sheet = NULL,
  to_sheet = NULL,
  .before = NULL,
  .after = NULL,
  call = caller_env()
) {
  maybe_string(to_sheet, call = call)
  x <- gs4_get(ssid)
  s <- lookup_sheet(from_sheet, sheets_df = x$sheets, call = call)
  gs4_bullets(c(
    v = "Duplicating sheet {.w_sheet {s$name}} in {.s_sheet {x$name}}."
  ))

  index <- resolve_index(x$sheets, .before, .after, call = call)
  dup_request <- new(
    "DuplicateSheetRequest",
    sourceSheetId = s$id,
    insertSheetIndex = index,
    newSheetName = to_sheet
  )

  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = list(duplicateSheet = dup_request)
    )
  )
  resp_raw <- request_make(req)
  resp <- gargle::response_process(resp_raw)
  to_name <- pluck(resp, "replies", 1, "duplicateSheet", "properties", "title")
  gs4_bullets(c(v = "Copied as {.w_sheet {to_name}}."))

  invisible(ssid)
}

sheet_copy_external <- function(
  from_ssid,
  from_sheet = NULL,
  to_ssid,
  to_sheet = NULL,
  .before = NULL,
  .after = NULL,
  call = caller_env()
) {
  from_x <- gs4_get(from_ssid)
  to_x <- gs4_get(to_ssid)
  maybe_string(to_sheet, "sheet_copy", call = call)

  from_s <- lookup_sheet(from_sheet, sheets_df = from_x$sheets, call = call)
  gs4_bullets(c(
    v = "Copying sheet {.w_sheet {from_s$name}} from \\
         {.s_sheet {from_x$name}} to {.s_sheet {to_x$name}}."
  ))

  req <- request_generate(
    "sheets.spreadsheets.sheets.copyTo",
    params = list(
      spreadsheetId = from_ssid,
      sheetId = from_s$id,
      destinationSpreadsheetId = as.character(to_ssid)
    )
  )
  resp_raw <- request_make(req)
  to_s <- gargle::response_process(resp_raw)

  # early exit if no need to relocate and/or rename copied sheet
  index <- resolve_index(to_x$sheets, .before, .after, call = call)
  if (is.null(index) && is.null(to_sheet)) {
    gs4_bullets(c(v = "Copied as {.w_sheet {to_s$title}}."))
    return(invisible(to_ssid))
  }

  sp <- new(
    "SheetProperties",
    sheetId = to_s$sheetId,
    title = to_sheet,
    index = index
  )

  update_req <- new(
    "UpdateSheetPropertiesRequest",
    properties = sp,
    fields = gargle::field_mask(sp)
  )

  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = to_ssid,
      requests = list(updateSheetProperties = update_req)
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)
  gs4_bullets(c(v = "Copied as {.w_sheet {to_sheet}}."))

  invisible(to_ssid)
}

# ---- nm_fun ----
me_ <- nm_fun("TEST-sheet_copy")

# ---- tests ----
test_that("internal copy works", {
  skip_if_offline()
  skip_if_no_token()

  ss <- local_ss(
    me_("internal"),
    sheets = list(iris = head(iris), chickwts = head(chickwts))
  )
  sheet_copy(ss, to_sheet = "xyz", .after = 1)
  out <- sheet_names(ss)
  expect_equal(out, c("iris", "xyz", "chickwts"))
})

test_that("external copy works", {
  skip_if_offline()
  skip_if_no_token()

  ss_source <- local_ss(
    me_("source"),
    sheets = list(iris = head(iris), chickwts = head(chickwts))
  )
  ss_dest <- local_ss(me_("dest"))

  sheet_copy(
    ss_source,
    from_sheet = "chickwts",
    to_ss = ss_dest,
    to_sheet = "chicks-two",
    .before = 1
  )
  out <- sheet_names(ss_dest)
  expect_equal(out, c("chicks-two", "Sheet1"))
})

#' Read cells from a Sheet
#'
#' This low-level function returns cell data in a tibble with one row per cell.
#' This tibble has integer variables `row` and `col` (referring to location
#' with the Google Sheet), an A1-style reference `loc`, and a `cell`
#' list-column. The flagship function [read_sheet()], a.k.a. [range_read()], is
#' what most users are looking for, rather than `range_read_cells()`.
#' [read_sheet()] is basically `range_read_cells()` (this function), followed by
#' [spread_sheet()], which looks after reshaping and column typing. But if you
#' really want raw cell data from the API, `range_read_cells()` is for you!
#'
#' @eval param_ss()
#' @eval param_sheet(
#'   action = "read",
#'   "Ignored if the sheet is specified via `range`. If neither argument",
#'   "specifies the sheet, defaults to the first visible sheet."
#' )
#' @template range
#' @template skip-read
#' @template n_max
#' @param cell_data How much detail to get for each cell. `"default"` retrieves
#'   the fields actually used when googlesheets4 guesses or imposes cell and
#'   column types. `"full"` retrieves all fields in the [`CellData`
#'   schema](https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#CellData).
#'   The main differences relate to cell formatting.
#' @param discard_empty Whether to discard cells that have no data. Literally,
#'   we check for an `effectiveValue`, which is one of the fields in the
#'   [`CellData`
#'   schema](https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#CellData).
#'
#' @seealso Wraps the `spreadsheets.get` endpoint:
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/get>
#'
#' @return A tibble with one row per cell in the `range`.
#' @export
#'
#' @examplesIf gs4_has_token()
#' range_read_cells(gs4_example("deaths"), range = "arts_data")
#'
#' # if you want detailed and exhaustive cell data, do this
#' range_read_cells(
#'   gs4_example("formulas-and-formats"),
#'   cell_data = "full",
#'   discard_empty = FALSE
#' )
range_read_cells <- function(
  ss,
  sheet = NULL,
  range = NULL,
  skip = 0,
  n_max = Inf,
  cell_data = c("default", "full"),
  discard_empty = TRUE
) {
  cell_data <- match.arg(cell_data)

  # range spec params are checked inside get_cells():
  # ss, sheet, range, skip, n_max
  out <- get_cells(
    ss = ss,
    sheet = sheet,
    range = range,
    skip = skip,
    n_max = n_max,
    col_names_in_sheet = FALSE,
    detail_level = cell_data,
    discard_empty = discard_empty
  )
  out$cell <- apply_ctype(out$cell)
  add_loc(out)
}

# I use this elsewhere during development, so handy to have in a function
add_loc <- function(df) {
  tibble::add_column(
    df,
    loc = as.character(glue("{cellranger::num_to_letter(df$col)}{df$row}")),
    .before = "cell"
  )
}

# ---- nm_fun ----
me_ <- nm_fun("TEST-range_flood")

# ---- tests ----
test_that("range_flood() works", {
  skip_if_offline()
  skip_if_no_token()

  dat <- tibble::tibble(x = rep(1, 3), y = rep(2, 3), z = rep(3, 3))
  ss <- local_ss(me_(), sheets = list(dat))

  # clear values and format
  range_flood(ss, range = "A:A")

  # reset values and reformat
  range_flood(ss, range = "B:B", cell = "hi")

  # reset values, leave format unchanged
  range_flood(ss, range = "C:C", cell = "bye", reformat = FALSE)

  out <- range_read_cells(ss, cell_data = "full", discard_empty = FALSE)

  expect_equal(
    purrr::map_chr(out$cell, "formattedValue", .default = ""),
    rep(c("", "hi", "bye"), 4)
  )

  column_A <- out[out$col == 1, ]
  fmts <- purrr::map(column_A$cell, "effectiveFormat")
  expect_true(all(purrr::map_lgl(fmts, is.null)))

  column_B <- out[out$col == 2, ]
  fmts <- purrr::map(column_B$cell, c("effectiveFormat", "backgroundColor"))
  expect_true(all(unlist(fmts) == 1))

  column_C_header <- out[out$col == 3 & out$row == 1, ]
  fmt <- purrr::pluck(
    column_C_header,
    "cell",
    1,
    "effectiveFormat",
    "backgroundColor"
  )
  expect_true(all(unlist(fmt) < 1))
})

test_that("cells() returns `row` and `col` as integer", {
  skip_if_offline()
  skip_if_no_token()

  out <- range_read_cells(
    test_sheet("googlesheets4-cell-tests"),
    range = "'range-experimentation'!A1:B2"
  )
  expect_true(is.integer(out$row))
  expect_true(is.integer(out$col))
})

test_that("slightly tricky `range`s work", {
  skip_if_offline()
  skip_if_no_token()

  out <- range_read_cells(
    test_sheet("googlesheets4-cell-tests"),
    range = "'range-experimentation'!B:D"
  )
  expect_true(all(grepl("^[BCD]", out$loc)))

  out <- range_read_cells(
    test_sheet("googlesheets4-cell-tests"),
    range = "'range-experimentation'!2:3"
  )
  expect_true(all(grepl("[23]$", out$loc)))

  out <- range_read_cells(
    test_sheet("googlesheets4-cell-tests"),
    range = "'range-experimentation'!B3:C"
  )
  expect_true(all(grepl("^[BC]", out$loc)))
  expect_true(all(grepl("[3-9]$", out$loc)))

  out <- range_read_cells(
    test_sheet("googlesheets4-cell-tests"),
    range = "'range-experimentation'!B3:5"
  )
  expect_true(all(grepl("^[BCDE]", out$loc)))
  expect_true(all(grepl("[345]$", out$loc)))
})

# https://github.com/tidyverse/googlesheets4/issues/4
test_that("full cell data and empties are within reach", {
  skip_if_offline()
  skip_if_no_token()

  out <- range_read_cells(
    test_sheet("googlesheets4-cell-tests"),
    sheet = "empties-and-formats",
    cell_data = "full",
    discard_empty = FALSE
  )

  # B2 is empty; make sure it's here
  expect_true("B2" %in% out$loc)

  # C2 is empty and orange; make sure it's here and format is available
  expect_no_error(
    cell <- out$cell[[which(out$loc == "C2")]]
  )
  expect_true(!is.null(cell$effectiveFormat))

  # C1 bears a note
  expect_no_error(
    cell <- out$cell[[which(out$loc == "C1")]]
  )
  note <- cell$note
  expect_true(!is.null(note))
  expect_match(note, "Note")
})

# https://github.com/tidyverse/googlesheets4/issues/78
test_that("formula cells are parsed based on effectiveValue", {
  skip_if_offline()
  skip_if_no_token()

  out <- range_read_cells(
    test_sheet("googlesheets4-cell-tests"),
    sheet = "formulas",
    range = "B:B",
    cell_data = "full",
    discard_empty = FALSE
  )

  expect_s3_class(out$cell[[which(out$loc == "B2")]], "CELL_TEXT")
  expect_s3_class(out$cell[[which(out$loc == "B3")]], "CELL_NUMERIC")
  expect_s3_class(out$cell[[which(out$loc == "B4")]], "CELL_BLANK")
  expect_s3_class(out$cell[[which(out$loc == "B5")]], "CELL_TEXT")
  expect_s3_class(out$cell[[which(out$loc == "B6")]], "CELL_BLANK")
})

#' Add a data validation rule to a cell range
#'
#' @description
#' *Note: not yet exported, still very alpha. Usage still requires using
#' low-level helpers.*
#'
#' `range_add_validation()` adds a data validation rule to a range of cells.
#'
#' @eval param_ss()
#' @eval param_sheet()
#' @param range Cells to apply data validation to. This `range` argument has
#'   important similarities and differences to `range` elsewhere (e.g.
#'   [range_read()]):
#'   * Similarities: Can be a cell range, using A1 notation ("A1:D3") or using
#'     the helpers in [`cell-specification`]. Can combine sheet name and cell
#'     range ("Sheet1!A5:A") or refer to a sheet by name (`range = "Sheet1"`,
#'     although `sheet = "Sheet1"` is preferred for clarity).
#'   * Difference: Can NOT be a named range.
#' @param rule An instance of `googlesheets4_schema_DataValidationRule`, which
#'   implements the
#'   [DataValidationRule](https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#datavalidationrule)
#'   schema.
#'
#' @template ss-return
#' @seealso Makes a `SetDataValidationRequest`:
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#setdatavalidationrequest>
#'
#' @keywords internal
#' @noRd
#'
#' @examplesIf gs4_has_token()
#' # create a data frame to use as initial data
#' df <- data.frame(
#'   id = 1:3,
#'   "Hungry?" = NA,
#'   ice_cream = NA,
#'   check.names = FALSE
#' )
#'
#' # create Sheet
#' ss <- gs4_create("range-add-validation-demo", sheets = list(df))
#'
#' # create a column that presents as a basic TRUE/FALSE checkbox
#' rule_checkbox <- googlesheets4:::new(
#'   "DataValidationRule",
#'   condition = googlesheets4:::new_BooleanCondition(type = "BOOLEAN"),
#'   inputMessage = "Please let us know if you are hungry.",
#'   strict = TRUE,
#'   showCustomUi = TRUE
#' )
#' googlesheets4:::range_add_validation(
#'   ss,
#'   range = "Sheet1!B2:B", rule = rule_checkbox
#' )
#'
#' # create a column that presents as a dropdown list
#' rule_dropdown_list <- googlesheets4:::new(
#'   "DataValidationRule",
#'   condition = googlesheets4:::new_BooleanCondition(
#'     type = "ONE_OF_LIST", values = c("vanilla", "chocolate", "strawberry")
#'   ),
#'   inputMessage = "Which ice cream flavor do you want?",
#'   strict = TRUE,
#'   showCustomUi = TRUE
#' )
#' googlesheets4:::range_add_validation(
#'   ss,
#'   range = "Sheet1!C2:C", rule = rule_dropdown_list
#' )
#'
#' read_sheet(ss)
#'
#' # clean up
#' gs4_find("range-add-validation-demo") %>%
#'   googledrive::drive_trash()
range_add_validation <- function(ss, sheet = NULL, range = NULL, rule) {
  ssid <- as_sheets_id(ss)
  maybe_sheet(sheet)
  check_range(range)
  if (!is.null(rule)) {
    stopifnot(inherits(rule, "googlesheets4_schema_DataValidationRule"))
  }

  x <- gs4_get(ssid)
  gs4_bullets(c(v = "Editing {.s_sheet {x$name}}."))

  # determine (work)sheet ------------------------------------------------------
  range_spec <- as_range_spec(
    range,
    sheet = sheet,
    sheets_df = x$sheets,
    nr_df = x$named_ranges
  )
  range_spec$sheet_name <- range_spec$sheet_name %||%
    first_visible_name(x$sheets)
  s <- lookup_sheet(range_spec$sheet_name, sheets_df = x$sheets)
  gs4_bullets(c(v = "Editing sheet {.w_sheet {range_spec$sheet_name}}."))

  # form batch update request --------------------------------------------------
  sdv_req <- list(
    setDataValidation = new(
      "SetDataValidationRequest",
      range = as_GridRange(range_spec),
      rule = rule
    )
  )

  # do it ----------------------------------------------------------------------
  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = list(sdv_req)
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)

  invisible(ssid)
}

# helpers ----
new_BooleanCondition <- function(type = "NOT_BLANK", values = NULL) {
  out <- new("BooleanCondition", type = type)

  # TODO: build enum checking into our schema-based construction
  schema <- attr(out, "schema")
  enum <- schema$enum[[which(schema$property == "type")]]
  stopifnot(type %in% enum$enum)

  if (length(values) < 1) {
    return(out)
  }

  needs_relative_date <- c(
    "DATE_BEFORE",
    "DATE_AFTER",
    "DATE_ON_OR_BEFORE",
    "DATE_ON_OR_AFTER"
  )
  if (type %in% needs_relative_date) {
    gs4_abort(
      "{.field relativeDate} not yet supported as a {.code conditionValue}.",
      .internal = TRUE
    )
  }
  patch(out, values = map(values, ~ list(userEnteredValue = as.character(.x))))
}

## this is the "cell getter" for range_read_cells() and read_sheet()
get_cells <- function(
  ss,
  sheet = NULL,
  range = NULL,
  col_names_in_sheet = TRUE,
  skip = 0,
  n_max = Inf,
  detail_level = c("default", "full"),
  discard_empty = TRUE,
  call = caller_env()
) {
  ssid <- as_sheets_id(ss)

  maybe_sheet(sheet, call = call)
  check_range(range, call = call)
  check_bool(col_names_in_sheet, call = call)
  check_non_negative_integer(skip, call = call)
  check_non_negative_integer(n_max, call = call)
  detail_level <- match.arg(detail_level)
  check_bool(discard_empty, call = call)

  ## retrieve spreadsheet metadata --------------------------------------------
  x <- gs4_get(ssid)
  gs4_bullets(c(v = "Reading from {.s_sheet {x$name}}."))

  ## prepare range specification for API --------------------------------------

  ## user's range, sheet, skip --> qualified A1 range, suitable for API
  range_spec <- as_range_spec(
    range,
    sheet = sheet,
    skip = skip,
    sheets_df = x$sheets,
    nr_df = x$named_ranges
  )
  # if we send no range, we get all cells from all sheets; not what we want
  effective_range <- as_A1_range(range_spec) %||% first_visible_name(x$sheets)
  gs4_bullets(c(v = "Range {.range {effective_range}}."))

  ## main GET -----------------------------------------------------------------
  resp <- read_cells_impl_(
    ssid,
    ranges = effective_range,
    detail_level = detail_level
  )
  out <- cells(resp)

  if (discard_empty) {
    # cells can be present, just because they bear a format (much like Excel)
    cell_is_empty <- map_lgl(out$cell, ~ is.null(pluck(.x, "effectiveValue")))
    out <- out[!cell_is_empty, ]
  }

  ## enforce geometry on the cell data frame ----------------------------------
  if (range_spec$shim) {
    range_spec$cell_limits <- range_spec$cell_limits %||%
      as_cell_limits(effective_range)
    out <- insert_shims(out, range_spec$cell_limits)
    ## guarantee:
    ## every row and every column spanned by user's range is represented by at
    ## least one cell, (could be placeholders w/ no content from API, though)
    ##
    ## NOTE:
    ## this does NOT imply that every spreadsheet cell spanned by user's range
    ## is represented by a cell in 'out' --> rectangling must be robust to holes
  } else if (n_max < Inf) {
    out <- enforce_n_max(out, n_max, col_names_in_sheet)
  }
  out
}

# https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/get
# I want a separate worker so there is a version of this available that
# accepts `fields` (or `includeGridData`), yet I don't want a user-facing
# function that exposes those details
read_cells_impl_ <- function(
  ssid,
  ranges,
  fields = NULL,
  detail_level = c("default", "full")
) {
  # there are 2 ways to control the level of detail re: cell data:
  #   1. Supply a field mask. What we currently do.
  #   2. Set `includeGridData` to true. This gets *everything* about the
  #      Spreadsheet and the Sheet(s). So far, this seems like TMI.
  detail_level <- match.arg(detail_level)
  cell_mask <- switch(
    detail_level,
    "default" = ".values(effectiveValue,formattedValue,effectiveFormat.numberFormat)",
    "full" = ""
  )
  default_fields <- c(
    "spreadsheetId",
    "properties.title",
    "sheets.properties(sheetId,title)",
    glue("sheets.data(startRow,startColumn,rowData{cell_mask})")
  )
  fields <- fields %||% glue_collapse(default_fields, sep = ",")

  req <- request_generate(
    "sheets.spreadsheets.get",
    params = list(
      spreadsheetId = ssid,
      ranges = ranges,
      fields = fields
    )
  )
  raw_resp <- request_make(req)
  gargle::response_process(raw_resp)
}

## input: an instance of Spreadsheet
## https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets#Spreadsheet
## output: a tibble with one row per non-empty cell (row, column, cell)
cells <- function(x = list()) {
  ## identify upper left cell of the rectangle
  ## values are absent in the response if equal to 0, hence the default
  ## return values are zero-based, hence we add 1
  start_row <- (pluck(x, "sheets", 1, "data", 1, "startRow") %||% 0) + 1
  start_column <- (pluck(x, "sheets", 1, "data", 1, "startColumn") %||% 0) + 1

  # TODO: make this an as_tibble method?
  # TODO: deal with the merged cells
  # TODO: ensure this returns integer columns where appropriate

  row_data <- x %>%
    pluck("sheets", 1, "data", 1, "rowData") %>%
    map("values")

  ## an empty row can be present as an explicit NULL
  ## within a non-empty row, an empty cell can be present as list()
  ## rows are ragged and appear to end at the last non-empty cell
  row_lengths <- map_int(row_data, length)
  n_rows <- length(row_data)

  tibble::tibble(
    row = rep.int(
      seq.int(from = start_row, length.out = n_rows),
      times = row_lengths
    ),
    col = as.integer(start_column + sequence(row_lengths) - 1),
    cell = purrr::flatten(row_data)
  )
}

insert_shims <- function(df, cell_limits) {
  ## emulating behaviour of readxl
  if (nrow(df) == 0) {
    return(df)
  }
  df$row <- as.integer(df$row)
  df$col <- as.integer(df$col)

  ## 1-based indices, referring to cell coordinates in the spreadsheet
  start_row <- cell_limits$ul[[1]]
  end_row <- cell_limits$lr[[1]]
  start_col <- cell_limits$ul[[2]]
  end_col <- cell_limits$lr[[2]]

  shim_up <- notNA(start_row) && start_row < min(df$row)
  shim_left <- notNA(start_col) && start_col < min(df$col)
  shim_down <- notNA(end_row) && end_row > max(df$row)
  shim_right <- notNA(end_col) && end_col > max(df$col)

  ## add placeholder to establish upper left corner
  if (shim_up || shim_left) {
    df <- tibble::add_row(
      df,
      row = start_row %|% min(df$row),
      col = start_col %|% min(df$col),
      cell = list(list()),
      .before = 1
    )
  }

  ## add placeholder to establish lower right corner
  if (shim_down || shim_right) {
    df <- tibble::add_row(
      df,
      row = end_row %|% max(df$row),
      col = end_col %|% max(df$col),
      cell = list(list())
    )
  }

  df
}

enforce_n_max <- function(out, n_max, col_names_in_sheet) {
  row_max <- realize_n_max(n_max, out$row, col_names_in_sheet)
  out[out$row <= row_max, ]
}

realize_n_max <- function(n_max, rows, col_names_in_sheet) {
  start_row <- min(rows)
  end_row <- max(rows)
  n_read <- end_row - start_row + 1
  to_read <- n_max + col_names_in_sheet
  if (n_read <= to_read) {
    Inf
  } else {
    start_row + to_read - 1
  }
}

# x = empty cell = not sent back by API
# O = occupied cell = present in API payload
#    A  B  C  D  E
# 1  x  x  x  x  x
# 2  x  O  O  O  x
# 3  x  O  O  x  x  <-- yes, it is intentional that D3 is empty
# 4  x  x  x  x  x

cell_df <- tibble::tribble(
  ~ row, ~ col, ~ cell,
      2,     2,   "B2",
      2,     3,   "C2",
      2,     4,   "D2",
      3,     2,   "B3",
      3,     3,   "C3"
)
# row and col should really be integer
cell_df$row <- as.integer(cell_df$row)
cell_df$col <- as.integer(cell_df$col)
# cell has to be a list-column for the tibble::add_row() in insert_shims()
# to work, with increased type-strictness coming to tibble v3.0
# TODO: maybe use an even more realistic cell-type of object here, when the
# helpers are better
cell_df$cell <- as.list(cell_df$cell)

limitize <- function(df) {
  cell_limits(c(min(df$row), min(df$col)), c(max(df$row), max(df$col)))
}

expect_shim <- function(rg) {
  expect_identical(
    limitize(insert_shims(cell_df, as_cell_limits(rg))),
    as_cell_limits(rg)
  )
}

test_that("observed data occupies range rectangle --> no shim needed", {
  expect_identical(
    insert_shims(cell_df, cell_limits = as_cell_limits("B2:D3")),
    cell_df
  )
})

test_that("can shim a single side", {
  ## up
  expect_shim("B1:D3")
  ## down
  expect_shim("B2:D4")
  ## left
  expect_shim("A2:D3")
  ## right
  expect_shim("B2:E3")
})

test_that("can shim two opposing sides", {
  ## row direction
  expect_shim("B1:D4")
  ## col direction
  expect_shim("A2:E3")
})

test_that("can shim on two perpendicular sides", {
  ## up and left
  expect_shim("A1:D3")
  ## up and right
  expect_shim("B1:E3")
  # down and left
  expect_shim("A2:D4")
  # down and right
  expect_shim("B2:E4")
})

test_that("can shim three sides", {
  ## all but bottom
  expect_shim("A1:E3")
  ## all but left
  expect_shim("B1:E4")
  ## all but top
  expect_shim("A2:E4")
  ## all but right
  expect_shim("A1:D4")
})

test_that("can shim four sides", {
  expect_shim("A1:E4")
})

library(tidyverse)
devtools::load_all() # I assume we're in googlesheets4 source
library(googledrive)

gs4_auth_testing()

ss <- test_sheet_create()
gs4_browse(ss)

# TODO: I created this worksheet in the browser, by copying from
# gs4_example("formulas-and-formats")
# add code here once possible
# most challenging cell is B4, the one that contains
# =IMAGE("https://www.google.com/images/srpr/logo3w.png")
# i.e. a formula that evaluates to an image

as_GridCoordinate <- function(x, ...) {
  UseMethod("as_GridCoordinate")
}

#' @export
as_GridCoordinate.default <- function(x, ...) {
  abort_unsupported_conversion(x, to = "GridCoordinate")
}

#' @export
as_GridCoordinate.range_spec <- function(x, ..., strict = TRUE) {
  grid_range <- as_GridRange(x)

  if (identical(names(grid_range), "sheetId")) {
    return(new("GridCoordinate", sheetId = grid_range$sheetId))
  }

  if (strict) {
    row_index_diff <- grid_range$endRowIndex - grid_range$startRowIndex
    col_index_diff <- grid_range$endColumnIndex - grid_range$startColumnIndex
    if (row_index_diff != 1 || col_index_diff != 1) {
      gs4_abort(c(
        "Range must identify exactly 1 cell:",
        x = "Invalid cell range: {.range {x$cell_range}}"
      ))
    }
  }

  grid_range <- grid_range %>%
    discard(is.null) %>%
    discard(is.na)
  new(
    "GridCoordinate",
    sheetId = grid_range$sheetId,
    rowIndex = grid_range$startRowIndex,
    columnIndex = grid_range$startColumnIndex
  )
}

test_that("we can make a GridCoordinate from a range_spec, simplest case", {
  sheets_df <- tibble::tibble(name = "abc", id = 123)

  spec <- new_range_spec(sheet_name = "abc", sheets_df = sheets_df)
  out <- as_GridCoordinate(spec)
  expect_equal(out$sheetId, 123)
  expect_length(out, 1)

  spec <- new_range_spec(
    sheet_name = "abc",
    cell_range = "G3",
    sheets_df = sheets_df
  )
  out <- as_GridCoordinate(spec)
  expect_equal(out$rowIndex, 2)
  expect_equal(out$columnIndex, 6)
})

test_that("we can (or won't) make a GridCoordinate from a mutli-cell range", {
  sheets_df <- tibble::tibble(name = "abc", id = 123)

  spec <- new_range_spec(
    sheet_name = "abc",
    cell_range = "A3:B4",
    sheets_df = sheets_df
  )
  expect_error(as_GridCoordinate(spec), "Invalid cell range")

  spec2 <- new_range_spec(
    sheet_name = "abc",
    cell_range = "A3",
    sheets_df = sheets_df
  )
  expect_equal(
    as_GridCoordinate(spec, strict = FALSE),
    as_GridCoordinate(spec2)
  )

  spec <- new_range_spec(
    sheet_name = "abc",
    cell_range = "A:B",
    sheets_df = sheets_df
  )
  out <- as_GridCoordinate(spec, strict = FALSE)
  expect_null(out$rowIndex)
  expect_equal(out$columnIndex, 0)

  spec <- new_range_spec(
    sheet_name = "abc",
    cell_range = "2:4",
    sheets_df = sheets_df
  )
  out <- as_GridCoordinate(spec, strict = FALSE)
  expect_equal(out$rowIndex, 1)
  expect_null(out$columnIndex)
})

# This file is the interface between googlesheets4 and the
# auth functionality in gargle.

# Initialization happens in .onLoad
.auth <- NULL

## The roxygen comments for these functions are mostly generated from data
## in this list and template text maintained in gargle.
gargle_lookup_table <- list(
  PACKAGE = "googlesheets4",
  YOUR_STUFF = "your Google Sheets",
  PRODUCT = "Google Sheets",
  API = "Sheets API",
  PREFIX = "gs4"
)

#' Authorize googlesheets4
#'
#' @eval gargle:::PREFIX_auth_description(gargle_lookup_table)
#' @eval gargle:::PREFIX_auth_details(gargle_lookup_table)
#' @eval gargle:::PREFIX_auth_params()
#'
#' @param scopes One or more API scopes. Each scope can be specified in full or,
#'   for Sheets API-specific scopes, in an abbreviated form that is recognized by
#'   [gs4_scopes()]:
#'   * "spreadsheets" = "https://www.googleapis.com/auth/spreadsheets"
#'     (the default)
#'   * "spreadsheets.readonly" =
#'     "https://www.googleapis.com/auth/spreadsheets.readonly"
#'   * "drive" = "https://www.googleapis.com/auth/drive"
#'   * "drive.readonly" = "https://www.googleapis.com/auth/drive.readonly"
#'   * "drive.file" = "https://www.googleapis.com/auth/drive.file"
#'
#'   See
#'   <https://developers.google.com/identity/protocols/oauth2/scopes#sheets> for
#'   details on the permissions for each scope.
#'
#' @family auth functions
#' @export
#'
#' @examplesIf rlang::is_interactive()
#' # load/refresh existing credentials, if available
#' # otherwise, go to browser for authentication and authorization
#' gs4_auth()
#'
#' # indicate the specific identity you want to auth as
#' gs4_auth(email = "jenny@example.com")
#'
#' # force a new browser dance, i.e. don't even try to use existing user
#' # credentials
#' gs4_auth(email = NA)
#'
#' # use a 'read only' scope, so it's impossible to edit or delete Sheets
#' gs4_auth(scopes = "spreadsheets.readonly")
#'
#' # use a service account token
#' gs4_auth(path = "foofy-83ee9e7c9c48.json")
gs4_auth <- function(
  email = gargle::gargle_oauth_email(),
  path = NULL,
  subject = NULL,
  scopes = "spreadsheets",
  cache = gargle::gargle_oauth_cache(),
  use_oob = gargle::gargle_oob_default(),
  token = NULL
) {
  gargle::check_is_service_account(path, hint = "gs4_auth_configure")
  scopes <- gs4_scopes(scopes)

  # I have called `gs4_auth(token = drive_token())` multiple times now,
  # without attaching googledrive. Expose this error noisily, before it gets
  # muffled by the `tryCatch()` treatment of `token_fetch()`.
  force(token)

  cred <- gargle::token_fetch(
    scopes = scopes,
    client = gs4_oauth_client() %||% gargle::tidyverse_client(),
    email = email,
    path = path,
    subject = subject,
    package = "googlesheets4",
    cache = cache,
    use_oob = use_oob,
    token = token
  )
  if (!inherits(cred, "Token2.0")) {
    gs4_abort(c(
      "Can't get Google credentials.",
      "i" = "Are you running {.pkg googlesheets4} in a non-interactive \\
             session? Consider:",
      "*" = "Call {.fun gs4_deauth} to prevent the attempt to get credentials.",
      "*" = "Call {.fun gs4_auth} directly with all necessary specifics.",
      "i" = "See gargle's \"Non-interactive auth\" vignette for more details:",
      "i" = "{.url https://gargle.r-lib.org/articles/non-interactive-auth.html}"
    ))
  }
  .auth$set_cred(cred)
  .auth$set_auth_active(TRUE)

  invisible()
}

#' Suspend authorization
#'
#' @eval gargle:::PREFIX_deauth_description_with_api_key(gargle_lookup_table)
#'
#' @family auth functions
#' @export
#' @examplesIf rlang::is_interactive()
#' gs4_deauth()
#' gs4_user()
#'
#' # get metadata on the public 'deaths' spreadsheet
#' gs4_example("deaths") %>%
#'   gs4_get()
gs4_deauth <- function() {
  .auth$set_auth_active(FALSE)
  .auth$clear_cred()
  invisible()
}

#' Produce configured token
#'
#' @eval gargle:::PREFIX_token_description(gargle_lookup_table)
#' @eval gargle:::PREFIX_token_return()
#'
#' @family low-level API functions
#' @export
#' @examplesIf gs4_has_token()
#' req <- request_generate(
#'   "sheets.spreadsheets.get",
#'   list(spreadsheetId = "abc"),
#'   token = gs4_token()
#' )
#' req
gs4_token <- function() {
  if (isFALSE(.auth$auth_active)) {
    return(NULL)
  }
  if (!gs4_has_token()) {
    gs4_auth()
  }
  httr::config(token = .auth$cred)
}

#' Is there a token on hand?
#'
#' @eval gargle:::PREFIX_has_token_description(gargle_lookup_table)
#' @eval gargle:::PREFIX_has_token_return()
#'
#' @family low-level API functions
#' @export
#'
#' @examples
#' gs4_has_token()
gs4_has_token <- function() {
  inherits(.auth$cred, "Token2.0")
}

#' Edit and view auth configuration
#'
#' @eval gargle:::PREFIX_auth_configure_description(gargle_lookup_table)
#' @eval gargle:::PREFIX_auth_configure_params()
#' @eval gargle:::PREFIX_auth_configure_return(gargle_lookup_table)
#'
#' @family auth functions
#' @export
#' @examples
#' # see and store the current user-configured OAuth client (probably `NULL`)
#' (original_client <- gs4_oauth_client())
#'
#' # see and store the current user-configured API key (probably `NULL`)
#' (original_api_key <- gs4_api_key())
#'
#' # the preferred way to configure your own client is via a JSON file
#' # downloaded from Google Developers Console
#' # this example JSON is indicative, but fake
#' path_to_json <- system.file(
#'   "extdata", "client_secret_installed.googleusercontent.com.json",
#'   package = "gargle"
#' )
#' gs4_auth_configure(path = path_to_json)
#'
#' # this is also obviously a fake API key
#' gs4_auth_configure(api_key = "the_key_I_got_for_a_google_API")
#'
#' # confirm the changes
#' gs4_oauth_client()
#' gs4_api_key()
#'
#' # restore original auth config
#' gs4_auth_configure(client = original_client, api_key = original_api_key)
gs4_auth_configure <- function(client, path, api_key, app = deprecated()) {
  if (lifecycle::is_present(app)) {
    lifecycle::deprecate_warn(
      "1.1.0",
      "gs4_auth_configure(app)",
      "gs4_auth_configure(client)"
    )
    gs4_auth_configure(client = app, path = path, api_key = api_key)
  }

  if (!missing(client) && !missing(path)) {
    gs4_abort(
      "Must supply exactly one of {.arg client} and {.arg path}, not both."
    )
  }
  stopifnot(missing(api_key) || is.null(api_key) || is_string(api_key))

  if (!missing(path)) {
    stopifnot(is_string(path))
    client <- gargle::gargle_oauth_client_from_json(path)
  }
  stopifnot(
    missing(client) ||
      is.null(client) ||
      inherits(client, "gargle_oauth_client")
  )

  if (!missing(client) || !missing(path)) {
    .auth$set_client(client)
  }

  if (!missing(api_key)) {
    .auth$set_api_key(api_key)
  }

  invisible(.auth)
}

#' @export
#' @rdname gs4_auth_configure
gs4_api_key <- function() {
  .auth$api_key
}

#' @export
#' @rdname gs4_auth_configure
gs4_oauth_client <- function() {
  .auth$client
}

#' Get info on current user
#'
#' @eval gargle:::PREFIX_user_description()
#' @eval gargle:::PREFIX_user_seealso()
#' @eval gargle:::PREFIX_user_return()
#'
#' @export
#' @examples
#' gs4_user()
gs4_user <- function() {
  if (!gs4_has_token()) {
    gs4_bullets(c(i = "Not logged in as any specific Google user."))
    return(invisible())
  }

  email <- gargle::token_email(gs4_token())
  gs4_bullets(c(i = "Logged in to {.pkg googlesheets4} as {.email {email}}."))
  invisible(email)
}

# use this as a guard whenever a googlesheets4 function calls a
# googledrive function that can make an API call
# goal is to expose (most) cases of being auth'ed as 2 different users
# which can lead to very puzzling failures
check_gs4_email_is_drive_email <- function() {
  if (googledrive::drive_has_token() && gs4_has_token()) {
    drive_email <- googledrive::drive_user()[["emailAddress"]]
    gs4_email <- with_gs4_quiet(gs4_user())
    if (drive_email != gs4_email) {
      gs4_bullets(c(
        "!" = "Authenticated as 2 different users with googledrive and \\
               googlesheets4:",
        " " = "googledrive: {.email {drive_email}}",
        " " = "googlesheets4: {.email {gs4_email}}",
        " " = "If you get a puzzling result, this is probably why.",
        "i" = "See the article \"Using googlesheets4 with googledrive\" \\
               for tips:",
        " " = "{.url https://googlesheets4.tidyverse.org/articles/articles/drive-and-sheets.html}"
      ))
    }
  }
}

#' Produce scopes specific to the Sheets API
#'
#' When called with no arguments, `gs4_scopes()` returns a named character
#' vector of scopes associated with the Sheets API. If `gs4_scopes(scopes =)` is
#' given, an abbreviated entry such as `"sheets.readonly"` is expanded to a full
#' scope (`"https://www.googleapis.com/auth/sheets.readonly"` in this case).
#' Unrecognized scopes are passed through unchanged.
#'
#' @inheritParams gs4_auth
#'
#' @seealso
#'   <https://developers.google.com/identity/protocols/oauth2/scopes#sheets> for
#'   details on the permissions for each scope.
#' @returns A character vector of scopes.
#' @family auth functions
#' @export
#' @examples
#' gs4_scopes("spreadsheets")
#' gs4_scopes("spreadsheets.readonly")
#' gs4_scopes("drive")
#' gs4_scopes()
gs4_scopes <- function(scopes = NULL) {
  if (is.null(scopes)) {
    sheets_scopes
  } else {
    resolve_scopes(user_scopes = scopes, package_scopes = sheets_scopes)
  }
}

sheets_scopes <- c(
  spreadsheets = "https://www.googleapis.com/auth/spreadsheets",
  spreadsheets.readonly = "https://www.googleapis.com/auth/spreadsheets.readonly",
  drive = "https://www.googleapis.com/auth/drive",
  drive.readonly = "https://www.googleapis.com/auth/drive.readonly",
  drive.file = "https://www.googleapis.com/auth/drive.file"
)

resolve_scopes <- function(user_scopes, package_scopes) {
  m <- match(user_scopes, names(package_scopes))
  ifelse(is.na(m), user_scopes, package_scopes[m])
}

# unexported helpers that are nice for internal use ----
gs4_auth_internal <- function(
  account = c("docs", "testing"),
  scopes = NULL,
  drive = TRUE
) {
  account <- match.arg(account)
  can_decrypt <- gargle::secret_has_key("GOOGLESHEETS4_KEY")
  online <- !is.null(curl::nslookup("sheets.googleapis.com", error = FALSE))
  if (!can_decrypt || !online) {
    gs4_abort(
      message = c(
        "Auth unsuccessful:",
        if (!can_decrypt) {
          c("x" = "Can't decrypt the {.field {account}} service account token.")
        },
        if (!online) {
          c(
            "x" = "We don't appear to be online. Or maybe the Sheets API is down?"
          )
        }
      ),
      class = "googlesheets4_auth_internal_error",
      can_decrypt = can_decrypt,
      online = online
    )
  }

  if (!is_interactive()) {
    local_gs4_quiet()
  }
  filename <- glue("googlesheets4-{account}.json")
  # TODO: revisit when I do PKG_scopes()
  # https://github.com/r-lib/gargle/issues/103
  scopes <- scopes %||% "https://www.googleapis.com/auth/drive"
  gs4_auth(
    scopes = scopes,
    path = gargle::secret_decrypt_json(
      system.file("secret", filename, package = "googlesheets4"),
      "GOOGLESHEETS4_KEY"
    )
  )
  gs4_user()
  if (drive) {
    googledrive::drive_auth(token = gs4_token())
    gs4_bullets(c(i = "Authed also with {.pkg googledrive}."))
  }
  invisible(TRUE)
}

gs4_auth_docs <- function(scopes = NULL, drive = TRUE) {
  gs4_auth_internal("docs", scopes = scopes, drive = drive)
}

gs4_auth_testing <- function(scopes = NULL, drive = TRUE) {
  gs4_auth_internal("testing", scopes = scopes, drive = drive)
}

local_deauth <- function(env = parent.frame()) {
  original_cred <- .auth$get_cred()
  original_auth_active <- .auth$auth_active
  gs4_bullets(c(i = "Going into deauthorized state."))
  withr::defer(
    gs4_bullets(c("i" = "Restoring previous auth state.")),
    envir = env
  )
  withr::defer(
    {
      .auth$set_cred(original_cred)
      .auth$set_auth_active(original_auth_active)
    },
    envir = env
  )
  gs4_deauth()
}

# deprecated functions ----

#' Get currently configured OAuth app (deprecated)
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' In light of the new [gargle::gargle_oauth_client()] constructor and class of
#' the same name, `gs4_oauth_app()` is being replaced by
#' [gs4_oauth_client()].
#' @keywords internal
#' @export
gs4_oauth_app <- function() {
  lifecycle::deprecate_warn(
    "1.1.0",
    "gs4_oauth_app()",
    "gs4_oauth_client()"
  )
  gs4_oauth_client()
}

test_that("gs4_auth_configure works", {
  old_client <- gs4_oauth_client()
  old_api_key <- gs4_api_key()
  withr::defer(
    gs4_auth_configure(client = old_client, api_key = old_api_key)
  )

  expect_no_error(gs4_oauth_client())
  expect_no_error(gs4_api_key())

  expect_snapshot(
    gs4_auth_configure(client = gargle::gargle_client(), path = "PATH"),
    error = TRUE
  )

  gs4_auth_configure(client = gargle::gargle_client())
  expect_s3_class(gs4_oauth_client(), "gargle_oauth_client")

  path_to_json <- system.file(
    "extdata",
    "client_secret_installed.googleusercontent.com.json",
    package = "gargle"
  )
  gs4_auth_configure(path = path_to_json)
  expect_s3_class(gs4_oauth_client(), "gargle_oauth_client")

  gs4_auth_configure(client = NULL)
  expect_null(gs4_oauth_client())

  gs4_auth_configure(api_key = "API_KEY")
  expect_identical(gs4_api_key(), "API_KEY")

  gs4_auth_configure(api_key = NULL)
  expect_null(gs4_api_key())
})

test_that("gs4_oauth_app() is deprecated", {
  withr::local_options(lifecycle_verbosity = "warning")
  expect_snapshot(absorb <- gs4_oauth_app())
})

test_that("gs4_auth_configure(app =) is deprecated in favor of client", {
  withr::local_options(lifecycle_verbosity = "warning")
  (original_client <- gs4_oauth_client())
  withr::defer(gs4_auth_configure(client = original_client))

  client <- gargle::gargle_oauth_client_from_json(
    system.file(
      "extdata",
      "client_secret_installed.googleusercontent.com.json",
      package = "gargle"
    ),
    name = "test-client"
  )
  expect_snapshot(
    gs4_auth_configure(app = client)
  )
  expect_equal(gs4_oauth_client()$name, "test-client")
  expect_equal(gs4_oauth_client()$id, "abc.apps.googleusercontent.com")
})

# gs4_scopes() ----
test_that("gs4_scopes() reveals Sheets scopes", {
  expect_snapshot(gs4_scopes())
})

test_that("gs4_scopes() substitutes actual scope for short form", {
  expect_equal(
    gs4_scopes(c(
      "spreadsheets",
      "drive",
      "drive.readonly"
    )),
    c(
      "https://www.googleapis.com/auth/spreadsheets",
      "https://www.googleapis.com/auth/drive",
      "https://www.googleapis.com/auth/drive.readonly"
    )
  )
})

test_that("gs4_scopes() passes unrecognized scopes through", {
  expect_equal(
    gs4_scopes(c(
      "email",
      "spreadsheets.readonly",
      "https://www.googleapis.com/auth/cloud-platform"
    )),
    c(
      "email",
      "https://www.googleapis.com/auth/spreadsheets.readonly",
      "https://www.googleapis.com/auth/cloud-platform"
    )
  )
})

#' Add one or more (work)sheets
#'
#' Adds one or more (work)sheets to an existing (spread)Sheet. Note that sheet
#' names must be unique.
#'
#' @eval param_ss()
#' @param sheet One or more new sheet names. If unspecified, one new sheet is
#'   added and Sheets autogenerates a name of the form "SheetN".
#' @param ... Optional parameters to specify additional properties, common to
#'   all of the new sheet(s). Not relevant to most users. Specify fields of the
#'   [`SheetProperties`
#'   schema](https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#SheetProperties)
#'    in `name = value` form.
#' @eval param_before_after("sheet(s)")
#'
#' @template ss-return
#'
#' @export
#' @family worksheet functions
#' @seealso
#' Makes a batch of `AddSheetRequest`s (one per sheet):
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#addsheetrequest>
#'
#' @examplesIf gs4_has_token()
#' ss <- gs4_create("add-sheets-to-me")
#'
#' # the only required argument is the target spreadsheet
#' ss %>% sheet_add()
#'
#' # but you CAN specify sheet name and/or position
#' ss %>% sheet_add("apple", .after = 1)
#' ss %>% sheet_add("banana", .after = "apple")
#'
#' # add multiple sheets at once
#' ss %>% sheet_add(c("coconut", "dragonfruit"))
#'
#' # keeners can even specify additional sheet properties
#' ss %>%
#'   sheet_add(
#'     sheet = "eggplant",
#'     .before = 1,
#'     gridProperties = list(
#'       rowCount = 3, columnCount = 6, frozenRowCount = 1
#'     )
#'   )
#'
#' # get an overview of the sheets
#' sheet_properties(ss)
#'
#' # clean up
#' gs4_find("add-sheets-to-me") %>%
#'   googledrive::drive_trash()
sheet_add <- function(ss, sheet = NULL, ..., .before = NULL, .after = NULL) {
  maybe_character(sheet)
  ssid <- as_sheets_id(ss)
  x <- gs4_get(ssid)
  index <- resolve_index(x$sheets, .before, .after)
  n_new <- if (is.null(sheet)) 1 else length(sheet)
  gs4_bullets(c(v = "Adding {n_new} sheet{?s} to {.s_sheet {x$name}}:"))

  ss <- sheet_add_impl_(ssid, sheet_name = sheet, index = index, ...)

  new_sheet_names <- setdiff(ss$sheets$name, x$sheets$name)
  gs4_bullets(
    bulletize(gargle_map_cli(new_sheet_names, template = "{.w_sheet <<x>>}"))
  )

  invisible(ssid)
}

sheet_add_impl_ <- function(ssid, sheet_name = NULL, index = NULL, ...) {
  sheet_name <- sheet_name %||% list(NULL)
  dots <- list2(...)
  requests <- map(
    sheet_name,
    ~ make_addSheet(title = .x, index = index, dots = dots)
  )
  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = requests,
      includeSpreadsheetInResponse = TRUE,
      responseIncludeGridData = FALSE
    )
  )
  resp_raw <- request_make(req)
  resp <- gargle::response_process(resp_raw)
  new_googlesheets4_spreadsheet(resp$updatedSpreadsheet)
}

resolve_index <- function(
  sheets_df,
  .before = NULL,
  .after = NULL,
  call = caller_env()
) {
  if (is.null(.before) && is.null(.after)) {
    return(NULL)
  }

  if (is.null(.after)) {
    s <- lookup_sheet(.before, sheets_df = sheets_df, call = call)
    return(s$index)
  }

  if (is.numeric(.after)) {
    .after <- min(.after, nrow(sheets_df))
  }
  s <- lookup_sheet(.after, sheets_df = sheets_df, call = call)
  s$index + 1
}

make_addSheet <- function(title = NULL, index = NULL, dots = list()) {
  if (length(title) + length(index) + length(dots) == 0) {
    # if sending no sheet properties, this must be NULL and not list()
    return(list(addSheet = NULL))
  }

  list(
    addSheet = new(
      "AddSheetRequest",
      properties = new("SheetProperties", title = title, index = index, !!!dots)
    )
  )
}

# ---- nm_fun ----
me_ <- nm_fun("TEST-sheet_add")

# ---- tests ----
test_that("sheet_add() rejects non-character `sheet`", {
  expect_snapshot(
    sheet_add(test_sheet("googlesheets4-cell-tests"), sheet = 3),
    error = TRUE
  )
})

test_that("sheet_add() works", {
  skip_if_offline()
  skip_if_no_token()

  ss <- local_ss(me_())

  expect_no_error(
    sheet_add(ss)
  )

  expect_no_error(
    sheet_add(ss, "apple", .after = 1)
  )

  expect_no_error(
    sheet_add(ss, "banana", .after = "apple")
  )

  expect_no_error(
    sheet_add(ss, c("coconut", "dragonfruit"))
  )

  expect_no_error(
    sheet_add(
      ss,
      sheet = "eggplant",
      .before = 1,
      gridProperties = list(
        rowCount = 3,
        columnCount = 6,
        frozenRowCount = 1
      )
    )
  )

  sheets_df <- sheet_properties(ss)

  expect_identical(
    sheets_df$name,
    c(
      "eggplant",
      "Sheet1",
      "apple",
      "banana",
      "Sheet2",
      "coconut",
      "dragonfruit"
    )
  )
  expect_identical(vlookup("eggplant", sheets_df, "name", "grid_rows"), 3L)
  expect_identical(vlookup("eggplant", sheets_df, "name", "grid_columns"), 6L)
})

#' Get Sheet metadata
#'
#' Retrieve spreadsheet-specific metadata, such as details on the individual
#' (work)sheets or named ranges.
#'   * `gs4_get()` complements [googledrive::drive_get()], which
#'     returns metadata that exists for any file on Drive.
#'
#' @eval param_ss()
#'
#' @return A list with S3 class `googlesheets4_spreadsheet`, for printing
#'   purposes.
#' @export
#' @seealso Wraps the `spreadsheets.get` endpoint:
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/get>
#'
#' @examplesIf gs4_has_token()
#' gs4_get(gs4_example("mini-gap"))
gs4_get <- function(ss) {
  resp <- gs4_get_impl_(as_sheets_id(ss))
  new_googlesheets4_spreadsheet(resp)
}

## I want a separate worker so there is a version of this available that
## accepts `fields`, yet I don't want a user-facing function with `fields` arg
gs4_get_impl_ <- function(ssid, fields = NULL) {
  fields <- fields %||%
    "spreadsheetId,properties,spreadsheetUrl,sheets.properties,sheets.protectedRanges,namedRanges"
  req <- request_generate(
    "sheets.spreadsheets.get",
    params = list(
      spreadsheetId = ssid,
      fields = fields
    )
  )
  raw_resp <- request_make(req)
  gargle::response_process(raw_resp)
}

# for development only
str1 <- function(x, ...) utils::str(x, ..., max.level = 1)

noNA <- Negate(anyNA)
allNA <- function(x) all(is.na(x))
notNA <- Negate(is.na)

isFALSE <- function(x) identical(x, FALSE)

is_string <- function(x) is.character(x) && length(x) == 1L

is_integerish <- function(x) {
  floor(x) == x
}

check_data_frame <- function(x, arg = caller_arg(x), call = caller_env()) {
  if (!is.data.frame(x)) {
    gs4_abort(
      c(
        "{.arg {arg}} must be a {.cls data.frame}:",
        x = "{.arg {arg}} has class {.cls {class(x)}}."
      ),
      call = call
    )
  }
  x
}

check_string <- function(x, arg = caller_arg(x), call = caller_env()) {
  check_character(x, arg = arg, call = call)
  check_length_one(x, arg = arg, call = call)
  x
}

maybe_string <- function(x, arg = caller_arg(x), call = caller_env()) {
  if (is.null(x)) {
    x
  } else {
    check_string(x, arg = arg, call = call)
  }
}

check_length_one <- function(x, arg = caller_arg(x), call = caller_env()) {
  if (length(x) != 1) {
    gs4_abort(
      "{.arg {arg}} must have length 1, not length {length(x)}.",
      call = call
    )
  }
  x
}

check_has_length <- function(x, arg = caller_arg(x), call = caller_env()) {
  if (length(x) < 1) {
    gs4_abort(
      "{.arg {arg}} must have length greater than zero.",
      call = call
    )
  }
  x
}

check_character <- function(x, arg = caller_arg(x), call = caller_env()) {
  if (!is.character(x)) {
    gs4_abort(
      c(
        "{.arg {arg}} must be {.cls character}:",
        x = "{.arg {arg}} has class {.cls {class(x)}}."
      ),
      call = call
    )
  }
  x
}

maybe_character <- function(x, arg = caller_arg(x), call = caller_env()) {
  if (is.null(x)) {
    x
  } else {
    check_character(x, arg = arg, call = call)
  }
}

check_non_negative_integer <- function(
  i,
  arg = caller_arg(i),
  call = caller_env()
) {
  if (
    length(i) != 1 || !is.numeric(i) || !is_integerish(i) || is.na(i) || i < 0
  ) {
    gs4_abort(
      c(
        "{.arg {arg}} must be a positive integer:",
        x = "{.arg {arg}} has class {.cls {class(i)}}."
      ),
      call = call
    )
  }
  i
}

maybe_non_negative_integer <- function(
  i,
  arg = caller_arg(i),
  call = caller_env()
) {
  if (is.null(i)) {
    i
  } else {
    check_non_negative_integer(i, arg = arg, call = call)
  }
}

check_bool <- function(bool, arg = caller_arg(bool), call = caller_env()) {
  if (!is_bool(bool)) {
    gs4_abort(
      "{.arg {arg}} must be either {.code TRUE} or {.code FALSE}.",
      call = call
    )
  }
  bool
}

maybe_bool <- function(bool, arg = caller_arg(bool), call = caller_env()) {
  if (is.null(bool)) {
    bool
  } else {
    check_bool(bool, arg = arg, call = call)
  }
}

vlookup <- function(this, data, key, value) {
  stopifnot(is_string(key), is_string(value))
  m <- match(this, data[[key]])
  data[[value]][m]
}

## avoid the name `trim_ws` because it's an argument of several functions in
## this package
ws_trim <- function(x) {
  sub("\\s*$", "", sub("^\\s*", "", x))
}

enforce_na <- function(x, na = "") {
  stopifnot(is.character(x), is.character(na))
  out <- x
  if (length(na) > 0) {
    out[x %in% na] <- NA_character_
  }
  if (!("" %in% na)) {
    out[is.na(x)] <- ""
  }
  out
}

groom_text <- function(x, na = "", trim_ws = TRUE) {
  if (isTRUE(trim_ws)) {
    x <- ws_trim(x)
  }
  enforce_na(x, na)
}

test_that("check_length_one() works", {
  expect_no_error(check_length_one(1))
  expect_error(check_length_one(1:2), "must have length 1")
  expect_error(check_length_one(letters), "letters")
})

test_that("check_character() works", {
  expect_no_error(check_character(letters))
  expect_error(check_character(1:2), "integer")
})


test_that("vlookup() works", {
  df <- tibble::tibble(
    i = 1:3,
    letters = letters[i],
    dupes = c("a", "c", "c"),
    fctr = factor(letters)
  )

  ## internal function, therefore it does not support unquoted variable names
  ## R <= 3.4.4 error msg is "object 'i' not found"
  ## R devel error msg is "is_string(key) is not TRUE"
  expect_error(vlookup("c", df, letters, i))

  expect_identical(vlookup("c", df, "letters", "i"), 3L)
  expect_identical(vlookup(c("a", "c"), df, "letters", "i"), c(1L, 3L))

  ## match() returns position of *first* match
  expect_identical(vlookup("c", df, "dupes", "i"), 2L)
  expect_identical(vlookup(c("c", "c"), df, "dupes", "i"), c(2L, 2L))

  expect_identical(vlookup("b", df, "fctr", "i"), 2L)
  expect_identical(vlookup(c("b", "c", "a"), df, "fctr", "i"), c(2L, 3L, 1L))
})

test_that("enforce_na() works", {
  expect_error(enforce_na(1), "is.character(x) is not TRUE", fixed = TRUE)
  expect_error(enforce_na("a", 1), "is.character(na) is not TRUE", fixed = TRUE)

  expect_identical(enforce_na(character()), character())

  expect_identical(
    enforce_na(c("a", "", "c")),
    c("a", NA, "c")
  )
  expect_identical(
    enforce_na(c("a", "", "c"), na = "c"),
    c("a", "", NA)
  )
  expect_identical(
    enforce_na(c("abc", "", "cab"), na = c("abc", "")),
    c(NA, NA, "cab")
  )
  expect_identical(
    enforce_na(c("a", "", "c"), na = character()),
    c("a", "", "c")
  )
})

#' @export
as_tibble.googlesheets4_schema_ProtectedRange <- function(x, ...) {
  grid_range <- new("GridRange", !!!pluck(x, "range"))
  grid_range <- as_tibble(grid_range)

  tibble::tibble(
    protected_range_id = glean_int(x, "protectedRangeId"),
    description = glean_chr(x, "description"),
    requesting_user_can_edit = glean_lgl(x, "requestingUserCanEdit"),
    warning_only = glean_lgl(x, "warningOnly"),
    has_unprotected_ranges = rlang::has_name(x, "unprotectedRanges"),
    editors = x$editors %||% list(),
    named_range_id = glean_chr(x, "namedRangeId"),
    !!!grid_range
  )
}

gs4_theme <- function() {
  list(
    span.field = list(transform = single_quote_if_no_color),
    # This is same as custom `.drivepath` style in googledrive
    span.s_sheet = list(color = "cyan", fmt = double_quote_weird_name),
    span.w_sheet = list(color = "green", fmt = single_quote_weird_name),
    span.range = list(color = "yellow", fmt = single_quote_weird_name),
    # since we're using color so much elsewhere, I think the standard bullet
    # should be "normal" color; matches what I do in googledrive
    ".bullets .bullet-*" = list(
      "text-exdent" = 2,
      before = function(x) paste0(cli::symbol$bullet, " ")
    )
  )
}

single_quote_weird_name <- function(x) {
  utils::getFromNamespace("quote_weird_name", "cli")(x)
}

# this is just the body of cli's quote_weird_name() but with a double quote
double_quote_weird_name <- function(x) {
  x2 <- utils::getFromNamespace("quote_weird_name0", "cli")(x)
  if (x2[[2]] || cli::num_ansi_colors() == 1) {
    x2[[1]] <- paste0('"', x2[[1]], '"')
  }
  x2[[1]]
}

single_quote_if_no_color <- function(x) quote_if_no_color(x, "'")
double_quote_if_no_color <- function(x) quote_if_no_color(x, '"')

quote_if_no_color <- function(x, quote = "'") {
  # TODO: if a better way appears in cli, use it
  # @gabor says: "if you want to have before and after for the no-color case
  # only, we can have a selector for that, such as:
  # span.field::no-color
  # (but, at the time I write this, cli does not support this yet)
  if (cli::num_ansi_colors() > 1) {
    x
  } else {
    paste0(quote, x, quote)
  }
}

# useful to me during development, so I can see how my messages look w/o color
local_no_color <- function(.envir = parent.frame()) {
  withr::local_envvar(c("NO_COLOR" = 1), .local_envir = .envir)
}

with_no_color <- function(code) {
  withr::with_envvar(c("NO_COLOR" = 1), code)
}

message <- function(...) {
  gs4_abort(
    "
    Internal error: use the UI functions in {.pkg googlesheets4} \\
    instead of {.fun message}",
    .internal = TRUE
  )
}

fr <- function(x) {
  cli::ansi_align(
    as.character(x),
    align = "right",
    width = max(cli::ansi_nchar(x))
  )
}

fl <- function(x) {
  cli::ansi_align(
    as.character(x),
    align = "left",
    width = max(cli::ansi_nchar(x))
  )
}

gs4_quiet <- function() {
  getOption("googlesheets4_quiet", default = NA)
}

#' @export
#' @rdname googlesheets4-configuration
#' @param env The environment to use for scoping
#' @examplesIf gs4_has_token()
#' # message: "Creating new Sheet ..."
#' (ss <- gs4_create("gs4-quiet-demo", sheets = "alpha"))
#'
#' # message: "Editing ..., Writing ..."
#' range_write(ss, data = data.frame(x = 1, y = "a"))
#'
#' # suppress messages for a small amount of code
#' with_gs4_quiet(
#'   ss %>% sheet_append(data.frame(x = 2, y = "b"))
#' )
#'
#' # message: "Writing ..., Appending ..."
#' ss %>% sheet_append(data.frame(x = 3, y = "c"))
#'
#' # suppress messages until end of current scope
#' local_gs4_quiet()
#' ss %>% sheet_append(data.frame(x = 4, y = "d"))
#'
#' # see that all the data was, in fact, written
#' read_sheet(ss)
#'
#' # clean up
#' gs4_find("gs4-quiet-demo") %>%
#'   googledrive::drive_trash()
local_gs4_quiet <- function(env = parent.frame()) {
  withr::local_options(list(googlesheets4_quiet = TRUE), .local_envir = env)
}

local_gs4_loud <- function(env = parent.frame()) {
  withr::local_options(list(googlesheets4_quiet = FALSE), .local_envir = env)
}

#' @export
#' @rdname googlesheets4-configuration
#' @param code Code to execute quietly
with_gs4_quiet <- function(code) {
  withr::with_options(list(googlesheets4_quiet = TRUE), code = code)
}

is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

gs4_bullets <- function(text, .envir = parent.frame()) {
  quiet <- gs4_quiet() %|% is_testing()
  if (quiet) {
    return(invisible())
  }
  cli::cli_div(theme = gs4_theme())
  # TODO: fix this: when I switched from cli::cli_bullets() to
  # cli::cli_inform(), my custom bullet styling was lost (suppressing colour)
  cli::cli_bullets(text, .envir = .envir)
}

#' Error conditions for the googlesheets4 package
#'
#' @param class Use only if you want to subclass beyond `googlesheets4_error`
#'
#' @keywords internal
#' @name gs4-errors
#' @noRd
NULL

gs4_abort <- function(
  message,
  ...,
  class = NULL,
  .envir = parent.frame(),
  call = caller_env()
) {
  cli::cli_div(theme = gs4_theme())
  cli::cli_abort(
    message = message,
    ...,
    class = c(class, "googlesheets4_error"),
    .envir = .envir,
    call = call
  )
}

# helpful in the default method of an as_{to} generic
# exists mostly to template the message
abort_unsupported_conversion <- function(from, to) {
  if (is.null(from)) {
    msg_from <- "{.code NULL}"
  } else {
    msg_from <- "something of class {.cls {class(from)}}"
  }
  msg <- glue(
    "
    Don't know how to make an instance of {.cls {to}} from <<msg_from>>.",
    .open = "<<",
    .close = ">>"
  )
  gs4_abort(msg)
}

test_that("gs4_quiet() falls back to NA if googlesheets4_quiet is unset", {
  withr::with_options(
    list(googlesheets4_quiet = NULL),
    expect_true(is.na(gs4_quiet()))
  )
})

test_that("gs4_abort() throws classed condition", {
  expect_error(gs4_abort("oops"), class = "googlesheets4_error")
  expect_gs4_error(gs4_abort("oops"))
  expect_gs4_error(gs4_abort("oops", class = "googlesheets4_foo"))
  expect_error(
    gs4_abort("oops", class = "googlesheets4_foo"),
    class = "googlesheets4_foo"
  )
})

test_that("abort_unsupported_conversion() works", {
  x <- structure(1, class = c("a", "b", "c"))
  expect_snapshot_error(
    abort_unsupported_conversion(x, "target_class")
  )
})

auth_success <- tryCatch(
  gs4_auth_testing(),
  googlesheets4_auth_internal_error = function(e) NULL
)

if (!isTRUE(auth_success)) {
  gs4_bullets(c(
    "!" = "Internal auth failed; calling {.fun gs4_deauth}."
  ))
  gs4_deauth()
}

skip_if_no_token <- function() {
  if (gs4_has_token()) {
    # hack to slow things down in CI
    Sys.sleep(3)
  } else {
    skip("No token")
  }
}

expect_gs4_error <- function(...) {
  expect_error(..., class = "googlesheets4_error")
}

local_ss <- function(name, ..., env = parent.frame()) {
  existing <- gs4_find(name)
  if (nrow(existing) > 0) {
    gs4_abort("A spreadsheet named {.s_sheet name} already exists.")
  }

  withr::defer(
    {
      trash_me <- gs4_find(name)
      if (nrow(trash_me) < 1) {
        cli::cli_warn(
          "The spreadsheet named {.s_sheet name} already seems to be deleted."
        )
      } else {
        quiet <- gs4_quiet() %|% is_testing()
        if (quiet) {
          googledrive::local_drive_quiet()
        }
        googledrive::drive_trash(trash_me)
      }
    },
    envir = env
  )
  gs4_create(name, ...)
}

#' @keywords internal
#' @import rlang
"_PACKAGE"

## usethis namespace: start
#' @importFrom gargle bulletize
#' @importFrom gargle gargle_map_cli
#' @importFrom glue glue
#' @importFrom glue glue_collapse
#' @importFrom glue glue_data
#' @importFrom googledrive as_id
#' @importFrom lifecycle deprecated
#' @importFrom purrr compact
#' @importFrom purrr discard
#' @importFrom purrr imap
#' @importFrom purrr keep
#' @importFrom purrr map
#' @importFrom purrr map_chr
#' @importFrom purrr map_dbl
#' @importFrom purrr map_int
#' @importFrom purrr map_lgl
#' @importFrom purrr map2
#' @importFrom purrr modify_if
#' @importFrom purrr pluck
#' @importFrom purrr pmap
#' @importFrom purrr pmap_chr
#' @importFrom purrr transpose
#' @importFrom purrr walk
#' @importFrom tibble as_tibble
## usethis namespace: end
NULL

#' Internal vctrs methods
#'
#' @import vctrs
#' @keywords internal
#' @name googlesheets4-vctrs
NULL

#' googlesheets4 configuration
#'
#' @description
#' Some aspects of googlesheets4 behaviour can be controlled via an option.
#'
#' @section Messages:
#'
#' The `googlesheets4_quiet` option can be used to suppress messages from
#' googlesheets4. By default, googlesheets4 always messages, i.e. it is *not*
#' quiet.
#'
#' Set `googlesheets4_quiet` to `TRUE` to suppress messages, by one of these
#' means, in order of decreasing scope:
#' * Put `options(googlesheets4_quiet = TRUE)` in a start-up file, such as
#'   `.Rprofile`, or in your R script
#' * Use `local_gs4_quiet()` to silence googlesheets4 in a specific scope
#' * Use `with_gs4_quiet()` to run a small bit of code silently
#'
#' `local_gs4_quiet()` and `with_gs4_quiet()` follow the conventions of the
#' the withr package (<https://withr.r-lib.org>).
#'
#' @section Auth:
#'
#' Read about googlesheets4's main auth function, [gs4_auth()]. It is powered
#' by the gargle package, which consults several options:
#' * Default Google user or, more precisely, `email`: see
#'   [gargle::gargle_oauth_email()]
#' * Whether or where to cache OAuth tokens: see
#'   [gargle::gargle_oauth_cache()]
#' * Whether to prefer "out-of-band" auth: see
#'   [gargle::gargle_oob_default()]
#' * Application Default Credentials: see [gargle::credentials_app_default()]
#'
#' @name googlesheets4-configuration
NULL

# used for building functions that construct Sheet names in tests ----
nm_fun <- function(context, user_run = TRUE) {
  user_run <- if (isTRUE(user_run)) nm_user_run() else NULL
  y <- purrr::compact(list(context, user_run))
  function(x = character()) as.character(glue_collapse(c(x, y), sep = "-"))
}

nm_user_run <- function() {
  if (as.logical(Sys.getenv("GITHUB_ACTIONS", unset = "false"))) {
    glue("gha-{Sys.getenv('GITHUB_WORKFLOW')}-{Sys.getenv('GITHUB_RUN_ID')}")
  } else {
    random_id <- ids::proquint(n = 1, n_words = 2)
    glue("{Sys.info()['user']}-{random_id}")
  }
}

# ---- nm_fun ----
me_ <- nm_fun("TEST-sheet_relocate")

# ---- tests ----
test_that("relocation works", {
  skip_if_offline()
  skip_if_no_token()

  sheet_names <- c("alfa", "bravo", "charlie", "delta", "echo", "foxtrot")
  ss <- local_ss(me_(), sheets = sheet_names)

  sheet_relocate(ss, "echo", .before = "bravo")
  sheet_relocate(ss, list("foxtrot", 4))
  sheet_relocate(ss, c("bravo", "alfa", "echo"), .after = 10)
  expect_equal(
    sheet_names(ss),
    c("foxtrot", "charlie", "delta", "bravo", "alfa", "echo")
  )
})

# ---- nm_fun ----
me_ <- nm_fun("TEST-sheet_rename")

# ---- tests ----
test_that("internal copy works", {
  skip_if_offline()
  skip_if_no_token()

  ss <- local_ss(
    me_(),
    sheets = list(iris = head(iris), chickwts = head(chickwts))
  )
  ss %>%
    sheet_rename(2, new_name = "poultry") %>%
    sheet_rename(1, new_name = "flowers")
  out <- sheet_names(ss)
  expect_equal(out, c("flowers", "poultry"))
})

## this file represents the interface with the cellranger package

#' Specify cells
#'
#' Many functions in googlesheets4 use a `range` argument to target specific
#' cells. The Sheets v4 API expects user-specified ranges to be expressed via
#' [its A1
#' notation](https://developers.google.com/sheets/api/guides/concepts#a1_notation),
#' but googlesheets4 accepts and converts a few alternative specifications
#' provided by the functions in the [cellranger][cellranger] package. Of course,
#' you can always provide A1-style ranges directly to functions like
#' [read_sheet()] or [range_read_cells()]. Why would you use the
#' [cellranger][cellranger] helpers? Some ranges are practically impossible to
#' express in A1 notation, specifically when you want to describe rectangles
#' with some bounds that are specified and others determined by the data.
#'
#' @name cell-specification
#'
#' @examplesIf gs4_has_token() && rlang::is_interactive()
#' ss <- gs4_example("mini-gap")
#'
#' # Specify only the rows or only the columns
#' read_sheet(ss, range = cell_rows(1:3))
#' read_sheet(ss, range = cell_cols("C:D"))
#' read_sheet(ss, range = cell_cols(1))
#'
#' # Specify upper or lower bound on row or column
#' read_sheet(ss, range = cell_rows(c(NA, 4)))
#' read_sheet(ss, range = cell_cols(c(NA, "D")))
#' read_sheet(ss, range = cell_rows(c(3, NA)))
#' read_sheet(ss, range = cell_cols(c(2, NA)))
#' read_sheet(ss, range = cell_cols(c("C", NA)))
#'
#' # Specify a partially open rectangle
#' read_sheet(ss, range = cell_limits(c(2, 3), c(NA, NA)), col_names = FALSE)
#' read_sheet(ss, range = cell_limits(c(1, 2), c(NA, 4)))
NULL

#' @importFrom cellranger cell_limits
#' @name cell_limits
#' @export
#' @rdname cell-specification
NULL

#' @importFrom cellranger cell_rows
#' @name cell_rows
#' @export
#' @rdname cell-specification
NULL

#' @importFrom cellranger cell_cols
#' @name cell_cols
#' @export
#' @rdname cell-specification
NULL

#' @importFrom cellranger anchored
#' @name anchored
#' @export
#' @rdname cell-specification
NULL

#' @param reformat Logical, indicates whether to reformat the affected cells.
#'   Currently googlesheets4 provides no real support for formatting, so
#'   `reformat = TRUE` effectively means that edited cells become unformatted.

devtools::load_all() # I assume we're in googlesheets4 source
library(googledrive)
library(tidyverse)

# googlesheets4:::gs4_auth_testing()

ss <- test_sheet_create("googlesheets4-col-types")
gs4_browse(ss)

df <- tibble(
  logical = c(TRUE, FALSE, NA, TRUE),
  character = c("apple", "banana", "cherry", "durian"),
  factor = factor(c("one", "two", "three", "four")),
  integer = 1:4,
  double = 4:1 - 2.5,
  date = as.Date(c("2003-06-06", "1982-12-05", "2014-02-14", "1999-08-27")),
  datetime = as.POSIXct(c("1978-05-31 04:24:32", "2006-07-19 23:27:37",
                          "2003-12-21 09:20:29", "1975-04-14 13:31:03"))
)

sheet_write(df, ss, sheet = "lots-of-types")

make_column <- function(df, ctype, ..., nr, guess_max = min(1000, nr)) {
  ## must resolve COL_GUESS here (vs when parsing) because need to know ctype
  ## here, when making the column
  ctype <- resolve_col_type(df$cell[df$row <= guess_max], ctype)
  parsed <- gs4_parse(df$cell, ctype, ...)
  if (is.null(parsed)) {
    return()
  }
  fodder <- rep_len(NA, length.out = nr)
  column <- switch(
    ctype,
    ## NAs must be numeric in order to initialize datetimes with a timezone
    CELL_DATE = as_Date(as.numeric(fodder)),
    ## TODO: time of day not really implemented yet
    CELL_TIME = as_POSIXct(as.numeric(fodder)),
    CELL_DATETIME = as_POSIXct(as.numeric(fodder)),
    COL_LIST = vector(mode = "list", length = nr),
    as.vector(fodder, mode = typeof(parsed))
  )
  if (ctype == "CELL_TEXT") {
    dots <- list2(...)
    column <- enforce_na(column, na = dots$na)
  }
  column[df$row] <- parsed
  column
}

resolve_col_type <- function(cell, ctype = "COL_GUESS") {
  if (ctype != "COL_GUESS") {
    return(ctype)
  }
  cell %>%
    ctype() %>%
    effective_cell_type() %>%
    consensus_col_type()
}

gs4_parse <- function(x, ctype, ...) {
  stopifnot(is_string(ctype))
  # fmt: skip
  parse_fun <- switch(
    ctype,
    COL_SKIP      = as_skip,
    CELL_LOGICAL  = as_logical,
    CELL_INTEGER  = as_integer,
    CELL_NUMERIC  = as_double,
    CELL_DATE     = as_date,
    # TODO: CELL_TIME not really implemented yet
    CELL_TIME     = as_datetime,
    CELL_DATETIME = as_datetime,
    CELL_TEXT     = as_character,
    COL_CELL      = as_cell,
    COL_LIST      = as_list,
    ## TODO: factor, duration
    gs4_abort(
      "Not a recognized column type: {.field {ctype}}",
      .internal = TRUE
    )
  )
  if (inherits(x, "SHEETS_CELL")) {
    x <- list(x)
  }
  parse_fun(x, ...)
}

as_skip <- function(cell, ...) NULL
as_cell <- function(cell, ...) cell

as_list <- function(cell, ...) {
  ctypes <- cell %>%
    ctype() %>%
    effective_cell_type() %>%
    blank_to_logical()
  map2(cell, ctypes, gs4_parse, ...)
}

## prepare to coerce to logical, integer, double
cell_content <- function(cell, na = "", trim_ws = TRUE) {
  switch(
    ctype(cell),
    CELL_BLANK = NA,
    CELL_LOGICAL = pluck(cell, "effectiveValue", "boolValue"),
    CELL_NUMERIC = pluck(cell, "effectiveValue", "numberValue"),
    CELL_DATE = NA_real_,
    CELL_TIME = NA_real_,
    CELL_DATETIME = NA_real_,
    CELL_TEXT = cell %>%
      pluck("effectiveValue", "stringValue") %>%
      groom_text(na = na, trim_ws = trim_ws)
  )
}

as_logical <- function(cell, na = "", trim_ws = TRUE) {
  cell %>%
    map(cell_content, na = na, trim_ws = trim_ws) %>%
    map_lgl(as.logical)
}

as_integer <- function(cell, na = "", trim_ws = TRUE) {
  cell %>%
    map(cell_content, na = na, trim_ws = trim_ws) %>%
    map_int(as.integer)
}

as_double <- function(cell, na = "", trim_ws = TRUE) {
  cell %>%
    map(cell_content, na = na, trim_ws = trim_ws) %>%
    map_dbl(as.double)
}

## prepare to coerce to date, time, datetime
cell_content_datetime <- function(cell, na = "", trim_ws = TRUE) {
  switch(
    ctype(cell),
    CELL_BLANK = NA,
    CELL_LOGICAL = NA,
    CELL_NUMERIC = NA,
    CELL_DATE = pluck(cell, "effectiveValue", "numberValue"),
    CELL_TIME = pluck(cell, "effectiveValue", "numberValue"),
    CELL_DATETIME = pluck(cell, "effectiveValue", "numberValue"),
    CELL_TEXT = NA
  )
}

as_datetime <- function(cell, na = "", trim_ws = TRUE) {
  cell %>%
    map(cell_content_datetime, na = na, trim_ws = trim_ws) %>%
    map_dbl(as.double) %>%
    map_dbl(`*`, 24 * 60 * 60) %>%
    as_POSIXct()
}

as_date <- function(cell, na = "", trim_ws = TRUE) {
  cell %>%
    map(cell_content_datetime, na = na, trim_ws = trim_ws) %>%
    map_dbl(as.double) %>%
    as_Date()
}

## TODO: not wired up yet (body is same as as_datetime)
# as_time <- function(cell, na = "", trim_ws = TRUE) {
#   cell %>%
#     map(cell_content_datetime, na = na, trim_ws = trim_ws) %>%
#     map_dbl(as.double) %>%
#     `*`(24 * 60 * 60) %>%
#     as_POSIXct()
# }

## prepare to coerce to character
cell_content_chr <- function(cell, na = "", trim_ws = TRUE) {
  fv <- pluck(cell, "formattedValue", .default = NA_character_)
  groom_text(fv, na = na, trim_ws = trim_ws)
}

as_character <- function(cell, na = "", trim_ws = TRUE) {
  cell %>%
    map(cell_content_chr, na = na, trim_ws = trim_ws) %>%
    map_chr(as.character)
}

as_Date <- function(x = NA_real_, origin = "1899-12-30", tz = "UTC", ...) {
  as.Date(x, origin = origin, tz = tz, ...)
}

as_POSIXct <- function(x = NA_real_, origin = "1899-12-30", tz = "UTC", ...) {
  as.POSIXct(x, origin = origin, tz = tz, ...)
}

test_that("resolve_col_type() passes `ctype` other than 'COL_GUESS' through", {
  expect_identical(resolve_col_type("a cell", "COL_ANYTHING"), "COL_ANYTHING")
})

test_that("resolve_col_type() implements coercion DAG for 'COL_GUESS'", {
  input <- c("l", "D")
  expect_identical(resolve_col_type(input, "COL_GUESS"), "COL_LIST")
})

test_that("blank cell doesn't trick resolve_col_type() into guessing COL_LIST", {
  input <- list(
    structure(1, class = c("CELL_BLANK", "SHEETS_CELL")),
    structure(1, class = c("CELL_TEXT", "SHEETS_CELL"))
  )
  expect_identical(resolve_col_type(input, "COL_GUESS"), "CELL_TEXT")
})

#' Visit a Sheet in a web browser
#'
#' Visits a Google Sheet in your default browser, if session is interactive.
#'
#' @inheritParams read_sheet
#'
#' @return The Sheet's browser URL, invisibly.
#' @export
#' @examples
#' gs4_example("mini-gap") %>% gs4_browse()
gs4_browse <- function(ss) {
  ## TO RECONSIDER AFTER AUTH: get the official link, if we're in auth state?
  # googledrive::drive_browse(as_sheets_id(ss))
  ssid <- as_sheets_id(ss)
  url <- glue("https://docs.google.com/spreadsheets/d/{ssid}")
  if (is_interactive()) {
    utils::browseURL(url)
  }
  invisible(url)
}

#' Example Sheets
#'
#' googlesheets4 makes a variety of world-readable example Sheets available for
#' use in documentation and reprexes. These functions help you access the
#' example Sheets. See `vignette("example-sheets", package = "googlesheets4")`
#' for more.
#'
#' @param matches A regular expression that matches the name of the desired
#'   example Sheet(s). `matches` is optional for the plural `gs4_examples()`
#'   and, if provided, it can match multiple Sheets. The singular
#'   `gs4_example()` requires `matches` and it must match exactly one Sheet.
#'
#' @return

#' * `gs4_example()`: a [sheets_id]
#' * `gs4_examples()`: a named vector of all built-in examples, with class
#'   [`drive_id`][googledrive::as_id]

#'
#' @name gs4_examples
#' @examplesIf gs4_has_token()
#' gs4_examples()
#' gs4_examples("gap")
#'
#' gs4_example("gapminder")
#' gs4_example("deaths")
NULL

#' @rdname gs4_examples
#' @export
gs4_examples <- function(matches) {
  many_sheets(
    needle = matches,
    haystack = example_and_test_sheets("example"),
    adjective = "example"
  )
}

#' @rdname gs4_examples
#' @export
gs4_example <- function(matches) {
  one_sheet(
    needle = matches,
    haystack = example_and_test_sheets("example"),
    adjective = "example"
  )
}

many_sheets <- function(needle, haystack, adjective, call = caller_env()) {
  out <- haystack

  if (!missing(needle)) {
    check_string(needle, call = call)
    sel <- grepl(needle, names(out), ignore.case = TRUE)
    if (!any(sel)) {
      gs4_abort(
        "Can't find {adjective} Sheet that matches {.q {needle}}.",
        call = call
      )
    }
    out <- as_id(out[sel])
  }

  out
}

one_sheet <- function(needle, haystack, adjective, call = caller_env()) {
  check_string(needle, call = call)
  out <- many_sheets(
    needle = needle,
    haystack = haystack,
    adjective = adjective,
    call = call
  )
  if (length(out) > 1) {
    gs4_abort(
      c(
        "Found multiple matching {adjective} Sheets:",
        bulletize(gargle_map_cli(names(out), template = "{.s_sheet <<x>>}")),
        i = "Make the {.arg matches} regular expression more specific."
      ),
      call = call
    )
  }
  as_sheets_id(out)
}

example_and_test_sheets <- function(purpose = NULL) {
  # inlining env_cache() logic, so I don't need bleeding edge rlang
  if (!env_has(.googlesheets4, "example_and_test_sheets")) {
    inventory_id <- "1dSIZ2NkEPDWiEbsg9G80Hr9Xe7HZglEAPwGhVa-OSyA"
    local_gs4_quiet()
    if (!gs4_has_token()) {
      # don't trigger auth just for this
      local_deauth()
    }
    dat <- read_sheet(as_sheets_id(inventory_id))
    env_poke(.googlesheets4, "example_and_test_sheets", dat)
  }
  dat <- env_get(.googlesheets4, "example_and_test_sheets")
  if (!is.null(purpose)) {
    dat <- dat[dat$purpose == purpose, ]
  }
  out <- dat$id
  names(out) <- dat$name
  as_id(out)
}

# test sheet management ----
test_sheets <- function(matches) {
  many_sheets(
    needle = matches,
    haystack = example_and_test_sheets("test"),
    adjective = "test"
  )
}

test_sheet <- function(matches = "googlesheets4-cell-tests") {
  one_sheet(
    needle = matches,
    haystack = example_and_test_sheets("test"),
    adjective = "test"
  )
}

test_sheet_create <- function(name = "googlesheets4-cell-tests") {
  stopifnot(is_string(name))

  user <- gs4_user()
  if (!grepl("^googlesheets4-sheet-keeper", user)) {
    user <- sub("@.+$", "", user)
    gs4_abort(
      "
      Must be auth'd as {.email googlesheets4-sheet-keeper}, \\
      not {.email {user}}."
    )
  }

  existing <- gs4_find()
  m <- match(name, existing$name)
  if (is.na(m)) {
    gs4_bullets(c(v = "Creating {.s_sheet {name}}."))
    ss <- gs4_create(name)
  } else {
    gs4_bullets(c(
      v = "Testing sheet named {.s_sheet {name}} already exists ... using that."
    ))
    ss <- existing$id[[m]]
  }
  ssid <- as_sheets_id(ss)

  # it's fiddly to check current sharing status, so just re-share
  gs4_bullets(c(
    v = 'Making sure "anyone with a link" can read {.s_sheet {name}}.'
  ))
  gs4_share(ssid)
  ssid
}

test_that("gs4_examples() lists all examples, in a named drive_id object", {
  skip_if_offline()
  skip_on_cran()

  examples <- gs4_examples()
  expect_true(is.character(examples))
  expect_true(length(examples) > 0)
  expect_true(is.character(names(examples)))
  expect_s3_class(examples, "drive_id")
})

test_that("gs4_example() returns a sheets_id", {
  skip_if_offline()
  skip_on_cran()

  expect_s3_class(gs4_example("deaths"), "sheets_id")
})

test_that("gs4_examples() requires a match if `matches` is supplied", {
  skip_if_offline()
  skip_on_cran()

  expect_error(gs4_examples("nope"), "Can't find")
})

test_that("gs4_example() requires `matches`", {
  skip_if_offline()
  skip_on_cran()

  expect_error(gs4_example(), "missing")
})

test_that("`matches` works in gs4_examples()", {
  skip_if_offline()
  skip_on_cran()

  examples <- gs4_examples("gap")
  expect_true(length(examples) > 0)
  expect_s3_class(examples, "drive_id")
})

test_that("`matches` works in gs4_example()", {
  skip_if_offline()
  skip_on_cran()

  expect_no_error(
    example <- gs4_example("gapminder")
  )
  expect_length(example, 1)
  expect_s3_class(example, "drive_id")
  expect_s3_class(example, "sheets_id")
})

test_that("gs4_example() requires a unique match", {
  skip_if_offline()
  skip_on_cran()

  expect_error(gs4_example("gap"), "multiple")
})

test_that("example functions work when deauth'd", {
  skip_if_offline()
  skip_on_cran()

  examples <- gs4_examples()
  gapminder <- gs4_example("gapminder")

  local_deauth()
  env_bind(.googlesheets4, example_and_test_sheets = zap())

  expect_equal(gs4_examples(), examples)
  expect_equal(gs4_example("gapminder"), gapminder)
})

test_that("range_spreadread() works", {
  skip_if_offline()
  skip_if_no_token()
  skip_if_not_installed("readr")

  # specify a sheet-qualified cell range
  read <- range_read(
    test_sheet("googlesheets4-cell-tests"),
    range = "'range-experimentation'!B:D"
  )
  speedread <- range_speedread(
    test_sheet("googlesheets4-cell-tests"),
    range = "'range-experimentation'!B:D",
    col_types = readr::cols() # suppress col spec printing
  )
  expect_equal(read, speedread, ignore_attr = TRUE)

  # specify col_types
  read <- range_read(
    gs4_example("deaths"),
    sheet = "other",
    range = "A5:F15",
    col_types = "??i?DD"
  )
  speedread <- range_speedread(
    gs4_example("deaths"),
    sheet = "other",
    range = "A5:F15",
    col_types = readr::cols(
      Age = readr::col_integer(),
      `Date of birth` = readr::col_date("%m/%d/%Y"),
      `Date of death` = readr::col_date("%m/%d/%Y")
    )
  )
  expect_equal(read, speedread, ignore_attr = TRUE)
})

#' (Over)write new data into a range
#'
#' @description
#'
#' Writes a data frame into a range of cells. Main differences from
#' [sheet_write()] (a.k.a. [write_sheet()]):
#' * Narrower scope. `range_write()` literally targets some cells, not a whole
#'   (work)sheet.
#' * The edited rectangle is not explicitly styled as a table. Nothing special
#'   is done re: formatting a header row or freezing rows.
#' * Column names can be suppressed. This means that, although `data` must
#'   be a data frame (at least for now), `range_write()` can actually be used
#'   to write arbitrary data.
#' * The target (spread)Sheet and (work)sheet must already exist. There is no
#'   ability to create a Sheet or add a worksheet.
#' * The target sheet dimensions are not "trimmed" to shrink-wrap the `data`.
#'   However, the sheet might gain rows and/or columns, in order to write
#'   `data` to the user-specified `range`.
#'
#' If you just want to add rows to an existing table, the function you probably
#' want is [sheet_append()].
#'
#' @section Range specification:
#' The `range` argument of `range_write()` is special, because the Sheets API
#' can implement it in 2 different ways:
#' * If `range` represents exactly 1 cell, like "B3", it is taken as the *start*
#'   (or upper left corner) of the targeted cell rectangle. The edited cells are
#'   determined implicitly by the extent of the `data` we are writing. This
#'   frees you from doing fiddly range computations based on the dimensions of
#'   the `data`.
#' * If `range` describes a rectangle with multiple cells, it is interpreted
#'   as the *actual* rectangle to edit. It is possible to describe a rectangle
#'   that is unbounded on the right (e.g. "B2:4"), on the bottom (e.g. "A4:C"),
#'   or on both the right and the bottom (e.g.
#'   `cell_limits(c(2, 3), c(NA, NA))`. Note that **all cells** inside the
#'   rectangle receive updated data and format. Important implication: if the
#'   `data` object isn't big enough to fill the target rectangle, the cells that
#'    don't receive new data are effectively cleared, i.e. the existing value
#'    and format are deleted.
#'
#' @eval param_ss()
#' @param data A data frame.
#' @eval param_sheet(
#'   action = "write into",
#'   "Ignored if the sheet is specified via `range`. If neither argument",
#'   "specifies the sheet, defaults to the first visible sheet."
#' )
#' @param range Where to write. This `range` argument has important similarities
#'   and differences to `range` elsewhere (e.g. [range_read()]):
#'   * Similarities: Can be a cell range, using A1 notation ("A1:D3") or using
#'     the helpers in [`cell-specification`]. Can combine sheet name and cell
#'     range ("Sheet1!A5:A") or refer to a sheet by name (`range = "Sheet1"`,
#'     although `sheet = "Sheet1"` is preferred for clarity).
#'   * Difference: Can NOT be a named range.
#'   * Difference: `range` can be interpreted as the *start* of the target
#'     rectangle (the upper left corner) or, more literally, as the actual
#'     target rectangle. See the "Range specification" section for details.
#' @param col_names Logical, indicates whether to send the column names of
#'   `data`.
#' @template reformat
#'
#' @template ss-return
#' @export
#' @family write functions
#' @seealso
#' If sheet size needs to change, makes an `UpdateSheetPropertiesRequest`:
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#UpdateSheetPropertiesRequest>
#'
#' The main data write is done via an `UpdateCellsRequest`:
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#updatecellsrequest>
#'
#' @examplesIf gs4_has_token()
#' # create a Sheet with some initial, empty (work)sheets
#' (ss <- gs4_create("range-write-demo", sheets = c("alpha", "beta")))
#'
#' df <- data.frame(
#'   x = 1:3,
#'   y = letters[1:3]
#' )
#'
#' #  write df somewhere other than the "upper left corner"
#' range_write(ss, data = df, range = "D6")
#'
#' # view your magnificent creation in the browser
#' gs4_browse(ss)
#'
#' # send data of disparate types to a 1-row rectangle
#' dat <- tibble::tibble(
#'   string = "string",
#'   logical = TRUE,
#'   datetime = Sys.time()
#' )
#' range_write(ss, data = dat, sheet = "beta", col_names = FALSE)
#'
#' # send data of disparate types to a 1-column rectangle
#' dat <- tibble::tibble(
#'   x = list(Sys.time(), FALSE, "string")
#' )
#' range_write(ss, data = dat, range = "beta!C5", col_names = FALSE)
#'
#' # clean up
#' gs4_find("range-write-demo") %>%
#'   googledrive::drive_trash()
range_write <- function(
  ss,
  data,
  sheet = NULL,
  range = NULL,
  col_names = TRUE, # not sure about this default
  reformat = TRUE
) {
  ssid <- as_sheets_id(ss)
  check_data_frame(data)
  maybe_sheet(sheet)
  check_range(range)
  check_bool(col_names)
  check_bool(reformat)

  x <- gs4_get(ssid)
  gs4_bullets(c(v = "Editing {.s_sheet {x$name}}."))

  # determine (work)sheet ------------------------------------------------------
  range_spec <- as_range_spec(
    range,
    sheet = sheet,
    sheets_df = x$sheets,
    nr_df = x$named_ranges
  )
  range_spec$sheet_name <- range_spec$sheet_name %||%
    first_visible_name(x$sheets)
  gs4_bullets(c(v = "Writing to sheet {.w_sheet {range_spec$sheet_name}}."))

  # initialize the batch update requests; store details on target sheet s ------
  requests <- list()
  s <- lookup_sheet(range_spec$sheet_name, sheets_df = x$sheets)

  # package the write location as `start` or `range` ---------------------------
  loc <- prepare_loc(range_spec)

  # figure out if we need to resize the sheet ----------------------------------
  dims_needed <- prepare_dims(loc, data, col_names)
  resize_req <- prepare_resize_request(
    s,
    nrow_needed = dims_needed$nrow,
    ncol_needed = dims_needed$ncol,
    exact = FALSE
  )

  if (!is.null(resize_req)) {
    new_dims <- pluck(
      resize_req,
      "updateSheetProperties",
      "properties",
      "gridProperties"
    )
    gs4_bullets(c(
      v = "Changing dims: ({s$grid_rows} x {s$grid_columns}) --> \\
           ({new_dims$rowCount %||% s$grid_rows} x \\
           {new_dims$columnCount %||% s$grid_columns})."
    ))
    requests <- c(requests, list(resize_req))
  }

  # pack the data, specify field mask ------------------------------------------
  fields <- if (reformat) {
    "userEnteredValue,userEnteredFormat"
  } else {
    "userEnteredValue"
  }
  data_req <- new(
    "UpdateCellsRequest",
    rows = as_RowData(data, col_names = col_names),
    fields = fields,
    !!!loc
  )
  requests <- c(requests, list(list(updateCells = data_req)))

  # do it ----------------------------------------------------------------------
  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = requests
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)

  invisible(ssid)
}

prepare_loc <- function(x) {
  if (is.null(x$cell_limits)) {
    if (is.null(x$cell_range)) {
      return(list(start = as_GridCoordinate(x)))
    }
    x$cell_limits <- limits_from_range(x$cell_range)
  }

  if (more_than_one_cell(x$cell_limits)) {
    list(range = as_GridRange(x))
  } else {
    list(start = as_GridCoordinate(x))
  }
}

more_than_one_cell <- function(cl) {
  if (anyNA(cl$ul) || anyNA(cl$lr)) {
    return(TRUE)
  }

  nrows <- cl$lr[1] - cl$ul[1] + 1
  ncols <- cl$lr[2] - cl$ul[2] + 1
  nrows > 1 || ncols > 1
}

prepare_dims <- function(write_loc, data, col_names) {
  # Here it is actually useful to us that the row and column indices inside
  # `write_loc` are zero-indexed. Recall that:
  #   * `start` is an instance of GridCoordinate
  #   * `range` is an instance of GridRange
  if (has_name(write_loc, "start")) {
    return(list(
      nrow = (write_loc$start$rowIndex %||% 0) + nrow(data) + col_names,
      ncol = (write_loc$start$columnIndex %||% 0) + ncol(data)
    ))
  }

  # we must be writing to a `range`
  # take explicit end indices literally
  # otherwise infer from start indices + size of `data`
  gr <- write_loc$range
  list(
    nrow = gr$endRowIndex %||%
      ((gr$startRowIndex %||% 0) + nrow(data) + col_names),
    ncol = gr$endColumnIndex %||%
      ((gr$startColumnIndex %||% 0) + ncol(data))
  )
}

# ---- nm_fun ----
me_ <- nm_fun("TEST-range_write")

# ---- tests ----
test_that("range_write() works", {
  skip_if_offline()
  skip_if_no_token()

  n <- 3
  m <- 5
  data <- suppressMessages(
    # silence messages about name repair
    tibble::as_tibble(
      matrix(head(letters, n * m), nrow = n, ncol = m),
      .name_repair = "unique"
    )
  )

  ss <- local_ss(me_(), sheets = list(foo = data))
  # this is intentional below: refer to sheet in various ways

  # write into existing cells --> no size change
  range_write(ss, data[3:2, ])
  props <- sheet_properties(ss)
  expect_equal(props$grid_rows, n + 1)
  expect_equal(props$grid_columns, m)
  df <- read_sheet(ss)
  expect_identical(df[1, ], df[3, ])

  # write into non-existing cells --> sheet must grow
  range_write(ss, data, range = "foo!F5")
  props <- sheet_properties(ss)
  expect_equal(props$grid_rows, (5 - 1) + n + 1)
  expect_equal(props$grid_columns, (which(LETTERS == "F") - 1) + m)
  df <- read_sheet(ss, range = cell_cols(c("F", NA)))
  expect_equal(df, data)

  # write into existing and non-existing cells --> need new columns
  range_write(ss, data[1:3], sheet = "foo", range = "I2:K5")
  props <- sheet_properties(ss)
  expect_equal(props$grid_columns, (which(LETTERS == "K")))
  df <- read_sheet(ss, range = "I2:K5")
  expect_equal(df, data[1:3])
})

# https://github.com/tidyverse/googlesheets4/issues/203
test_that("we can write a hole-y tibble containing NULLs", {
  skip_if_offline()
  skip_if_no_token()

  dat_write <- tibble::tibble(A = list(NULL, "HI"), B = month.abb[1:2])

  ss <- local_ss(me_("write-NULL"), sheets = dat_write)
  write_sheet(dat_write, ss, sheet = 1)

  dat_read <- read_sheet(ss)
  expect_equal(dat_read$A, c(NA, "HI"))
  expect_equal(dat_read$B, dat_write$B)

  dat_read <- read_sheet(ss, col_types = "Lc")
  expect_equal(dat_read$A, dat_write$A)
  expect_equal(dat_read$B, dat_write$B)
})

# ---- helpers ----
test_that("prepare_loc() makes the right call re: `start` vs. `range`", {
  expect_loc <- function(x, loc) {
    sheets_df <- tibble::tibble(name = "Sheet1", index = 0, id = 123)
    out <- prepare_loc(as_range_spec(x, sheets_df = sheets_df))
    expect_named(out, loc)
  }

  expect_loc(NULL, "start")
  expect_loc("Sheet1", "start")
  expect_loc("D4", "start")
  expect_loc("B5:B5", "start")
  expect_loc(cell_limits(c(5, 2), c(5, 2)), "start")

  expect_loc("B4:G9", "range")
  expect_loc("A2:F", "range")
  expect_loc("A2:5", "range")
  expect_loc("C:E", "range")
  expect_loc("5:7", "range")
  expect_loc(cell_limits(c(2, 4), c(NA, NA)), "range")
})

test_that("prepare_dims() works when write_loc is a `start` (a GridCoordinate)", {
  n <- 3
  m <- 5
  data <- suppressMessages(
    # silence messages about name repair
    tibble::as_tibble(
      matrix(head(letters, n * m), nrow = n, ncol = m),
      .name_repair = "unique"
    )
  )

  expect_dims <- function(loc, col_names, dims) {
    expect_equal(prepare_dims(loc, data, col_names = col_names), dims)
  }

  # no row or column info --> default offset is 0 (remember these are 0-indexed)
  loc <- list(start = new("GridCoordinate", sheetId = 123))
  expect_dims(loc, col_names = TRUE, list(nrow = n + 1, ncol = m))
  expect_dims(loc, col_names = FALSE, list(nrow = n, ncol = m))

  # row offset
  loc <- list(start = new("GridCoordinate", sheetId = 123, rowIndex = 2))
  expect_dims(loc, col_names = TRUE, list(nrow = 2 + n + 1, ncol = m))
  expect_dims(loc, col_names = FALSE, list(nrow = 2 + n, ncol = m))

  # column offset
  loc <- list(start = new("GridCoordinate", sheetId = 123, columnIndex = 3))
  expect_dims(loc, col_names = TRUE, list(nrow = n + 1, ncol = 3 + m))
  expect_dims(loc, col_names = FALSE, list(nrow = n, ncol = 3 + m))

  # row and column offset
  loc <- list(
    start = new("GridCoordinate", sheetId = 123, rowIndex = 2, columnIndex = 3)
  )
  expect_dims(loc, col_names = TRUE, list(nrow = 2 + n + 1, ncol = 3 + m))
  expect_dims(loc, col_names = FALSE, list(nrow = 2 + n, ncol = 3 + m))
})

test_that("prepare_dims() works when write_loc is a `range` (a GridRange)", {
  n <- 3
  m <- 5
  data <- suppressMessages(
    # silence messages about name repair
    tibble::as_tibble(
      matrix(head(letters, n * m), nrow = n, ncol = m),
      .name_repair = "unique"
    )
  )

  expect_dims <- function(x, col_names, dims) {
    sheets_df <- tibble::tibble(name = "Sheet1", index = 0)
    loc <- prepare_loc(as_range_spec(x, sheets_df = sheets_df))
    expect_equal(prepare_dims(loc, data, col_names = col_names), dims)
  }

  # fully specified range; lower right cell is all that matters
  expect_dims(
    "B4:G9",
    col_names = TRUE,
    list(nrow = 9, ncol = which(LETTERS == "G"))
  )
  expect_dims(
    "B4:G9",
    col_names = FALSE,
    list(nrow = 9, ncol = which(LETTERS == "G"))
  )

  # range is open on the bottom
  # get row extent from upper left of range + data, column extent from range
  expect_dims(
    "B3:D",
    col_names = TRUE,
    list(nrow = 2 + n + 1, ncol = which(LETTERS == "D"))
  )
  expect_dims(
    "B3:D",
    col_names = FALSE,
    list(nrow = 2 + n, ncol = which(LETTERS == "D"))
  )

  # range is open on the right
  # get row extent from range, column extent from range + data
  expect_dims(
    "C3:5",
    col_names = TRUE,
    list(nrow = 5, ncol = which(LETTERS == "C") + m - 1)
  )
  expect_dims(
    "C3:5",
    col_names = FALSE,
    list(nrow = 5, ncol = which(LETTERS == "C") + m - 1)
  )

  # range is open on left (trivially) and on the right
  # get row extent from range, column extent from the data
  expect_dims("5:7", col_names = TRUE, list(nrow = 7, ncol = m))
  expect_dims("5:7", col_names = FALSE, list(nrow = 7, ncol = m))

  # range is open on the top (trivially) and bottom
  # get row extent from data, column extent from range
  expect_dims(
    "B:H",
    col_names = TRUE,
    list(nrow = n + 1, ncol = which(LETTERS == "H"))
  )
  expect_dims(
    "B:H",
    col_names = FALSE,
    list(nrow = n, ncol = which(LETTERS == "H"))
  )

  # range is open on the bottom and right
  # get row extent from range + data, column extent from range + data
  expect_dims(
    cell_limits(c(2, 4), c(NA, NA)),
    col_names = TRUE,
    list(nrow = 2 + n + 1 - 1, ncol = 4 + m - 1)
  )
  expect_dims(
    cell_limits(c(2, 4), c(NA, NA)),
    col_names = FALSE,
    list(nrow = 2 + n - 1, ncol = 4 + m - 1)
  )
})

#' Add a named range
#'
#' Adds a named range. Not really ready for showtime yet, so not exported. But
#' I need it to (re)create the 'deaths' example Sheet.
#'
#' @noRd
#'
#' @eval param_ss()
#' @param name Name for the new named range.
#' @eval param_sheet(action = "SOMETHING")
#' @template range
#'
#' @template ss-return
#' @keywords internal
#' @examplesIf gs4_has_token()
#' dat <- data.frame(x = 1:3, y = letters[1:3])
#' ss <- gs4_create("range-add-named-demo", sheets = list(alpha = dat))
#'
#' ss %>%
#'   range_add_named("two_rows", sheet = "alpha", range = "A2:B3")
#'
#' # notice the 'two_rows' named range reported here
#' ss
#'
#' # clean up
#' gs4_find("range-add-named-demo") %>%
#'   googledrive::drive_trash()
range_add_named <- function(ss, name, sheet = NULL, range = NULL) {
  ssid <- as_sheets_id(ss)
  name <- check_string(name)
  maybe_sheet(sheet)
  check_range(range)

  x <- gs4_get(ssid)
  gs4_bullets(c(v = "Working in {.s_sheet {x$name}}."))

  # determine (work)sheet ------------------------------------------------------
  range_spec <- as_range_spec(
    range,
    sheet = sheet,
    sheets_df = x$sheets,
    nr_df = x$named_ranges
  )
  range_spec$sheet_name <- range_spec$sheet_name %||%
    first_visible_name(x$sheets)

  # form batch update request --------------------------------------------------
  req <- list(
    addNamedRange = new(
      "AddNamedRangeRequest",
      namedRange = as_NamedRange(range_spec, name = name)
    )
  )

  # do it ----------------------------------------------------------------------
  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = list(req)
    )
  )
  resp_raw <- request_make(req)
  reply <- gargle::response_process(resp_raw)
  reply <- pluck(reply, "replies", 1, "addNamedRange", "namedRange")
  reply <- new("NamedRange", !!!reply)
  # TODO: this would not be so janky if new_googlesheets4_spreadsheet() were
  #       factored in a way I could make better use of its logic
  reply <- as.list(as_tibble(reply))
  reply$sheet_name <- vlookup(
    reply$sheet_id,
    data = x$sheets,
    key = "id",
    value = "name"
  )
  A1_range <- qualified_A1(reply$sheet_name, do.call(make_cell_range, reply))
  gs4_bullets(c(
    v = "Created new range named {.range {reply$name}} \\
         representing {.range {A1_range}}."
  ))

  invisible(ssid)
}

# given the files owned by the googlesheets4-sheet-keeper service account,
# create/update an inventory file consulted by gs4_examples()

library(here)
library(googledrive)
library(tidyverse)
library(googlesheets4)

# auth with the special-purpose service account
gs4_auth(
  path = "~/.R/gargle/googlesheets4-sheet-keeper.json",
  scopes = "https://www.googleapis.com/auth/drive"
)
gs4_user()
drive_auth(token = gs4_token())
drive_user()

# exclude the inventory Sheet ... too meta!
dat <- drive_find(q = "not name contains 'gs4_example_and_test_sheets'")

if (anyDuplicated(dat$name)) {
  stop("Duplicated file names! You are making a huge mistake.")
}

dat <- dat %>%
  mutate(
    purpose = if_else(str_detect(name, "^googlesheets4"), "test", "example")
  ) %>%
  select(name, purpose, id) %>%
  arrange(purpose, name)

# record in local csv, because the visibility afforded by a plain old csv file
# is useful to me, e.g. easy to see change over time
write_csv(
  dat,
  file = here("inst", "extdata", "example_and_test_sheets.csv")
)

# initial creation

# ss <- gs4_create(
#   "gs4_example_and_test_sheets",
#   sheets = list(gs4_example_and_test_sheets = dat)
# )
# drive_share_anyone(ss)
# drive_publish(ss)

# as_id(ss)
# 1dSIZ2NkEPDWiEbsg9G80Hr9Xe7HZglEAPwGhVa-OSyA

# in future, write over the data to update
ssid <- as_sheets_id("1dSIZ2NkEPDWiEbsg9G80Hr9Xe7HZglEAPwGhVa-OSyA")
ss <- gs4_get(ssid)
ss

# range_autofit(ss)
drive_browse(ssid)

# creates an array of instances of RowData
as_RowData <- function(df, col_names = TRUE) {
  df_cells <- purrr::modify(df, as_CellData)
  df_rows <- pmap(df_cells, list)
  if (col_names) {
    df_rows <- c(list(as_CellData(names(df))), df_rows)
  }
  map(df_rows, ~ list(values = unname(.x)))
}

#' Flood or clear a range of cells
#'
#' `range_flood()` "floods" a range of cells with the same content.
#' `range_clear()` is a wrapper that handles the common special case of
#' clearing the cell value. Both functions, by default, also clear the format,
#' but this can be specified via `reformat`.
#'
#' @eval param_ss()
#' @eval param_sheet(action = "write into")
#' @template range
#' @param cell The value to fill the cells in the `range` with. If unspecified,
#'   the default of `NULL` results in clearing the existing value.
#' @template reformat
#'
#' @template ss-return
#' @export
#' @family write functions
#' @seealso Makes a `RepeatCellRequest`:
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#repeatcellrequest>
#'
#' @examplesIf gs4_has_token()
#' # create a data frame to use as initial data
#' df <- gs4_fodder(10)
#'
#' # create Sheet
#' ss <- gs4_create("range-flood-demo", sheets = list(df))
#'
#' # default behavior (`cell = NULL`): clear value and format
#' range_flood(ss, range = "A1:B3")
#'
#' # clear value but preserve format
#' range_flood(ss, range = "C1:D3", reformat = FALSE)
#'
#' # send new value
#' range_flood(ss, range = "4:5", cell = ";-)")
#'
#' # send formatting
#' # WARNING: use these unexported, internal functions at your own risk!
#' # This not (yet) officially supported, but it's possible.
#' blue_background <- googlesheets4:::CellData(
#'   userEnteredFormat = googlesheets4:::new(
#'     "CellFormat",
#'     backgroundColor = googlesheets4:::new(
#'       "Color",
#'       red = 159 / 255, green = 183 / 255, blue = 196 / 255
#'     )
#'   )
#' )
#' range_flood(ss, range = "I:J", cell = blue_background)
#'
#' # range_clear() is a shortcut where `cell = NULL` always
#' range_clear(ss, range = "9:9")
#' range_clear(ss, range = "10:10", reformat = FALSE)
#'
#' # clean up
#' gs4_find("range-flood-demo") %>%
#'   googledrive::drive_trash()
range_flood <- function(
  ss,
  sheet = NULL,
  range = NULL,
  cell = NULL,
  reformat = TRUE
) {
  ssid <- as_sheets_id(ss)
  maybe_sheet(sheet)
  check_range(range)
  check_bool(reformat)

  x <- gs4_get(ssid)
  gs4_bullets(c(v = "Editing {.s_sheet {x$name}}."))

  # determine (work)sheet ------------------------------------------------------
  range_spec <- as_range_spec(
    range,
    sheet = sheet,
    sheets_df = x$sheets,
    nr_df = x$named_ranges
  )
  range_spec$sheet_name <- range_spec$sheet_name %||%
    first_visible_name(x$sheets)
  s <- lookup_sheet(range_spec$sheet_name, sheets_df = x$sheets)
  gs4_bullets(c(v = "Editing sheet {.w_sheet {range_spec$sheet_name}}."))

  # prepare cell and field mask ------------------------------------------------
  # TODO: adapt here when CellData becomes a vctrs class
  if (is_CellData(cell)) {
    fields <- gargle::field_mask(cell)
  } else {
    cell <- as_CellData(cell %||% NA)[[1]]
    fields <- if (reformat) {
      "userEnteredValue,userEnteredFormat"
    } else {
      "userEnteredValue"
    }
  }

  # form batch update request --------------------------------------------------
  repeat_req <- list(
    repeatCell = new(
      "RepeatCellRequest",
      range = as_GridRange(range_spec),
      cell = cell,
      fields = fields
    )
  )

  # do it ----------------------------------------------------------------------
  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = list(repeat_req)
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)

  invisible(ssid)
}

#' @rdname range_flood
#' @export
range_clear <- function(ss, sheet = NULL, range = NULL, reformat = TRUE) {
  range_flood(
    ss = ss,
    sheet = sheet,
    range = range,
    reformat = reformat
  )
}

#' Append rows to a sheet
#'
#' Adds one or more new rows after the last row with data in a (work)sheet,
#' increasing the row dimension of the sheet if necessary.
#'
#' @eval param_ss()
#' @param data A data frame.
#' @eval param_sheet(action = "append to")
#'
#' @template ss-return
#' @export
#' @family write functions
#' @family worksheet functions
#' @seealso Makes an `AppendCellsRequest`:
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#AppendCellsRequest>
#'
#' @examplesIf gs4_has_token()
#' # we will recreate the table of "other" deaths from this example Sheet
#' (deaths <- gs4_example("deaths") %>%
#'   range_read(range = "other_data", col_types = "????DD"))
#'
#' # split the data into 3 pieces, which we will send separately
#' deaths_one <- deaths[1:5, ]
#' deaths_two <- deaths[6, ]
#' deaths_three <- deaths[7:10, ]
#'
#' # create a Sheet and send the first chunk of data
#' ss <- gs4_create("sheet-append-demo", sheets = list(deaths = deaths_one))
#'
#' # append a single row
#' ss %>% sheet_append(deaths_two)
#'
#' # append remaining rows
#' ss %>% sheet_append(deaths_three)
#'
#' # read and check against the original
#' deaths_replica <- range_read(ss, col_types = "????DD")
#' identical(deaths, deaths_replica)
#'
#' # clean up
#' gs4_find("sheet-append-demo") %>%
#'   googledrive::drive_trash()
sheet_append <- function(ss, data, sheet = 1) {
  check_data_frame(data)
  ssid <- as_sheets_id(ss)
  check_sheet(sheet)

  x <- gs4_get(ssid)
  gs4_bullets(c(v = "Writing to {.s_sheet {x$name}}."))

  s <- lookup_sheet(sheet, sheets_df = x$sheets)
  gs4_bullets(c(v = "Appending {nrow(data)} row{?s} to {.w_sheet {s$name}}."))

  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = prepare_rows(s$id, data),
      responseIncludeGridData = FALSE
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)

  invisible(ssid)
}

prepare_rows <- function(sheet_id, df) {
  list(
    appendCells = new(
      "AppendCellsRequest",
      sheetId = sheet_id,
      rows = as_RowData(df, col_names = FALSE), # an array of instances of RowData
      fields = "userEnteredValue,userEnteredFormat.numberFormat"
    )
  )
}

A1_char_class <- "[a-zA-Z0-9:$]"
compound_rx <- glue("(?<sheet>^.+)!(?<cell_range>{A1_char_class}+$)")
letter_part <- "[$]?[A-Za-z]{1,3}"
number_part <- "[$]?[0-9]{1,8}"
A1_rx <- glue("^{letter_part}{number_part}$|^{letter_part}$|^{number_part}$")
A1_decomp <- glue("(?<column>{letter_part})?(?<row>{number_part})?")

qualified_A1 <- function(sheet_name = NULL, cell_range = NULL) {
  n_missing <- is.null(sheet_name) + is.null(cell_range)
  if (n_missing == 2) {
    return()
  }
  sep <- if (n_missing == 0) "!" else ""
  # API docs: "For simplicity, it is safe to always surround the sheet name
  # with single quotes."
  as.character(
    glue("{sq_escape(sheet_name) %||% ''}{sep}{cell_range %||% ''}")
  )
}

as_sheets_range <- function(x) {
  stopifnot(inherits(x, what = "cell_limits"))
  # TODO: we don't show people providing sheet name via cell_limits
  #       so I proceed as if sheet is always specified elsewhere
  x$sheet <- NA_character_
  x <- resolve_limits(x)
  limits <- x[c("ul", "lr")]

  if (noNA(unlist(limits))) {
    return(cellranger::as.range(x, fo = "A1"))
  }

  # cellranger::as.range() does the wrong thing for everything below here,
  # i.e. returns NA
  # But we can make valid A1 ranges for the Sheets API in many cases.
  # Until cellranger is capable, we must do it in googlesheets4.

  if (allNA(unlist(limits))) {
    return(NULL)
  }

  row_limits <- map_int(limits, 1)
  col_limits <- map_int(limits, 2)

  if (allNA(col_limits) && noNA(row_limits)) {
    return(paste0(row_limits, collapse = ":"))
  }
  if (allNA(row_limits) && noNA(col_limits)) {
    return(paste0(cellranger::num_to_letter(col_limits), collapse = ":"))
  }

  if (noNA(limits$ul) && sum(is.na(limits$lr)) == 1) {
    ul <- paste0(cellranger::num_to_letter(col_limits[1]), row_limits[1])
    lr <- if (is.na(col_limits[2])) {
      row_limits[2]
    } else {
      cellranger::num_to_letter(col_limits[2])
    }
    return(paste0(c(ul, lr), collapse = ":"))
  }

  # if resolve_limits() is doing its job, we should never get here
  gs4_abort(c(
    "Can't express these {.cls cell_limits} as an A1 range:",
    # cell_limits doesn't have a format method :(
    x = utils::capture.output(print(x))
  ))
}

# think of cell_limits like so:
# ul = upper left  |  lr = lower right
# -----------------+------------------
#      start_row              end_row
#      start_col              end_col
# if start is specified, then so must be the end
#
# here we replace end_row or end_col in such cases with an actual number
#
# if provided, sheet_data is a list with two named elements:
#   * `grid_rows` = max row extent
#   * `grid_columns` = max col extent
# probably obtained like so:
# df <- gs4_get()$sheets
# df[df$name == sheet, c("grid_rows", "grid_columns")]
resolve_limits <- function(cell_limits, sheet_data = NULL) {
  # If no sheet_data, use theoretical maxima.
  # https://workspaceupdates.googleblog.com/2022/03/ten-million-cells-google-sheets.html
  # Rows: Max number of cells is 10 million. So that must be the maximum
  #       number of rows (imagine a spreadsheet with 1 sheet and 1 column).
  # Columns: Max col is "ZZZ" = cellranger::letter_to_num("ZZZ") = 18278
  MAX_ROW <- sheet_data$grid_rows %||% 10000000L
  MAX_COL <- sheet_data$grid_columns %||% 18278L

  limits <- c(cell_limits$ul, cell_limits$lr)
  if (noNA(limits) || allNA(limits)) {
    # rectangle is completely specified or completely unspecified
    return(cell_limits)
  }

  rlims <- function(cl) map_int(cl[c("ul", "lr")], 1)
  clims <- function(cl) map_int(cl[c("ul", "lr")], 2)

  # i:j, ?:j, i:?
  if (allNA(clims(cell_limits))) {
    cell_limits$ul[1] <- cell_limits$ul[1] %|% 1L
    cell_limits$lr[1] <- cell_limits$lr[1] %|% MAX_ROW
    return(cell_limits)
  }

  # X:Y, ?:Y, X:?
  if (allNA(rlims(cell_limits))) {
    cell_limits$ul[2] <- cell_limits$ul[2] %|% 1L
    cell_limits$lr[2] <- cell_limits$lr[2] %|% MAX_COL
    return(cell_limits)
  }

  # complete ul
  cell_limits$ul[1] <- cell_limits$ul[1] %|% 1L
  cell_limits$ul[2] <- cell_limits$ul[2] %|% 1L

  if (allNA(cell_limits$lr)) {
    # populate col of lr
    cell_limits$lr[2] <- cell_limits$lr[2] %|% MAX_COL
  }

  cell_limits
}

## Note: this function is NOT vectorized, x is scalar
as_cell_limits <- function(x) {
  check_character(x)
  check_length_one(x)
  ## match against <sheet name>!<A1 cell reference or range>?
  parsed <- rematch2::re_match(x, compound_rx)

  ## successful match (and parse)
  if (notNA(parsed$`.match`)) {
    cell_limits <- limits_from_range(parsed$cell_range)
    cell_limits$sheet <- parsed$sheet
    return(cell_limits)
  }

  ## failed to match
  ## two possibilities:
  ##   * An A1 cell reference or range
  ##   * Name of a sheet or named region
  if (all(grepl(A1_rx, strsplit(x, split = ":")[[1]]))) {
    limits_from_range(x)
  } else {
    ## TO THINK: I am questioning if this should even be allowed
    ## perhaps you MUST use sheet argument for this, not range?
    ## to be clear: we're talking about passing a sheet name or name of a
    ## named range, without a '!A1:C4' type of range as suffix
    cell_limits(sheet = x)
  }
  ## TODO: above is still not sophisticated enough to detect that
  ## A, AA, AAA (strings of length less than 4) and
  ## 1, 12, ..., 1234567 (numbers with less than 8 digits)
  ## are not, I believe, valid ranges
}

limits_from_range <- function(x) {
  x_split <- strsplit(x, ":")[[1]]
  if (!length(x_split) %in% 1:2) {
    gs4_abort("Invalid range: {.range {x}}")
  }
  if (!all(grepl(A1_rx, x_split))) {
    gs4_abort("Invalid range: {.range {x}}")
  }
  corners <- rematch2::re_match(x_split, A1_decomp)
  if (anyNA(corners$.match)) {
    gs4_abort("Invalid range: {.range {x}}")
  }
  corners$column <- ifelse(
    nzchar(corners$column),
    corners$column,
    NA_character_
  )
  corners$row <- ifelse(nzchar(corners$row), corners$row, NA_character_)
  corners$row <- as.integer(corners$row)
  if (nrow(corners) == 1) {
    corners <- corners[c(1, 1), ]
  }
  cellranger::cell_limits(
    ul = c(
      corners$row[1] %|% NA_integer_,
      cellranger::letter_to_num(corners$column[1]) %|% NA_integer_
    ),
    lr = c(
      corners$row[2] %|% NA_integer_,
      cellranger::letter_to_num(corners$column[2]) %|% NA_integer_
    )
  )
}

check_range <- function(range = NULL, call = caller_env()) {
  if (is.null(range) || inherits(range, "cell_limits") || is_string(range)) {
    return(range)
  }
  gs4_abort(
    "{.arg range} must be {.code NULL}, a string, or a {.cls cell_limits} \\
     object.",
    call = call
  )
}

## the `...` are used to absorb extra variables when this is used inside pmap()
make_cell_range <- function(
  start_row,
  end_row,
  start_column,
  end_column,
  sheet_name,
  ...
) {
  cl <- cellranger::cell_limits(
    ul = c(start_row, start_column),
    lr = c(end_row, end_column),
    sheet = glue::single_quote(sheet_name)
  )
  as_sheets_range(cl)
}

## A pair of functions for the (un)escaping of spreadsheet names
## for use in range strings like 'Sheet1'!A2:D4
sq_escape <- function(x) {
  if (is.null(x)) {
    return()
  }
  ## if string already starts and ends with single quote, pass it through
  is_not_quoted <- !map_lgl(x, ~ grepl("^'.*'$", .x))
  ## duplicate each single quote and protect string with single quotes
  x[is_not_quoted] <- paste0("'", gsub("'", "''", x[is_not_quoted]), "'")
  x
}

sq_unescape <- function(x) {
  if (is.null(x)) {
    return()
  }
  ## only modify if string starts and ends with single quote
  is_quoted <- map_lgl(x, ~ grepl("^'.*'$", .x))
  ## strip leading and trailing single quote and substitute 1 single quote
  ## for every pair of single quotes
  x[is_quoted] <- gsub("''", "'", sub("^'(.*)'$", "\\1", x[is_quoted]))
  x
}

# sq_escape() and sq_unescape() ----
test_that("sq_escape() and sq_unescape() pass NULL through", {
  expect_null(sq_escape(NULL))
  expect_null(sq_unescape(NULL))
})

test_that("sq_escape() does nothing if string already single-quoted", {
  x <- c("'abc'", "'ab'c'", "''")
  expect_identical(sq_escape(x), x)
})

test_that("sq_escape() duplicates single quotes and adds to start, end", {
  expect_identical(
    sq_escape(c("abc", "'abc", "abc'", "'a'bc", "'")),
    c("'abc'", "'''abc'", "'abc'''", "'''a''bc'", "''''")
  )
})

test_that("sq_unescape() does nothing if string is not single-quoted", {
  x <- c("abc", "'abc", "abc'", "a'bc", "'a'bc")
  expect_identical(sq_unescape(x), x)
})

test_that("sq_unescape() strips outer single quotes, de-duplicates inner", {
  expect_identical(
    sq_unescape(c("'abc'", "'''abc'", "'abc'''", "'''a''bc'", "''''")),
    c("abc", "'abc", "abc'", "'a'bc", "'")
  )
})

# qualified_A1() ----
test_that("qualified_A1 works", {
  expect_null(qualified_A1())
  expect_identical(qualified_A1("foo"), "'foo'")
  expect_identical(qualified_A1("foo", "A1"), "'foo'!A1")
  expect_identical(qualified_A1("'foo'"), "'foo'")
  expect_identical(qualified_A1(cell_range = "A1"), "A1")
})

# lookup_sheet_name() ----
test_that("lookup_sheet_name() requires sheet to be length-1 character or numeric", {
  expect_error(lookup_sheet_name(c("a", "b")), "length 1")
  expect_error(lookup_sheet_name(1:2), "length 1")
  expect_error(lookup_sheet_name(TRUE), "must be either")
})

test_that("lookup_sheet_name() errors if number is incompatible with sheet names", {
  sheets_df <- tibble::tibble(name = c("a", "foo", "z"))
  expect_error(lookup_sheet_name(4, sheets_df), "out-of-bounds")
  expect_error(lookup_sheet_name(0, sheets_df), "out-of-bounds")
})

test_that("lookup_sheet_name() consults sheet names, if given", {
  sheets_df <- tibble::tibble(name = c("a", "foo", "z"))
  expect_identical(lookup_sheet_name("foo", sheets_df), "foo")
  expect_error(
    lookup_sheet_name("nope", sheets_df),
    class = "googlesheets4_error_sheet_not_found"
  )
})

test_that("lookup_sheet_name() works with a number", {
  sheets_df <- tibble::tibble(name = c("a", "foo", "z"))
  expect_identical(lookup_sheet_name(2, sheets_df), "foo")
})

# resolve_limits() ----
test_that("resolve_limits() leaves these cases unchanged", {
  expect_no_change <- function(cl) expect_identical(resolve_limits(cl), cl)

  expect_no_change(cell_limits(c(2, 2), c(3, 3)))
  expect_no_change(cell_limits(c(NA, NA), c(NA, NA)))
  expect_no_change(cell_limits(c(2, NA), c(3, NA)))
  expect_no_change(cell_limits(c(NA, 2), c(NA, 3)))
  expect_no_change(cell_limits(c(2, 2), c(3, NA)))
  expect_no_change(cell_limits(c(2, 2), c(NA, 3)))
})

test_that("resolve_limits() completes a row- or column-only range", {
  expect_identical(
    resolve_limits(cell_limits(c(2, NA), c(NA, NA))),
    cell_limits(c(2, NA), c(10000000, NA))
  )
  expect_identical(
    # I now think it's a bug that cell_limits() fills in this start row
    resolve_limits(cell_limits(c(NA, NA), c(3, NA))),
    cell_limits(c(1, NA), c(3, NA))
  )
  expect_identical(
    resolve_limits(cell_limits(c(NA, 2), c(NA, NA))),
    cell_limits(c(NA, 2), c(NA, 18278))
  )
  expect_identical(
    # I now think it's a bug that cell_limits() fills in this start column
    resolve_limits(cell_limits(c(NA, NA), c(NA, 3))),
    cell_limits(c(NA, 1), c(NA, 3))
  )
})

test_that("resolve_limits() completes upper left cell", {
  expect_identical(
    resolve_limits(cell_limits(c(2, NA), c(NA, 3))),
    cell_limits(c(2, 1), c(NA, 3))
  )
  expect_identical(
    resolve_limits(cell_limits(c(NA, 2), c(3, NA))),
    cell_limits(c(1, 2), c(3, NA))
  )
  expect_identical(
    resolve_limits(cell_limits(c(NA, NA), c(3, 3))),
    cell_limits(c(1, 1), c(3, 3))
  )
  expect_identical(
    resolve_limits(cell_limits(c(2, NA), c(3, 3))),
    cell_limits(c(2, 1), c(3, 3))
  )
  expect_identical(
    resolve_limits(cell_limits(c(NA, 2), c(3, 3))),
    cell_limits(c(1, 2), c(3, 3))
  )
})

test_that("resolve_limits() populates column of lower right cell", {
  expect_identical(
    resolve_limits(cell_limits(c(2, 2), c(NA, NA))),
    cell_limits(c(2, 2), c(NA, 18278))
  )
})

# as_sheets_range ----

## "case numbers" refer to output produced by:
# tidyr::crossing(
#   start_row = c(NA, "start_row"), start_col = c(NA, "start_col"),
#   end_row = c(NA, "end_row"), end_col = c(NA, "end_col")
# )

test_that("as_sheets_range() works when all limits are given", {
  #  1 start_row start_col end_row end_col
  expect_identical(
    as_sheets_range(cell_limits(c(2, 2), c(3, 3))),
    "B2:C3"
  )
})

test_that("as_sheets_range() returns NULL when all limits are NA", {
  # 16 NA        NA        NA      NA
  expect_null(as_sheets_range(cell_limits()))
})

test_that("as_sheets_range() deals with row-only range", {
  #  6 start_row NA        end_row NA
  expect_identical(
    as_sheets_range(cell_limits(c(2, NA), c(3, NA))),
    "2:3"
  )
})

test_that("as_sheets_range() deals with column-only range", {
  # 11 NA        start_col NA      end_col
  expect_identical(
    as_sheets_range(cell_limits(c(NA, 2), c(NA, 3))),
    "B:C"
  )
})

test_that("as_sheets_range() deals when one of lr limits is missing", {
  #  2 start_row start_col end_row NA
  #  3 start_row start_col NA      end_col
  expect_identical(
    as_sheets_range(cell_limits(c(2, 2), c(3, NA))),
    "B2:3"
  )
  expect_identical(
    as_sheets_range(cell_limits(c(2, 2), c(NA, 3))),
    "B2:C"
  )
})

# TODO: I disabled this when switching to testthat 3e
# removing the mock of cellranger::cell_limits() didn't seem to hurt anything,
# which seems odd
#
# I have to think about cellranger soon, so revisit this when that happens
#
# commenting out, not skipping, because this is the only with_mock()

#test_that("as_sheets_range() errors for limits that should be fixed by resolve_limits()", {
# I think cellranger::cell_limits() should do much less.
# Already planning here for such a change there.
# Here's a very crude version of what I have in mind.
# cl <- function(ul, lr) {
#   structure(
#     list(ul = as.integer(ul), lr = as.integer(lr), sheet = NA_character_),
#     class = c("cell_limits", "list")
#   )
# }
#   with_mock(
#     resolve_limits = function(x) x,
#     `cellranger:::cell_limits` = function(ul, lr, sheet) cl(ul, lr), {
#       #  5 start_row NA        end_row end_col
#       expect_error(as_sheets_range(cell_limits(c(2, NA), c(3, 3))))
#       #  9 NA        start_col end_row end_col
#       expect_error(as_sheets_range(cell_limits(c(NA, 2), c(3, 3))))
#       # 13 NA        NA        end_row end_col
#       expect_error(as_sheets_range(cell_limits(c(NA, NA), c(3, 3))))
#       #  8 start_row NA        NA      NA
#       expect_error(as_sheets_range(cell_limits(c(2, NA), c(NA, NA))))
#       # 14 NA        NA        end_row NA
#       expect_error(as_sheets_range(cell_limits(c(NA, NA), c(2, NA))))
#       # 12 NA        start_col NA      NA
#       expect_error(as_sheets_range(cell_limits(c(NA, 2), c(NA, NA))))
#       # 15 NA        NA        NA      end_col
#       expect_error(as_sheets_range(cell_limits(c(NA, NA), c(NA, 3))))
#       # 10 NA        start_col end_row NA
#       expect_error(as_sheets_range(cell_limits(c(NA, 2), c(3, NA))))
#       #  7 start_row NA        NA      end_col
#       expect_error(as_sheets_range(cell_limits(c(2, NA), c(NA, 3))))
#       #  4 start_row start_col NA      NA
#       expect_error(as_sheets_range(cell_limits(c(2, 2), c(NA, NA))))
#     }
#   )
# })

#' Generate a Google Sheets API request
#'
#' @description Generate a request, using knowledge of the [Sheets
#'   API](https://developers.google.com/sheets/api/) from its Discovery
#'   Document (`https://www.googleapis.com/discovery/v1/apis/sheets/v4/rest`). Use
#'   [request_make()] to execute the request. Most users should, instead, use
#'   higher-level wrappers that facilitate common tasks, such as reading or
#'   writing worksheets or cell ranges. The functions here are intended for
#'   internal use and for programming around the Sheets API.
#'
#' @description `request_generate()` lets you provide the bare minimum of input.
#'   It takes a nickname for an endpoint and:
#'   * Uses the API spec to look up the `method`, `path`, and `base_url`.
#'   * Checks `params` for validity and completeness with respect to the
#'   endpoint. Uses `params` for URL endpoint substitution and separates
#'   remaining parameters into those destined for the body versus the query.
#'   * Adds an API key to the query if and only if `token = NULL`.
#'
#' @param endpoint Character. Nickname for one of the selected Sheets API v4
#'   endpoints built into googlesheets4. Learn more in [gs4_endpoints()].
#' @param params Named list. Parameters destined for endpoint URL substitution,
#'   the query, or the body.
#' @param key API key. Needed for requests that don't contain a token. The need
#'   for an API key in the absence of a token is explained in Google's document
#'   "Credentials, access, security, and identity"
#'   (`https://support.google.com/googleapi/answer/6158857?hl=en&ref_topic=7013279`).
#'   In order of precedence, these sources are consulted: the formal `key`
#'   argument, a `key` parameter in `params`, a user-configured API key set up
#'   with [gs4_auth_configure()] and retrieved with [gs4_api_key()].
#' @param token Set this to `NULL` to suppress the inclusion of a token. Note
#'   that, if auth has been de-activated via [gs4_deauth()],
#'   `gs4_token()` will actually return `NULL`.
#'
#' @return `list()`\cr Components are `method`, `url`, `body`, and `token`,
#'   suitable as input for [request_make()].
#' @export
#' @family low-level API functions
#' @seealso [gargle::request_develop()], [gargle::request_build()],
#'   [gargle::request_make()]
#' @examples
#' req <- request_generate(
#'   "sheets.spreadsheets.get",
#'   list(spreadsheetId = gs4_example("deaths")),
#'   key = "PRETEND_I_AM_AN_API_KEY",
#'   token = NULL
#' )
#' req
request_generate <- function(
  endpoint = character(),
  params = list(),
  key = NULL,
  token = gs4_token()
) {
  ept <- .endpoints[[endpoint]]
  if (is.null(ept)) {
    gs4_abort(c("Endpoint not recognized:", x = "{.field {endpoint}}"))
  }

  # if there are problems in `params`, such as a nonexistent item,
  # let's complain now
  force(params)

  ## modifications specific to googlesheets4 package
  params$key <- key %||%
    params$key %||%
    gs4_api_key() %||%
    gargle::tidyverse_api_key()

  req <- gargle::request_develop(
    endpoint = ept,
    params = params,
    base_url = attr(.endpoints, which = "base_url", exact = TRUE)
  )
  gargle::request_build(
    path = req$path,
    method = req$method,
    params = req$params,
    body = req$body,
    token = token,
    base_url = req$base_url
  )
}

test_that("can generate a basic request", {
  req <- request_generate(
    "sheets.spreadsheets.get",
    list(spreadsheetId = "abc123"),
    token = NULL
  )
  expect_identical(req$method, "GET")
  expect_match(
    req$url,
    "^https://sheets.googleapis.com/v4/spreadsheets/abc123\\?key=.+"
  )
  expect_null(req$token)
})

# currently just for development
# I'm generally auth'd as:
#  * as a service acct (which means I can't look at anything in the browser)
#  * with Drive and Sheets scope
#  * with googlesheets4 and googledrive
# so this is helpful for quickly granting anyone or myself specifically
# permission to read or write a Sheet I'm fiddling with in the browser or the
# API explorer
#
# Note defaults: role = "reader", type = "anyone"
# --> "anyone with the link" can view
#
# examples:
# gs4_share(ss)
# gs4_share(ss, type = "user", emailAddress = "jane@example.com")
# gs4_share(ss, type = "user", emailAddress = "jane@example.com", role = "writer")
gs4_share <- function(
  ss,
  ...,
  role = c(
    "reader",
    "commenter",
    "writer",
    "owner",
    "organizer"
  ),
  type = c("anyone", "user", "group", "domain")
) {
  check_gs4_email_is_drive_email()
  role <- match.arg(role)
  type <- match.arg(type)
  googledrive::drive_share(
    file = as_sheets_id(ss),
    role = role,
    type = type,
    ...
  )
}

new <- function(id, ...) {
  schema <- .tidy_schemas[[id]]
  if (is.null(schema)) {
    gs4_abort("Can't find a tidy schema with id {.field {id}}.")
  }
  dots <- list2(...)
  dots <- discard(dots, is.null)

  check_against_schema(dots, schema = schema)

  structure(
    dots,
    # explicit 'list' class is a bit icky but makes jsonlite happy
    # in various vctrs futures, this could need revisiting
    class = c(id_as_class(id), "googlesheets4_schema", "list"),
    schema = schema
  )
}

# TODO: if it proves necessary, this could do more meaningful checks
check_against_schema <- function(x, schema = NULL, id = NA_character_) {
  schema <- schema %||%
    .tidy_schemas[[id %|% id_from_class(x)]] %||%
    attr(x, "schema")
  if (is.null(schema)) {
    gs4_abort(
      "
      Trying to check an object of class {.cls {class(x)}}, \\
      but can't get a schema."
    )
  }
  stopifnot(is_dictionaryish(x))
  unexpected <- setdiff(names(x), schema$property)
  if (length(unexpected) > 0) {
    gs4_abort(c(
      "Properties not recognized for the {.field {attr(schema, 'id')}} schema:",
      bulletize(gargle_map_cli(unexpected), bullet = "x")
    ))
  }
  x
}

id_as_class <- function(id) glue("googlesheets4_schema_{id}")

id_from_class <- function(x) {
  m <- grepl("^googlesheets4_schema_", class(x))
  if (!any(m)) {
    return(NA_character_)
  }
  m <- which(m)[1]
  sub("^googlesheets4_schema_", "", class(x)[m])
}

# patch ----
patch <- function(x, ...) {
  UseMethod("patch")
}

#' @export
patch.default <- function(x, ...) {
  gs4_abort(
    "
    Don't know how to {.fun patch} an object of class {.cls {class(x)}}."
  )
}

#' @export
patch.googlesheets4_schema <- function(x, ...) {
  dots <- list2(...)
  dots <- discard(dots, is.null)
  x[names(dots)] <- dots
  check_against_schema(x)
}

test_that("new() errors for non-existing id", {
  expect_error(new("I_don't_exist"), "Can't find")
})

test_that("new() works (and doesn't require data)", {
  out <- new("Spreadsheet")
  expect_length(out, 0)
  expect_s3_class(out, "googlesheets4_schema_Spreadsheet")
  expect_s3_class(out, "googlesheets4_schema")
  expect_s3_class(attr(out, "schema"), "tbl_df")
})

test_that("new() accepts data expected for schema", {
  out <- new("Spreadsheet", spreadsheetId = "abc")
  expect_identical(out$spreadsheetId, "abc")
})

test_that("new() rejects data not expected for schema", {
  expect_snapshot(
    new("Spreadsheet", foofy = "blah"),
    error = TRUE
  )
  expect_snapshot(
    new("Spreadsheet", foofy = "blah", foo = "bar"),
    error = TRUE
  )
})

test_that("new() ignores NULL-valued inputs", {
  out <- new("GridRange", sheetId = 123, startRowIndex = 2, endRowIndex = NULL)
  expect_false(has_name(out, "endRowIndex"))
})

test_that("patch() fails informatively for non-schema input", {
  expect_error(patch(1), "Don't know how")
})

test_that("patch() with no data passes input through", {
  out <- new("Spreadsheet", spreadsheetId = "abc")
  expect_identical(out, patch(out))
})

test_that("patch() accepts data expected for schema", {
  expect_identical(
    new("Spreadsheet", spreadsheetId = "abc"),
    new("Spreadsheet") %>% patch(spreadsheetId = "abc")
  )
})

test_that("patch() rejects data not expected for schema", {
  x <- new("Spreadsheet")
  expect_error(patch(x, foofy = "blah"), "not recognized")
})

test_that("patch() overwrites existing data", {
  x <- new("Spreadsheet", spreadsheetId = "abc")
  x <- patch(x, spreadsheetId = "xyz")
  expect_identical(x$spreadsheetId, "xyz")
  expect_length(x, 1)
})

test_that("patch() retains classes", {
  x <- new("Spreadsheet")
  classes_in <- class(x)
  x <- patch(x, spreadsheetId = "abc")
  classes_out <- class(x)
  expect_identical(classes_in, classes_out)
})

test_that("patch() ignores NULL-valued inputs", {
  out <- new("GridRange", sheetId = 123) %>%
    patch(startRowIndex = 2, endRowIndex = NULL)
  expect_false(has_name(out, "endRowIndex"))
})

test_that("check_against_schema() errors when no schema can be found", {
  x <- structure(
    list(google_thing = "a"),
    class = c("googlesheets4_schema_SomeThing", "googlesheets4_schema", "list")
  )
  expect_snapshot(
    check_against_schema(x),
    error = TRUE
  )
})

test_that("id_from_class() works when schema class is present", {
  x <- structure(
    list(google_thing = "a"),
    class = c("googlesheets4_schema_SomeThing", "googlesheets4_schema", "list")
  )
  expect_equal(id_from_class(x), "SomeThing")
})

test_that("check_against_schema() errors if names aren't unique", {
  expect_error(
    check_against_schema(
      list(spreadsheetId = "abc", spreadsheetId = "def"),
      id = "Spreadsheet"
    ),
    "is_dictionaryish(x) is not TRUE",
    fixed = TRUE
  )
})

test_that("id_from_class() returns NA when schema class is absent", {
  x <- structure(list(google_thing = "a"), class = "list")
  expect_equal(id_from_class(x), NA_character_)
})

#' Read Sheet as CSV
#'
#' @description
#' This function uses a quick-and-dirty method to read a Sheet that bypasses the
#' Sheets API and, instead, parses a CSV representation of the data. This can be
#' much faster than [range_read()] -- noticeably so for "large" spreadsheets.
#' There are real downsides, though, so we recommend this approach only when the
#' speed difference justifies it. Here are the limitations we must accept to get
#' faster reading:
#' * Only formatted cell values are available, not underlying values or details
#'   on the formats.
#' * We can't target a named range as the `range`.
#' * We have no access to the data type of a cell, i.e. we don't know that it's
#'   logical, numeric, or datetime. That must be re-discovered based on the
#'   CSV data (or specified by the user).
#' * Auth and error handling have to be handled a bit differently internally,
#'   which may lead to behaviour that differs from other functions in
#'   googlesheets4.
#'
#' Note that the Sheets API is still used to retrieve metadata on the target
#' Sheet, in order to support range specification. `range_speedread()` also
#' sends an auth token with the request, unless a previous call to
#' [gs4_deauth()] has put googlesheets4 into a de-authorized state.
#'
#' @inheritParams range_read_cells
#' @param ... Passed along to the CSV parsing function (currently
#'   `readr::read_csv()`).
#'
#' @return A [tibble][tibble::tibble-package]
#' @export
#'
#' @examplesIf gs4_has_token()
#' if (require("readr")) {
#'   # since cell type is not available, use readr's col type specification
#'   range_speedread(
#'     gs4_example("deaths"),
#'     sheet = "other",
#'     range = "A5:F15",
#'     col_types = cols(
#'       Age = col_integer(),
#'       `Date of birth` = col_date("%m/%d/%Y"),
#'       `Date of death` = col_date("%m/%d/%Y")
#'     )
#'   )
#' }
#'
#' # write a Sheet that, by default, is NOT world-readable
#' (ss <- sheet_write(chickwts))
#'
#' # demo that range_speedread() sends a token, which is why we can read this
#' range_speedread(ss)
#'
#' # clean up
#' googledrive::drive_trash(ss)
range_speedread <- function(ss, sheet = NULL, range = NULL, skip = 0, ...) {
  check_installed("readr", "to use `range_speedread()`.")

  ssid <- as_sheets_id(ss)
  maybe_sheet(sheet)
  check_range(range)
  check_non_negative_integer(skip)

  x <- gs4_get(ssid)

  params <- list(
    spreadsheet_id = unclass(ssid),
    path = "export",
    format = "csv"
  )
  sheet_msg <- ""
  range_msg <- ""
  range_spec <- as_range_spec(
    range,
    sheet = sheet,
    skip = skip,
    sheets_df = x$sheets,
    nr_df = x$named_ranges
  )
  if (!is.null(range_spec$named_range)) {
    gs4_abort("{.fun range_speedread} cannot work with a named range.")
  }
  if (!is.null(range_spec$cell_limits)) {
    range_spec$cell_range <- as_sheets_range(range_spec$cell_limits)
  }
  if (!is.null(range_spec$cell_range)) {
    params[["range"]] <- range_spec$cell_range
    range_msg <- ", range {.range {range_spec$cell_range}}"
  }
  if (!is.null(range_spec$sheet_name)) {
    s <- lookup_sheet(range_spec$sheet_name, sheets_df = x$sheets)
    params[["gid"]] <- s$id
    sheet_msg <- ", sheet {.w_sheet {range_spec$sheet_name}}"
  }
  msg <- glue(
    "
    Reading from {.s_sheet {x$name}}<<sheet_msg>><<range_msg>>.",
    .open = "<<",
    .close = ">>"
  )
  gs4_bullets(c(v = msg))

  token <- gs4_token() %||% list()

  req <- gargle::request_build(
    path = "spreadsheets/d/{spreadsheet_id}/{path}",
    params = params,
    base_url = "https://docs.google.com"
  )
  gs4_bullets(c(i = "Export URL: {.url {req$url}}"))

  response <- httr::GET(req$url, config = token)
  stopifnot(identical(httr::http_type(response), "text/csv"))
  readr::read_csv(httr::content(response, type = "raw"), ...)
}

#  input: a named list, usually an instance of googlesheets4_schema_Spreadsheet
# output: instance of googlesheets4_spreadsheet, which is actually useful
new_googlesheets4_spreadsheet <- function(x = list()) {
  # fmt: skip
  ours_theirs <- list(
    spreadsheet_id  = "spreadsheetId",
    spreadsheet_url = "spreadsheetUrl",
    name            = list("properties", "title"),
    locale          = list("properties", "locale"),
    time_zone       = list("properties", "timeZone")
  )
  out <- map(ours_theirs, ~ pluck(x, !!!.x, .default = "<unknown>"))

  if (!is.null(x$sheets)) {
    sheets <- map(x$sheets, ~ new("Sheet", !!!.x))
    sheet_properties <- map(sheets, as_tibble)
    out$sheets <- do.call(rbind, sheet_properties)

    protected_ranges <- map(sheets, "protectedRanges")
    protected_ranges <- purrr::flatten(protected_ranges)
    protected_ranges <- map(protected_ranges, ~ new("ProtectedRange", !!!.x))
    protected_ranges <- map(protected_ranges, as_tibble)
    out$protected_ranges <- do.call(rbind, protected_ranges)
  }

  if (!is.null(x$namedRanges)) {
    named_ranges <- map(x$namedRanges, ~ new("NamedRange", !!!.x))
    named_ranges <- map(named_ranges, as_tibble)
    named_ranges <- do.call(rbind, named_ranges)

    # if there is only 1 sheet, sheetId might not be sent!
    # https://github.com/tidyverse/googlesheets4/issues/29
    needs_sheet_id <- is.na(named_ranges$sheet_id)
    if (any(needs_sheet_id)) {
      # if sheetId is missing, I assume it's the "first" (visible?) sheet
      named_ranges$sheet_id[needs_sheet_id] <- first_visible_id(out$sheets)
    }
    named_ranges$sheet_name <- vlookup(
      named_ranges$sheet_id,
      data = out$sheets,
      key = "id",
      value = "name"
    )

    # https://github.com/tidyverse/googlesheets4/issues/175
    # dysfunctional named ranges are possible and should not prevent us from
    # dealing with a Sheet
    possibly_make_cell_range <- purrr::possibly(
      make_cell_range,
      otherwise = NA_character_
    )
    named_ranges$cell_range <- pmap_chr(named_ranges, possibly_make_cell_range)
    named_ranges$A1_range <- qualified_A1(
      named_ranges$sheet_name,
      named_ranges$cell_range
    )
    named_ranges$A1_range[is.na(named_ranges$cell_range)] <- NA_character_

    out$named_ranges <- named_ranges
  }

  structure(out, class = c("googlesheets4_spreadsheet", "list"))
}

#' @export
format.googlesheets4_spreadsheet <- function(x, ...) {
  cli::cli_div(theme = gs4_theme())
  # fmt: skip
  meta <- list(
  `Spreadsheet name` = cli::format_inline("{.s_sheet {x$name}}"),
                  ID = as.character(x$spreadsheet_id),
              Locale = x$locale,
         `Time zone` = x$time_zone,
       `# of sheets` = if (rlang::has_name(x, "sheets")) {
         as.character(nrow(x$sheets))
       } else {
         "<unknown>"
       }
  )
  if (!is.null(x$named_ranges)) {
    meta <- c(meta, `# of named ranges` = as.character(nrow(x$named_ranges)))
  }
  if (!is.null(x$protected_ranges)) {
    meta <- c(
      meta,
      `# of protected ranges` = as.character(nrow(x$protected_ranges))
    )
  }
  out <- c(
    cli::cli_format_method(
      cli::cli_h1("<googlesheets4_spreadsheet>")
    ),
    glue("{fr(names(meta))}: {fl(meta)}")
  )

  if (!is.null(x$sheets)) {
    col1 <- fr(c(
      "(Sheet name)",
      sapply(
        gargle::gargle_map_cli(x$sheets$name, template = "{.w_sheet <<x>>}"),
        cli::format_inline
      )
    ))
    col2 <- c(
      "(Nominal extent in rows x columns)",
      glue_data(x$sheets, "{grid_rows} x {grid_columns}")
    )
    out <- c(
      out,
      cli::cli_format_method(
        cli::cli_h1("<sheets>")
      ),
      glue_data(list(col1 = col1, col2 = col2), "{col1}: {col2}")
    )
  }

  if (!is.null(x$named_ranges)) {
    col1 <- fr(c(
      "(Named range)",
      sapply(
        gargle::gargle_map_cli(
          x$named_ranges$name,
          template = "{.range <<x>>}"
        ),
        cli::format_inline
      )
    ))
    col2 <- fl(c("(A1 range)", x$named_ranges$A1_range))
    out <- c(
      out,
      cli::cli_format_method(
        cli::cli_h1("<named ranges>")
      ),
      glue_data(list(col1 = col1, col2 = col2), "{col1}: {col2}")
    )
  }

  out
}

#' @export
print.googlesheets4_spreadsheet <- function(x, ...) {
  cat(format(x), sep = "\n")
  invisible(x)
}

#' (Over)write new data into a Sheet
#'
#' @description
#'
#' This is one of the main ways to write data with googlesheets4. This function
#' writes a data frame into a (work)sheet inside a (spread)Sheet. The target
#' sheet is styled as a table:
#'   * Special formatting is applied to the header row, which holds column
#'     names.
#'   * The first row (header row) is frozen.
#'   * The sheet's dimensions are set to "shrink wrap" the `data`.
#'
#' If no existing Sheet is specified via `ss`, this function delegates to
#' [`gs4_create()`] and the new Sheet's name is randomly generated. If that's
#' undesirable, call [`gs4_create()`] directly to get more control.
#'
#' If no `sheet` is specified or if `sheet` doesn't identify an existing sheet,
#' a new sheet is added to receive the `data`. If `sheet` specifies an existing
#' sheet, it is effectively overwritten! All pre-existing values, formats, and
#' dimensions are cleared and the targeted sheet gets new values and dimensions
#' from `data`.
#'
#' This function goes by two names, because we want it to make sense in two
#' contexts:
#' * `write_sheet()` evokes other table-writing functions, like
#'   `readr::write_csv()`. The `sheet` here technically refers to an individual
#'   (work)sheet (but also sort of refers to the associated Google
#'   (spread)Sheet).
#' * `sheet_write()` is the right name according to the naming convention used
#'   throughout the googlesheets4 package.
#'
#' `write_sheet()` and `sheet_write()` are equivalent and you can use either one.
#'
#' @param data A data frame. If it has zero rows, we send one empty pseudo-row
#'   of data, so that we can apply the usual table styling. This empty row goes
#'   away (gets filled, actually) the first time you send more data with
#'   [sheet_append()].
#' @eval param_ss()
#' @eval param_sheet(action = "write into")
#'
#' @template ss-return
#' @export
#' @family write functions
#' @family worksheet functions
#'
#' @examplesIf gs4_has_token()
#' df <- data.frame(
#'   x = 1:3,
#'   y = letters[1:3]
#' )
#'
#' # specify only a data frame, get a new Sheet, with a random name
#' ss <- write_sheet(df)
#' read_sheet(ss)
#'
#' # clean up
#' googledrive::drive_trash(ss)
#'
#' # create a Sheet with some initial, placeholder data
#' ss <- gs4_create(
#'   "sheet-write-demo",
#'   sheets = list(alpha = data.frame(x = 1), omega = data.frame(x = 1))
#' )
#'
#' # write df into its own, new sheet
#' sheet_write(df, ss = ss)
#'
#' # write mtcars into the sheet named "omega"
#' sheet_write(mtcars, ss = ss, sheet = "omega")
#'
#' # get an overview of the sheets
#' sheet_properties(ss)
#'
#' # view your magnificent creation in the browser
#' gs4_browse(ss)
#'
#' # clean up
#' gs4_find("sheet-write-demo") %>%
#'   googledrive::drive_trash()
sheet_write <- function(data, ss = NULL, sheet = NULL) {
  data_quo <- enquo(data)
  data <- eval_tidy(data_quo)
  check_data_frame(data)

  # no Sheet provided --> call gs4_create() ---------------------------------
  if (is.null(ss)) {
    if (quo_is_symbol(data_quo)) {
      sheet <- sheet %||% as_name(data_quo)
    }
    if (is.null(sheet)) {
      return(gs4_create(sheets = data))
    } else {
      check_string(sheet)
      return(gs4_create(sheets = list2(!!sheet := data)))
    }
  }

  # finish checking inputs -----------------------------------------------------
  ssid <- as_sheets_id(ss)
  maybe_sheet(sheet)

  # retrieve spreadsheet metadata ----------------------------------------------
  x <- gs4_get(ssid)
  gs4_bullets(c(v = "Writing to {.s_sheet {x$name}}."))

  # no `sheet` ... but maybe we can name the sheet after the data --------------
  if (is.null(sheet) && quo_is_symbol(data_quo)) {
    candidate <- as_name(data_quo)
    # accept proposed name iff it does not overwrite existing sheet
    if (!is.null(candidate)) {
      m <- match(candidate, x$sheets$name)
      sheet <- if (is.na(m)) candidate else NULL
    }
  }

  # initialize the batch update requests and the target sheet s ----------------
  requests <- list()
  s <- NULL

  # ensure there's a target sheet, ready to receive data -----------------------
  if (!is.null(sheet)) {
    s <- tryCatch(
      lookup_sheet(sheet, sheets_df = x$sheets),
      googlesheets4_error_sheet_not_found = function(cnd) NULL
    )
  }
  if (is.null(s)) {
    x <- sheet_add_impl_(ssid, sheet_name = sheet)
    s <- lookup_sheet(nrow(x$sheets), sheets_df = x$sheets)
  } else {
    # create request to clear the data and formatting in pre-existing sheet
    requests <- c(
      requests,
      list(bureq_clear_sheet(s$id))
    )
  }
  gs4_bullets(c(v = "Writing to sheet {.w_sheet {s$name}}."))

  # create request to write data frame into sheet ------------------------------
  requests <- c(
    requests,
    prepare_df(s$id, data)
  )

  # do it ----------------------------------------------------------------------
  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = requests,
      responseIncludeGridData = FALSE
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)

  invisible(ssid)
}

#' @rdname sheet_write
#' @export
write_sheet <- sheet_write

#' @export
as_tibble.googlesheets4_schema_SheetProperties <- function(x, ...) {
  # fmt: skip
  tibble::tibble(
    # TODO: open question whether I should explicitly unescape title here
    name         =  glean_chr(x, "title"),
    index        =  glean_int(x, "index"),
    id           =  glean_int(x, "sheetId"),
    type         =  glean_chr(x, "sheetType"),
    visible      = !glean_lgl(x, "hidden", .default = FALSE),
    grid_rows    =  glean_int(x, c("gridProperties", "rowCount")),
    grid_columns =  glean_int(x, c("gridProperties", "columnCount"))
  )
}

library(tidyverse)
library(googledrive)
library(googlesheets4)
library(readxl)

googlesheets4:::gs4_auth_docs()
gs4_user()

x <- gs4_find()

if ("deaths" %in% x$name) {
  stop("Yo, 'deaths' already exists. Are you sure you want to do this?")
}

# to work on an existing 'deaths' Sheet
# ss <- gs4_find("deaths") %>% as_sheets_id()

# to create one from scratch
deaths_xlsx <- readxl_example("deaths.xlsx")
ss <- drive_upload(deaths_xlsx, name = "deaths", type = "spreadsheet")

googlesheets4:::gs4_share(ss)

gs4_browse(ss)

googlesheets4:::range_add_named(ss, name = "arts_data", range = "arts!A5:F15")
googlesheets4:::range_add_named(ss, name = "other_data", range = "other!A5:F15")

ss

unclass(ss)

library(tidyverse)
devtools::load_all() # I assume we're in googlesheets4 source
library(googledrive)

gs4_auth_testing()

ss <- test_sheet_create()
gs4_browse(ss)

# TODO: I created this worksheet in the browser; add code here once possible

# I riffed on the original Sheet provided by @nadnudus in
# https://github.com/tidyverse/googlesheets4/issues/4
# ssid <- as_sheets_id("1UbdlyITXLvsxQt6kpszu5gfiDmF5Q-wOrNC7l4E9jOg")
# gs4_browse(ss)
# I copied the "legend" worksheet.
# Added a note to C1.
# Added a comment (as rstudio jenny) to C2.

library(devtools)
library(fs)
library(here)
library(purrr)

# setup and/or clean
here("examples") %>%
  dir_create() %>%
  dir_ls() %>%
  file_delete()

rd_files <- here("man") %>% dir_ls(regex = "[.][Rr]d$")

do_one <- function(x) {
  print(path_file(x))
  tmp <- file_temp()
  tools::Rd2ex(x, out = tmp)
  if (file_exists(tmp)) {
    cat(
      readLines(tmp),
      file = here("examples", "googlesheets4-examples.R"),
      append = TRUE, sep = "\n"
    )
  } else {
    print("  ^ no examples!!")
  }
}
walk(rd_files, do_one)

#' Relocate one or more (work)sheets
#'
#' @description
#' Move (work)sheets around within a (spread)Sheet. The outcome is most
#' predictable for these common and simple use cases:
#' * Reorder and move one or more sheets to the front.
#' * Move a single sheet to a specific (but arbitrary) location.
#' * Move multiple sheets to the back with `.after = 100` (`.after` can be
#'   any number greater than or equal to the number of sheets).
#'
#' If your relocation task is more complicated and you are puzzled by the
#' result, break it into a sequence of simpler calls to
#' `sheet_relocate()`.
#'
#' @eval param_ss()
#' @eval param_sheet(
#'   action = "relocate",
#'   "You can pass a vector to move multiple sheets at once or even a list,",
#'   "if you need to mix names and positions."
#' )
#' @param .before,.after Specification of where to locate the sheets(s)
#'   identified by `sheet`. Exactly one of `.before` and `.after` must be
#'   specified. Refer to an existing sheet by name (via a string) or by position
#'   (via a number).
#'
#' @template ss-return
#' @export
#' @family worksheet functions
#' @seealso
#' Constructs a batch of `UpdateSheetPropertiesRequest`s (one per sheet):
#' * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#UpdateSheetPropertiesRequest>
#'
#' @examplesIf gs4_has_token()
#' sheet_names <- c("alfa", "bravo", "charlie", "delta", "echo", "foxtrot")
#' ss <- gs4_create("sheet-relocate-demo", sheets = sheet_names)
#' sheet_names(ss)
#'
#' # move one sheet, forwards then backwards
#' ss %>%
#'   sheet_relocate("echo", .before = "bravo") %>%
#'   sheet_names()
#' ss %>%
#'   sheet_relocate("echo", .after = "delta") %>%
#'   sheet_names()
#'
#' # reorder and move multiple sheets to the front
#' ss %>%
#'   sheet_relocate(list("foxtrot", 4)) %>%
#'   sheet_names()
#'
#' # put the sheets back in the original order
#' ss %>%
#'   sheet_relocate(sheet_names) %>%
#'   sheet_names()
#'
#' # reorder and move multiple sheets to the back
#' ss %>%
#'   sheet_relocate(c("bravo", "alfa", "echo"), .after = 10) %>%
#'   sheet_names()
#'
#' # clean up
#' gs4_find("sheet-relocate-demo") %>%
#'   googledrive::drive_trash()
sheet_relocate <- function(
  ss,
  sheet,
  .before = if (is.null(.after)) 1,
  .after = NULL
) {
  ssid <- as_sheets_id(ss)
  walk(sheet, check_sheet)
  maybe_sheet(.before)
  maybe_sheet(.after)

  x <- gs4_get(ssid)
  gs4_bullets(c(v = "Relocating sheets in {.s_sheet {x$name}}."))

  if (!is.null(.before)) {
    sheet <- rev(sheet)
  }
  requests <- map(
    sheet,
    ~ make_UpdateSheetPropertiesRequest(
      sheet = .x,
      .before = .before,
      .after = .after,
      sheets_df = x$sheets,
      call = quote(sheet_relocate())
    )
  )
  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = requests
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)

  invisible(ssid)
}

make_UpdateSheetPropertiesRequest <- function(
  sheet,
  .before,
  .after,
  sheets_df,
  call = caller_env()
) {
  s <- lookup_sheet(sheet, sheets_df = sheets_df, call = call)
  index <- resolve_index(sheets_df, .before, .after, call = call)
  sp <- new("SheetProperties", sheetId = s$id, index = index)
  update_req <- new(
    "UpdateSheetPropertiesRequest",
    properties = sp,
    fields = gargle::field_mask(sp)
  )
  list(updateSheetProperties = update_req)
}

#' Get data about (work)sheets
#'
#' Reveals full metadata or just the names for the (work)sheets inside a
#' (spread)Sheet.
#'
#' @eval param_ss()
#'
#' @return
#'   * `sheet_properties()`: A tibble with one row per (work)sheet.
#'   * `sheet_names()`: A character vector of (work)sheet names.
#' @export
#' @family worksheet functions
#' @examplesIf gs4_has_token()
#' ss <- gs4_example("gapminder")
#' sheet_properties(ss)
#' sheet_names(ss)
sheet_properties <- function(ss) {
  x <- gs4_get(ss)
  pluck(x, "sheets")
}

#' @export
#' @rdname sheet_properties
sheet_names <- function(ss) {
  sheet_properties(ss)$name
}

# ctype = cell or column type
# most types are valid for a cell or a column
# however, a couple are valid only for cells or only for a column

#                       Type can be   Type can be  Type can be
# shortcode             discovered    guessed for  imposed on
#    = ctype            from a cell   a column     a column
# fmt: skip
.ctypes <- c(
  `_` = "COL_SKIP",      # --          no           yes
  `-` = "COL_SKIP",
        "CELL_BLANK",    # yes         no           no
  l   = "CELL_LOGICAL",  # yes         yes          yes
  i   = "CELL_INTEGER",  # no          no           yes
  d   = "CELL_NUMERIC",  # yes         yes          yes
  n   = "CELL_NUMERIC",  #
  D   = "CELL_DATE",     # yes         no           yes
  t   = "CELL_TIME",     # yes         no           yes
  `T` = "CELL_DATETIME", # yes         yes          yes
  c   = "CELL_TEXT",     # yes         yes          yes
  C   = "COL_CELL",      # --          no           yes
  L   = "COL_LIST",      # --          yes          yes
  `?` = "COL_GUESS"      # --          --           --
)

# TODO: add to above:
# CELL_DURATION
# COL_FACTOR

# this generic is "dumb": it only reports ctype
# it doesn't implement any logic about guessing, coercion, etc.
ctype <- function(x, ...) {
  UseMethod("ctype")
}

#' @export
ctype.NULL <- function(x, ...) {
  abort_unsupported_conversion(x, to = "ctype")
}

#' @export
ctype.SHEETS_CELL <- function(x, ...) {
  out <- class(x)[[1]]
  if (out %in% .ctypes) {
    out
  } else {
    NA_character_
  }
}

#' @export
ctype.character <- function(x, ...) .ctypes[x]

#' @export
ctype.list <- function(x, ...) {
  out <- rep_along(x, NA_character_)
  is_SHEETS_CELL <- map_lgl(x, inherits, what = "SHEETS_CELL")
  out[is_SHEETS_CELL] <- map_chr(x[is_SHEETS_CELL], ctype)
  out
}

#' @export
ctype.default <- function(x, ...) {
  abort_unsupported_conversion(x, to = "ctype")
}

# fmt: skip
.discovered_to_effective_type <- c(
  # If discovered   Then effective
  # cell type is:   cell type is:
  CELL_BLANK       = "CELL_BLANK",
  CELL_LOGICAL     = "CELL_LOGICAL",
  CELL_INTEGER     = "CELL_NUMERIC",  ## integers are jsonlite being helpful
  CELL_NUMERIC     = "CELL_NUMERIC",
  CELL_DATE        = "CELL_DATETIME", ## "date" is just a format in Sheets
  CELL_TIME        = "CELL_DATETIME", ## "time" is just a format in Sheets
  CELL_DATETIME    = "CELL_DATETIME",
  CELL_TEXT        = "CELL_TEXT"
)

# input:  cell type, presumably discovered
# output: effective cell type
#
# Where do we use this?
#   * To choose cell-specific parser when col type is COL_LIST == "L"
#   * Pre-processing cell types prior to forming a consensus for an entire
#     column when col type is COL_GUESS = "?"
# This is the where we store type-guessing fiddliness that is specific to
# Google Sheets.
effective_cell_type <- function(ctype) .discovered_to_effective_type[ctype]

# input:  a ctype
# output: vector of ctypes that can hold such input with no data loss, going
#         from most generic (list) to most specific (type of that cell)
# examples:
# CELL_LOGICAL --> COL_LIST, CELL_NUMERIC, CELL_INTEGER, CELL_LOGICAL
# CELL_DATE --> COL_LIST, CELL_DATETIME, CELL_DATE
# CELL_BLANK --> NULL
# fmt: skip
admissible_types <- function(x) {
  z <- c(
    CELL_LOGICAL  = "CELL_INTEGER",
    CELL_INTEGER  = "CELL_NUMERIC",
    CELL_NUMERIC  = "COL_LIST",

    CELL_DATE     = "CELL_DATETIME",
    CELL_DATETIME = "COL_LIST",

    CELL_TIME     = "COL_LIST",

    CELL_TEXT     = "COL_LIST"
  )
  if (x[[1]] == "COL_LIST") {
    return(x)
  }
  if (!x[[1]] %in% names(z)) {
    return()
  }
  c(admissible_types(z[[x[[1]]]]), x)
}

# find the most specific ctype that is admissible for a pair of ctypes
# the limiting case is COL_LIST
# HOWEVER use ctypes that are good for cells, i.e. "two blanks make a blank"
upper_type <- function(x, y) {
  upper_bound(admissible_types(x), admissible_types(y)) %||% "CELL_BLANK"
}

# find the most specific ctype that is admissible for a set of ctypes
# HOWEVER use ctypes that are good for columns, i.e. "two blanks make a
# logical"
consensus_col_type <- function(ctype) {
  out <- Reduce(upper_type, unique(ctype), init = "CELL_BLANK")
  blank_to_logical(out)
}

blank_to_logical <- function(ctype) {
  modify_if(ctype, ~ identical(.x, "CELL_BLANK"), ~"CELL_LOGICAL")
}

# input: an instance of CellData
# https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets#CellData
# returns same, but applies a class vector:
#   [1] a ctype, inspired by the CellType enum in readxl
#   [2] SHEETS_CELL
apply_ctype <- function(cell_list, na = "", trim_ws = TRUE) {
  ctypes <- map_chr(cell_list, infer_ctype, na = na, trim_ws = trim_ws)
  map2(cell_list, ctypes, ~ structure(.x, class = c(.y, "SHEETS_CELL")))
}

infer_ctype <- function(cell, na = "", trim_ws = TRUE) {
  # Blank cell criteria
  #   * cell is NULL or list()
  #   * cell has no effectiveValue
  #   * formattedValue matches an `na` string
  if (
    length(cell) == 0 ||
      length(cell[["effectiveValue"]]) == 0 ||
      is_na_string(cell[["formattedValue"]], na = na, trim_ws = trim_ws)
  ) {
    return("CELL_BLANK")
  }

  effective_type <- .extended_value[[names(cell[["effectiveValue"]])]]

  if (!identical(effective_type, "number")) {
    return(switch(
      effective_type,
      error = "CELL_BLANK",
      string = "CELL_TEXT",
      boolean = "CELL_LOGICAL",
      formula = {
        cli::cli_warn(
          "
          Internal warning from googlesheets4: \\
          Cell has formula as effectiveValue. \\
          I thought this was impossible!"
        )
        "CELL_TEXT"
      },
      gs4_abort(
        "Unhandled effective_type: {.field {effective_type}}",
        .internal = TRUE
      )
    ))
  }
  # only numeric cells remain

  nf_type <- pluck(
    cell,
    "effectiveFormat",
    "numberFormat",
    "type",
    # in theory, should consult hosting spreadsheet for a default format
    # if that's absent, should consult locale (of spreadsheet? user? unclear)
    # for now, I punt on this
    .default = "NUMBER"
  )
  .number_types[[nf_type]]
}

# userEnteredValue and effectiveValue hold an instance of ExtendedValue
# https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets#ExtendedValue
# {
#   // Union field value can be only one of the following:
#   "numberValue": number,
#   "stringValue": string,
#   "boolValue": boolean,
#   "formulaValue": string,
#   "errorValue": {
#     object(ErrorValue)
#   },
#   // End of list of possible types for union field value.
# }
# fmt: skip
.extended_value <- c(
   numberValue = "number",
   stringValue = "string",
     boolValue = "boolean",
  formulaValue = "formula",  # hypothesis: this is impossible in effectiveValue
    errorValue = "error"
)

# fmt: skip
.number_types <- c(
  TEXT       = "CELL_NUMERIC",
  NUMBER     = "CELL_NUMERIC",
  PERCENT    = "CELL_NUMERIC",
  CURRENCY   = "CELL_NUMERIC",
  SCIENTIFIC = "CELL_NUMERIC",
  # on the R side, all of the above are treated as numeric
  # no current reason to distinguish them, for col type guessing or coercion
  DATE       = "CELL_DATE",
  TIME       = "CELL_TIME",
  DATE_TIME  = "CELL_DATETIME"
)

is_na_string <- function(x, na = "", trim_ws = TRUE) {
  if (length(na) == 0) {
    return(FALSE)
  }
  fv <- if (trim_ws) ws_trim(x) else x
  any(fv == na)
}

# compares x[i] to y[i] and returns the last element where they are equal
# example:
# upper_bound(c("a", "b"), c("a", "b", "c")) is "b"
upper_bound <- function(x, y) {
  nx <- length(x)
  ny <- length(y)
  # these brackets make covr happy
  if (nx + ny == 0) {
    return()
  }
  if (nx == 0) {
    return(y[[ny]])
  }
  if (ny == 0) {
    return(x[[nx]])
  }
  comp <- seq_len(min(nx, ny))
  # TODO: if our DAG were more complicated, I think this would need to be
  # based on a set operation
  res <- x[comp] == y[comp]
  if (!any(res)) {
    return()
  }
  x[[max(which(res))]]
}

test_that("ctype() errors for unanticipated inputs", {
  expect_error(ctype(NULL))
  expect_error(ctype(data.frame(cell = "cell")))
})

test_that("ctype() works on a SHEET_CELL, when it should", {
  expect_identical(
    ctype(structure(1, class = c("wut", "SHEETS_CELL"))),
    NA_character_
  )
  expect_identical(
    ctype(structure(1, class = c("CELL_NUMERIC", "SHEETS_CELL"))),
    "CELL_NUMERIC"
  )
})

test_that("ctype() works on shortcodes, when it should", {
  expect_equal(
    unname(ctype(c("?", "-", "n", "z", "D"))),
    c("COL_GUESS", "COL_SKIP", "CELL_NUMERIC", NA, "CELL_DATE")
  )
})

test_that("ctype() works on lists, when it should", {
  list_of_cells <- list(
    structure(1, class = c("CELL_NUMERIC", "SHEETS_CELL")),
    "nope",
    NULL,
    structure(1, class = c("wut", "SHEETS_CELL")),
    structure(1, class = c("CELL_TEXT", "SHEETS_CELL"))
  )
  expect_equal(
    ctype(list_of_cells),
    c("CELL_NUMERIC", NA, NA, NA, "CELL_TEXT")
  )
})

test_that("effective_cell_type() doesn't just pass ctype through", {
  ## neither the API nor JSON has a proper way to convey integer-ness
  expect_equal(unname(effective_cell_type("CELL_INTEGER")), "CELL_NUMERIC")
  ## conversion to date or time is lossy, so never guess that
  expect_equal(unname(effective_cell_type("CELL_DATE")), "CELL_DATETIME")
  expect_equal(unname(effective_cell_type("CELL_TIME")), "CELL_DATETIME")
})

test_that("consensus_col_type() implements our type coercion DAG", {
  expect_identical(
    consensus_col_type(c("CELL_TEXT", "CELL_TEXT")),
    "CELL_TEXT"
  )
  expect_identical(
    consensus_col_type(c("CELL_LOGICAL", "CELL_NUMERIC")),
    "CELL_NUMERIC"
  )
  expect_identical(
    consensus_col_type(c("CELL_LOGICAL", "CELL_DATE")),
    "COL_LIST"
  )
  expect_identical(
    consensus_col_type(c("CELL_DATE", "CELL_DATETIME")),
    "CELL_DATETIME"
  )
  expect_identical(
    consensus_col_type(c("CELL_TEXT", "CELL_BLANK")),
    "CELL_TEXT"
  )
  expect_identical(consensus_col_type("CELL_TEXT"), "CELL_TEXT")
  expect_identical(consensus_col_type("CELL_BLANK"), "CELL_LOGICAL")
})

#' Change the size of a (work)sheet
#'
#' Changes the number of rows and/or columns in a (work)sheet.
#'
#' @eval param_ss()
#' @eval param_sheet(action = "resize")
#' @param nrow,ncol Desired number of rows or columns, respectively. The default
#'   of `NULL` means to leave unchanged.
#' @param exact Logical, indicating whether to impose `nrow` and `ncol` exactly
#'   or to treat them as lower bounds. If `exact = FALSE`,
#'   `sheet_resize()` can only add cells. If `exact = TRUE`, cells can be
#'   deleted and their contents are lost.
#'
#' @template ss-return
#' @export
#' @family worksheet functions
#' @seealso Makes an `UpdateSheetPropertiesRequest`:
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#UpdateSheetPropertiesRequest>
#'
#' @examplesIf gs4_has_token()
#' # create a Sheet with the default initial worksheet
#' (ss <- gs4_create("sheet-resize-demo"))
#'
#' # see (work)sheet dims
#' sheet_properties(ss)
#'
#' # no resize occurs
#' sheet_resize(ss, nrow = 2, ncol = 6)
#'
#' # reduce sheet size
#' sheet_resize(ss, nrow = 5, ncol = 7, exact = TRUE)
#'
#' # add rows
#' sheet_resize(ss, nrow = 7)
#'
#' # add columns
#' sheet_resize(ss, ncol = 10)
#'
#' # add rows and columns
#' sheet_resize(ss, nrow = 9, ncol = 12)
#'
#' # re-inspect (work)sheet dims
#' sheet_properties(ss)
#'
#' # clean up
#' gs4_find("sheet-resize-demo") %>%
#'   googledrive::drive_trash()
sheet_resize <- function(
  ss,
  sheet = NULL,
  nrow = NULL,
  ncol = NULL,
  exact = FALSE
) {
  ssid <- as_sheets_id(ss)
  maybe_sheet(sheet)
  maybe_non_negative_integer(nrow)
  maybe_non_negative_integer(ncol)
  check_bool(exact)

  x <- gs4_get(ssid)
  s <- lookup_sheet(sheet, sheets_df = x$sheets)
  gs4_bullets(c(
    v = "Resizing sheet {.w_sheet {s$name}} in {.s_sheet {x$name}}."
  ))

  bureq <- prepare_resize_request(
    s,
    nrow_needed = nrow,
    ncol_needed = ncol,
    exact = exact
  )

  if (is.null(bureq)) {
    gs4_bullets(c(
      i = "No need to change existing dims ({s$grid_rows} x {s$grid_columns})."
    ))
    return(invisible(ssid))
  }

  new_grid_properties <- pluck(
    bureq,
    "updateSheetProperties",
    "properties",
    "gridProperties"
  )
  new_nrow <- pluck(new_grid_properties, "rowCount") %||% s$grid_rows
  new_ncol <- pluck(new_grid_properties, "columnCount") %||% s$grid_columns

  gs4_bullets(c(
    v = "Changing dims: ({s$grid_rows} x {s$grid_columns}) --> \\
         ({new_nrow} x {new_ncol})."
  ))

  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = list(bureq)
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)

  invisible(ssid)
}

prepare_resize_request <- function(
  sheet_info,
  nrow_needed,
  ncol_needed,
  exact = FALSE
) {
  nrow_sheet <- sheet_info$grid_rows
  ncol_sheet <- sheet_info$grid_columns

  new_dims <- c(
    make_dim_patch(nrow_sheet, nrow_needed, "nrow", exact),
    make_dim_patch(ncol_sheet, ncol_needed, "ncol", exact)
  )

  if (length(new_dims) == 0) {
    NULL
  } else {
    bureq_set_grid_properties(
      sheetId = sheet_info$id,
      nrow = new_dims$nrow,
      ncol = new_dims$ncol,
      frozenRowCount = NULL
    )
  }
}

make_dim_patch <- function(current, target, nm, exact = FALSE) {
  out <- list()
  if (is.null(target)) {
    return(out)
  }
  patch_needed <- (isTRUE(exact) && current != target) || current < target
  if (patch_needed) {
    out[[nm]] <- target
  }
  out
}

# ---- nm_fun ----
me_ <- nm_fun("TEST-sheet_resize")

# ---- tests ----
test_that("sheet_resize() works", {
  skip_if_offline()
  skip_if_no_token()

  ss <- local_ss(me_())
  local_gs4_loud()

  # no resize occurs
  expect_message(sheet_resize(ss, nrow = 2, ncol = 6), "No need") %>%
    suppressMessages()

  # reduce sheet size
  suppressMessages(sheet_resize(ss, nrow = 5, ncol = 7, exact = TRUE))
  props <- sheet_properties(ss)
  expect_equal(props$grid_rows, 5)
  expect_equal(props$grid_columns, 7)
})

test_that("prepare_resize_request() works for resize & no resize", {
  n <- 3
  m <- 5
  sheet_info <- list(grid_rows = n, grid_columns = m)

  # (n - 1, n, n + 1) x (m - 1, m, m + 1) x (TRUE, FALSE)
  # 3 * 3 * 2 = 18 combinations

  # exact = FALSE
  df <- expand.grid(
    nrow_needed = n + -1:1,
    ncol_needed = m + -1:1,
    exact = FALSE
  )
  req <- pmap(df, prepare_resize_request, sheet_info = sheet_info)
  grid_properties <- purrr::map(
    req,
    c("updateSheetProperties", "properties", "gridProperties")
  )

  # sheet is big enough --> no resize request
  purrr::walk(
    grid_properties[df$nrow_needed <= n & df$ncol_needed <= m],
    expect_null
  )

  # not enough rows
  purrr::walk(
    grid_properties[df$nrow_needed > n],
    ~ expect_true(has_name(.x, "rowCount"))
  )

  # not enough columns
  purrr::walk(
    grid_properties[df$ncol_needed > m],
    ~ expect_true(has_name(.x, "columnCount"))
  )

  # exact = TRUE
  df <- expand.grid(
    nrow_needed = n + -1:1,
    ncol_needed = m + -1:1,
    exact = TRUE
  )
  req <- pmap(df, prepare_resize_request, sheet_info = sheet_info)
  grid_properties <- purrr::map(
    req,
    c("updateSheetProperties", "properties", "gridProperties")
  )

  # sheet has correct size --> no resize request
  purrr::walk(
    grid_properties[df$nrow_needed == n & df$ncol_needed == m],
    expect_null
  )

  # not enough rows or too many rows
  purrr::walk(
    grid_properties[df$nrow_needed != n],
    ~ expect_true(has_name(.x, "rowCount"))
  )

  # not enough columns or too many columns
  purrr::walk(
    grid_properties[df$ncol_needed != m],
    ~ expect_true(has_name(.x, "columnCount"))
  )
})

#' ---
#' output: github_document
#' ---

#+ error = TRUE

devtools::load_all(".")

req <- request_generate(
  "spreadsheets.get",
  ## ID of 'googlesheets4-design-exploration', but replaced last 2 chars w/ '-'
  list(spreadsheetId = "1xTUxWGcFLtDIHoYJ1WsjQuLmpUtBf--8Bcu5lQ302--"),
  token = NULL
)
raw_resp <- request_make(req)
response_process(raw_resp)

ct <- httr::content(raw_resp)
str(ct)

## range_spec is an "internal-use only" S3 class ----
new_range_spec <- function(...) {
  l <- list2(...)
  structure(
    list(
      sheet_name = l$sheet_name %||% NULL,
      named_range = l$named_range %||% NULL,
      cell_range = l$cell_range %||% NULL,
      cell_limits = l$cell_limits %||% NULL,
      shim = FALSE,
      sheets_df = l$sheets_df %||% NULL,
      nr_df = l$nr_df %||% NULL
    ),
    # useful when debugging range specification, but otherwise this is TMI
    # .input = l$.input,
    class = "range_spec"
  )
}

as_range_spec <- function(x, ...) {
  UseMethod("as_range_spec")
}

#' @export
as_range_spec.default <- function(x, ...) {
  gs4_abort(c(
    "Can't make a range suitable for the Sheets API from the supplied \\
     {.arg range}.",
    x = "{.arg range} has class {.cls {class(x)}}.",
    i = "{.arg range} must be {.code NULL}, a string, or \\
         a {.cls cell_limits} object."
  ))
}

## as_range_spec.character ----

# anticipated inputs to the character method for x (= range)
# **** means "doesn't matter, never consulted"
#
# sheet   range         skip
# --------------------------------------
# ****    Sheet1!A1:B2  ****
# ****    Named_range   ****
# ****    Sheet1        i     weird, but I guess we roll with it (re-dispatch)
#         A1:B2         ****
# Sheet1  A1:B2         ****
# 3       A1:B2         ****
#' @export
as_range_spec.character <- function(
  x,
  ...,
  sheet = NULL,
  skip = 0,
  sheets_df = NULL,
  nr_df = NULL
) {
  check_length_one(x)

  out <- new_range_spec(
    sheets_df = sheets_df,
    nr_df = nr_df,
    .input = list(
      sheet = sheet,
      range = x,
      skip = skip
    )
  )

  m <- rematch2::re_match(x, compound_rx)

  # range looks like: Sheet1!A1:B2
  if (notNA(m[[".match"]])) {
    out$sheet_name <- lookup_sheet_name(m$sheet, sheets_df)
    out$cell_range <- m$cell_range
    out$shim <- TRUE
    return(out)
  }

  # check if range matches a named range
  m <- match(x, nr_df$name)
  if (notNA(m)) {
    out$named_range <- x
    return(out)
  }

  # check if range matches a sheet name
  # API docs: "When a named range conflicts with a sheet's name, the named range
  # is preferred."
  m <- match(x, sheets_df$name)
  if (notNA(m)) {
    # Re-dispatch as if provided as `sheet`. Which it should have been.
    return(as_range_spec(NULL, sheet = x, skip = skip, sheets_df = sheets_df))
  }

  # range must be in A1 notation
  m <- grepl(A1_rx, strsplit(x, split = ":")[[1]])
  if (!all(m)) {
    gs4_abort(c(
      "{.arg range} doesn't appear to be a range in A1 notation, a named \\
       range, or a sheet name:",
      x = "{.range {x}}"
    ))
  }
  out$cell_range <- x
  if (!is.null(sheet)) {
    out$sheet_name <- lookup_sheet_name(sheet, sheets_df)
  }
  out$shim <- TRUE
  out
}

## as_range_spec.NULL ----

# anticipated inputs to the NULL method for x (= range)
#
# sheet       skip
# --------------------------------------
#             0     This is what "nothing" looks like. Send nothing.
# Sheet1 / 2  0     Send sheet name.
#             >0    Express skip request in cell_limits object and re-dispatch.
# Sheet1 / 2  >0    <same as previous>
#' @export
as_range_spec.NULL <- function(
  x,
  ...,
  sheet = NULL,
  skip = 0,
  sheets_df = NULL
) {
  out <- new_range_spec(
    sheets_df = sheets_df,
    .input = list(sheet = sheet, skip = skip)
  )

  if (skip < 1) {
    if (!is.null(sheet)) {
      out$sheet_name <- lookup_sheet_name(sheet, sheets_df)
    }
    return(out)
  }

  as_range_spec(
    cell_rows(c(skip + 1, NA)),
    sheet = sheet,
    sheets_df = sheets_df,
    shim = FALSE
  )
}

## as_range_spec.cell_limits ----

# anticipated inputs to the cell_limits method for x (= range)
#
# sheet       range
# --------------------------------------
#             cell_limits   Send A1 representation of cell_limits. Let the API
#                           figure out the sheet. API docs imply it will be the
#                           "first visible sheet".
# Sheet1 / 2  cell_limits   Resolve sheet name, make A1 range, send combined
#                           result.
#' @export
as_range_spec.cell_limits <- function(
  x,
  ...,
  shim = TRUE,
  sheet = NULL,
  sheets_df = NULL
) {
  out <- new_range_spec(
    sheets_df = sheets_df,
    .input = list(sheet = sheet, range = x, shim = shim)
  )
  out$cell_limits <- x
  if (!is.null(sheet)) {
    out$sheet_name <- lookup_sheet_name(sheet, sheets_df)
  }
  out$shim <- shim
  out
}

#' @export
format.range_spec <- function(x, ...) {
  is_df <- names(x) %in% c("sheets_df", "nr_df")
  x[is_df & !map_lgl(x, is.null)] <- "<provided>"
  glue("{fr(names(x))}: {x}")
}

#' @export
print.range_spec <- function(x, ...) {
  cat(format(x), sep = "\n")
  invisible(x)
}

as_A1_range <- function(x) {
  stopifnot(inherits(x, "range_spec"))

  if (!is.null(x$named_range)) {
    return(x$named_range)
  }

  if (!is.null(x$cell_limits)) {
    x$cell_range <- as_sheets_range(x$cell_limits)
  }

  qualified_A1(x$sheet_name, x$cell_range)
}

# has been useful during development, at times
# sheets_A1_range <- function(ss,
#                             sheet = NULL,
#                             range = NULL,
#                             skip = 0) {
#   ssid <- as_sheets_id(ss)
#   maybe_sheet(sheet)
#   check_range(range)
#   check_non_negative_integer(skip)
#
#   # retrieve spreadsheet metadata ----------------------------------------------
#   x <- gs4_get(ssid)
#   gs4_bullets(c(i = "Spreadsheet name: {.s_sheet {x$name}}"))
#
#   # range specification --------------------------------------------------------
#   range_spec <- as_range_spec(
#     range, sheet = sheet, skip = skip,
#     sheets_df = x$sheets, nr_df = x$named_ranges
#   )
#   A1_range <- as_A1_range(range_spec)
#   gs4_bullets(c(i = "A1 range {.range {A1_range}}"))
#
#   range_spec
# }

# as_range_spec() ----
test_that("as_range_spec() rejects hopeless input", {
  expect_error(as_range_spec(3), "Can't make a range")
})

test_that("as_range_spec() can deal with nothingness", {
  spec <- as_range_spec(NULL)
  expect_true(all(map_lgl(spec, ~ is.null(.x) || isFALSE(.x))))
})

test_that("as_range_spec() partitions 'Sheet1!A1:B2'", {
  sheets_df <- tibble::tibble(name = "Sheet1")

  spec <- as_range_spec("Sheet1!A1:B2", sheets_df = sheets_df)
  expect_identical(spec$sheet_name, "Sheet1")
  expect_identical(spec$cell_range, "A1:B2")
  expect_true(spec$shim)

  spec <- as_range_spec("'Sheet1'!A5:A", sheets_df = sheets_df)
  # make sure we store unescaped name in range_spec
  expect_identical(spec$sheet_name, "Sheet1")
  expect_identical(spec$cell_range, "A5:A")
  expect_true(spec$shim)
})

test_that("as_range_spec() seeks a named range, then a sheet name", {
  nr_df <- tibble::tibble(name = c("a", "thingy", "z"))

  spec <- as_range_spec("thingy", nr_df = nr_df)
  expect_null(spec$sheet_name)
  expect_identical(spec$named_range, "thingy")
  expect_false(spec$shim)

  spec <- as_range_spec("thingy", nr_df = nr_df, sheets_df = nr_df)
  expect_null(spec$sheet_name)
  expect_identical(spec$named_range, "thingy")
  expect_false(spec$shim)

  spec <- as_range_spec(
    "thingy",
    nr_df = tibble::tibble(name = letters[1:3]),
    sheets_df = nr_df
  )
  expect_null(spec$named_range)
  expect_identical(spec$sheet_name, "thingy")
  expect_false(spec$shim)
})

test_that("A1 range is detected, w/ or w/o sheet", {
  spec <- as_range_spec("1:2")
  expect_identical(spec$cell_range, "1:2")
  expect_true(spec$shim)

  sheets_df <- tibble::tibble(name = LETTERS[1:3])
  spec <- as_range_spec("1:2", sheet = 3, sheets_df = sheets_df)
  expect_identical(spec$sheet_name, "C")
  expect_identical(spec$cell_range, "1:2")
  expect_true(spec$shim)

  spec <- as_range_spec("1:2", sheet = "B", sheets_df = sheets_df)
  expect_identical(spec$sheet_name, "B")
  expect_identical(spec$cell_range, "1:2")
  expect_true(spec$shim)
})

test_that("skip is converted to equivalent cell limits", {
  spec <- as_range_spec(x = NULL, skip = 1)
  expect_equal(spec$cell_limits, cell_rows(c(2, NA)))
})

test_that("cell_limits input works, w/ or w/o sheet", {
  spec <- as_range_spec(cell_rows(1:2))
  expect_equal(spec$cell_limits, cell_rows(1:2))
  expect_true(spec$shim)

  sheets_df <- tibble::tibble(name = LETTERS[1:3])

  spec <- as_range_spec(cell_rows(1:2), sheet = 3, sheets_df = sheets_df)
  expect_equal(spec$sheet_name, "C")
  expect_equal(spec$cell_limits, cell_rows(1:2))
  expect_true(spec$shim)

  spec <- as_range_spec(cell_rows(1:2), sheet = "B", sheets_df = sheets_df)
  expect_equal(spec$sheet_name, "B")
  expect_equal(spec$cell_limits, cell_rows(1:2))
  expect_true(spec$shim)
})

test_that("invalid range is rejected", {
  # no named ranges or sheet names for lookup --> interpret as A1
  expect_error(
    as_range_spec("thingy"),
    "doesn't appear to be"
  )

  expect_error(
    as_range_spec("thingy", nr_names = "nope", sheet_names = "nah"),
    "doesn't appear to be"
  )
})

test_that("unresolvable sheet raises error", {
  expect_gs4_error(as_range_spec("A5:A", sheet = 3), "Can't look up")
  expect_gs4_error(as_range_spec(x = NULL, sheet = 3), "Can't look up")
  sheets_df <- tibble::tibble(name = LETTERS[1:3])
  expect_error(
    as_range_spec(x = NULL, sheet = "nope", sheets_df = sheets_df),
    class = "googlesheets4_error_sheet_not_found"
  )
  expect_error(
    as_range_spec("A5:A", sheet = "nope", sheets_df = sheets_df),
    class = "googlesheets4_error_sheet_not_found"
  )
  expect_error(
    as_range_spec("nope!A5:A", sheets_df = sheets_df),
    class = "googlesheets4_error_sheet_not_found"
  )
})

# as_A1_range() ----
test_that("as_A1_range() works", {
  expect_null(as_A1_range(new_range_spec()))

  expect_equal(as_A1_range(new_range_spec(sheet_name = "Sheet1")), "'Sheet1'")

  expect_equal(as_A1_range(new_range_spec(named_range = "abc")), "abc")

  expect_equal(as_A1_range(new_range_spec(cell_range = "B3:D9")), "B3:D9")
  expect_equal(
    as_A1_range(new_range_spec(sheet_name = "Sheet1", cell_range = "A1")),
    "'Sheet1'!A1"
  )

  rs <- new_range_spec(cell_limits = cell_cols(3:5))
  expect_equal(as_A1_range(rs), "C:E")

  rs <- new_range_spec(sheet_name = "Sheet1", cell_limits = cell_rows(2:3))
  expect_equal(as_A1_range(rs), "'Sheet1'!2:3")
})

lookup_sheet <- function(
  sheet = NULL,
  sheets_df,
  visible = NA,
  call = caller_env()
) {
  maybe_sheet(sheet, call = call)
  if (is.null(sheets_df)) {
    gs4_abort(
      "Can't look up, e.g., sheet name or id without sheet metadata.",
      call = call
    )
  }

  if (isTRUE(visible)) {
    sheets_df <- sheets_df[sheets_df$visible, ]
  }

  if (is.null(sheet)) {
    first_sheet <- which.min(sheets_df$index)
    return(as.list(sheets_df[first_sheet, ]))
  }
  # sheet is a string or an integer

  if (is.character(sheet)) {
    sheet <- sq_unescape(sheet)
    m <- match(sheet, sheets_df$name)
    if (is.na(m)) {
      gs4_abort(
        c("Can't find a sheet with this name:", x = "{.w_sheet {sheet}}"),
        sheet = sheet,
        # there is some usage where we throw this error, but it is OK
        # and we use tryCatch()
        # that's why we apply the sub-class
        class = "googlesheets4_error_sheet_not_found",
        call = call
      )
    }
    return(as.list(sheets_df[m, ]))
  }
  # sheet is an integer

  m <- as.integer(sheet)
  if (!(m %in% seq_len(nrow(sheets_df)))) {
    gs4_abort(
      c(
        "There {?is/are} {nrow(sheets_df)} sheet{?s}:",
        x = "Requested sheet number is out-of-bounds: {m}"
      ),
      call = call
    )
  }
  as.list(sheets_df[m, ])
}

first_sheet <- function(sheets_df, visible = NA) {
  lookup_sheet(sheet = NULL, sheets_df = sheets_df, visible = visible)
}

first_visible <- function(sheets_df) first_sheet(sheets_df, visible = TRUE)

first_visible_id <- function(sheets_df) {
  first_sheet(sheets_df, visible = TRUE)$id
}

first_visible_name <- function(sheets_df) {
  first_sheet(sheets_df, visible = TRUE)$name
}

lookup_sheet_name <- function(sheet, sheets_df) {
  s <- lookup_sheet(sheet = sheet, sheets_df = sheets_df)
  s$name
}

check_sheet <- function(sheet, arg = caller_arg(sheet), call = caller_env()) {
  check_length_one(sheet, arg = arg, call = call)
  if (!is.character(sheet) && !is.numeric(sheet)) {
    gs4_abort(
      c(
        "{.arg {arg}} must be either {.cls character} (sheet name) or \\
         {.cls numeric} (sheet number):",
        x = "{.arg {arg}} has class {.cls {class(sheet)}}."
      ),
      call = call
    )
  }
  sheet
}

maybe_sheet <- function(
  sheet = NULL,
  arg = caller_arg(sheet),
  call = caller_env()
) {
  if (is.null(sheet)) {
    sheet
  } else {
    check_sheet(sheet, arg = arg, call = call)
  }
}

#' Normalize user input re: (work)sheet names and/or data
#'
#' @param sheets_quo Quosure containing user input re: how to populate
#'   (work)sheets.
#'
#' @return A list with 2 equal-sized components, `name` and `value`. Size =
#'   number of (work)sheets.
#' @keywords internal
#' @noRd
enlist_sheets <- function(sheets_quo) {
  sheets <- eval_tidy(sheets_quo)

  null_along <- function(x) vector(mode = "list", length = length(x))

  if (is.null(sheets)) {
    return(NULL)
  }

  if (is.character(sheets)) {
    return(list(name = sheets, value = null_along(sheets)))
  }

  if (inherits(sheets, "data.frame")) {
    if (quo_is_symbol(sheets_quo)) {
      return(list(name = as_name(sheets_quo), value = list(sheets)))
    } else {
      return(list(name = list(NULL), value = list(sheets)))
    }
  }

  if (is_list(sheets)) {
    nms <- if (is_named(sheets)) names(sheets) else null_along(sheets)
    return(list(name = nms, value = unname(sheets)))
  }

  # we should never get here, so not a user-facing message
  gs4_abort("Invalid input for (work)sheet(s).")
}

test_that("enlist_sheets() works", {
  df1 <- data.frame(x = 1L)
  df2 <- data.frame(x = 2L)
  df_list <- list(df1 = df1, df2 = df2)
  f <- function(sheets = NULL) enlist_sheets(enquo(sheets))

  expect_null(f())
  expect_identical(
    f(c("string_1", "string_2")),
    list(name = c("string_1", "string_2"), value = list(NULL, NULL))
  )
  expect_identical(
    f(df1),
    list(name = "df1", value = list(data.frame(x = 1L)))
  )
  expect_identical(
    f(list(df1, df2)),
    list(name = list(NULL, NULL), value = list(df1, df2))
  )
  expect_identical(
    f(list(df1 = df1, df2 = df2)),
    list(name = c("df1", "df2"), value = list(df1, df2))
  )
  expect_identical(
    f(df_list),
    f(list(df1 = df1, df2 = df2))
  )
  expect_identical(
    f(data.frame(x = 1L)),
    list(name = list(NULL), value = list(data.frame(x = 1L)))
  )
})

#' @param skip Minimum number of rows to skip before reading anything, be it
#'   column names or data. Leading empty rows are automatically skipped, so this
#'   is a lower bound. Ignored if `range` is given.

new_formula <- function(x = character()) {
  vec_assert(x, character())
  new_vctr(x, class = "googlesheets4_formula")
}

#' Class for Google Sheets formulas
#'
#' In order to write a formula into Google Sheets, you need to store it as an
#' object of class `googlesheets4_formula`. This is how we distinguish a
#' "regular" character string from a string that should be interpreted as a
#' formula. `googlesheets4_formula` is an S3 class implemented using the [vctrs
#' package](https://vctrs.r-lib.org/articles/s3-vector.html).
#'
#' @param x Character.
#'
#' @return An S3 vector of class `googlesheets4_formula`.
#' @export
#' @family write functions
#'
#' @examplesIf gs4_has_token()
#' dat <- data.frame(x = c(1, 5, 3, 2, 4, 6))
#'
#' ss <- gs4_create("gs4-formula-demo", sheets = dat)
#' ss
#'
#' summaries <- tibble::tribble(
#'   ~desc, ~summaries,
#'   "max", "=max(A:A)",
#'   "sum", "=sum(A:A)",
#'   "min", "=min(A:A)",
#'   "sparkline", "=SPARKLINE(A:A, {\"color\", \"blue\"})"
#' )
#'
#' # explicitly declare a column as `googlesheets4_formula`
#' summaries$summaries <- gs4_formula(summaries$summaries)
#' summaries
#'
#' range_write(ss, data = summaries, range = "C1", reformat = FALSE)
#'
#' miscellany <- tibble::tribble(
#'   ~desc, ~example,
#'   "hyperlink", "=HYPERLINK(\"http://www.google.com/\",\"Google\")",
#'   "image", "=IMAGE(\"https://www.google.com/images/srpr/logo3w.png\")"
#' )
#' miscellany$example <- gs4_formula(miscellany$example)
#' miscellany
#'
#' sheet_write(miscellany, ss = ss)
#'
#' # clean up
#' gs4_find("gs4-formula-demo") %>%
#'   googledrive::drive_trash()
gs4_formula <- function(x = character()) {
  x <- vec_cast(x, character())
  new_formula(x)
}

#' @importFrom methods setOldClass
setOldClass(c("googlesheets4_formula", "vctrs_vctr"))

#' @export
vec_ptype_abbr.googlesheets4_formula <- function(x, ...) {
  "fmla"
}

#' @method vec_ptype2 googlesheets4_formula
#' @export vec_ptype2.googlesheets4_formula
#' @export
#' @rdname googlesheets4-vctrs
vec_ptype2.googlesheets4_formula <- function(x, y, ...) {
  UseMethod("vec_ptype2.googlesheets4_formula", y)
}

#' @method vec_ptype2.googlesheets4_formula default
#' @export
vec_ptype2.googlesheets4_formula.default <- function(
  x,
  y,
  ...,
  x_arg = "x",
  y_arg = "y"
) {
  vec_default_ptype2(x, y, x_arg = x_arg, y_arg = y_arg)
}

#' @method vec_ptype2.googlesheets4_formula googlesheets4_formula
#' @export
vec_ptype2.googlesheets4_formula.googlesheets4_formula <- function(x, y, ...) {
  new_formula()
}

#' @method vec_ptype2.googlesheets4_formula character
#' @export
vec_ptype2.googlesheets4_formula.character <- function(x, y, ...) character()

#' @method vec_ptype2.character googlesheets4_formula
#' @export
vec_ptype2.character.googlesheets4_formula <- function(x, y, ...) character()

#' @method vec_cast googlesheets4_formula
#' @export vec_cast.googlesheets4_formula
#' @export
#' @rdname googlesheets4-vctrs
vec_cast.googlesheets4_formula <- function(x, to, ...) {
  UseMethod("vec_cast.googlesheets4_formula")
}

#' @method vec_cast.googlesheets4_formula default
#' @export
vec_cast.googlesheets4_formula.default <- function(x, to, ...) {
  vec_default_cast(x, to)
}

#' @method vec_cast.googlesheets4_formula googlesheets4_formula
#' @export
vec_cast.googlesheets4_formula.googlesheets4_formula <- function(x, to, ...) {
  x
}

#' @method vec_cast.googlesheets4_formula character
#' @export
vec_cast.googlesheets4_formula.character <- function(x, to, ...) {
  gs4_formula(x)
}

#' @method vec_cast.character googlesheets4_formula
#' @export
vec_cast.character.googlesheets4_formula <- function(x, to, ...) {
  vec_data(x)
}

test_that("constructors return length-0 vector when called with no arguments", {
  expect_length(new_formula(), 0)
  expect_length(gs4_formula(), 0)
})

test_that("low-level constructor errors for non-character input", {
  expect_error(new_formula(1:3), class = "vctrs_error_assert_ptype")
})

test_that("user-friendly constructor works for coercible input", {
  expect_s3_class(
    gs4_formula(factor("=sum(A:A)")),
    "googlesheets4_formula"
  )
})

test_that("common type of googlesheets4_formula and character is character", {
  expect_identical(
    vctrs::vec_ptype2(character(), gs4_formula()),
    character()
  )
  expect_identical(
    vctrs::vec_ptype2(gs4_formula(), character()),
    character()
  )
})

test_that("googlesheets4_formula and character are coercible", {
  expect_identical(
    vctrs::vec_cast("=sum(A:A)", gs4_formula()),
    gs4_formula("=sum(A:A)")
  )
  expect_identical(
    vctrs::vec_cast(gs4_formula("=sum(A:A)"), character()),
    "=sum(A:A)"
  )
  expect_identical(
    vctrs::vec_cast(gs4_formula("=sum(A:A)"), gs4_formula()),
    gs4_formula("=sum(A:A)")
  )
})

test_that("can concatenate googlesheets4_formula", {
  expect_identical(
    vctrs::vec_c(
      gs4_formula("=sum(A:A)"),
      gs4_formula("=sum(B:B)")
    ),
    gs4_formula(c("=sum(A:A)", "=sum(B:B)"))
  )
})

test_that("googlesheets4_formula can have missing elements", {
  out <- vctrs::vec_c(
    gs4_formula("=sum(A:A)"),
    NA,
    gs4_formula("=min(B2:G7"),
    NA
  )
  expect_s3_class(out, "googlesheets4_formula")
  expect_true(all(is.na(out[c(2, 4)])))
})

#' @export
as_tibble.googlesheets4_schema_NamedRange <- function(x, ...) {
  grid_range <- new("GridRange", !!!pluck(x, "range"))
  grid_range <- as_tibble(grid_range)

  tibble::tibble(
    name = glean_chr(x, "name"),
    id = glean_chr(x, "namedRangeId"),
    !!!grid_range
  )
}

as_NamedRange <- function(x, ...) {
  UseMethod("as_NamedRange")
}

#' @export
as_NamedRange.default <- function(x, ...) {
  abort_unsupported_conversion(x, to = "NamedRange")
}

#' @export
as_NamedRange.range_spec <- function(x, ..., name) {
  new("NamedRange", name = name, range = as_GridRange(x))
}

#' Find Google Sheets
#'
#' Finds your Google Sheets. This is a very thin wrapper around
#' [googledrive::drive_find()], that specifies you want to list Drive files
#' where `type = "spreadsheet"`. Therefore, note that this will require auth for
#' googledrive! See the article [Using googlesheets4 with
#' googledrive](https://googlesheets4.tidyverse.org/articles/articles/drive-and-sheets.html)
#' if you want to coordinate auth between googlesheets4 and googledrive. This
#' function will emit an informational message if you are currently logged in
#' with both googlesheets4 and googledrive, but as different users.
#'
#' @param ... Arguments (other than `type`, which is hard-wired as `type =
#'   "spreadsheet"`) that are passed along to [googledrive::drive_find()].
#'
#' @inherit googledrive::drive_find return
#' @export
#'
#' @examplesIf gs4_has_token()
#' # see all your Sheets
#' gs4_find()
#'
#' # see 5 Sheets, prioritized by creation time
#' x <- gs4_find(order_by = "createdTime desc", n_max = 5)
#' x
#'
#' # hoist the creation date, using other packages in the tidyverse
#' # x %>%
#' #   tidyr::hoist(drive_resource, created_on = "createdTime") %>%
#' #   dplyr::mutate(created_on = as.Date(created_on))
gs4_find <- function(...) {
  check_gs4_email_is_drive_email()
  googledrive::drive_find(..., type = "spreadsheet")
}

#' Freeze rows or columns in a (work)sheet
#'
#' @description
#' *Note: not yet exported.*
#'
#' Sets the number of frozen rows or column for a (work)sheet.
#'
#' @eval param_ss()
#' @eval param_sheet()
#' @param nrow,ncol Desired number of frozen rows or columns, respectively. The
#'   default of `NULL` means to leave unchanged.
#'
#' @template ss-return
#' @seealso Makes an `UpdateSheetPropertiesRequest`:
#'   * <# https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#UpdateSheetPropertiesRequest>
#'
#' @keywords internal
#' @noRd
#'
#' @examplesIf gs4_has_token()
#' # create a data frame to use as initial data
#' # intentionally has lots of rows and columns
#' dat <- gs4_fodder(25)
#'
#' # create Sheet
#' ss <- gs4_create("sheet-freeze-example", sheets = list(dat))
#'
#' # look at it in the browser
#' gs4_browse(ss)
#'
#' # freeze first 2 columns
#' sheet_freeze(ss, ncol = 2)
#'
#' # clean up
#' gs4_find("sheet-freeze-example") %>%
#'   googledrive::drive_trash()
sheet_freeze <- function(ss, sheet = NULL, nrow = NULL, ncol = NULL) {
  ssid <- as_sheets_id(ss)
  maybe_sheet(sheet)
  maybe_non_negative_integer(nrow)
  maybe_non_negative_integer(ncol)

  if (is.null(nrow) && is.null(ncol)) {
    gs4_bullets(c(i = "Nothing to be done."))
    return(invisible(ssid))
  }

  dims <- c(
    if (!is.null(nrow)) cli::pluralize("{nrow} row{?s}"),
    if (!is.null(ncol)) cli::pluralize("{ncol} column{?s}")
  )
  dims <- glue_collapse(dims, sep = " and ")

  x <- gs4_get(ssid)
  s <- lookup_sheet(sheet, sheets_df = x$sheets)
  gs4_bullets(c(
    v = "Freezing {dims} on sheet {.w_sheet {s$name}} in {.s_sheet {x$name}}."
  ))

  freeze_req <- bureq_set_grid_properties(
    sheetId = s$id,
    frozenRowCount = nrow,
    frozenColumnCount = ncol
  )

  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = freeze_req
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)

  invisible(ssid)
}

# executed interactively
# think of this as notes, in the form of code

# transferring ownership of the offical example and test Sheets to the
# sheet keeper service account
# and revoking write permission for myself and the docs service account

# objective is to make it impossible for development and checking mishaps to
# delete example Sheets

library(tidyverse)
library(googledrive)
library(googlesheets4)

sheet_keeper_email <-
  "googlesheets4-sheet-keeper@gargle-169921.iam.gserviceaccount.com"

# example Sheets started out as owned by the docs service account
# test Sheets started out as owned by the testing service account
# one of my personal Google identities had the 'write' role on each, as well
# googlesheets4:::gs4_auth_docs()
googlesheets4:::gs4_auth_testing()
gs4_find()

# target_sheets <- googlesheets4:::.gs4_examples
target_sheets <- googlesheets4:::.test_sheets

dat <- target_sheets %>%
  as_id() %>%
  drive_get() %>%
  drive_reveal("permissions")

# purely for exploration
dat_explore <- dat %>%
  hoist(permissions_resource, perms = list("permissions")) %>%
  select(-ends_with("_resource")) %>%
  unnest_auto(perms) %>%
  rename(drive_id = id) %>%
  unnest_wider(perms)
View(dat_explore)
dat_explore %>%
  select(name, id, type, role, emailAddress, displayName) %>%
  View()

# determine which Sheets are not yet owned by sheet keeper
to_transfer <- dat_explore %>%
  filter(role == "owner", displayName != "googlesheets4-sheet-keeper") %>%
  # select(name, displayName)
  pull(name)

# transfer ownership to sheet keeper
dat %>%
  filter(name %in% to_transfer) %>%
  drive_share(
    role = "owner",
    type = "user",
    emailAddress = sheet_keeper_email,
    transferOwnership = TRUE
  )

# revoking write access from the docs or testing account and my personal
# account
# must auth as sheet keeper now
gs4_auth(
  path = "~/.R/gargle/googlesheets4-sheet-keeper.json",
  scopes = "https://www.googleapis.com/auth/drive"
)
gs4_user()
drive_auth(token = gs4_token())
drive_user()

dat <- gs4_find() %>%
  filter(name %in% names(target_sheets)) %>%
  drive_reveal("permissions")

View(dat$drive_resource)

dat_explore <- dat %>%
  hoist(permissions_resource, perms = list("permissions")) %>%
  select(-permissions_resource) %>%
  unnest_auto(perms) %>%
  rename(drive_id = id) %>%
  unnest_wider(perms) %>%
  rename(permission_id = id)
View(dat_explore)
dat_explore %>%
  filter(role == "writer", type == "user") %>%
  select(name, type, role, emailAddress, drive_id, permission_id) %>%
  View()

delete_one_permission <- function(drive_id, permission_id) {
  request <- googledrive::request_generate(
    endpoint = "drive.permissions.delete",
    params = list(fileId = drive_id, permissionId = permission_id)
  )
  response <- googledrive::request_make(request, encode = "json")
  gargle::response_process(response)
}

to_delete <- dat_explore %>%
  filter(role == "writer", type == "user") %>%
  select(drive_id, permission_id)

out <- map2(to_delete$drive_id, to_delete$permission_id, delete_one_permission)


# 2021-06-05 I decided to publish some sheets, so they could be embedded in a
# distill website
drive_user()
deaths <- drive_get("deaths")
deaths %>%
  drive_reveal("published")
deaths <- deaths %>%
  drive_publish()
View(deaths$revision_resource)


ff <- drive_get("formulas-and-formats")
ff %>%
  drive_reveal("published")
ff <- ff %>%
  drive_publish()

library(tidyverse)
library(googledrive)
library(googlesheets4)
library(gapminder)

googlesheets4:::gs4_auth_docs()

# if I were making the gapminder sheet from scratch, here's what I would do now:
ss <- gs4_create(
  "gapminder-reboot",
  sheets = split(gapminder, gapminder$continent)
)
drive_trash(ss)
# but I am not doing this -- I want to keep the existing Sheet ID
# instead I will edit it (and mini-gap) in situ

## Update gapminder example Sheet ----
ss <- gs4_find("gapminder") %>% as_sheets_id()
# gs4_browse(ss)

gapminder_split <- split(gapminder, gapminder$continent)
sheet_write(gapminder_split$Africa,   ss = ss, sheet = "Africa")
sheet_write(gapminder_split$Americas, ss = ss, sheet = "Americas")
sheet_write(gapminder_split$Asia,     ss = ss, sheet = "Asia")
sheet_write(gapminder_split$Europe,   ss = ss, sheet = "Europe")
sheet_write(gapminder_split$Oceania,  ss = ss, sheet = "Oceania")

## Update mini-gap example Sheet ----
mini_gap <- gapminder %>%
  arrange(year) %>%
  group_by(continent) %>%
  slice(1:5) %>%
  split(.$continent)

(ss <- gs4_find("mini-gap") %>% as_sheets_id())
# gs4_browse(ss)

sheet_write(mini_gap$Africa,   ss = ss, sheet = "Africa")
sheet_write(mini_gap$Americas, ss = ss, sheet = "Americas")
sheet_write(mini_gap$Asia,     ss = ss, sheet = "Asia")
sheet_write(mini_gap$Europe,   ss = ss, sheet = "Europe")
sheet_write(mini_gap$Oceania,  ss = ss, sheet = "Oceania")

# https://developers.google.com/sheets/api/samples/formatting#format_a_header_row
# returns: a wrapped instance of RepeatCellRequest
bureq_header_row <- function(
  row = 1,
  sheetId = NULL,
  backgroundColor = 0.92,
  horizontalAlignment = "CENTER",
  bold = TRUE
) {
  row <- row - 1 # indices are zero-based; intervals are half open: [start, end)
  grid_range <- new(
    "GridRange",
    startRowIndex = row,
    endRowIndex = row + 1,
    sheetId = sheetId
  )

  cell_format <- new(
    "CellFormat",
    horizontalAlignment = horizontalAlignment,
    backgroundColor = new(
      "Color",
      # I want a shade of grey
      red = backgroundColor,
      green = backgroundColor,
      blue = backgroundColor
    ),
    textFormat = new(
      "TextFormat",
      bold = bold
    )
  )
  cell_data <- new_CellData(userEnteredFormat = cell_format)
  # hard-coding because we really do want to reset child properties not
  # explicitly set here
  # example: TextFormat's other children, like fontFamily or underline
  # example: Color's other child, alpha
  fields <- "userEnteredFormat(horizontalAlignment,backgroundColor,textFormat)"

  list(
    repeatCell = new(
      "RepeatCellRequest",
      range = grid_range,
      cell = cell_data,
      fields = fields
    )
  )
}

# based on this, except I clear everything by sending 'fields = "*"'
# https://developers.google.com/sheets/api/samples/sheet#clear_a_sheet_of_all_values_while_preserving_formats
# returns: a wrapped instance of RepeatCellRequest
bureq_clear_sheet <- function(sheetId) {
  list(
    repeatCell = new(
      "RepeatCellRequest",
      range = new("GridRange", sheetId = sheetId),
      fields = "*"
    )
  )
}

# https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#UpdateSheetPropertiesRequest
bureq_set_grid_properties <- function(
  sheetId,
  nrow = NULL,
  ncol = NULL,
  frozenRowCount = 1,
  frozenColumnCount = NULL
) {
  gp <- new("GridProperties", rowCount = nrow, columnCount = ncol)
  if (!is.null(frozenRowCount) && frozenRowCount > 0) {
    gp <- patch(gp, frozenRowCount = frozenRowCount)
  }
  if (!is.null(frozenColumnCount) && frozenColumnCount > 0) {
    gp <- patch(gp, frozenColumnCount = frozenColumnCount)
  }
  if (length(gp) == 0) {
    return(NULL)
  }

  sp <- new("SheetProperties", sheetId = sheetId, gridProperties = gp)
  list(
    updateSheetProperties = new(
      "UpdateSheetPropertiesRequest",
      properties = sp,
      fields = gargle::field_mask(sp)
    )
  )
}

# https://developers.google.com/sheets/api/samples/rowcolumn#automatically_resize_a_column
# https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#AutoResizeDimensionsRequest
bureq_auto_resize_dimensions <- function(
  sheetId,
  dimension = c("COLUMNS", "ROWS"),
  start = NULL,
  end = NULL
) {
  dimension <- match.arg(dimension)
  # https://developers.google.com/sheets/api/reference/rest/v4/DimensionRange
  # A range along a single dimension on a sheet. All indexes are zero-based.
  # Indexes are half open: the start index is inclusive and the end index is
  # exclusive. Missing indexes indicate the range is unbounded on that side.
  dimension_range <- new(
    "DimensionRange",
    sheetId = sheetId,
    dimension = dimension
  )
  if (!is.null(start) && notNA(start)) {
    check_non_negative_integer(start)
    dimension_range <- patch(dimension_range, startIndex = start - 1)
  }
  if (!is.null(end) && notNA(end)) {
    check_non_negative_integer(end)
    dimension_range <- patch(dimension_range, endIndex = end)
  }
  list(
    autoResizeDimensions = new(
      "AutoResizeDimensionsRequest",
      dimensions = dimension_range
    )
  )
}

# functions to help reduce duplication and increase consistency in the docs

### ss ----
param_ss <- function(..., pname = "ss") {
  template <- glue(
    "
    @param {pname} \\
    Something that identifies a Google Sheet:
    * its file id as a string or [`drive_id`][googledrive::as_id]
    * a URL from which we can recover the id
    * a one-row [`dribble`][googledrive::dribble], which is how googledrive
      represents Drive files
    * an instance of `googlesheets4_spreadsheet`, which is what [gs4_get()]
      returns

    Processed through [as_sheets_id()]."
  )
  dots <- list2(...)
  if (length(dots) > 0) {
    template <- c(template, dots)
  }
  glue_collapse(template, sep = " ")
}

### sheet ----
param_sheet <- function(..., action = "act on", pname = "sheet") {
  template <- glue(
    "
    @param {pname} \\
    Sheet to {action}, in the sense of \"worksheet\" or \"tab\". \\
    You can identify a sheet by name, with a string, or by position, \\
    with a number.
    "
  )
  dots <- list2(...)
  if (length(dots) > 0) {
    template <- c(template, dots)
  }
  glue_collapse(template, sep = " ")
}

param_before_after <- function(sheet_text) {
  glue(
    "
    @param .before,.after \\
    Optional specification of where to put the new {sheet_text}. \\
    Specify, at most, one of `.before` and `.after`. Refer to an existing \\
    sheet by name (via a string) or by position (via a number). If \\
    unspecified, Sheets puts the new {sheet_text} at the end.
    "
  )
}

#' Rename a (work)sheet
#'
#' Changes the name of a (work)sheet.
#'
#' @eval param_ss()
#' @eval param_sheet(
#'   action = "rename",
#'   "Defaults to the first visible sheet."
#' )
#' @param new_name New name of the sheet, as a string. This is required.
#'
#' @template ss-return
#' @export
#' @family worksheet functions
#' @seealso Makes an `UpdateSheetPropertiesRequest`:
#'   * <https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#UpdateSheetPropertiesRequest>
#'
#' @examplesIf gs4_has_token()
#' ss <- gs4_create(
#'   "sheet-rename-demo",
#'   sheets = list(cars = head(cars), chickwts = head(chickwts))
#' )
#' sheet_names(ss)
#'
#' ss %>%
#'   sheet_rename(1, new_name = "automobiles") %>%
#'   sheet_rename("chickwts", new_name = "poultry")
#'
#' # clean up
#' gs4_find("sheet-rename-demo") %>%
#'   googledrive::drive_trash()
sheet_rename <- function(ss, sheet = NULL, new_name) {
  ssid <- as_sheets_id(ss)
  maybe_sheet(sheet)
  check_string(new_name)

  x <- gs4_get(ssid)
  s <- lookup_sheet(sheet, sheets_df = x$sheets)
  gs4_bullets(c(
    v = "Renaming sheet {.w_sheet {s$name}} to {.w_sheet {new_name}}."
  ))

  sp <- new("SheetProperties", sheetId = s$id, title = new_name)
  update_req <- new(
    "UpdateSheetPropertiesRequest",
    properties = sp,
    fields = gargle::field_mask(sp)
  )

  req <- request_generate(
    "sheets.spreadsheets.batchUpdate",
    params = list(
      spreadsheetId = ssid,
      requests = list(updateSheetProperties = update_req)
    )
  )
  resp_raw <- request_make(req)
  gargle::response_process(resp_raw)

  invisible(ssid)
}

devtools::load_all() # I assume we're in googlesheets4 source

googlesheets4:::gs4_auth_testing()

ss <- test_sheet_create("googlesheets4-cell-tests")
gs4_browse(ss)

df <- gs4_fodder(5)
sheet_write(df, ss, sheet = "range-experimentation")

schema_rectangle <- function(s) {
  if (!"tidyverse" %in% .packages()) {
    stop("Attach the tidyverse package before using schema_rectangle()")
  }
  schema <- pluck(.schemas, s)
  if (schema$type != "object") {
    msg <- glue::glue(
      "Schema must be of type {sq('object')}, not {sq(schema$type)}"
    )
    stop(msg)
  }

  properties <- pluck(schema, "properties")
  scaffold <- list(
    description      = "Just a placeholder",
    type             = "scaffold",
    "$ref"           = "SCHEMA",
    items            = list("$ref" = "SCHEMA"),
    format           = "FORMAT",
    enum             = letters[1:3],
    enumDescriptions = LETTERS[1:3]
  )
  df <- tibble(properties = c(scaffold = list(scaffold), properties))

  df <- df %>%
    mutate(property = names(properties)) %>%
    select(property, everything()) %>%
    unnest_wider(properties) %>%
    select(-description) %>%
    mutate(type = replace_na(type, "object")) %>%
    rename(instance_of = "$ref")

  # workaround for https://github.com/tidyverse/tidyr/issues/806
  repair <- function(x) {
    map_if(x, ~ inherits(.x, "vctrs_unspecified"), ~ vctrs::unspecified(0))
  }
  df <- modify_if(df, is_list, repair)

  df <- df %>%
    hoist(items, array_of = "$ref")

  make_enum_tibble <- function(x, y) {
    tibble(
      enum     = x %||% character(),
      enumDesc = y %||% character()
    )
  }

  df <- df %>%
    mutate(new = map2(enum, enumDescriptions, make_enum_tibble)) %>%
    select(-starts_with("enum")) %>%
    rename(enum = new) %>%
    mutate(type = if_else(map_lgl(enum, ~ nrow(.x) > 0), "enum", type))

  attr(df, "id") <- s

  df %>%
    filter(property != "scaffold") %>%
    arrange(property)
}

# articles that I only want to compile intentionally

library(knitr)

knit(
  # the leading `_` keeps pkgdown from re-rendering the original
  "vignettes/articles/_dates-and-times.Rmd.orig.Rmd",
  "vignettes/articles/dates-and-times.Rmd"
)

# making some changes to
# test Sheet: googlesheets4-col-types
# sheet: NAs

# I want the ID to remain the same, so will modify existing Sheet
library(googlesheets4)
library(googledrive)
library(tidyverse)

gs4_auth(
  path = "~/.R/gargle/googlesheets4-sheet-keeper.json",
  scopes = "https://www.googleapis.com/auth/drive"
)
gs4_user()

examples <- googlesheets4:::.test_sheets
ssid <- examples["googlesheets4-col-types"]
(ss <- gs4_get(ssid))

dat <- tribble(
  ~...Missing,  ~...NA,  ~space, ~empty_string, ~truly_empty, ~complete,
        "one",   "one",   "one",         "one",        "one",     "one",
    "Missing",    "NA",     " ",            "",           NA,     "two",
      "three", "three", "three",       "three",      "three",   "three"
)

write_sheet(dat, ssid, sheet = "NAs")

range_read(ssid)

library(tidyverse)

source(
  system.file("discovery-doc-ingest", "ingest-functions.R", package = "gargle")
)

# if my use of schemas works out well, maybe this will migrate upstream into
# gargle and join the other ingest helpers
source(here::here("data-raw", "schema-rectangling.R"))

existing <- list_discovery_documents("sheets")
if (length(existing) > 1) {
  rlang::warn("MULTIPLE DISCOVERY DOCUMENTS FOUND. FIX THIS!")
}

if (length(existing) < 1) {
  rlang::inform("Downloading Discovery Document")
  x <- download_discovery_document("sheets:v4")
} else {
  msg <- glue::glue("
    Using existing Discovery Document:
      * {existing}
    ")
  rlang::inform(msg)
  x <- here::here("data-raw", existing)
}

dd <- read_discovery_document(x)

methods      <- get_raw_methods(dd)
more_methods <- get_raw_methods(pluck(dd, "resources", "spreadsheets"))
methods <- c(methods, more_methods)

methods <- methods %>% map(groom_properties,  dd)
methods <- methods %>% map(add_schema_params, dd)
methods <- methods %>% map(add_global_params, dd)

.endpoints <- methods
attr(.endpoints, "base_url") <- dd$rootUrl
# View(.endpoints)

# I'm exploring the pros/cons of working with these more properly, as opposed
# to the "flattened" or "inlined" representation currently in .endpoints
.schemas <- pluck(dd, "schemas")

these <- c(
  "Spreadsheet",
  "SpreadsheetProperties",
  "Sheet",
  "SheetProperties",
  "NamedRange",
  "GridRange",
  "RepeatCellRequest",
  "CellData",
  "RowData",
  "CellFormat",
  "Color",
  "TextFormat",
  "UpdateCellsRequest",
  "UpdateSheetPropertiesRequest",
  "GridCoordinate",
  "DeleteDimensionRequest",
  "InsertDimensionRequest",
  "DimensionRange",
  "GridProperties",
  "AppendCellsRequest",
  "AutoResizeDimensionsRequest",
  "DeleteRangeRequest",
  "DuplicateSheetRequest",
  "AddSheetRequest",
  "UpdateNamedRangeRequest",
  "AddNamedRangeRequest",
  "SetDataValidationRequest",
  "DataValidationRule",
  "BooleanCondition",
  "ConditionValue",
  "AddProtectedRangeRequest",
  "UpdateProtectedRangeRequest",
  "DeleteProtectedRangeRequest",
  "ProtectedRange",
  "Editors"
)

.tidy_schemas <- these %>%
  set_names() %>%
  map(schema_rectangle)
# View(.tidy_schemas)

fs::dir_create(here::here("data-raw", "schemas"))
write_one <- function(data, id) {
  withr::local_options(list(width = 150, crayon.enabled = FALSE))
  pillar:::num_colors(forget = TRUE)
  sink(here::here("data-raw", "schemas", id))
  cat("#", id, " \n")
  print(data, n = Inf, width = Inf)
  sink()
}
iwalk(.tidy_schemas, write_one)

usethis::use_data(
  .endpoints, .schemas, .tidy_schemas,
  internal = TRUE, overwrite = TRUE
)

---
title: "Find and identify Sheets"
---

```{r setup, include = FALSE}
can_decrypt <- gargle:::secret_can_decrypt("googlesheets4")
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  error = TRUE,
  purl = can_decrypt,
  eval = can_decrypt
)
options(tibble.print_min = 5L, tibble.print_max = 5L)
```

```{r eval = !can_decrypt, echo = FALSE, comment = NA}
message("No token available. Code chunks will not be evaluated.")
```

There are many ways to get your hands on your Sheets, in order to work with them via googlesheets4. They basically range from "ugly, but low effort" to "more humane, but more effort".

## Attach googlesheets4

```{r}
library(googlesheets4)
```

## Auth

As a regular, interactive user, you can just let googlesheets4 prompt you for anything it needs re: auth.

Since this article is compiled noninteractively on a server, we have arranged for googlesheets4 to use a service account token (not shown).


```{r include = FALSE}
# happens in .onLoad() when IN_PKGDOWN, but need this for local dev/preview
googlesheets4:::gs4_auth_docs(drive = TRUE) 

# attempt to reduce quota exhaustion problems
if (identical(Sys.getenv("IN_PKGDOWN"), "true")) Sys.sleep(30)
```

## Use a URL

When you visit a Sheet in the browser, you can copy that URL to your clipboard. Such URLs look something like this:

```
https://docs.google.com/spreadsheets/d/1U6Cf_qEOhiR9AZqTqS3mbMF3zt2db48ZP5v3rkrAEJY/edit#gid=780868077
```

which breaks down like this:

```
https://docs.google.com/spreadsheets/d/SPREADSHEET_ID/edit#gid=SHEET_ID
```

Notice that this URL contains a (spread)Sheet ID and a (work)sheet ID. This URL happens to link to [the official example Sheet that holds Gapminder data](https://docs.google.com/spreadsheets/d/1U6Cf_qEOhiR9AZqTqS3mbMF3zt2db48ZP5v3rkrAEJY/edit#gid=780868077).

googlesheets4 accepts such a URL as the `ss` argument (think "spreadsheet") of many functions:

```{r}
ugly_url <- "https://docs.google.com/spreadsheets/d/1U6Cf_qEOhiR9AZqTqS3mbMF3zt2db48ZP5v3rkrAEJY/edit#gid=780868077"
read_sheet(ugly_url)
```

At this time, although the URL may contain both a (spread)Sheet and a (work)sheet, we only extract the (spread)Sheet ID. If the function targets a specific (work)sheet, that is typically specified via arguments like `range` or `sheet` or a default of "first (visible) sheet".

These URLs are not particularly nice to look at in your code, though.

## Use a Sheet ID

You can extract the Sheet ID from a URL with `as_sheets_id()` (which is what we are doing internally to make the URL work in the first place):

```{r}
ssid <- as_sheets_id(ugly_url)
class(ssid)
unclass(ssid)
```

`as_sheets_id()` is a generic function, which means it knows what to do with a few different types of input. For character string input, `as_sheets_id()` passes a string through, unless it looks like a URL. If it looks like a URL, we use a regular expression to extract the Sheet ID. The returned string bears the classes `sheets_id` and `drive_id` (for playing nicely with googledrive).

Why did we call `unclass(ssid)` above to see the naked Sheet ID? Because, by default, when you print an instance of `sheets_id`, we attempt to show you some current metadata about the Sheet.

```{r}
ssid
```

This is the same metadata you'll see when you call `gs4_get()` (but you must call `gs4_get()` explicitly if you want to *store* the returned metadata).

googlesheets4 also accepts a Sheet ID as the `ss` argument of many functions:

```{r}
read_sheet(ssid)
```

I think in a script or app that will endure for a while it is better to refer to a Sheet by its ID than by its URL. The Sheet ID is nicer to look at, it is complete, it is minimal.

```{r}
ssid <- "1U6Cf_qEOhiR9AZqTqS3mbMF3zt2db48ZP5v3rkrAEJY"
read_sheet(ssid)
```

Note this demonstration that a Sheet ID also works when provided as a plain, old string, i.e. it does not **have** to have the `sheets_id` class. In some contexts, you might even prefer to store it as a string, in order to bypass the special printing behaviour for `sheets_id`.

When the Sheet is specified via a character string, googlesheets4 assumes it is a Sheet ID (or an ID-containing URL). This is NOT the case for googledrive, which assumes a character string is a file name or path. Therefore, for maximum happiness, in a mixed googlesheets4 / googledrive workflow, it's a good idea to be explicit and declare a string to be a Sheet ID, when that is the case.

```{r}
ssid <- as_sheets_id("1U6Cf_qEOhiR9AZqTqS3mbMF3zt2db48ZP5v3rkrAEJY")
read_sheet(ssid)
```

If your script or app targets a specific Sheet, the most efficient and robust way to address it is by its ID.

## Use the Sheet's name (uses googledrive)

A big feature of the googledrive package is the ability to navigate between human-friendly file names and machine-friendly file IDs. Both the Drive and Sheets APIs require the use of IDs, so the illusion that you can identify a Drive file by name is provided by the googledrive package. (A Google Sheet is just a special case of a Drive file ... a file that happens to be a spreadsheet.)

If you need to refer to a Sheet by name, i.e. if you need to lookup its file ID based on its name, you must use the googledrive package for that. There are other reasons for using these two packages together: the Sheets API has an intentionally narrow focus on spreadsheet operations involving worksheets and cells. General whole-file operations, like copy / rename / move / share, must be done via the Drive API and, within R, via googledrive. See the article [Using googlesheets4 with googledrive](https://googlesheets4.tidyverse.org/articles/articles/drive-and-sheets.html) for more.

It's time to attach googledrive (in addition to the already-attached googlesheets4):

```{r, message = FALSE}
library(googledrive)
```

(In our hidden auth chunk, we actually put a shared token into force for both googlesheets4 and googledrive, anticipating this moment.)

The Sheet we've been working with is named "gapminder" and is owned by the account we've logged in as here. We can use `googledrive::drive_get()` to identify a file by name:

```{r}
(gap_dribble <- drive_get("gapminder"))
```

`drive_get()` returns a [`dribble`](https://googledrive.tidyverse.org/reference/dribble.html), a "Drive tibble", where each row holds info on a Drive file. `as_sheets_id()` also accepts a one-row `dribble`, so we can get right into our normal googlesheets4 workflows:

```{r}
gap_id <- as_sheets_id(gap_dribble)
unclass(gap_id)
gap_id
```

Since we generally call `as_sheets_id()` on whatever the user provides as `ss`, you can even pass `gap_dribble` straight into googlesheets4 functions.

```{r}
sheet_properties(gap_dribble)
```


Two important things to note:

  * googledrive requires auth for functions like `drive_get()` and
    `drive_find()` (see below). At first, you can just react to the interactive
    auth prompts and **make sure you auth as the same user** with googledrive
    and googlesheets4. Once you get tired of doing auth for both packages,
    read the article [Using googlesheets4 with googledrive](https://googlesheets4.tidyverse.org/articles/articles/drive-and-sheets.html). Coordinated auth should get even easier in the future.
    - In our hidden auth here, we have taken special measures to use a shared
      token for googledrive and googlesheets4.
  * Remember to use `drive_get()` on a Sheet name **that you have**. If you
    don't have a Sheet named "gapminder", the code above won't yield anything.
    As a rough rule of thumb, if you don't see it at [spreadsheets.google.com](http://spreadsheets.google.com), you can't
    `drive_get()` it either.

## List your Sheets (uses googledrive)

What if you want to see all of your Sheets? Or all the Sheets with "gap" in their name?

`googledrive::drive_find()` is the workhorse function for these tasks for general Drive files. It has lots of bells and whistles and we can use one of them to narrow the search to Google Sheets:

```{r}
drive_find(type = "spreadsheet")
```

This is so handy that we've made `gs4_find()` in googlesheets4, which is a shortcut for `drive_find(type = "spreadsheet")`:

```{r}
gs4_find()
```

See the examples for `drive_find()` and `gs4_find()` for more ideas about how to search Drive effectively for your Sheets.

---
title: "googlesheets4 auth"
---

```{r setup, include = FALSE}
can_decrypt <- gargle:::secret_can_decrypt("googlesheets4")
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  error = TRUE,
  purl = can_decrypt,
  eval = can_decrypt
)
options(tibble.print_min = 4L, tibble.print_max = 4L)
```

```{r eval = !can_decrypt, echo = FALSE, comment = NA}
message("No token available. Code chunks will not be evaluated.")
```

googlesheets4 will, by default, help you interact with Sheets as an authenticated Google user. The package facilitates this process upon first need.

## `gs4_deauth()`

If you don't need to access private Sheets, use `gs4_deauth()` to indicate there is no need for a token. This puts googlesheets4 into a de-authorized mode.

Here's how an R script might look if all you plan to do is read Sheets that are world-readable or readable by "anyone with a link":

```{r eval = FALSE}
library(googlesheets4)

gs4_deauth()

# imagine this is the URL or ID of a Sheet readable by anyone (with a link)
ss <- "?????"
dat <- read_sheet(ss)
```

## Default auth behaviour and beyond

As soon as googlesheets4 needs a token, it tries to discover one. If it fails, it engages with you interactively to help you get a token. Once successful, that token is remembered for subsequent use in that R session.

Users can take control of auth proactively via the [`gs4_auth*()` family of functions](https://googlesheets4.tidyverse.org/reference/index.html#section-auth). Examples of what you can control or provide:

  * The email address of the Google identity you want to use.
  * Whether to cache tokens and where.
  * Whether to use out-of-band auth.
  * A service account token.
  * The OAuth app and/or API key.
  
Auth is actually handled by the gargle package ([gargle.r-lib.org](https://gargle.r-lib.org)), similar to googledrive, bigrquery, and gmailr, and gargle's documentation and articles are the definitive guide to more advanced topics.

## Multi-package auth

It is common to use googlesheets4 together with the googledrive package ([googledrive.tidyverse.org](https://googledrive.tidyverse.org)). See the article [Using googlesheets4 with googledrive](https://googlesheets4.tidyverse.org/articles/articles/drive-and-sheets.html) for advice on how to streamline auth in this case.

---
title: "Function and class names"
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This article explains conventions in googlesheets4 around:

  * Function names
  * Classes

# Prefixes: gs4, sheet, range

Almost all functions in googlesheets4 have one of these prefixes:

  * `gs4_`, referring variously to the googlesheets4 package, v4 of the Google 
    Sheets API, or to operations on one or more (spread)Sheets
  * `sheet_`, referring to an operation on one or more (work)sheets
  * `range_`, referring to an operation on a range of cells

This table summarizes what the `gs4_`, `sheet_`, `range_` mean conceptually and which arguments you can expect to see in the function signature.

| prefix | `ss`  | `sheet` | `range` | scope            |
|--------|-------|---------|---------|------------------|
| gs4_   | yes   | no      | no      | a (spread)Sheet  |
| sheet_ | yes   | yes     | no      | a (work)sheet    |
| range_ | yes   | yes     | yes     | a range of cells |

Take this table with a grain of salt.
There are a few violations of it in function signatures, when justified, but it's true in broad strokes.
And remember that `gs4_` is also used for general, package-level functions.

# Previous use of `sheets_` prefix

When googlesheets4 first appeared on CRAN and up to v0.1.1 (released 2020-03-21), it had a nearly universal `sheets_` prefix.
In version 0.2.0 (released 2020-05-08), googlesheets4 gained a tremendous amount of writing and editing functionality.
At that time, it became clear that the `sheets_` scheme was an impediment to generating descriptive, predictable function names.
For example, you can delete a (spread)Sheet, a (work)sheet, or a cell range.
Which one of these functions will be named `sheets_delete()`?
How do we name the others?
There is no good answer to this, which is why we adopted the 3 prefix strategy.

Any function that existed in v0.1.1 was formally deprecated in v0.2.0 and was removed in v1.0.0.
Below are tables documenting the name changes, that also cover functions that only existed in dev versions, but that may have seen use in the community.
As time goes on, this article becomes more of historical note.

The dev version of googlesheets4 was bumped from 0.1.1.9000 to 0.1.1.9100 when the `sheets_` naming scheme was abandoned.
The last commit under the old scheme was also tagged as "v0.1.1.9000", which means one can install from that specific state via `devtools::install_github("tidyverse/googlesheets4@v0.1.1.9000")`.

## Auth and the Sheet API v4 surface 

```{r echo = FALSE}
dat <- tibble::tribble(
  ~ "<= v0.1.1.9000",              ~ ">= v0.1.1.9100",
  "`sheets_api_key()` (*)",        "`gs4_api_key()`",
  "`sheets_auth()` (*)",           "`gs4_auth()`",
  "`sheets_auth_configure()` (*)", "`gs4_auth_configure()`",
  "`sheets_deauth()` (*)",         "`gs4_deauth()`",
  "`sheets_endpoints()` (*)",      "`gs4_endpoints()`",
  "`sheets_has_token()` (*)",      "`gs4_has_token()`",
  "`sheets_oauth_app()` (*)",      "`gs4_oauth_app()`",
  "`sheets_token()` (*)",          "`gs4_token()`",
  "`sheets_user()` (*)",           "`gs4_user()`"
)
knitr::kable(dat)
```

(*) indicates functions in the CRAN versions <= v0.1.1.

## (Spread)sheet scope

```{r echo = FALSE}
dat <- tibble::tribble(
  ~ "<= v0.1.1.9000",       ~ ">= v0.1.1.9100",
  "`sheets_browse()` (*)",   "`gs4_browse()`",
  "`sheets_create()`",       "`gs4_create()`",
  "`sheets_find()` (*)",     "`gs4_find()`",
  "`sheets_fodder()`",       "`gs4_fodder()`",
  "`sheets_formula()`",      "`gs4_formula()`",
  "`sheets_example()` (*)",  "`gs4_example()`",
  "`sheets_examples()` (*)", "`gs4_examples()`",
  "`sheets_get()` (*)",      "`gs4_get()`",
  "`sheets_random()`",       "`gs4_random()`"
)
knitr::kable(dat)
```

(*) indicates functions in the CRAN versions <= v0.1.1.

## (Work)sheet scope

```{r echo = FALSE}
dat <- tibble::tribble(
  ~ "<= v0.1.1.9000",            ~ ">= v0.1.1.9100",
  "`sheets_append()`",           "`sheet_append()`",          
  "`sheets_sheet_add()`",        "`sheet_add()`",
  "`sheets_sheet_copy()`",       "`sheet_copy()`",
  "`sheets_sheet_delete()`",     "`sheet_delete()`",
  "`sheets_sheet_names()`",      "`sheet_names()`",
  "`sheets_sheet_properties()`", "`sheet_properties()`",
  "`sheets_sheet_relocate()`",   "`sheet_relocate()`",
  "`sheets_sheet_rename()`",     "`sheet_rename()`",
  "`sheets_sheet_resize()`",     "`sheet_resize()`",
  "`sheets_sheets()` (*)",       "`sheet_names()`",
  "`sheets_write()`",            "`sheet_write()`"
)
knitr::kable(dat)
```

(*) indicates functions in the CRAN versions <= v0.1.1.

## Range scope

```{r echo = FALSE}
dat <- tibble::tribble(
  ~ "<= v0.1.1.9000",            ~ ">= v0.1.1.9100",
  "`sheets_auto_resize_dims()`", "`range_autofit()`",          
  "`sheets_cells()` (*)",        "`range_read_cells()`",
  "`sheets_clear()`",            "`range_clear()`",
  "`sheets_edit()`",             "`range_write()`",
  "`sheets_flood()`",            "`range_flood()`",
  "`sheets_read()` (*)",         "`range_read()`",
  "`sheets_speedread()`",        "`range_speedread()`"
)
knitr::kable(dat)
```

(*) indicates functions in the CRAN versions <= v0.1.1.

Note: "range" can mean two things in the Sheets API:

  * An A1-style range, given as a string. Examples: `A1:B3` (a cell range), `Africa` (a worksheet name), `Africa!A1:B3` (a sheet-qualified cell range), `arts_data` (a named range). Some API endpoints require this, believe it or not.
  * A cell rectangle described by start/end row/column, packaged as an instance of a schema, such as `GridRange`. Most API endpoints use this.

Fun fact: The "cell rectangle" type of range is almost a sub-case of the A1-style range, except there are rectangles open on one or more sides that can be described via `GridRange` that cannot be expressed as an A1-range string.
The mostly-developer-facing article [Range specification](https://googlesheets4.tidyverse.org/articles/articles/range-specification.html) documents all this and how I approach it internally.

The `range_` prefix encompasses both types of ranges and each function has to indicate what sorts of `range` it supports, which is determined by the behaviour of the associated API endpoint.

# Classes

Any user facing class that is related to a schema should be named like `googlesheets4_schema_name`, where the schema name is in lower_snake_case.

The internal schema-derived classes should be like `googlesheets4_schema_SchemaName` / `googlesheets4_schema` / `list`. Use the `googlesheets4_schema` prefix and retain the API's UpperCamelCase.

This release is prompted by a request from Kurt Hornik and I am under a deadline of 2023-03-22 (I was given a few days of wiggle room due to travel). That request related to making the signature of a method match the generic.

I'm also updating my email address as maintainer.

## R CMD check results

0 errors | 0 warnings | 1 note

   New maintainer:
     Jennifer Bryan <jenny@posit.co>
   Old maintainer(s):
     Jennifer Bryan <jenny@rstudio.com>

## revdepcheck results

We checked 10 reverse dependencies, comparing R CMD check results across CRAN and dev versions of this package.

 * We saw 0 new problems
 * We failed to check 0 packages

---
title: "Using googlesheets4 with googledrive"
---

```{r setup, include = FALSE}
can_decrypt <- gargle:::secret_can_decrypt("googlesheets4")
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  error = TRUE,
  purl = can_decrypt,
  eval = can_decrypt
)
```

```{r eval = !can_decrypt, echo = FALSE, comment = NA}
message("No token available. Code chunks will not be evaluated.")
```

```{r eval = FALSE, include = FALSE}
# article assume current user has access to a Sheet named "chicken-sheet"
# execute this if it seems to have gone missing
x <- drive_find("chicken")
if (nrow(x) != 1) {
  x <- gs4_example("chicken-sheet")
  y <- drive_cp(x, name = "chicken-sheet")
}
```

## Why use googlesheets4 and googledrive together?

googlesheets4 wraps the Sheets API v4, which lets you read, write, and format data in Sheets. The Sheets API is very focused on spreadsheet-oriented data and metadata, i.e. (work)sheets and cells.

The Sheets API offers practically no support for file-level operations, other than basic spreadsheet creation. There is no way to delete, copy, or rename a Sheet or to place it in a folder or to change its sharing permissions. We must use the Drive API for all of this, which is wrapped by the googledrive package (<https://googledrive.tidyverse.org>).

Another reason to use the googlesheets4 and googledrive packages together is for ease of file (Sheet) identification. The googlesheets4 package requires you to specify the target Sheet by its ID, not by its *name*. That's because the underlying APIs only accept file IDs. But the googledrive package offers lots of support for navigating between human-friendly file names and their associated IDs. This support applies to all files on Drive and, specifically, to Sheets.

Therefore, it is common to use googledrive and googlesheets4 together in a script or app.

## Coordinating auth

How does auth work if you're using googlesheets4 **and** googledrive? The path of least resistance is to do nothing and just let each package deal with its own auth. This works fine! But it's a bit clunky and you need to make sure you're using the same Google identity (email) with each package/API.

It can be nicer to be proactive about auth and use the same token for your googledrive and googlesheets4 work. Below we show a couple of ways to do this.

## Auth with googledrive first, then googlesheets4

Outline:

  * Make sure auth happens first with googledrive, probably by calling `googledrive::drive_auth()` yourself. The default scope is `"https://www.googleapis.com/auth/drive"`, which is sufficient for all your Drive and Sheets work.
  * Tell googlesheets4 to use the same token as googledrive.

First attach both packages.

```{r results = 'hide', message = FALSE}
library(googledrive)
library(googlesheets4)
```

Do auth first with googledrive. Remember `googledrive::drive_auth()` accepts additional arguments, e.g. to specify a Google identity via `email =` or to use a service account via `path =`. Then direct googlesheets4 to use the same token as googledrive.

```{r eval = FALSE}
drive_auth()
gs4_auth(token = drive_token())
```

```{r include = FALSE}
# happens in .onLoad() when IN_PKGDOWN, but need this for local dev/preview
googlesheets4:::gs4_auth_docs(drive = TRUE) 

# attempt to reduce quota exhaustion problems
if (identical(Sys.getenv("IN_PKGDOWN"), "true")) Sys.sleep(30)
```

Now you can use googledrive functions, like `googledrive::drive_find()` or `googledrive::drive_get()`, to list files or find them by name, path, or other property. Then, once you've identified the target file, use googlesheets4 to do spreadsheet-specific tasks.

```{r}
drive_find("chicken")

ss <- drive_get("chicken-sheet")

gs4_get(ss)

read_sheet(ss)
```

If you ever want to confirm the currently authenticated user, both packages provide a `*_user()` function that reveals some info:

```{r}
drive_user()
gs4_user()
```

We are using a service account to render this article. But if you've used the default OAuth flow, this should correspond to the email of the Google account you logged in with.

```{r include = FALSE}
gs4_deauth()
drive_deauth()
detach("package:googledrive", character.only = TRUE)
detach("package:googlesheets4", character.only = TRUE)
```

## Auth with googlesheets4 first, then googledrive

Outline:

  * Proactively auth with googlesheets4 and specify the `"https://www.googleapis.com/auth/drive"` scope. The default googlesheets4 scope is `"https://www.googleapis.com/auth/spreadsheets"`, which is insufficient for general work with the Drive API.
  * Tell googledrive to use the same token as googlesheets4.

First attach both packages.

```{r results = 'hide', message = FALSE}
library(googlesheets4)
library(googledrive)
```

Do auth first with googlesheets4, specifying a Drive scope. Remember `gs4_auth()` accepts additional arguments, e.g. to specify a Google identity via `email =` or to use a service account via `path =`. Then direct googledrive to use the same token as googlesheets4.

```{r eval = FALSE}
gs4_auth(scope = "https://www.googleapis.com/auth/drive")
drive_auth(token = gs4_token())
```

```{r include = FALSE}
googlesheets4:::gs4_auth_docs(drive = TRUE)
```

Now you can use googledrive functions to list files or find them by name, path, or other property. Then, once you've identified the target file, use googlesheets4 to do spreadsheet-specific tasks.

```{r}
drive_find("chicken")

ss <- drive_get("chicken-sheet")

gs4_get(ss)

read_sheet(ss)
```

```{r include = FALSE}
gs4_deauth()
drive_deauth()
detach("package:googledrive", character.only = TRUE)
detach("package:googlesheets4", character.only = TRUE)
```

## Scope savvy

If you only need "read" access to Drive or Sheets, the conservative thing to do is to specify a read-only scope. This is a great way to limit the damage anyone can do with the token -- you or someone else -- through carelessness or malice. If you are storing a token on a remote or shared location, it is wise to use the most conservative scope that still gets the job done.

Here are various scopes relevant to googledrive and googlesheets4 and what they would allow.

`drive` scope allows reading and writing with Drive and Sheets APIs. This scope is the most powerful and, therefore, the most dangerous.

```{r eval = FALSE}
PACKAGE_auth(
  ...,
  scopes = "https://www.googleapis.com/auth/drive",
  ...
)
```

`drive.readonly` still allows file identification via Drive and can be combined with `spreadsheets` if you plan to edit, create, or delete Sheets.

```{r eval = FALSE}
PACKAGE_auth(
  ...,
  scopes = c(
    "https://www.googleapis.com/auth/drive.readonly",
    "https://www.googleapis.com/auth/spreadsheets"
  ),
  ...
)
```

If you are just using Drive to identify Sheets and are only reading from those Sheets, the `drive.readonly` scope is sufficient and means you can't modify anything by accident.

```{r eval = FALSE}
PACKAGE_auth(
  ...,
  scopes = "https://www.googleapis.com/auth/drive.readonly",
  ...
)
```

If you are not using Drive at all, i.e. you always identify Sheets by file ID, and you are only reading from those Sheets, you only need googlesheets4 and `spreadsheets.readonly` is sufficient.

```{r eval = FALSE}
gs4_auth(
  ...,
  scopes = "https://www.googleapis.com/auth/spreadsheets.readonly",
  ...
)
```

---
title: "Fun with googledrive and readxl"
---

This article demonstrates how to use googlesheets4, googledrive, and readxl together. We demonstrate a roundtrip for data that starts and ends in R, but travels in spreadsheet form, via Google Sheets.

```{r setup, include = FALSE}
can_decrypt <- gargle:::secret_can_decrypt("googlesheets4")
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  error = TRUE,
  purl = can_decrypt,
  eval = can_decrypt
)
```

```{r eval = !can_decrypt, echo = FALSE, comment = NA}
message("No token available. Code chunks will not be evaluated.")
```

## Attach packages

```{r}
library(googlesheets4)
library(googledrive)
library(readxl)
```

## Auth

```{r include = FALSE}
# happens in .onLoad() when IN_PKGDOWN, but need this for local dev/preview
googlesheets4:::gs4_auth_docs(drive = TRUE) 

# attempt to reduce quota exhaustion problems
if (identical(Sys.getenv("IN_PKGDOWN"), "true")) Sys.sleep(30)
```

As a regular, interactive user, you can just let googlesheets4 prompt you for anything it needs re: auth.

Since this article is compiled noninteractively on a server, we activate a service token here, in a hidden chunk. We are also using a shared token for Sheets and Drive. You can read how to do that in your own work in the article [Using googlesheets4 with googledrive](https://googlesheets4.tidyverse.org/articles/articles/drive-and-sheets.html).

## Create a private Sheet from csv with the Drive API

Put the iris data into a csv file.

```{r}
(iris_tempfile <- tempfile(pattern = "iris-", fileext = ".csv"))
write.csv(iris, iris_tempfile, row.names = FALSE)
```

Use `googledrive::drive_upload()` to upload the csv and simultaneously convert to a Sheet.

```{r}
(iris_ss <- drive_upload(iris_tempfile, type = "spreadsheet"))

# visit the new Sheet in the browser, in an interactive session!
drive_browse(iris_ss)
```

Read data from the private Sheet into R.
```{r}
read_sheet(iris_ss, range = "B1:D6")
```

## Create a local xlsx from a Sheet with the Drive API

Download the Sheet as an Excel workbook.

```{r}
(iris_xlsxfile <- sub("[.]csv", ".xlsx", iris_tempfile))
drive_download(iris_ss, path = iris_xlsxfile, overwrite = TRUE)
```

## Read xlsx with readxl

Read the iris data back in via `readxl::read_excel()`.

```{r}
if (requireNamespace("readxl", quietly = TRUE)) {
  readxl::read_excel(iris_xlsxfile)  
}
```

## Clean up

```{r}
file.remove(iris_tempfile, iris_xlsxfile)
drive_trash(iris_ss)
```

---
title: "Dates and times"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Dates and times}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

<!-- dates-and-times.Rmd is generated from dates-and-times.Rmd.orig.Rmd. Please edit that file -->

```{r setup, include = FALSE}
can_decrypt <- gargle:::secret_can_decrypt("googlesheets4")
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  error = TRUE,
  purl = can_decrypt,
  eval = can_decrypt
)
```

```{r eval = !can_decrypt, echo = FALSE, comment = NA}
message("No token available. Code chunks will not be evaluated.")
```

This article provides advice on reading and writing datetimes with Google Sheets, specifically around the matter of time zones.

A related issue is how datetimes are formatted for presentation in the Sheet itself. You can read more about these formats in the [Date and time format patterns section of the Sheets API docs](https://developers.google.com/sheets/api/guides/formats#date_and_time_format_patterns). At the time of writing, googlesheets4 provides no user-friendly way to address these formats, although it may do so in the future.

## Attach packages and do auth

Attach googlesheets4.

```{r}
library(googlesheets4)
```

Since we eventually create and edit Sheets, we also auth here in a hidden chunk. If you run this code, you should expect auth to happen.
 
```{r include = FALSE}
# happens in .onLoad() when IN_PKGDOWN, but need this for local, intentional
# precompilation
googlesheets4:::gs4_auth_docs(drive = TRUE)
```

The lubridate package ([lubridate.tidyverse.org](https://lubridate.tidyverse.org)) is useful for this exploration, so we attach it now.

```{r}
library(lubridate, warn.conflicts = FALSE)
```

## How to work with time zones in a Google Sheet

You don't.

Literally, you can't.

I know this sounds very harsh, but it is the truth. Google Sheets offer essentially no support for time zones and your life will be simpler if you just make peace with this and accept that you will be looking at UTC times in Sheets.

A short demo: in R, capture the current time as `tt` and reveal the current time zone. Note that `tt` is displayed in R according to this time zone.

```{r}
(tt <- Sys.time())

Sys.timezone()
```

Write the `tt` datetime to a Sheet, configured with the same time zone as the local R session, and create another cell that captures the exact text presented in the browser UI for `tt`. Read this back into R.

```{r}
dat <- tibble::tibble(
  datetime = tt,
  as_displayed = gs4_formula("=TO_TEXT(A2)")
)

(ss <- gs4_create(
  "no-time-zone-effect",
  sheets = dat,
  timeZone = Sys.timezone()
))

read_sheet(ss)
```

Note that the `tt` datetime is displayed differently in Sheets than it is locally in R. Sheets presents datetimes in Coordinated Universal Time (time zone Etc/UTC), even if the Sheet's metadata specifies a different time zone, such as America/Vancouver.

```{r}
with_tz(tt, "Etc/UTC")
```

If you want to understand more about datetimes in R, in Sheets, and how you can sort of hack around this time zone problem, keep reading.

```{r include = FALSE}
gs4_find("no-time-zone-effect") %>% googledrive::drive_trash()
```

## Need-to-know basics of datetimes

Datetimes are a complicated topic. Here we dramatically oversimplify things, in the name of making a reader who is new to all of this at least minimally functional.

The main system used to represent times in the computing world is Unix epoch time:

> A moment in time is represented by the number of seconds that have elapsed since 1 January 1970 00:00:00 UTC.

The "UTC" part stands for "Coordinated Universal Time". Yes, the order of the letters is strangely different from the words! It's a great metaphor for this entire subject, because nothing is as simple as you'd like. Just accept it and move on. UTC is what you may already think of as "Greenwich Mean Time", if you've ever encountered that term.

There are three wrinkles we must acknowledge, even when oversimplifying:

1. Time zones. We don't all live in Greenwich, England, so local times are
   described by an offset from UTC.
   - I live in Vancouver, British Columbia, which is 8 hours behind UTC
     (UTC−08:00). Or, at least, it is part of the year ...
2. Daylight savings time. Lots of places change their clocks twice a year. UTC
   does not! UTC "just is". This means that your UTC offset is one number part
   of the year and another number part of the year.
   - My offset is -08:00 during standard time, but is -07:00 during daylight
     savings (roughly March - October).
3. Spreadsheets are special. Spreadsheets use a different form of epoch time.
   Their epoch or origin is usually around 1900 and they keep track of how many
   *days* have elapsed since the epoch (not seconds, like Unix epoch time).
   These are sometimes called "serial dates" or "serial numbers".
   - Google Sheets use an epoch of 30 December 1899 00:00:00 UTC.   
   - Horrors I will spare you: different epochs for different spreadsheet
     applications (or versions thereof) and the Lotus 1-2-3 leap year bug.
   
You can read more in the [Sheets API docs about Date/Time serial numbers](https://developers.google.com/sheets/api/guides/concepts#datetime_serial_numbers).

## Datetimes in R

R uses Unix epoch time.

R uses the [POSIXct](https://rdrr.io/r/base/DateTimeClasses.html) class to represent datetimes. (Yes, there's also `POSIXlt`, but I recommend and will focus on `POSIXct`.)

If you ask for the current time, R prints it formatted for your time zone (or, at least, it tries). You can also ask R to reveal what it thinks your time zone is.

```{r}
Sys.time()

Sys.timezone()
```

The time zone is **purely matter of display**, but it's a really nice touch! It is comforting to get a time printed by R that matches your experience of what time it is, based on looking at the clock on your wall ("clock time").

lubridate's `with_tz()` function lets you explicitly associate a datetime with a time zone, e.g. your own or any other time zone recognized by your system. And this, in turn, affects how the time is formatted for human eyeballs.

```{r}
tt <- Sys.time()

with_tz(tt, tzone = "America/Vancouver")

with_tz(tt, tzone = "America/Denver")

with_tz(tt, tzone = "Etc/UTC")
```

**Google Sheets have no ability to display a moment in time according to a specified time zone.** Yes, each Sheet has an associated time zone, but this is not what it influences, even though you might expect or hope for that.

When we read datetimes out of a Google Sheet, we must:

  * Convert from days to seconds.
  * Adjust for Unix epoch versus spreadsheet epoch.
  * Use the resulting number to construct an instance of `POSIXct`.
  
When we write a datetime to a Google Sheet, we must:

  * Convert from seconds to days.
  * Adjust for the spreadsheet epoch versus Unix epoch.
  * Use the resulting number to create an instance of the
  [`CellData`](https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#celldata)
  schema.

## Datetimes in Google Sheets

Google Sheets use a spreadsheet-y version of epoch time. A datetime cell holds a so-called serial number, which is the elapsed days since the epoch of 30 December 1899 00:00:00 UTC. This number is then displayed in a more human-friendly way, according to a special token string. Currently googlesheets4 doesn't offer any explicit support for dealing with these [format strings](https://developers.google.com/sheets/api/guides/formats#date_and_time_format_patterns), although one day it probably will.

Let's gain some intuition by looking at datetimes shortly after the epoch and inspecting the underlying serial numbers. In a hidden chunk, we create a Sheet and read it into R.

```{r include = FALSE}
# The constructed formulas are very special because of a very weird thing!
# The =DATE() function can only be used with dates on or after 1 Jan 1900,
# even though the Sheets epoch is 30 Dec 1899.
# https://support.google.com/docs/answer/3092969?hl=en&ref_topic=3105385
#   "Google Sheets uses the 1900 date system. The first date is 1/1/1900.
#    Between 0 and 1899, Google Sheets adds that value to 1900 to calculate
#    the year. For example, DATE(119,2,1) will create a date of 2/1/2019."
# Therefore, to construct a moment in 1899, I choose to construct the date part
# from parsing a string, then I can use =TIME() to apply a time to that.
```

```{r echo = FALSE}
dat <- tibble::tibble(
  datetime = gs4_formula(c(
    '=(TO_DATE(DATEVALUE("1899-12-30"))+TIME(12,0,0))',
    '=(TO_DATE(DATEVALUE("1899-12-31"))+TIME(18,0,0))'
  )),
  serial_number = gs4_formula(c("=ARRAYFORMULA(TO_PURE_NUMBER(A2:A))", NA))
)
ss <- gs4_create("near-the-epoch", sheets = dat)
read_sheet(ss)
```

1899-12-30 12:00:00 is noon on the day that is the Google Sheets epoch. Its underlying serial number is 0.5, because one half-day has elapsed since the epoch. 1899-12-31 18:00:00 is 6pm in the evening on the day after the epoch. Its underlying serial number is 1.75, because it's one plus three-quarters of a day since the epoch.

```{r include = FALSE}
gs4_find("near-the-epoch") %>% googledrive::drive_trash()
```

Every Google Sheet has an associated time zone. It is included in the metadata returned by `gs4_get()`, like the locale, and is revealed by default when we print a Sheets ID.

```{r}
(meta <- gs4_example("gapminder") %>%
   gs4_get())

meta$time_zone
```

However, this time zone has a very different impact -- much *less* impact -- on the user experience than the time zone in R.

The Sheets' time zone does **not** influence the display of datetimes. There is no way to request that a datetime be displayed according to a specific time zone -- not via the Sheet's time zone, not via the format string, and not via a Sheets function.

Datetimes in Google Sheets are fundamentally UTC-based and always display as such.

If you want to see "9:14 am" in your Sheet, you must make sure the serial number in that cell represents 9:14 in the morning, UTC time.

As far as I can tell, here is the only effect of a Sheet's time zone: The formulas `=NOW()` and `=TODAY()` take the local clock time or date, according to the the Sheet's time zone, and construct the UTC moment or date that will display as that time or date. Therefore `=NOW()`, especially, is almost misleading! It does not capture the current moment, in UTC, but instead fabricates a UTC moment that matches current local clock time.

This suggests various hacks if you truly, deeply want to see specific clock times in your Sheet, for non-UTC time zones.

Starting with the UTC moments, you must determine and apply the offset yourself. At a very crude level, this can be done from first principles with datetime arithmetic in the Sheet ("Vancouver is −08:00, so subtract 8 hours"). But then there's daylight savings time and other complexities ("Except, during DST, subtract 7 hours."). In reality, no mere mortal will *ever* get this right, in general. If you doubt me, please watch the YouTube video [Computerphile's "The Problem with Time & Timezones"](https://youtu.be/-5wpm-gesOY).

You need to use external, authoritative offset information, either within R or in the Sheet. Below, we show how to do this in R. In Sheets, people tend to use Google Apps script and solutions based on [moment.js](https://momentjs.com).

## Worked example

Let's make all of this concrete. We construct a data frame in R with a datetime and various versions of it that explore time zone issues. We also include a couple of Google Sheet formulas, to trigger some datetime work once the data is written into a Sheet. We sketch the construction of this data frame here, with considerable abuse of notation (mixing R code and Sheets formulas):

| what | datetime | serial_number |
|---------------------|-------------------------------------------------------------------------------------------------------------------------------------|---------------------------------------|
| moment | `Sys.time()` | `=ARRAYFORMULA(TO_PURE_NUMBER(B2:B))` |
| moment_ny | `with_tz(moment, tzone = "America/New_York")` |  |
| moment_utc | `with_tz(moment, tzone = "Etc/UTC")` |  |
| moment_ny_force_utc | `force_tz(moment_ny, tzone = "Etc/UTC")` |  |
| =NOW() | `=NOW()` |  |
| =(DATE(moment_utc) + TIME(moment_utc)) | `=(DATE({year},{month},{day})+time({hour},{minute},{round(second, 1)}))` where `year`, `month`, etc. are computed from `moment_utc` |  |

Capture the current `moment` in time, with `Sys.time()`, which has no explicit time zone. Store versions of `moment` with explicit time zones: America/New_York and Etc/UTC. Use `lubridate::force_tz()` to create a new moment in time: the moment in UTC that has the same clock time as the original `moment` in New York.

The first Sheets formula we use is `=NOW()`, which you might expect to be the equivalent of R's `Sys.time()`. But it's more like `force_tz(Sys.time(), tzone = "Etc/UTC")`. The second formula we construct is more elaborate. It uses datetime functions in Sheets to explicitly construct `moment_utc` in the Sheet. The last column uses `=TO_PURE_NUMBER()` to reveal the underlying serial numbers for all of the datetimes.

```{r define-populate-sheets, include = FALSE}
populate_sheets <- function(moment, ss) {
  force(moment)
  moment_ny        <-  with_tz(moment,    tzone = "America/New_York")
  moment_utc       <-  with_tz(moment,    tzone = "Etc/UTC")
  moment_ny_utc_force <- force_tz(moment_ny, tzone = "Etc/UTC")

  # https://github.com/tidyverse/lubridate/issues/882
  # use moment_utc, so I know for sure that time zone is UTC
  fixed_formula <- gs4_formula(glue::glue(
    "=(DATE({year},{month},{day})+time({hour},{minute},{round(second, 1)}))",
    year = year(moment_utc), month = month(moment_utc), day = day(moment_utc),
    hour = hour(moment_utc), minute = minute(moment_utc),
    second = second(moment_utc)
  ))
  now_formula <- gs4_formula("=NOW()")
  serial_number <- "=ARRAYFORMULA(TO_PURE_NUMBER(B2:B))"
  length(serial_number) <- 6

  dat <- tibble::tibble(
    what = c(
      "moment", "moment_ny", "moment_utc", "moment_ny_utc_force",
      "=NOW()", "=DATE(moment_utc)"
    ),
    datetime = list(
      moment, moment_ny, moment_utc, moment_ny_utc_force,
      now_formula, fixed_formula
    ),
    serial_number = gs4_formula(serial_number)
  )
  
  range_iso <- function(ss, range) {
    fmt <- googlesheets4:::CellData(
      userEnteredFormat = googlesheets4:::new(
        "CellFormat",
        numberFormat = list(type = "DATE_TIME", pattern = "yyyy-mm-dd hh:mm:ss")
      )
    )
    range_flood(ss, sheet = 1, range = range, cell = fmt)
  }

  for (i in seq_along(ss)) {
    sheet_write(dat, ss = ss[i], sheet = 1)
    range_autofit(ss[i], sheet = 1)
    range_iso(ss[i], range = "B2:B")
  }
  dat
}
```

Create 3 Sheets, with different approaches to the time zone:

* No explicit specification of time zone. It's hard to say what you'll get here!
* America/New_York
* Etc/UTC

```{r}
ss_xx  <- gs4_create("tz-default")
ss_ny  <- gs4_create("tz-america-new-york", timeZone = "America/New_York")
ss_utc <- gs4_create("tz-etc-utc", timeZone = "Etc/UTC")

show_timezone <- function(ss) gs4_get(ss)$time_zone

show_timezone(ss_xx)
show_timezone(ss_ny)
show_timezone(ss_utc)
```

```{r eval = FALSE, include = FALSE}
# useful when developing and the sheets already exist
(ss_xx <- gs4_find("tz-default") %>% as_sheets_id())
(ss_ny <- gs4_find("tz-america-new-york") %>% as_sheets_id())
(ss_utc <- gs4_find("tz-etc-utc") %>% as_sheets_id())
```

```{r eval = FALSE, include = FALSE}
# useful during development, in order to see these sheets in the browser
gs4_share(ss_xx)
gs4_share(ss_ny)
gs4_share(ss_utc)
```

```{r eval = FALSE, include = FALSE}
# useful during development, e.g. for capturing screenshots
gs4_browse(ss_xx)
gs4_browse(ss_ny)
gs4_browse(ss_utc)
```

Capture the current `moment` with `Sys.time()`, construct the data frame described above, and write it into each of the prepared Google Sheets.

```{r}
dat <- populate_sheets(Sys.time(), c(ss_xx, ss_ny, ss_utc))
```

First, let's look at `dat`, the data frame we sent.

```{r}
dat
```

That's hard to parse since the `datetime` column is a list-column. Here's a different look, with the most natural character representation of that column.

```{r echo = FALSE}
dat2 <- dat
dat2$datetime <- vapply(
  dat$datetime, function(x) format(x, usetz = TRUE), character(1)
)
dat2
```

Read the Sheets back into R, the Sheet with no explicit time zone set.

```{r}
read_sheet(ss_xx) %>% as.data.frame()
read_sheet(ss_ny) %>% as.data.frame()
read_sheet(ss_utc) %>% as.data.frame()
```

Main conclusions:

* All 3 versions of `moment` result in the same serial number in all 3 Sheets. Lesson: in R, time zone is merely a matter of display and, in Sheets, there is only UTC.
* `moment_ny_utc_force` (forcing `moment`s NY clock time into UTC) results in the same serial number in all 3 Sheets. Lesson: If you want to see a specify clock time in the Sheet, force this on the R side, before writing to Sheets. But realize that you have fudged the datetime data in order to get the desired display.
* `=NOW()` is one of the few things affected by a Sheet's time zone (along with `=TODAY()`. It allows you to force the Sheet's clock time into UTC.

Clean up.

```{r}
gs4_find("tz-") %>% 
  googledrive::drive_trash()
```

---
title: "Messages and errors in googlesheets4"
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  error = TRUE
)
```

```{r setup}
library(googlesheets4)
```

*In a hidden chunk here, I "export" the internal helpers covered below.*

```{r include = FALSE}
gs4_bullets <- googlesheets4:::gs4_bullets

gs4_abort <- googlesheets4:::gs4_abort
abort_unsupported_conversion <- googlesheets4:::abort_unsupported_conversion

bulletize <- gargle:::bulletize
```

## User-facing messages

Everything should be emitted by helpers in `utils-ui.R`: specifically, `gs4_bullets()` and, for errors, `gs4_abort()`.
These helpers are wrappers around `cli::cli_inform()` and `cli::cli_abort()`, respectively.

```{r}
gs4_bullets(c(
        "noindent",
  " " = "indent",
  "*" = "bullet",
  ">" = "arrow",
  "v" = "Doing good stuff for YOU the user",
  "x" = "Nope nope nope",
  "!" = "You might want to know about this",
  "i" = "The more you know!"
))
```

The helpers encourage consistent styling and make it possible to selectively silence messages coming from googlesheets4.
The googlesheets4 message helpers:

-   Use the [cli package](https://cli.r-lib.org/index.html) to get interpolation, inline markup, and pluralization.
-   Use googlesheets4's custom theme, which has styles for Sheet name, worksheet name, and range.
-   Eventually route through `rlang::inform()`, which is important because `inform()` prints to standard output in interactive sessions. This means that informational messages won't have the same "look" as errors and can generally be more stylish, at least in IDEs like RStudio.
-   Are under the control of the `"googlesheets4_quiet"` option. If it's unset, the default is to show messages (unless we're testing, i.e. the environment variable `TESTTHAT` is `"true"`). `googlesheets4_quiet = TRUE` will suppress messages. There are withr-style convenience helpers: `local_gs4_quiet()` and `with_gs4_quiet()`.

### Inline styling

How we use the inline classes:

-   `.s_sheet` for the name of Google Sheet (custom)
-   `.w_sheet` for the name of a worksheet (custom)
-   `.range` for an A1-style range or named range (custom)
-   `.code` for a column in a data frame and for reserved words, such as `NULL`, `TRUE`, and `NA`
-   `.arg`, `.fun`, `.path`, `.cls`, `.url`. `.email` for their usual purpose

*These may not demo well via pkgdown, but the interactive experience is nice.*

```{r}
nm <- "name-of-a-Google-Sheet"
gs4_bullets(c(v = "Creating new Sheet: {.s_sheet {nm}}"))

nm <- "name-of-a-worksheet"
gs4_bullets(c(v = "Protecting cells on sheet: {.w_sheet {nm}}"))

rg <- "A3:B20"
gs4_bullets(c(v = "Writing to the range: {.range {rg}}"))
```

Above, I don't include a period (`.`) at the end of a message with the form: `Description of thing: THING`.
I view it as a special case of a bullet list of simple items, which also don't get periods. (`bulletize()` comes from gargle.)

```{r}
gs4_bullets(c(
  "We're going to list some things:",
  bulletize(gargle::gargle_map_cli(month.abb[1:4]))
))
```

Other messages that are complete sentence, or at least aspire to be, **do** get a period.

```{r}
gs4_bullets("Doing the stuff you asked me to do.")

gs4_bullets(c(
  "You probably need to do one of these things:",
  "*" = "Call {.fun some_function}.",
  "*" = "Provide more specific input via {.arg some_argument}."
))
```

Most relevant cli docs:

-   [CLI inline markup](https://cli.r-lib.org/reference/inline-markup.html)
-   [Building a Semantic CLI](https://cli.r-lib.org/articles/semantic-cli.html)

### Pluralization

[cli's pluralization](https://cli.r-lib.org/articles/pluralization.html) is awesome!

```{r}
nm <- "name-of-a-Google-Sheet"
n_new <- 1
gs4_bullets(c(v = "Adding {n_new} sheet{?s} to {.s_sheet {nm}}"))

n_new <- 3
gs4_bullets(c(v = "Adding {n_new} sheet{?s} to {.s_sheet {nm}}"))
```

### Collapsing

[Collapsing lists of things](https://cli.r-lib.org/articles/semantic-cli.html#inline-lists-of-items) is great!
Also more pluralization.

```{r}
new_sheet_names <- c("apple", "banana", "cherry")
gs4_bullets(c("New sheet{?s}: {.w_sheet {new_sheet_names}}"))

new_sheet_names <- "kumquat"
gs4_bullets(c("New sheet{?s}: {.w_sheet {new_sheet_names}}"))
```

### Tricky stuff

If you want to see some tricky examples of building up a message from parts, look here:

-   `range_speedread()`

## Errors

Use `gs4_abort()` instead of `rlang::abort()` or `stop()`.
So far, I'm not really using `...` to put data in the condition, but I could start when/if there's a reason to.

`abort_unsupported_conversion()` is a wrapper around `gs4_abort()`.

```{r}
x <- structure(1, class = c("a", "b", "c"))
abort_unsupported_conversion(x, to = "SheetThing")
```

`abort_unsupported_conversion()` exists to standardize a recurring type of error message, usually encountered during development, not by end-users.
I use it a lot in the default method of an `as_{to}()` generic.

There's not much to add here re: errors, now that `cli::cli_abort()` exists.
Error messages are formed just like informational messages now.

---
title: "Example Sheets"
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>", 
  error = TRUE
)
```

An index of the official example and test Sheets. They are world-readable, so we do `gs4_deauth()`.

```{r setup}
library(googlesheets4)

gs4_deauth()
```

## Example Sheets

```{r include = FALSE}
# the non-interactive wrapper prevents tons of browser tabs opening when I
# have to tinker with this code interactively
sheets <- gs4_examples()
rlang::with_interactive(
  value = FALSE,
  example_urls <- purrr::map_chr(sheets, gs4_browse)
)
links <- glue::glue("[{names(example_urls)}]({unname(example_urls)})")
dat <- tibble::tibble(
  `name (these are links)` = links,
  id = unclass(sheets)
)
```

```{r echo = FALSE}
knitr::kable(dat)
```

### How to get hold of the example Sheets yourself

`gs4_examples()` returns a named vector of Sheet IDs. It is also an instance of `drive_id`, to make good things happen with googledrive.

```{r}
gs4_examples()

gs4_examples("gap")

gs4_examples("and")
```

`gs4_example()` returns exactly one Sheet ID (or errors). It is an instance of `drive_id`. It is also an instance of `sheets_id`, which means printing will (try to) reveal current metadata about the Sheet.

```{r}
gs4_example("chicken")

gs4_example("gap")
```

Here's a handy snippet to open all the example Sheets in browser tabs.

```{r eval = FALSE}
lapply(gs4_examples(), gs4_browse)

# for tidyversers
gs4_examples() %>% 
  purrr::walk(gs4_browse)
```

## Test Sheets

These are more developer-facing, but it's convenient to do them here too.

```{r include = FALSE}
sheets <- googlesheets4:::test_sheets()
rlang::with_interactive(
  value = FALSE,
  test_urls <- purrr::map_chr(sheets, gs4_browse)
)
links <- glue::glue("[{names(test_urls)}]({unname(test_urls)})")
dat <- tibble::tibble(
  `name (these are links)` = links,
  id = unclass(sheets)
)
```

```{r echo = FALSE}
knitr::kable(dat)
```

# string with invalid character is rejected

    Code
      as_sheets_id("abc&123")
    Condition
      Error in `validate_drive_id()`:
      ! A <drive_id> must match this regular expression: `^[a-zA-Z0-9_-]+$`
      Invalid input:
      x 'abc&123'

# invalid inputs are caught

    Code
      as_sheets_id(letters[1:2])
    Condition
      Error in `validate_sheets_id()`:
      ! A <sheets_id> object can't have length greater than 1.
      x Actual input has length 2.

# multi-row dribble is rejected

    Code
      as_sheets_id(d)
    Condition
      Error in `as_sheets_id()`:
      ! <dribble> input must have exactly 1 row.
      x Actual input has 2 rows.

# dribble with non-Sheet file is rejected

    Code
      as_sheets_id(d)
    Condition
      Error in `as_sheets_id()`:
      ! <dribble> input must refer to a Google Sheet, i.e. a file with MIME type 'application/vnd.google-apps.spreadsheet'.
      i File name: "chicken.txt"
      i File id: '1wOLeWVRkTb6lDmLRiOhg9iKM7DlN762Y'
      x MIME TYPE: 'text/plain'

# sheets_id print method reveals metadata

    Code
      print(gs4_example("gapminder"))
    Output
      
      -- <googlesheets4_spreadsheet> -------------------------------------------------
       Spreadsheet name: "gapminder"                                 
                     ID: 1U6Cf_qEOhiR9AZqTqS3mbMF3zt2db48ZP5v3rkrAEJY
                 Locale: en_US                                       
              Time zone: America/Los_Angeles                         
            # of sheets: 5                                           
      # of named ranges: 1                                           
      
      -- <sheets> --------------------------------------------------------------------
      (Sheet name): (Nominal extent in rows x columns)
          'Africa': 625 x 6
        'Americas': 301 x 6
            'Asia': 397 x 6
          'Europe': 361 x 6
         'Oceania': 25 x 6
      
      -- <named ranges> --------------------------------------------------------------
      (Named range): (A1 range)        
           'canada': 'Americas'!A38:F49

# sheets_id print method doesn't error for nonexistent ID

    Code
      as_sheets_id("12345")
    Output
      
      -- <googlesheets4_spreadsheet> -------------------------------------------------
      Spreadsheet name: "<unknown>"
                    ID: 12345      
                Locale: <unknown>  
             Time zone: <unknown>  
           # of sheets: <unknown>  
      
      Unable to get metadata for this Sheet. Error details:
      Client error: (404) NOT_FOUND
      * A specified resource is not found, or the request is rejected by undisclosed
        reasons, such as whitelisting.
      * Requested entity was not found.

# can print public sheets_id if deauth'd

    Code
      print(gs4_example("mini-gap"))
    Output
      
      -- <googlesheets4_spreadsheet> -------------------------------------------------
      Spreadsheet name: "mini-gap"                                  
                    ID: 1k94ZVVl6sdj0AXfK9MQOuQ4rOhd1PULqpAu2_kr9MAU
                Locale: en_US                                       
             Time zone: America/Los_Angeles                         
           # of sheets: 5                                           
      
      -- <sheets> --------------------------------------------------------------------
      (Sheet name): (Nominal extent in rows x columns)
          'Africa': 6 x 6
        'Americas': 6 x 6
            'Asia': 6 x 6
          'Europe': 6 x 6
         'Oceania': 6 x 6

# sheets_id print does not error for lack of cred

    Code
      print(gs4_example("mini-gap"))
    Output
      
      -- <googlesheets4_spreadsheet> -------------------------------------------------
      Spreadsheet name: "<unknown>"                                 
                    ID: 1k94ZVVl6sdj0AXfK9MQOuQ4rOhd1PULqpAu2_kr9MAU
                Locale: <unknown>                                   
             Time zone: <unknown>                                   
           # of sheets: <unknown>                                   
      
      Unable to get metadata for this Sheet. Error details:
      Can't get Google credentials.
      i Are you running googlesheets4 in a non-interactive session? Consider:
      * Call `gs4_deauth()` to prevent the attempt to get credentials.
      * Call `gs4_auth()` directly with all necessary specifics.
      i See gargle's "Non-interactive auth" vignette for more details:
      i <https://gargle.r-lib.org/articles/non-interactive-auth.html>

# new() rejects data not expected for schema

    Code
      new("Spreadsheet", foofy = "blah")
    Condition
      Error in `check_against_schema()`:
      ! Properties not recognized for the 'Spreadsheet' schema:
      x 'foofy'

---

    Code
      new("Spreadsheet", foofy = "blah", foo = "bar")
    Condition
      Error in `check_against_schema()`:
      ! Properties not recognized for the 'Spreadsheet' schema:
      x 'foofy'
      x 'foo'

# check_against_schema() errors when no schema can be found

    Code
      check_against_schema(x)
    Condition
      Error in `check_against_schema()`:
      ! Trying to check an object of class <googlesheets4_schema_SomeThing/googlesheets4_schema/list>, but can't get a schema.

---
title: "Range specification"
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This is article serves mostly as internal documentation of how various representations of cell ranges relate to each other in googlesheets4.

```{r, R.options = list(tidyverse.quiet = TRUE)}
library(tidyverse)
library(googlesheets4)

gs4_deauth()
```

# User-specified range

Several googlesheets4 functions allow the user to specify which cells to read or write. This range specification comes via the `sheet`, `range`, and (possibly) `skip` arguments. Examples of functions with this interface: `range_read()`, `range_read_cells()`, and `range_write()`.

  * `range` is the primary and most powerful argument and takes precedence over
    all others. `range` can be:
      - an [A1-style spreadsheet range](https://developers.google.com/sheets/api/guides/concepts#a1_notation),
        with or without a sheet name, such as "A3:D7" or "arts!A5:F15"
      - a (work)sheet name, such as "arts"
      - a named range, such as "arts_data"
      - a `cell_limits` object made with helpers from the cellranger package
  * `sheet` can be used to specify a (work)sheet by name, such as "arts", or
    by position, such as 3
  * `skip` is an optional argument that appears in `range_read()` and 
    `range_read_cells()`, to be compatible with functions like `read.table()`,
    `readr::read_delim()`, and `readxl::read_excel()`.
    
(You might think `n_max` should also be mentioned here, but for technical reasons, in general, `n_max` can only be enforced after we've read data. So it's not part of the range ingest problem.)

Loose ends and hangnails:

  * The story around (work)sheet visibility is muddy. But I'm letting that be,
    because I have yet to have a user interaction that had anything to do with
    sheet (in)visibility.
  * The problem of a completely unspecified range. In a very hand-wavy sense,
    "no range" means "all cells" or "all non-empty cells" or "all non-empty cells in the first (visible?) sheet". That is obviously very imprecise, which
    means that I have to think about how to handle "no range" for each
    individual endpoint or function. Sometimes I inject some info (such as the
    first visible worksheet), so we don't retrieve data that we will just throw away. This is all very special case-y right now.
  * `cell_limits` objects can hold sheet information and I have made no
    arrangements for that.
  * Should `range_write()` have a skip argument, just for internal consistency?

# Range spec

`range_spec` is an internal S3 class that is typically used inside any function that accepts the `(sheet, range, skip)` trio.

We need some sort of intermediate storage, in order to translate between the various ways the user can express their range and the requirements of different Sheets API endpoints (which differ more than you'd expect!). We generally require metadata about the associated Sheet in order to form a `range_spec`, because we potentially need to lookup the (work)`sheet` by position or determine whether a name refers to a named range or a (work)`sheet`.

The internal generic `as_range_spec()` dispatches on primary argument `x`, which maps to `range`. Its job is to transform user-supplied `(sheet, range, skip)` into:

  * `sheet_name`: (Work)sheet name, can be `NULL`.
  * `named_range`: Name of range, can be `NULL`. But if not `NULL`, then this is
    the complete specification of the range, i.e. it cannot be combined with any
    other information.
  * `cell_range` and/or `cell_limits`: Ways to specify a rectangle of cells, can
    be `NULL`. If user specifies `skip`, that is re-expressed as `cell_limits`. At intake, at most one of `cell_range` and `cell_limits` is populated. But internally, we might populate one from the other, as we prepare the range to
    meet the requirements of a specific API endpoint.
  * `shim`: indicates whether the user specified a specific cell rectangle
     (`shim = TRUE`) or we filled in some specifics pragmatically
     (`shim = FALSE`), which is necessary to express some partially open
     rectangles in A1-notation. This aspect of the user-supplied data needs to
     be retained to inform downstream post-processing.
  * (`sheets_df` and `nr_df` are retained in a `range_spec`, if they were
    provided to `as_range_spec()`.)
  * Loose end: I wonder if `range_spec` should have a field for (work)sheet id.
  
Here's how various user-specified ranges are stored as a `range_spec`.

```{r include = FALSE}
# switched from read_csv because https://github.com/tidyverse/readr/issues/1237
df <- read.csv(text = '
    "sheet",                          "range", "skip"
     "NULL",                           "NULL",    "0"
   "Africa",                           "NULL",    "0"
        "4",                           "NULL",    "0"
     "NULL",                           "NULL",    "2"
   "Africa",                           "NULL",    "3"
        "4",                           "NULL",    "4"
"<ignored>",                     "Asia!A1:B2",    "0"
"<ignored>",                         "canada",    "0"
"<ignored>",                         "Europe",    "3"
     "NULL",                          "A1:B2",    "0"
  "Oceania",                          "A1:B2",    "0"
        "2",                          "A1:B2",    "0"
     "NULL", "cell_limits(c(2, 3), c(NA, 5))",   "0"
 "Americas", "cell_limits(c(2, 3), c(NA, 5))",   "0"
', strip.white = TRUE)

escape_brackets <- function(x) {
  #str_replace_all(x, c("<" = "\\\\<", ">" = "\\\\>"))
  str_replace_all(x, c("<" = "&lt;", ">" = "&gt;"))
}

formatize <- function(x) paste(escape_brackets(format(x)), collapse = "<br>")
```

```{r include = FALSE}
ss <- gs4_example("gapminder")
ss_meta <- gs4_get(ss)

df <- df %>% 
  mutate(
    sheet_orig = escape_brackets(sheet),
    range_orig = range,
    sheet = map(sheet, parse_guess),
    range = map(range, parse_guess)
  )
df$sheet[df$sheet == "NULL"] <- list(NULL)
df$range[df$range == "NULL"] <- list(NULL)
df$range <- modify_at(
  df$range,
  str_which(df$range_orig, "^cell_limits"),
  ~ eval(parse(text = .x))
)

df <- df %>%
  rename(x = range) %>% # make the generic happy
  mutate(range_spec = pmap(.,
    googlesheets4:::as_range_spec,
    sheets_df = ss_meta$sheets,
    nr_df     = ss_meta$named_ranges
  )) %>% 
  rename(range = x)
```

```{r echo = FALSE}
df %>% 
  select(-sheet, -range) %>% 
  rename(range = range_orig, sheet = sheet_orig) %>% 
  mutate(range_spec = map_chr(range_spec, formatize)) %>% 
  select(sheet, range, skip, everything()) %>% 
  knitr::kable()
```

## Cell rectangles

When a `range_spec` is first formed, it may hold specifics for a cell rectangle, although it does not have to. This info appears in one of these fields:

  * `cell_range`: an [A1-style spreadsheet range](https://developers.google.com/sheets/api/guides/concepts#a1_notation),
    with NO sheet information
    - Examples: "A4", "D:G", "3:7", "B3:H", "C4:G16"
  * a `cell_limits` object, in the cellranger sense
    - Example: `cell_limits(ul = c(2, 3), lr = c(NA, 5)))`
  
A `cell_limits` object is, overall, a much more sane way to hold this information, because explicit `NA`s can be used to convey unboundedness. Some types of unboundedness simply can't be conveyed in A1 notation, even though those partially open rectangles are as legitimate as those that can be recorded in A1 notation.

In any case, between user expectations and API idiosyncrasies, we have to be able to translate between `cell_range` and `cell_limits`.

### A1 range from `cell_limits`

The reading functions `range_read()` and `range_read_cells()` hit the `sheets.spreadsheets.get` endpoint. Bizarrely, this endpoint requires the range to be specified in A1-notation. If the user specifies the cell rectangle in A1-notation, things are easy and we essentially just pass that along. But users can describe certain partially open cell rectangles via `cell_limits` that can't be literally expressed in A1-notation. The table below shows how all possible combinations of row and cell limits are translated into an A1-style range, using technical limits on the number of rows and columns in a Google Sheet.

```{r echo = FALSE}
f <- function(start_col, start_row, end_col, end_row) {
  cellranger::cell_limits(
    ul = c(start_row, start_col), lr = c(end_row, end_col)
  )
}

g <- function(col) {
  ifelse(is.na(col), "?", cellranger::num_to_letter(col))
}

h <- function(row) ifelse(is.na(row), "?", row)

foo <- function(start_col2, start_row2, end_col2, end_row2, ...) {
  paste0(start_col2, start_row2, ":", end_col2, end_row2)
}

df <- crossing(
  start_col = c(NA, 2), start_row = c(NA, 2),
  end_col = c(NA, 4), end_row = c(NA, 4)
)

df <- df %>% 
  mutate(cell_limits = pmap(., f)) %>% 
  mutate(start_col2 = g(start_col), start_row2 = h(start_row)) %>% 
  mutate(end_col2 = g(end_col), end_row2 = h(end_row)) %>% 
  mutate(naive_range = pmap(., foo)) %>%
  mutate(range = map(cell_limits, googlesheets4:::as_sheets_range),
         range = map_chr(range, 1, .default = "&lt;NULL&gt;"))
```

```{r echo = FALSE}
knitr::kable(
  select(
    df,
    start_col, start_row, end_col, end_row,
    naive_range, range)
)
```

### `cell_limits` from A1 range

When editing a sheet, we send an [`UpdateCellsRequest`](https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#updatecellsrequest), which takes the location of the write via a *union field*. We must send one of the following:

  * `start`: An instance of `GridCoordinate` which identifies one cell. This is
    where the write starts and the cells hit are determined by the structure of
    the accompanying `RowData`. Think of `start` as specifying the upper left
    corner of a target rectangle. Fields:
    - `sheetId`: ID of the (work)sheet. The only required field.
    - `rowIndex`: Zero-based row index.  Unspecified means 0.
    - `columnIndex`: Zero-based column index.  Unspecified means 0.
  * `range`: An instance of `GridRange` which identifies a rectangle of cells.
    Fields:
    - `sheetId`: ID of the (work)sheet. The only required field.
    - `startRowIndex`: Zero-based inclusive start row. Unspecified means 0.
    - `endRowIndex`: Zero-based exclusive end row. Unspecified means unbounded.
    - `startColumnIndex`: Zero-based inclusive start column. Unspecified means
      0.
    - `endColumnIndex`: Zero-based exclusive end column. Unspecified means
      unbounded.
    
    All the start/end row/column indices are optional. If omitted, it means
    the rectangle is unbounded on that side. Unbounded on the left or top is
    boring, because that just amounts to starting at the first row or column.
    But unbounded on the right or bottom is a more interesting scenario because of this:
    
    If the `RowData` doesn't cover the entire `range` rectangle, the targetted
    cells that don't receive new data are cleared, according to the accompanying
    field mask in `fields`.

`range_write()` is the only writing function in googlesheets4 that lets the user target a specific range. This function is why we need the ability to convert A1-notation to `cell_limits` and to decide whether user's range should be sent via `start` or `range`. `gs4_create()`, `sheet_write()`, and `sheet_append()` all offer a higher-level API, focused on the holistic management of (work)sheets that hold a single data table.

Here's a table that translates the A1-style ranges created above into `cell_limits`. Some scenarios are covered more than once in this input, because above we were tackling a different (harder) problem. But that's harmless and these inputs are good, in terms of covering all relevant scenarios. Note I dropped the row corresponding to "no input".

```{r, include = FALSE}
# apparently I didn't know about format methods when I wrote cellranger :(
formaticate <- function(x) escape_brackets(capture.output(print(x)))
```

```{r echo = FALSE}
df %>%
  select(range) %>% 
  filter(!str_detect(range, "NULL")) %>% 
  mutate(cell_limits = map(range, googlesheets4:::limits_from_range)) %>% 
  mutate(cell_limits = map_chr(cell_limits, formaticate)) %>% 
  knitr::kable()
```

How to decide whether to send user's range via `start` or `range`? Possible scenarios for `(sheet, range)` (remember: so far, `range_write()` doesn't offer `skip`):

  * `sheet` not given: populate with first (visible) sheet.
  * No `range`: Send sheet ID as `GridCoordinate` via `start`. We will
    overwrite data (and usually formats) in cells covered by the accompanying
    `data`. But we don't touch the rest of the sheet. Why not send as `range`?
    That feels too aggressive because it clears the data (and usually formats)
    of any cells in the sheet that are not covered by the `data`. But if user wants that, there's always `range_write()`.
  * `range` as A1 notation: Re-express it as `cell_limits`.
  * `range` as `cell_limits` breaks down into a couple scenarios:
    - `cell_limits` contains at least 1 `NA`: send as `GridRange` via `range`.
      Analysis of the "all `NA`" edge case: we send just the sheet ID as a
      `GridRange`, so we will clear the targeted fields for the whole sheet.
      This seems OK here, because no one can do this by accident, i.e. you
      have to go out of your way to make an explicit `cell_limits` object
      containing only `NA`s. Why would you do this instead of `range_write()`?
      Maybe you need to suppress the inclusion of `col_names` or you don't
      like the table formatting applied by `range_write()`.
    - `cell_limits` has no `NA`s: If rectangle includes more than 1 cell, send
      as `GridRange` via `range`. If rectangle is exactly 1 cell, send as
      `GridCoordinate` via `start`.

# col_names must be logical or character and have length

    Code
      wrapper_fun(1:3)
    Condition
      Error in `wrapper_fun()`:
      ! `col_names` must be <character>:
      x `col_names` has class <integer>.

---

    Code
      wrapper_fun(factor("a"))
    Condition
      Error in `wrapper_fun()`:
      ! `col_names` must be <character>:
      x `col_names` has class <factor>.

---

    Code
      wrapper_fun(character())
    Condition
      Error in `wrapper_fun()`:
      ! `col_names` must have length greater than zero.

# logical col_names must be TRUE or FALSE

    Code
      wrapper_fun(NA)
    Condition
      Error in `wrapper_fun()`:
      ! `col_names` must be either `TRUE` or `FALSE`.

---

    Code
      wrapper_fun(c(TRUE, FALSE))
    Condition
      Error in `wrapper_fun()`:
      ! `col_names` must be either `TRUE` or `FALSE`.

# sheet_add() rejects non-character `sheet`

    Code
      sheet_add(test_sheet("googlesheets4-cell-tests"), sheet = 3)
    Condition
      Error in `sheet_add()`:
      ! `sheet` must be <character>:
      x `sheet` has class <numeric>.

# gs4_auth_configure works

    Code
      gs4_auth_configure(client = gargle::gargle_client(), path = "PATH")
    Condition
      Error in `gs4_auth_configure()`:
      ! Must supply exactly one of `client` and `path`, not both.

# gs4_oauth_app() is deprecated

    Code
      absorb <- gs4_oauth_app()
    Condition
      Warning:
      `gs4_oauth_app()` was deprecated in googlesheets4 1.1.0.
      i Please use `gs4_oauth_client()` instead.

# gs4_auth_configure(app =) is deprecated in favor of client

    Code
      gs4_auth_configure(app = client)
    Condition
      Warning:
      The `app` argument of `gs4_auth_configure()` is deprecated as of googlesheets4 1.1.0.
      i Please use the `client` argument instead.

# gs4_scopes() reveals Sheets scopes

    Code
      gs4_scopes()
    Output
                                                 spreadsheets 
               "https://www.googleapis.com/auth/spreadsheets" 
                                        spreadsheets.readonly 
      "https://www.googleapis.com/auth/spreadsheets.readonly" 
                                                        drive 
                      "https://www.googleapis.com/auth/drive" 
                                               drive.readonly 
             "https://www.googleapis.com/auth/drive.readonly" 
                                                   drive.file 
                 "https://www.googleapis.com/auth/drive.file" 

---
title: "How to create a googlesheets4 reprex"
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

"reprex" is short for "reproducible example". A reprex is very helpful when reporting a bug or requesting a new feature. The [reprex package](https://reprex.tidyverse.org) helps with some of the fiddly mechanics of making a self-contained, well-formatted reprex.

You might worry about some awkward things when using reprex with googlesheets4:

  * What about auth? What if you need to reprex with a private Sheet?
  * How do you make a Sheet readable by everyone or specific individuals, like
    a package maintainer?

Here we show various ways this can work, if rough order of preference.

## Use a public example Sheet

If you can make your point with one of the example Sheets, exposed via `gs4_examples()` and `gs4_example()`, do so! You can call `gs4_deauth()` explicitly in your reprex to shut down any attempt to get a token.

```{r eval = FALSE}
library(googlesheets4)

gs4_deauth()

# put the googlesheets4 code you want to reproduce below here
# the following is just an example, replace it with your own code
gs4_example("mini-gap") %>% 
  gs4_get()
```

## Create a Sheet and make it world-readable

If you can create an example Sheet that makes your point and make it readable by "anyone with a link", do so! You can call `gs4_deauth()` explicitly in your reprex to shut down any attempt to get a token.

How do you make a Sheet world-readable? Do this setup once! It should not be part of your reprex. Two options:

  * In the browser: *Share* (big green button) > *Advanced* > *Who has access ... Change* > *Link sharing: On - Anyone with the link*
  * Using googledrive:
  
    ```{r eval = FALSE}
    library(googledrive)
    
    x <- drive_get("YOUR_SHEET_NAME")
    
    drive_share(x, role = "reader", type = "anyone")
    ```

Now make a reprex just like we do with the official example Sheets:  

```{r eval = FALSE}
library(googlesheets4)

gs4_deauth()

# put the googlesheets4 code you want to reproduce below here
# the following is just an example, replace it with your own code
gs4_get("YOUR_SHEET_ID_OR_URL_GOES_HERE")
```

### Grant access to specific user(s)

If you can't create a world-readable example Sheet, perhaps you can still share one with specific individuals, such as a package maintainer. How to share with specific user(s):

  * In the browser: *Share* (big green button) > *People* > *Enter names or email addresses*.
  * Using googledrive:
  
    ```{r eval = FALSE}
    library(googledrive)
    
    x <- drive_get("YOUR_SHEET_NAME")
    
    drive_share(
      x, 
      role = "reader", type = "user",
      email_address = "jane_package_maintainer@example.org"
    )
    ```

See the next section for advice on your reprex code.

## Use a private Sheet

`reprex::reprex()` eventually runs your code in a fresh, non-interactive R session. You won't be there to do anything about auth, like select the right identity or approve the use of a cached token.

This general situation is documented in the gargle vignette [Non-interactive auth](https://gargle.r-lib.org/articles/non-interactive-auth.html). But here's the short version:

  * Develop a code snippet that works interactively for you, in a clean R
    session.
  * Note who you are logged in as. `gs4_user()` reveals this if you're not
    sure.
  * Insert an explicit call to `gs4_auth(email = "SOMEONE@example.org)`
    into your code (see below).
  * Use `reprex::reprex()`.
  
Here's an example of a snippet suitable for `reprex::reprex()`, assuming the user has successfully run it once interactively, so there's a cached token for "jane_doe@example.com".

```{r eval = FALSE}
library(googlesheets4)

gs4_auth(email = "jane_doe@example.com")

# put the googlesheets4 code you want to reproduce below here
# the following is just an example, replace it with your own code
ssid <- "some_very_long_string_of_letters_and_digits"
gs4_get(ssid)
```

If you're reluctant to reveal your email address and/or the spreadsheet id, you can use special comments to create a hidden chunk and a visible body chunk. If the Sheet is private and no one else will be able to access it anyway, this is still a good option to show exactly what you're seeing locally.

```{r eval = FALSE}
library(googlesheets4)

#+ include = FALSE
# code here is executed but the results won't appear in rendered reprex
gs4_auth(email = "jane_doe@example.com")
ssid <- "some_very_long_string_of_letters_and_digits"

#+ include = TRUE
# put the googlesheets4 code you want to reproduce below here
# the following is just an example, replace it with your own code
gs4_get(ssid)
```

---
title: "Read Sheets"
description: >
  Specifying the range and column types, getting detailed cell data
---

```{r setup, include = FALSE}
can_decrypt <- gargle:::secret_can_decrypt("googlesheets4")
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  error = TRUE,
  purl = can_decrypt,
  eval = can_decrypt
)
```

```{r eval = !can_decrypt, echo = FALSE, comment = NA}
message("No token available. Code chunks will not be evaluated.")
```

```{r}
library(googlesheets4)
```

Basic Sheet reading is shown in the [Get started](https://googlesheets4.tidyverse.org/articles/googlesheets4.html) article. Here we show how to target a specific (work)sheet or cell range, how to deal with column types, and how to get detailed cell data.

## Auth

As a regular, interactive user, you can just let googlesheets4 prompt you for anything it needs re: auth.

Since this article is compiled noninteractively on a server, we have arranged for googlesheets4 to use a service account token (not shown).

```{r include = FALSE}
# happens in .onLoad() when IN_PKGDOWN, but need this for local dev/preview
googlesheets4:::gs4_auth_docs(drive = TRUE) 

# attempt to reduce quota exhaustion problems
if (identical(Sys.getenv("IN_PKGDOWN"), "true")) Sys.sleep(30)
```

## `read_sheet()` and `range_read()` are synonyms

The main "read" function of the googlesheets4 package goes by two names, because we want it to make sense in two contexts:

* `read_sheet()` evokes other table-reading functions, like
 `readr::read_csv()` and `readxl::read_excel()`. The `sheet` in this case
  refers to a Google (spread)Sheet.
  
* `range_read()` is technically the right name according to the naming
  convention used throughout the googlesheets4 package, because we can read from
  an arbitrary cell range.
  
`read_sheet()` and `range_read()` are synonyms and you can use either one. Throughout this article, we're going to use `range_read()`.

Note: The first release of googlesheets used a `sheets_` prefix everywhere, so we had `sheets_read()`. It still works, but it's deprecated and will go away
rather swiftly.

## Specify the range and column types

Here we read from the "mini-gap" and "deaths" example Sheets to show some of the different ways to specify (work)sheet and cell ranges. 

```{r}
range_read(gs4_example("mini-gap"), sheet = 2)

range_read(gs4_example("mini-gap"), sheet = "Oceania", n_max = 3)

range_read(gs4_example("deaths"), skip = 4, n_max = 10)

```

The example below shows the use of `range` to specify both the (work)sheet and an A1-style cell range.

It also demonstrates how `col_types` gives control of column types, similar to how `col_types` works in readr and readxl. Note that currently there is only support for the "shortcode" style of column specification and we plan to align better with readr's capabilities in a future release.

* For the full list of column types and how to specify them, see the [Column specification](https://googlesheets4.tidyverse.org/reference/range_read.html#column-specification) section of the help for `range_read()`.

```{r}

range_read(
  gs4_example("deaths"), range = "other!A5:F15", col_types = "?ci??D"
)

```

```{r include = FALSE}
deaths_url <- rlang::with_interactive(value = FALSE,
  gs4_browse(gs4_example("deaths"))
)
deaths_hyperlink <- glue::glue("[here]({deaths_url})")
```

If you looked at the "deaths" spreadsheet in the browser (it's `r deaths_hyperlink`), you know that it has some of the typical features of real world spreadsheets: the main data rectangle has prose intended for human-consumption before and after it. That's why we have to specify the range when we read from it.

We've designated the data rectangles as [named ranges](https://support.google.com/docs/answer/63175?co=GENIE.Platform%3DDesktop&hl=en), which provides a very slick way to read them -- definitely less brittle and mysterious than approaches like `range = "other!A5:F15"` or `skip = 4, n_max = 10`. A named range can be passed via the `range =` argument:

```{r}
gs4_example("deaths") %>% 
  range_read(range = "arts_data")
```

The named ranges, if any exist, are part of the information returned by `gs4_get()`.

## Detailed cell data

`range_read_cells()` returns a data frame with one row per cell and it gives access to raw cell data sent by the Sheets API.

```{r}
(df <- range_read_cells(gs4_example("deaths"), range = "E5:E7"))
df$cell[[3]]
```

Specify `cell_data = "full", discard_empty = FALSE` to get even more data if you, for example, need access to cell formulas or formatting.

`spread_sheet()` converts data in the "one row per cell" form into the data frame you get from `range_read()`, which involves reshaping and column typing.

```{r}
df %>% spread_sheet(col_types = "D")
## is same as ...
range_read(gs4_example("deaths"), range = "E5:E7", col_types ="D")
```

## When speed matters

If your Sheet is so large that the speed of `range_read()` is causing problems, consider `range_speedread()`. It uses a special URL that allows a Sheet to be read as comma-separated values (CSV). Access via this URL doesn't use the Sheets API (although `range_speedread()` still makes an API call to retrieve Sheet metadata). As an example, on a Sheet with around 57,000 rows and 25 columns (over 1.4 million cells), `range_speedread()` takes ~5 seconds, whereas `range_read()` takes closer to 3 minutes. Why wouldn't we always take the faster option?!? Because the speed difference is imperceptible for many Sheets and there are some downsides (described later).

`range_speedread()` has much the same interface as `range_read()`.

```{r}
gs4_example("gapminder") %>% 
  range_speedread(sheet = "Oceania", n_max = 3)
```

The output above reveals that, under the hood, `range_speedread()` calls an external function for CSV parsing (namely, `readr::read_csv()`). An important consequence is that all arguments around column type specification are passed along to the CSV-parsing function. Here is a demo using readr-style column specification:

```{r}
gs4_example("deaths") %>% 
  range_speedread(
    range = "other!A5:F15",
    col_types = readr::cols(
      Age = readr::col_integer(),
      `Date of birth` = readr::col_date("%m/%d/%Y"),
      `Date of death` = readr::col_date("%m/%d/%Y")
    )
  )
```

Compare that to how we would read the same data with `range_read()`:

```{r}
gs4_example("deaths") %>% 
  range_read(range = "other_data", col_types = "??i?DD")
```

This example highlights two important differences:

  * `range = "other!A5:F15"` versus `range = "other_data"`: `range_speedread()`
    can't access a named range, whereas `range_read()` can.
  * `readr::col_date("%m/%d/%Y")` vs `D`: `range_speedread()` must parse a
    character representation of all cell data, including datetimes, whereas 
    `range_read()` has access to the actual cell data and its type.

What's the speed difference for something like the Africa sheet in the
"gapminder" example Sheet? (around 625 rows x 6 columns, or 3700 cells)

```{r}
system.time(
  gs4_example("gapminder") %>% range_speedread(sheet = "Africa")
)
system.time(
  gs4_example("gapminder") %>% range_read(sheet = "Africa")
)
```

The modest difference above shows that the speed difference is unlikely to be a gamechanger in many settings.

Summary of how to think about `range_speedread()` vs `range_read()`:

  * Both use auth (literally, send a token), unless there was a prior to call to
    `gs4_deauth()`.
  * `range_speedread()` is faster, but it's not noticeable for typical Sheets.
  * `range_speedread()` uses readr-style column type specification, which is
    actually more flexible than what `range_read()` currently does. In future
    googlesheets4 releases, we will adopt readr-style column type specification.
  * `range_speedread()` requires more detailed column type specification,
    because it cannot access unformatted cell data and the actual cell type, as
    `range_read()` can.
  * `range_speedread()` can't access full cell data, e.g., formatting.
  * `range_speedread()` can't work with named ranges.

---
title: "Dates and times"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Dates and times}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

<!-- dates-and-times.Rmd is generated from dates-and-times.Rmd.orig.Rmd. Please edit that file -->





This article provides advice on reading and writing datetimes with Google Sheets, specifically around the matter of time zones.

A related issue is how datetimes are formatted for presentation in the Sheet itself. You can read more about these formats in the [Date and time format patterns section of the Sheets API docs](https://developers.google.com/sheets/api/guides/formats#date_and_time_format_patterns). At the time of writing, googlesheets4 provides no user-friendly way to address these formats, although it may do so in the future.

## Attach packages and do auth

Attach googlesheets4.


```r
library(googlesheets4)
```

Since we eventually create and edit Sheets, we also auth here in a hidden chunk. If you run this code, you should expect auth to happen.
 


The lubridate package ([lubridate.tidyverse.org](https://lubridate.tidyverse.org)) is useful for this exploration, so we attach it now.


```r
library(lubridate, warn.conflicts = FALSE)
```

## How to work with time zones in a Google Sheet

You don't.

Literally, you can't.

I know this sounds very harsh, but it is the truth. Google Sheets offer essentially no support for time zones and your life will be simpler if you just make peace with this and accept that you will be looking at UTC times in Sheets.

A short demo: in R, capture the current time as `tt` and reveal the current time zone. Note that `tt` is displayed in R according to this time zone.


```r
(tt <- Sys.time())
#> [1] "2020-04-23 12:48:57 PDT"

Sys.timezone()
#> [1] "America/Vancouver"
```

Write the `tt` datetime to a Sheet, configured with the same time zone as the local R session, and create another cell that captures the exact text presented in the browser UI for `tt`. Read this back into R.


```r
dat <- tibble::tibble(
  datetime = tt,
  as_displayed = gs4_formula("=TO_TEXT(A2)")
)

(ss <- gs4_create(
  "no-time-zone-effect",
  sheets = dat,
  timeZone = Sys.timezone()
))
#> Creating new Sheet: "no-time-zone-effect"
#>   Spreadsheet name: no-time-zone-effect
#>                 ID: 1rH4nOOFBFvqCLEcYqaPjMIPj7EHfcT21tLTRL2FbHx0
#>             Locale: en_US
#>          Time zone: America/Vancouver
#>        # of sheets: 1
#> 
#> (Sheet name): (Nominal extent in rows x columns)
#>          dat: 2 x 2

read_sheet(ss)
#> Reading from "no-time-zone-effect"
#> Range "dat"
#> # A tibble: 1 x 2
#>   datetime            as_displayed       
#>   <dttm>              <chr>              
#> 1 2020-04-23 19:48:57 2020-04-23 19:48:58
```

Note that the `tt` datetime is displayed differently in Sheets than it is locally in R. Sheets presents datetimes in Coordinated Universal Time (time zone Etc/UTC), even if the Sheet's metadata specifies a different time zone, such as America/Vancouver.


```r
with_tz(tt, "Etc/UTC")
#> [1] "2020-04-23 19:48:57 UTC"
```

If you want to understand more about datetimes in R, in Sheets, and how you can sort of hack around this time zone problem, keep reading.



## Need-to-know basics of datetimes

Datetimes are a complicated topic. Here we dramatically oversimplify things, in the name of making a reader who is new to all of this at least minimally functional.

The main system used to represent times in the computing world is Unix epoch time:

> A moment in time is represented by the number of seconds that have elapsed since 1 January 1970 00:00:00 UTC.

The "UTC" part stands for "Coordinated Universal Time". Yes, the order of the letters is strangely different from the words! It's a great metaphor for this entire subject, because nothing is as simple as you'd like. Just accept it and move on. UTC is what you may already think of as "Greenwich Mean Time", if you've ever encountered that term.

There are three wrinkles we must acknowledge, even when oversimplifying:

1. Time zones. We don't all live in Greenwich, England, so local times are
   described by an offset from UTC.
   - I live in Vancouver, British Columbia, which is 8 hours behind UTC
     (UTC−08:00). Or, at least, it is part of the year ...
2. Daylight savings time. Lots of places change their clocks twice a year. UTC
   does not! UTC "just is". This means that your UTC offset is one number part
   of the year and another number part of the year.
   - My offset is -08:00 during standard time, but is -07:00 during daylight
     savings (roughly March - October).
3. Spreadsheets are special. Spreadsheets use a different form of epoch time.
   Their epoch or origin is usually around 1900 and they keep track of how many
   *days* have elapsed since the epoch (not seconds, like Unix epoch time).
   These are sometimes called "serial dates" or "serial numbers".
   - Google Sheets use an epoch of 30 December 1899 00:00:00 UTC.   
   - Horrors I will spare you: different epochs for different spreadsheet
     applications (or versions thereof) and the Lotus 1-2-3 leap year bug.
   
You can read more in the [Sheets API docs about Date/Time serial numbers](https://developers.google.com/sheets/api/guides/concepts#datetime_serial_numbers).

## Datetimes in R

R uses Unix epoch time.

R uses the [POSIXct](https://rdrr.io/r/base/DateTimeClasses.html) class to represent datetimes. (Yes, there's also `POSIXlt`, but I recommend and will focus on `POSIXct`.)

If you ask for the current time, R prints it formatted for your time zone (or, at least, it tries). You can also ask R to reveal what it thinks your time zone is.


```r
Sys.time()
#> [1] "2020-04-23 12:49:04 PDT"

Sys.timezone()
#> [1] "America/Vancouver"
```

The time zone is **purely matter of display**, but it's a really nice touch! It is comforting to get a time printed by R that matches your experience of what time it is, based on looking at the clock on your wall ("clock time").

lubridate's `with_tz()` function lets you explicitly associate a datetime with a time zone, e.g. your own or any other time zone recognized by your system. And this, in turn, affects how the time is formatted for human eyeballs.


```r
tt <- Sys.time()

with_tz(tt, tzone = "America/Vancouver")
#> [1] "2020-04-23 12:49:04 PDT"

with_tz(tt, tzone = "America/Denver")
#> [1] "2020-04-23 13:49:04 MDT"

with_tz(tt, tzone = "Etc/UTC")
#> [1] "2020-04-23 19:49:04 UTC"
```

**Google Sheets have no ability to display a moment in time according to a specified time zone.** Yes, each Sheet has an associated time zone, but this is not what it influences, even though you might expect or hope for that.

When we read datetimes out of a Google Sheet, we must:

  * Convert from days to seconds.
  * Adjust for Unix epoch versus spreadsheet epoch.
  * Use the resulting number to construct an instance of `POSIXct`.
  
When we write a datetime to a Google Sheet, we must:

  * Convert from seconds to days.
  * Adjust for the spreadsheet epoch versus Unix epoch.
  * Use the resulting number to create an instance of the
  [`CellData`](https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#celldata)
  schema.

## Datetimes in Google Sheets

Google Sheets use a spreadsheet-y version of epoch time. A datetime cell holds a so-called serial number, which is the elapsed days since the epoch of 30 December 1899 00:00:00 UTC. This number is then displayed in a more human-friendly way, according to a special token string. Currently googlesheets4 doesn't offer any explicit support for dealing with these [format strings](https://developers.google.com/sheets/api/guides/formats#date_and_time_format_patterns), although one day it probably will.

Let's gain some intuition by looking at datetimes shortly after the epoch and inspecting the underlying serial numbers. In a hidden chunk, we create a Sheet and read it into R.




```
#> Creating new Sheet: "near-the-epoch"
#> Reading from "near-the-epoch"
#> Range "dat"
#> # A tibble: 2 x 2
#>   datetime            serial_number
#>   <dttm>                      <dbl>
#> 1 1899-12-30 12:00:00          0.5 
#> 2 1899-12-31 18:00:00          1.75
```

1899-12-30 12:00:00 is noon on the day that is the Google Sheets epoch. Its underlying serial number is 0.5, because one half-day has elapsed since the epoch. 1899-12-31 18:00:00 is 6pm in the evening on the day after the epoch. Its underlying serial number is 1.75, because it's one plus three-quarters of a day since the epoch.



Every Google Sheet has an associated time zone. It is included in the metadata returned by `gs4_get()`, like the locale, and is revealed by default when we print a Sheets ID.


```r
(meta <- gs4_example("gapminder") %>%
   gs4_get())
#>   Spreadsheet name: gapminder
#>                 ID: 1U6Cf_qEOhiR9AZqTqS3mbMF3zt2db48ZP5v3rkrAEJY
#>             Locale: en_US
#>          Time zone: America/Los_Angeles
#>        # of sheets: 5
#> 
#> (Sheet name): (Nominal extent in rows x columns)
#>       Africa: 625 x 6
#>     Americas: 301 x 6
#>         Asia: 397 x 6
#>       Europe: 361 x 6
#>      Oceania: 25 x 6
#> 
#> (Named range): (A1 range)        
#>        canada: 'Americas'!A38:F49

meta$time_zone
#> [1] "America/Los_Angeles"
```

However, this time zone has a very different impact -- much *less* impact -- on the user experience than the time zone in R.

The Sheets' time zone does **not** influence the display of datetimes. There is no way to request that a datetime be displayed according to a specific time zone -- not via the Sheet's time zone, not via the format string, and not via a Sheets function.

Datetimes in Google Sheets are fundamentally UTC-based and always display as such.

If you want to see "9:14 am" in your Sheet, you must make sure the serial number in that cell represents 9:14 in the morning, UTC time.

As far as I can tell, here is the only effect of a Sheet's time zone: The formulas `=NOW()` and `=TODAY()` take the local clock time or date, according to the the Sheet's time zone, and construct the UTC moment or date that will display as that time or date. Therefore `=NOW()`, especially, is almost misleading! It does not capture the current moment, in UTC, but instead fabricates a UTC moment that matches current local clock time.

This suggests various hacks if you truly, deeply want to see specific clock times in your Sheet, for non-UTC time zones.

Starting with the UTC moments, you must determine and apply the offset yourself. At a very crude level, this can be done from first principles with datetime arithmetic in the Sheet ("Vancouver is −08:00, so subtract 8 hours"). But then there's daylight savings time and other complexities ("Except, during DST, subtract 7 hours."). In reality, no mere mortal will *ever* get this right, in general. If you doubt me, please watch the YouTube video [Computerphile's "The Problem with Time & Timezones"](https://youtu.be/-5wpm-gesOY).

You need to use external, authoritative offset information, either within R or in the Sheet. Below, we show how to do this in R. In Sheets, people tend to use Google Apps script and solutions based on [moment.js](https://momentjs.com).

## Worked example

Let's make all of this concrete. We construct a data frame in R with a datetime and various versions of it that explore time zone issues. We also include a couple of Google Sheet formulas, to trigger some datetime work once the data is written into a Sheet. We sketch the construction of this data frame here, with considerable abuse of notation (mixing R code and Sheets formulas):

| what | datetime | serial_number |
|---------------------|-------------------------------------------------------------------------------------------------------------------------------------|---------------------------------------|
| moment | `Sys.time()` | `=ARRAYFORMULA(TO_PURE_NUMBER(B2:B))` |
| moment_ny | `with_tz(moment, tzone = "America/New_York")` |  |
| moment_utc | `with_tz(moment, tzone = "Etc/UTC")` |  |
| moment_ny_force_utc | `force_tz(moment_ny, tzone = "Etc/UTC")` |  |
| =NOW() | `=NOW()` |  |
| =(DATE(moment_utc) + TIME(moment_utc)) | `=(DATE({year},{month},{day})+time({hour},{minute},{round(second, 1)}))` where `year`, `month`, etc. are computed from `moment_utc` |  |

Capture the current `moment` in time, with `Sys.time()`, which has no explicit time zone. Store versions of `moment` with explicit time zones: America/New_York and Etc/UTC. Use `lubridate::force_tz()` to create a new moment in time: the moment in UTC that has the same clock time as the original `moment` in New York.

The first Sheets formula we use is `=NOW()`, which you might expect to be the equivalent of R's `Sys.time()`. But it's more like `force_tz(Sys.time(), tzone = "Etc/UTC")`. The second formula we construct is more elaborate. It uses datetime functions in Sheets to explicitly construct `moment_utc` in the Sheet. The last column uses `=TO_PURE_NUMBER()` to reveal the underlying serial numbers for all of the datetimes.



Create 3 Sheets, with different approaches to the time zone:

* No explicit specification of time zone. It's hard to say what you'll get here!
* America/New_York
* Etc/UTC


```r
ss_xx  <- gs4_create("tz-default")
#> Creating new Sheet: "tz-default"
ss_ny  <- gs4_create("tz-america-new-york", timeZone = "America/New_York")
#> Creating new Sheet: "tz-america-new-york"
ss_utc <- gs4_create("tz-etc-utc", timeZone = "Etc/UTC")
#> Creating new Sheet: "tz-etc-utc"

show_timezone <- function(ss) gs4_get(ss)$time_zone

show_timezone(ss_xx)
#> [1] "Etc/GMT"
show_timezone(ss_ny)
#> [1] "America/New_York"
show_timezone(ss_utc)
#> [1] "Etc/UTC"
```







Capture the current `moment` with `Sys.time()`, construct the data frame described above, and write it into each of the prepared Google Sheets.


```r
dat <- populate_sheets(Sys.time(), c(ss_xx, ss_ny, ss_utc))
#> Writing to "tz-default"
#> Writing to sheet "Sheet1"
#> Editing "tz-default"
#> Resizing one or more columns in "Sheet1"
#> Editing "tz-default"
#> Editing sheet "Sheet1"
#> Writing to "tz-america-new-york"
#> Writing to sheet "Sheet1"
#> Editing "tz-america-new-york"
#> Resizing one or more columns in "Sheet1"
#> Editing "tz-america-new-york"
#> Editing sheet "Sheet1"
#> Writing to "tz-etc-utc"
#> Writing to sheet "Sheet1"
#> Editing "tz-etc-utc"
#> Resizing one or more columns in "Sheet1"
#> Editing "tz-etc-utc"
#> Editing sheet "Sheet1"
```

First, let's look at `dat`, the data frame we sent.


```r
dat
#> # A tibble: 6 x 3
#>   what                datetime   serial_number                      
#>   <chr>               <list>     <fmla>                             
#> 1 moment              <dttm [1]> =ARRAYFORMULA(TO_PURE_NUMBER(B2:B))
#> 2 moment_ny           <dttm [1]> NA                                 
#> 3 moment_utc          <dttm [1]> NA                                 
#> 4 moment_ny_utc_force <dttm [1]> NA                                 
#> 5 =NOW()              <fmla>     NA                                 
#> 6 =DATE(moment_utc)   <fmla>     NA
```

That's hard to parse since the `datetime` column is a list-column. Here's a different look, with the most natural character representation of that column.


```
#> # A tibble: 6 x 3
#>   what             datetime                     serial_number                      
#>   <chr>            <chr>                        <fmla>                             
#> 1 moment           2020-04-23 12:49:20 PDT      =ARRAYFORMULA(TO_PURE_NUMBER(B2:B))
#> 2 moment_ny        2020-04-23 15:49:20 EDT      NA                                 
#> 3 moment_utc       2020-04-23 19:49:20 UTC      NA                                 
#> 4 moment_ny_utc_f… 2020-04-23 15:49:20 UTC      NA                                 
#> 5 =NOW()           =NOW()                       NA                                 
#> 6 =DATE(moment_ut… =(DATE(2020,4,23)+time(19,4… NA
```

Read the Sheets back into R, the Sheet with no explicit time zone set.


```r
read_sheet(ss_xx) %>% as.data.frame()
#> Reading from "tz-default"
#> Range "Sheet1"
#>                  what            datetime serial_number
#> 1              moment 2020-04-23 19:49:20      43944.83
#> 2           moment_ny 2020-04-23 19:49:20      43944.83
#> 3          moment_utc 2020-04-23 19:49:20      43944.83
#> 4 moment_ny_utc_force 2020-04-23 15:49:20      43944.66
#> 5              =NOW() 2020-04-23 19:49:22      43944.83
#> 6   =DATE(moment_utc) 2020-04-23 19:49:20      43944.83
read_sheet(ss_ny) %>% as.data.frame()
#> Reading from "tz-america-new-york"
#> Range "Sheet1"
#>                  what            datetime serial_number
#> 1              moment 2020-04-23 19:49:20      43944.83
#> 2           moment_ny 2020-04-23 19:49:20      43944.83
#> 3          moment_utc 2020-04-23 19:49:20      43944.83
#> 4 moment_ny_utc_force 2020-04-23 15:49:20      43944.66
#> 5              =NOW() 2020-04-23 15:49:23      43944.66
#> 6   =DATE(moment_utc) 2020-04-23 19:49:20      43944.83
read_sheet(ss_utc) %>% as.data.frame()
#> Reading from "tz-etc-utc"
#> Range "Sheet1"
#>                  what            datetime serial_number
#> 1              moment 2020-04-23 19:49:20      43944.83
#> 2           moment_ny 2020-04-23 19:49:20      43944.83
#> 3          moment_utc 2020-04-23 19:49:20      43944.83
#> 4 moment_ny_utc_force 2020-04-23 15:49:20      43944.66
#> 5              =NOW() 2020-04-23 19:49:25      43944.83
#> 6   =DATE(moment_utc) 2020-04-23 19:49:20      43944.83
```

Main conclusions:

* All 3 versions of `moment` result in the same serial number in all 3 Sheets. Lesson: in R, time zone is merely a matter of display and, in Sheets, there is only UTC.
* `moment_ny_utc_force` (forcing `moment`s NY clock time into UTC) results in the same serial number in all 3 Sheets. Lesson: If you want to see a specify clock time in the Sheet, force this on the R side, before writing to Sheets. But realize that you have fudged the datetime data in order to get the desired display.
* `=NOW()` is one of the few things affected by a Sheet's time zone (along with `=TODAY()`. It allows you to force the Sheet's clock time into UTC.

Clean up.


```r
gs4_find("tz-") %>% 
  googledrive::drive_trash()
#> Files trashed:
#>   * tz-etc-utc: 1xUiNXYsk6nDMmgDq8E5FhBwJA6Zg1XCyhkuQISI91As
#>   * tz-america-new-york: 1RYjk25FKJH2QRpfT1fHGmomeO2BQ5VqErWQADLGeFJI
#>   * tz-default: 1X9Zqe--GLw6X4jajyHVb9XmKh-AiLMwZKk_g5GoJl3Y
```

---
title: "Write Sheets"
description: >
  Targeting different parts of a Sheet, writing formulas, addressing cell format
---

```{r setup, include = FALSE}
can_decrypt <- gargle:::secret_can_decrypt("googlesheets4")
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  error = TRUE,
  purl = can_decrypt,
  eval = can_decrypt
)
```

```{r eval = !can_decrypt, echo = FALSE, comment = NA}
message("No token available. Code chunks will not be evaluated.")
```

```{r}
library(googlesheets4)
```

Basic Sheet writing is shown in the [Get started](https://googlesheets4.tidyverse.org/articles/googlesheets4.html) article. Here we explore the different uses for the writing functions:

  * `gs4_create()` emphasizes the creation of a new (spread)Sheet.
    - But it can create (work)sheets and write data.
  * `sheet_write()` emphasizes writing a data frame into a (work)sheet.
    - But it can create a new (spread)Sheet or a new (work)sheet.
  * `sheet_append()` is about adding rows to an existing data table.
  * `gs4_create()`, `sheet_write()`, and `sheet_append()` implement
    holistic operations for representing a single R data frame as a table in
    a (work)sheet within a Google Sheet.
    - `gs4_create()` and `sheet_write()` both impose this mentality via
      specific formatting, such as special treatment of the column header row.
    - All 3 functions aim to shrink-wrap the data.
  * `range_write()` is the one function that helps you write arbitrary data
    into an arbitrary range, although it is still oriented around a data frame.
  * `range_flood()` writes common content into all of the cells in a range.
    - An important special case is *clearing* a range of cells of their
      existing values and, optionally, formatting.
  * `range_delete()` deletes a range of cells and shifts other cells into the
    deleted area.

## Auth

As a regular, interactive user, you can just let googlesheets4 prompt you for anything it needs re: auth.

Since this article is compiled noninteractively on a server, we have arranged for googlesheets4 to use a service account token (not shown).

```{r include = FALSE}
# happens in .onLoad() when IN_PKGDOWN, but need this for local dev/preview
googlesheets4:::gs4_auth_docs(drive = TRUE)

# attempt to reduce quota exhaustion problems
if (identical(Sys.getenv("IN_PKGDOWN"), "true")) Sys.sleep(30)
```

## `gs4_create()`

Create a brand new Sheet with `gs4_create()`. You can specify the new Sheet's `name` (or accept a randomly generated name).

```{r}
ss1 <- gs4_create("sheets-create-demo-1")
ss1
```

Every Sheet *must* have at least one (work)sheet, so Google Sheets automatically creates an empty "Sheet1".

You can control the names and content of the initial (work)sheets with the `sheets` argument.

### Send sheet names

Use a character vector to specify the names of one or more empty sheets.

```{r}
ss2 <- gs4_create(
  "sheets-create-demo-2",
  sheets = c("alpha", "beta")
)
ss2
```

These sheets have no values and get their dimensions from Sheets default behaviour.

### Send a data frame

If you provide a data frame, it is used to populate the cells of a sheet and to set sheet dimensions (number of rows and columns). The header row also gets special treatment. The sheet inherits the name of the data frame, where possible.

```{r}
my_data <- data.frame(x = 1:3, y = letters[1:3])

ss3 <- gs4_create(
  "sheets-create-demo-3",
  sheets = my_data
)
ss3
```

### Send multiple data frames

If you provide a list of data frames, each is used to populate the cells of one sheet and to set sheet dimensions (number of rows and columns). The header row also gets special treatment. The sheets inherit the names from the list, if it has names.

```{r}
my_data_frames <- list(iris = head(iris), chickwts = head(chickwts))

ss4 <- gs4_create(
  "sheets-create-demo-4",
  sheets = my_data_frames
)
ss4
```

### Write metadata

Most users won't need to do this, but `gs4_create()` can set additional Sheet-level metadata, such as locale or time zone. To really make use of this feature, you need to read up on the [`spreadsheets.create` endpoint](https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/create).

Notice how the default empty Sheet here is named "Feuille 1", since we request a French locale.

```{r}
ss5 <- gs4_create(
  "sheets-create-demo-5",
  locale = "fr_FR",
  timeZone = "Europe/Paris"
)
ss5
```

I would only do this if you have specific evidence that the default behaviour with respect to locale and time zone is problematic for your use case.

### Clean up

Trash all the Sheets created above. This actually requires googledrive, since it is not possible to trash or delete Sheets through the Sheets API. In our hidden auth process, described earlier, we put a shared token into force for both Sheets and Drive. You can read how to do that in your own work in the article [Using googlesheets4 with googledrive](https://googlesheets4.tidyverse.org/articles/articles/drive-and-sheets.html).

```{r}
gs4_find("sheets-create-demo") %>%
  googledrive::drive_trash()
```

## `write_sheet()`, a.k.a. `sheet_write()`

`write_sheet()` is aliased to `sheet_write()` and is meant to evoke `readr::write_csv()` or `writexl::write_xlsx()`. Whereas `gs4_create()` emphasizes the *target Sheet*, `sheet_write()` emphasizes the *data you want to write*.

The only required argument for `sheet_write()` is the data.

```{r}
df <- data.frame(x = 1:3, y = letters[1:3])

random_ss <- sheet_write(df)
random_ss
```

This creates a new (work)sheet inside a new (spread)Sheet and returns its ID. You'll notice the new Sheet has a randomly generated name. If that is a problem, use `gs4_create()` instead, which affords more control over various aspects of the new Sheet.

Let's start over: we delete that Sheet and call `gs4_create()`, so we can specify the new Sheet's name. Then we'll modify it with `sheet_write()`. We send one sheet name, "chickwts", to prevent the creation of "Sheet1", but we send no data.

```{r}
googledrive::drive_trash(random_ss)

ss1 <- gs4_create(
  "write-sheets-demo-1",
  sheets = "chickwts"
)
ss1
```

`sheet_write()` allows us to write the actual `chickwts` data into the sheet by that name. 
```{r}
sheet_write(chickwts, ss = ss1, sheet = "chickwts")
ss1
```

`sheet_write()` can also write data into a new sheet, if `sheet` implicitly or explicitly targets a non-existent sheet.

```{r}
# explicitly make a new sheet named "iris"
sheet_write(iris, ss = ss1, sheet = "iris")

# implicitly make a new sheet named "mtcars"
sheet_write(mtcars, ss = ss1)
```

If no sheet name is given and it can't be determined from `data`, a name of the form "SheetN" is automatically generated by Sheets.

```{r}
sheet_write(data.frame(x = 1:2, y = 3:4), ss = ss1)
ss1
```

### Clean up

```{r}
gs4_find("write-sheets-demo") %>%
  googledrive::drive_trash()
```

## `sheet_append()`

```{r include = FALSE}
# attempt to reduce quota exhaustion problems
if (identical(Sys.getenv("IN_PKGDOWN"), "true")) Sys.sleep(100)
```

`sheet_append()` can add one or more rows of data to an existing (work)sheet.

Let's recreate the table of "other" deaths from an example Sheet, but without the annoying text above and below the data. First we bring that data into a local data frame and chop it into pieces.

```{r}
(deaths <- gs4_example("deaths") %>%
   range_read(range = "other_data", col_types = "????DD"))

deaths_zero  <- deaths[integer(), ] # "scaffolding" data frame with 0 rows
deaths_one   <- deaths[1:5, ] 
deaths_two   <- deaths[6, ]
deaths_three <- deaths[7:10, ]
```

We use `gs4_create()` to create a new (spread)Sheet and initialize a new (work)sheet with the "deaths" column headers, but no data. A second empty row is created (it must be, in order for us to freeze the top row), but it will soon be filled when we append.

```{r}
ss <- gs4_create("sheets-append-demo", sheets = list(deaths = deaths_zero))
ss
```

If you're following along, I recommend you open this Sheet in a web browser with `gs4_browse()` and revisit as we go along, to see how the initial empty row gets consumed and how additional rows are added to the targetted sheet automatically.

```{r}
gs4_browse(ss)
```

Send the data, one or more rows at a time. Keep inspecting in the browser if you're doing this yourself.

```{r}
ss

# send several rows
ss %>% sheet_append(deaths_one)
ss

# send a single row
ss %>% sheet_append(deaths_two)
ss

# send remaining rows
ss %>% sheet_append(deaths_three)
ss
```

Now the big reveal: have we successfully rebuilt that data through incremental updates?

```{r}
deaths_replica <- range_read(ss, col_types = "????DD")
identical(deaths, deaths_replica)
```

Gosh I hope that's still `TRUE` as it was the last time I checked this article!

### Clean up

```{r}
gs4_find("sheets-append-demo") %>%
  googledrive::drive_trash()
```

## `range_write()`

`range_write()` is the least opinionated writing function. It writes data into a range. It does no explicit formatting, although it can effectively apply formatting by *clearing* existing formats via `reformat = TRUE` (the default).

We focus here on the geometry of `range_write()`, i.e. which cells are edited.

In a hidden chunk, we've created a demo Sheet `ss_edit` and we've filled the cells with "-". We've also created `read_this()`, a wrapper around `range_read()` that names columns by letter, like spreadsheets do.

```{r include = FALSE}
library(tidyverse)

n <-  7

clear <- function(n) {
  suppressMessages(
    as_tibble(matrix("-", nrow = n, ncol = n), .name_repair = "unique")
  )
}

populate <- function(ss, sheet = NULL, range = NULL, col_names = FALSE) {
  suppressMessages(
    range_write(clear(n), ss = ss, sheet = sheet, range = range, col_names = col_names)
  )
}

read_this <- function(ss) {
  range_read(
    ss = ss,
    col_names = FALSE,
    .name_repair = ~ LETTERS[seq_along(.x)]
  )
}

ss_edit <- gs4_create("sheets-edit-demo")
#googlesheets4:::gs4_share(ss_edit) %>% gs4_browse()
sheet_resize(ss_edit, nrow = n, ncol = n, exact = TRUE)
populate(ss_edit)
```

Here's the initial state of `ss_edit`:

```{r}
read_this(ss_edit)
```

`df` is a small data frame we'll send as the `data` argument of `range_write()`:

```{r}
(df <- tibble(V1 = head(LETTERS,3), V2 = tail(LETTERS, 3)))
```

If we do not specify the range, `df` is written into the upper left corner and only affects cells spanned by `df`. To see where we've written, focus on the cells are NOT "x".

```{r}
range_write(ss_edit, data = df) %>% read_this()
```

(Here, and between all subsequent chunks, we reset `ss_edit` to its initial state.)

```{r, include = FALSE}
populate(ss_edit)
```

If we target a **single cell** with `range`, it specifies the upper left corner of the target area. The cells written are determined by the extent of the data.

```{r}
range_write(ss_edit, data = df, range = "C2") %>% read_this()
```

```{r, include = FALSE}
populate(ss_edit)
```

If `range` specifies multiple cells (it can even be unbounded on some sides), it is taken literally and all covered cells are written. If `range` is larger than the data, this results in some cells being *cleared* of their values. In this example, the `range` is "too big" for the data, so the remaining cells are cleared of their existing "-" value.

```{r}
range_write(ss_edit, data = df, range = "D4:G7") %>% read_this()
```

```{r, include = FALSE}
populate(ss_edit)
```

Here's another case where the `range` is bigger than it needs to be and it's unbounded on the bottom and top:

```{r}
range_write(ss_edit, data = df, range = "B:E") %>% read_this()
```

```{r, include = FALSE}
populate(ss_edit)
```

Here's another `range` that's unbounded on the left and "too big":

```{r}
range_write(ss_edit, data = df, range = "B2:6") %>% read_this()
```

```{r, include = FALSE}
populate(ss_edit)
```

The target sheet will be expanded, if necessary, if and only if `range` is a single cell (i.e. it gives the upper left corner).

```{r}
range_write(ss_edit, data = df, range = "G6") %>% read_this()
```

Although the `data` argument of `range_write()` must be a data frame, note that this does not actually limit what you can write:

 * Use `col_names = FALSE` to suppress sending the column names.
 * By definition, each variable of a data frame can be of different type.
 * By using a list-column, each row of a data frame column can be of different
   type.
   
The examples for `range_write()` show writing data of disparate type to a 1-row or a 1-column region.

### Clean up

```{r}
gs4_find("sheets-edit-demo") %>%
  googledrive::drive_trash()
```

## Write formulas

All the writing functions can write *formulas* into cells, if you indicate this in the R object you are writing, i.e. in the data frame. The `gs4_formula()` function marks a character vector as containing Sheets formulas, as opposed to regular character strings.

Here's a demo that also shows off using the Google Translate API inside a Sheet.

```{r}
lang_dat <- tibble::tribble(
       ~ english, ~ to,
           "dog", "es",
   "hello world", "ko",
  "baby turtles", "th" 
)
lang_dat$translated <- gs4_formula(
  '=GoogleTranslate(INDIRECT("R[0]C[-2]", FALSE), "en", INDIRECT("R[0]C[-1]", FALSE))'
)

(ss <- gs4_create("sheets-formula-demo", sheets = lang_dat))
```

Now we can read the data back out, complete with translations!

```{r}
range_read(ss)
```

### Clean up

```{r}
gs4_find("sheets-formula-demo") %>%
  googledrive::drive_trash()
```

## `range_flood()`

`range_flood()` "floods" all cells in a range with common content. By default, it floods them with no content, i.e. it clears cells of their existing value and format. Note that `range_clear()` is a convenience wrapper for this special case.

First, we create a data frame and initialize a new Sheet with that data.

```{r}
df <- gs4_fodder(10)
(ss <- gs4_create("range-flood-demo", sheets = list(df)))
```

To begin, each cell holds its A1-coordinates as its value. If you viewed this in the browser, you'd see the header row has some special formatting, including a gray background.

```{r}
range_read(ss)
```

By default, `range_flood()` sends an empty cell (and clears existing formatting).

```{r}
range_flood(ss, range = "A1:B3")
range_read(ss)
```

We can't easily demonstrate this here, but if you are working with a sheet in the browser, you can experiment with `reformat = TRUE/FALSE` to clear the existing format (or not).

We can also send a new value into a range of cells.

```{r}
range_flood(ss, range = "4:5", cell = ";-)")
range_read(ss)
```

### Clean up

```{r}
gs4_find("range-flood-demo") %>%
  googledrive::drive_trash()
```

## `range_delete()`

`range_delete()` deletes a range of cells and shifts other cells into the deleted area. If you're trying to delete an entire (work)sheet or change its actual dimensions or extent, you should, instead, use `sheet_delete()` or
`sheet_resize()`. If you just want to clear values or formats, use `range_clear()`.

First, we create a data frame and initialize a new Sheet with that data.

```{r}
df <- gs4_fodder(4)
(ss <- gs4_create("sheets-delete-demo", sheets = list(df)))
```

To begin, each cell holds its A1-coordinates as its value. If you viewed this in the browser, you'd see the header row has some special formatting, including a gray background.

```{r}
range_read(ss)
```

Let's delete some rows.

```{r}
range_delete(ss, range = "2:3")
range_read(ss)
```

Let's delete a column.

```{r}
range_delete(ss, range = "C")
range_read(ss)
```

Let's delete a bounded cell range. Since it's unclear how to fill this space, we must specify `shift`. We'll fill the space by shifting remaining cells to the left. Remember that `range` continues to refer to actual rows and columns and, due to our deletions, no longer match with the values which reflect the original locations.

```{r}
range_delete(ss, range = "B2:B3", shift = "left")
range_read(ss)
```

### Clean up

```{r}
gs4_find("sheets-delete-demo") %>%
  googledrive::drive_trash()
```

404-nonexistent-sheet.R
================
jenny
Fri Mar 30 16:05:19 2018

``` r
devtools::load_all(".")
```

    ## Loading googlesheets4

``` r
req <- request_generate(
  "spreadsheets.get",
  ## ID of 'googlesheets4-design-exploration', but replaced last 2 chars w/ '-'
  list(spreadsheetId = "1xTUxWGcFLtDIHoYJ1WsjQuLmpUtBf--8Bcu5lQ302--"),
  token = NULL
)
raw_resp <- request_make(req)
response_process(raw_resp)
```

    ## Error: HTTP error 404
    ## * message: 'Requested entity was not found.'
    ## * status: 'NOT_FOUND'

``` r
ct <- httr::content(raw_resp)
str(ct)
```

    ## List of 1
    ##  $ error:List of 3
    ##   ..$ code   : int 404
    ##   ..$ message: chr "Requested entity was not found."
    ##   ..$ status : chr "NOT_FOUND"

---
title: "Get started with googlesheets4"
---

```{r setup, include = FALSE}
can_decrypt <- gargle:::secret_can_decrypt("googlesheets4")
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  error = TRUE,
  purl = can_decrypt,
  eval = can_decrypt
)
options(tibble.print_min = 5L, tibble.print_max = 5L)
```

```{r eval = !can_decrypt, echo = FALSE, comment = NA}
message("No token available. Code chunks will not be evaluated.")
```

```{r index-auth, include = FALSE}
googlesheets4:::gs4_auth_docs()
```

```{r}
library(googlesheets4)
```

This article takes a quick tour of the main features of googlesheets4. 
Remember to see [the articles](https://googlesheets4.tidyverse.org/articles/index.html) for more detailed treatment of all these topics and more.

## `read_sheet()`, a.k.a. `range_read()`

`read_sheet()` is the main "read" function and should evoke `readr::read_csv()` and `readxl::read_excel()`. It's an alias for `range_read()`, which is the correct name for this function according to the scheme for naming googlesheets4 functions. You can use them interchangeably. googlesheets4 is pipe-friendly (and reexports `%>%`), but works just fine without the pipe.

`read_sheet()` is designed to "just work", for most purposes, most of the time. It can read straight from a Sheets browser URL:

```{r}
read_sheet("https://docs.google.com/spreadsheets/d/1U6Cf_qEOhiR9AZqTqS3mbMF3zt2db48ZP5v3rkrAEJY/edit#gid=780868077")
```

However, these URLs are not pleasant to work with. More often, you will want to identify a Sheet by its ID:

```{r}
read_sheet("1U6Cf_qEOhiR9AZqTqS3mbMF3zt2db48ZP5v3rkrAEJY")
```

or by its name, which requires an assist from the googledrive package ([googledrive.tidyverse.org](https://googledrive.tidyverse.org)):

<!-- remove the 'message = i' later -->

```{r message=4}
library(googledrive)

drive_get("gapminder") %>% 
  read_sheet()
```

Note that the name-based approach above will only work if **you** have access to a Sheet named "gapminder". Sheet names cannot be used as absolute identifiers; only a Sheet ID can play that role.

For more Sheet identification concepts and strategies, see the article [Find and Identify Sheets](https://googlesheets4.tidyverse.org/articles/articles/find-identify-sheets.html). See the article [Read Sheets](https://googlesheets4.tidyverse.org/articles/articles/read-sheets.html) for more about reading from a specific (work)sheet or ranges, setting column type, and getting low-level cell data.

## Example Sheets and `gs4_browse()`

We've made a few Sheets available to "anyone with a link", for use in examples and docs. Two helper functions make it easy to get your hands on these file IDs.

  * `gs4_examples()` lists all the example Sheets and it can also filter by
    matching names to a regular expression.
  * `gs4_example()` requires a regular expression and returns exactly 1 Sheet
    ID (or throws an error).
    
```{r}
gs4_example("chicken-sheet") %>% 
  read_sheet()
```

If you'd like to see a Sheet in the browser, including our example Sheets, use `gs4_browse()`:

```{r}
gs4_example("deaths") %>%
  gs4_browse()
```

## Sheet metadata

`gs4_get()` exposes Sheet metadata, such as details on worksheets and named ranges.

```{r}
ss <- gs4_example("deaths")

gs4_get(ss)

sheet_properties(ss)

sheet_names(ss)
```

`sheet_properties()` and `sheet_names()` are two members of a larger family of functions for dealing with the (work)sheets within a (spread)Sheet.

The metadata exposed by `gs4_get()` is also revealed whenever you print an object that is (or can be converted to) a `sheets_id` (an S3 class we use to mark Sheet IDs).

`gs4_get()` is related to `googledrive::drive_get()`. Both functions return metadata about a file on Google Drive, such as its ID and name. However, `gs4_get()` reveals additional metadata that is specific to Drive files that happen to be Sheets, such as info about worksheets and named ranges.

## Writing Sheets

*The writing functions are the most recent additions and may still see some refinements re: user interface and which function does what. We're very interested to hear how these functions feel in terms of ergonomics.*

`sheet_write()` writes a data frame into a Sheet. The only required argument is the data.

```{r}
df <- data.frame(x = 1:3, y = letters[1:3])

ss <- sheet_write(df)
ss
```

You'll notice the new (spread)Sheet has a randomly generated name. If that is a problem, use `gs4_create()` instead, which affords more control over various aspects of the new Sheet.

Let's start over: we delete that Sheet and call `gs4_create()`, so we can specify the new Sheet's name.

```{r}
googledrive::drive_trash(ss)

ss <- gs4_create("testy-hedgehog", sheets = df)
ss
```

`sheet_write()` can write to new or existing (work)sheets in this Sheet. Let's write the `chickwts` data to a new sheet in `ss`.

```{r}
sheet_write(chickwts, ss)
ss
```

We can also use `sheet_write()` to replace the data in an existing sheet.

```{r}
sheet_write(data.frame(x = 4:10, letters[4:10]), ss, sheet = "df")
read_sheet(ss, sheet = "df")
```

`sheet_append()` adds one or more rows to an existing sheet.

```{r}
ss %>% sheet_append(data.frame(x = 11, letters[11]), sheet = "df")
read_sheet(ss, sheet = "df")
```

A related function -- `range_write()` -- writes arbitrary data, into an arbitrary range. It has a very different "feel" from `gs4_create()`, `sheet_write()`, and `sheet_append()`, all of which assume we're writing or growing a table of data in a (work)sheet. `range_write()` is much more surgical and limited. `range_write()` makes fewer assumptions about what it's writing and why.

There is also a family of `sheet_*()` functions that do pure (work)sheet operations, such as add and delete.

We take one last look at the sheets we created in `ss`, then clean up.

```{r}
sheet_properties(ss)

googledrive::drive_trash(ss)
```

The article [Write Sheets](https://googlesheets4.tidyverse.org/articles/articles/write-sheets.html) has even more detail.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sheet_rename.R
\name{sheet_rename}
\alias{sheet_rename}
\title{Rename a (work)sheet}
\usage{
sheet_rename(ss, sheet = NULL, new_name)
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}

\item{sheet}{Sheet to rename, in the sense of "worksheet" or "tab". You can identify a sheet by name, with a string, or by position, with a number. Defaults to the first visible sheet.}

\item{new_name}{New name of the sheet, as a string. This is required.}
}
\value{
The input \code{ss}, as an instance of \code{\link{sheets_id}}
}
\description{
Changes the name of a (work)sheet.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
ss <- gs4_create(
  "sheet-rename-demo",
  sheets = list(cars = head(cars), chickwts = head(chickwts))
)
sheet_names(ss)

ss \%>\%
  sheet_rename(1, new_name = "automobiles") \%>\%
  sheet_rename("chickwts", new_name = "poultry")

# clean up
gs4_find("sheet-rename-demo") \%>\%
  googledrive::drive_trash()
\dontshow{\}) # examplesIf}
}
\seealso{
Makes an \code{UpdateSheetPropertiesRequest}:
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#UpdateSheetPropertiesRequest}
}

Other worksheet functions: 
\code{\link{sheet_add}()},
\code{\link{sheet_append}()},
\code{\link{sheet_copy}()},
\code{\link{sheet_delete}()},
\code{\link{sheet_properties}()},
\code{\link{sheet_relocate}()},
\code{\link{sheet_resize}()},
\code{\link{sheet_write}()}
}
\concept{worksheet functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_auth.R
\name{gs4_auth}
\alias{gs4_auth}
\title{Authorize googlesheets4}
\usage{
gs4_auth(
  email = gargle::gargle_oauth_email(),
  path = NULL,
  scopes = "https://www.googleapis.com/auth/spreadsheets",
  cache = gargle::gargle_oauth_cache(),
  use_oob = gargle::gargle_oob_default(),
  token = NULL
)
}
\arguments{
\item{email}{Optional. Allows user to target a specific Google identity. If
specified, this is used for token lookup, i.e. to determine if a suitable
token is already available in the cache. If no such token is found, \code{email}
is used to pre-select the targetted Google identity in the OAuth chooser.
Note, however, that the email associated with a token when it's cached is
always determined from the token itself, never from this argument. Use \code{NA}
or \code{FALSE} to match nothing and force the OAuth dance in the browser. Use
\code{TRUE} to allow email auto-discovery, if exactly one matching token is
found in the cache. Specify just the domain with a glob pattern, e.g.
\code{"*@example.com"}, to create code that "just works" for both
\code{alice@example.com} and \code{bob@example.com}. Defaults to the option named
"gargle_oauth_email", retrieved by \code{\link[gargle:gargle_oauth_email]{gargle_oauth_email()}}.}

\item{path}{JSON identifying the service account, in one of the forms
supported for the \code{txt} argument of \code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}} (typically, a
file path or JSON string).}

\item{scopes}{A character vector of scopes to request. Pick from those listed
at \url{https://developers.google.com/identity/protocols/oauth2/scopes}.

For certain token flows, the
\code{"https://www.googleapis.com/auth/userinfo.email"} scope is unconditionally
included. This grants permission to retrieve the email address associated
with a token; gargle uses this to index cached OAuth tokens. This grants no
permission to view or send email and is generally considered a low-value
scope.}

\item{cache}{Specifies the OAuth token cache. Defaults to the option named
\code{"gargle_oauth_cache"}, retrieved via \code{\link[gargle:gargle_oauth_cache]{gargle_oauth_cache()}}.}

\item{use_oob}{Whether to prefer out-of-band authentication. Defaults to the
value returned by \code{\link[gargle:gargle_oob_default]{gargle_oob_default()}}.}

\item{token}{A token with class \link[httr:Token-class]{Token2.0} or an object of
httr's class \code{request}, i.e. a token that has been prepared with
\code{\link[httr:config]{httr::config()}} and has a \link[httr:Token-class]{Token2.0} in the
\code{auth_token} component.}
}
\description{
Authorize googlesheets4 to view and manage your Google Sheets. This function is a
wrapper around \code{\link[gargle:token_fetch]{gargle::token_fetch()}}.

By default, you are directed to a web browser, asked to sign in to your
Google account, and to grant googlesheets4 permission to operate on your
behalf with Google Sheets. By default, with your permission, these user
credentials are cached in a folder below your home directory, from where
they can be automatically refreshed, as necessary. Storage at the user
level means the same token can be used across multiple projects and
tokens are less likely to be synced to the cloud by accident.

If you are interacting with R within a browser (applies to RStudio Server,
Posit Workbench, and Posit Cloud), you need a variant of this flow,
known as out-of-band auth ("oob"). If this does not happen
automatically, you can request it yourself with \code{use_oob = TRUE} or,
more persistently, by setting an option via
\code{options(gargle_oob_default = TRUE)}.
}
\details{
Most users, most of the time, do not need to call \code{gs4_auth()}
explicitly -- it is triggered by the first action that requires
authorization. Even when called, the default arguments often suffice.
However, when necessary, this function allows the user to explicitly:
\itemize{
\item Declare which Google identity to use, via an email address. If there
are multiple cached tokens, this can clarify which one to use. It can
also force googlesheets4 to switch from one identity to another. If
there's no cached token for the email, this triggers a return to the
browser to choose the identity and give consent. You can specify just
the domain by using a glob pattern. This means that a script
containing \code{email = "*@example.com"} can be run without further
tweaks on the machine of either \code{alice@example.com} or
\code{bob@example.com}.
\item Use a service account token or workload identity federation.
\item Bring their own \link[httr:Token-class]{Token2.0}.
\item Specify non-default behavior re: token caching and out-of-bound
authentication.
\item Customize scopes.
}

For details on the many ways to find a token, see
\code{\link[gargle:token_fetch]{gargle::token_fetch()}}. For deeper control over auth, use
\code{\link[=gs4_auth_configure]{gs4_auth_configure()}} to bring your own OAuth client or API key.
Read more about gargle options, see \link[gargle:gargle_options]{gargle::gargle_options}.
}
\examples{
\dontshow{if (rlang::is_interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# load/refresh existing credentials, if available
# otherwise, go to browser for authentication and authorization
gs4_auth()

# force use of a token associated with a specific email
gs4_auth(email = "jenny@example.com")

# use a 'read only' scope, so it's impossible to edit or delete Sheets
gs4_auth(
  scopes = "https://www.googleapis.com/auth/spreadsheets.readonly"
)

# use a service account token
gs4_auth(path = "foofy-83ee9e7c9c48.json")
\dontshow{\}) # examplesIf}
}
\seealso{
Other auth functions: 
\code{\link{gs4_auth_configure}()},
\code{\link{gs4_deauth}()}
}
\concept{auth functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_auth.R
\name{gs4_has_token}
\alias{gs4_has_token}
\title{Is there a token on hand?}
\usage{
gs4_has_token()
}
\value{
Logical.
}
\description{
Reports whether googlesheets4 has stored a token, ready for use in downstream
requests.
}
\examples{
gs4_has_token()
}
\seealso{
Other low-level API functions: 
\code{\link{gs4_token}()},
\code{\link{request_generate}()},
\code{\link{request_make}()}
}
\concept{low-level API functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_get.R
\name{gs4_get}
\alias{gs4_get}
\title{Get Sheet metadata}
\usage{
gs4_get(ss)
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}
}
\value{
A list with S3 class \code{googlesheets4_spreadsheet}, for printing
purposes.
}
\description{
Retrieve spreadsheet-specific metadata, such as details on the individual
(work)sheets or named ranges.
\itemize{
\item \code{gs4_get()} complements \code{\link[googledrive:drive_get]{googledrive::drive_get()}}, which
returns metadata that exists for any file on Drive.
}
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
gs4_get(gs4_example("mini-gap"))
\dontshow{\}) # examplesIf}
}
\seealso{
Wraps the \code{spreadsheets.get} endpoint:
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/get}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/range_flood.R
\name{range_flood}
\alias{range_flood}
\alias{range_clear}
\title{Flood or clear a range of cells}
\usage{
range_flood(ss, sheet = NULL, range = NULL, cell = NULL, reformat = TRUE)

range_clear(ss, sheet = NULL, range = NULL, reformat = TRUE)
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}

\item{sheet}{Sheet to write into, in the sense of "worksheet" or "tab". You can identify a sheet by name, with a string, or by position, with a number.}

\item{range}{A cell range to read from. If \code{NULL}, all non-empty cells are
read. Otherwise specify \code{range} as described in \href{https://developers.google.com/sheets/api/guides/concepts#a1_notation}{Sheets A1 notation}
or using the helpers documented in \link{cell-specification}. Sheets uses
fairly standard spreadsheet range notation, although a bit different from
Excel. Examples of valid ranges: \code{"Sheet1!A1:B2"}, \code{"Sheet1!A:A"},
\code{"Sheet1!1:2"}, \code{"Sheet1!A5:A"}, \code{"A1:B2"}, \code{"Sheet1"}. Interpreted
strictly, even if the range forces the inclusion of leading, trailing, or
embedded empty rows or columns. Takes precedence over \code{skip}, \code{n_max} and
\code{sheet}. Note \code{range} can be a named range, like \code{"sales_data"}, without
any cell reference.}

\item{cell}{The value to fill the cells in the \code{range} with. If unspecified,
the default of \code{NULL} results in clearing the existing value.}

\item{reformat}{Logical, indicates whether to reformat the affected cells.
Currently googlesheets4 provides no real support for formatting, so
\code{reformat = TRUE} effectively means that edited cells become unformatted.}
}
\value{
The input \code{ss}, as an instance of \code{\link{sheets_id}}
}
\description{
\code{range_flood()} "floods" a range of cells with the same content.
\code{range_clear()} is a wrapper that handles the common special case of
clearing the cell value. Both functions, by default, also clear the format,
but this can be specified via \code{reformat}.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
# create a data frame to use as initial data
df <- gs4_fodder(10)

# create Sheet
ss <- gs4_create("range-flood-demo", sheets = list(df))

# default behavior (`cell = NULL`): clear value and format
range_flood(ss, range = "A1:B3")

# clear value but preserve format
range_flood(ss, range = "C1:D3", reformat = FALSE)

# send new value
range_flood(ss, range = "4:5", cell = ";-)")

# send formatting
# WARNING: use these unexported, internal functions at your own risk!
# This not (yet) officially supported, but it's possible.
blue_background <- googlesheets4:::CellData(
  userEnteredFormat = googlesheets4:::new(
    "CellFormat",
    backgroundColor = googlesheets4:::new(
      "Color",
      red = 159 / 255, green = 183 / 255, blue = 196 / 255
    )
  )
)
range_flood(ss, range = "I:J", cell = blue_background)

# range_clear() is a shortcut where `cell = NULL` always
range_clear(ss, range = "9:9")
range_clear(ss, range = "10:10", reformat = FALSE)

# clean up
gs4_find("range-flood-demo") \%>\%
  googledrive::drive_trash()
\dontshow{\}) # examplesIf}
}
\seealso{
Makes a \code{RepeatCellRequest}:
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#repeatcellrequest}
}

Other write functions: 
\code{\link{gs4_create}()},
\code{\link{gs4_formula}()},
\code{\link{range_delete}()},
\code{\link{range_write}()},
\code{\link{sheet_append}()},
\code{\link{sheet_write}()}
}
\concept{write functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_find.R
\name{gs4_find}
\alias{gs4_find}
\title{Find Google Sheets}
\usage{
gs4_find(...)
}
\arguments{
\item{...}{Arguments (other than \code{type}, which is hard-wired as \code{type = "spreadsheet"}) that are passed along to \code{\link[googledrive:drive_find]{googledrive::drive_find()}}.}
}
\value{
An object of class \code{\link[googledrive]{dribble}}, a tibble with one row per file.
}
\description{
Finds your Google Sheets. This is a very thin wrapper around
\code{\link[googledrive:drive_find]{googledrive::drive_find()}}, that specifies you want to list Drive files
where \code{type = "spreadsheet"}. Therefore, note that this will require auth for
googledrive! See the article \href{https://googlesheets4.tidyverse.org/articles/articles/drive-and-sheets.html}{Using googlesheets4 with googledrive}
if you want to coordinate auth between googlesheets4 and googledrive. This
function will emit an informational message if you are currently logged in
with both googlesheets4 and googledrive, but as different users.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
# see all your Sheets
gs4_find()

# see 5 Sheets, prioritized by creation time
x <- gs4_find(order_by = "createdTime desc", n_max = 5)
x

# hoist the creation date, using other packages in the tidyverse
# x \%>\%
#   tidyr::hoist(drive_resource, created_on = "createdTime") \%>\%
#   dplyr::mutate(created_on = as.Date(created_on))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/googlesheets4-package.R, R/utils-ui.R
\name{googlesheets4-configuration}
\alias{googlesheets4-configuration}
\alias{local_gs4_quiet}
\alias{with_gs4_quiet}
\title{googlesheets4 configuration}
\usage{
local_gs4_quiet(env = parent.frame())

with_gs4_quiet(code)
}
\arguments{
\item{env}{The environment to use for scoping}

\item{code}{Code to execute quietly}
}
\description{
Some aspects of googlesheets4 behaviour can be controlled via an option.
}
\section{Messages}{


The \code{googlesheets4_quiet} option can be used to suppress messages from
googlesheets4. By default, googlesheets4 always messages, i.e. it is \emph{not}
quiet.

Set \code{googlesheets4_quiet} to \code{TRUE} to suppress messages, by one of these
means, in order of decreasing scope:
\itemize{
\item Put \code{options(googlesheets4_quiet = TRUE)} in a start-up file, such as
\code{.Rprofile}, or in your R script
\item Use \code{local_gs4_quiet()} to silence googlesheets4 in a specific scope
\item Use \code{with_gs4_quiet()} to run a small bit of code silently
}

\code{local_gs4_quiet()} and \code{with_gs4_quiet()} follow the conventions of the
the withr package (\url{https://withr.r-lib.org}).
}

\section{Auth}{


Read about googlesheets4's main auth function, \code{\link[=gs4_auth]{gs4_auth()}}. It is powered
by the gargle package, which consults several options:
\itemize{
\item Default Google user or, more precisely, \code{email}: see
\code{\link[gargle:gargle_options]{gargle::gargle_oauth_email()}}
\item Whether or where to cache OAuth tokens: see
\code{\link[gargle:gargle_options]{gargle::gargle_oauth_cache()}}
\item Whether to prefer "out-of-band" auth: see
\code{\link[gargle:gargle_options]{gargle::gargle_oob_default()}}
\item Application Default Credentials: see \code{\link[gargle:credentials_app_default]{gargle::credentials_app_default()}}
}
}

\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
# message: "Creating new Sheet ..."
(ss <- gs4_create("gs4-quiet-demo", sheets = "alpha"))

# message: "Editing ..., Writing ..."
range_write(ss, data = data.frame(x = 1, y = "a"))

# suppress messages for a small amount of code
with_gs4_quiet(
  ss \%>\% sheet_append(data.frame(x = 2, y = "b"))
)

# message: "Writing ..., Appending ..."
ss \%>\% sheet_append(data.frame(x = 3, y = "c"))

# suppress messages until end of current scope
local_gs4_quiet()
ss \%>\% sheet_append(data.frame(x = 4, y = "d"))

# see that all the data was, in fact, written
read_sheet(ss)

# clean up
gs4_find("gs4-quiet-demo") \%>\%
  googledrive::drive_trash()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_endpoints.R
\name{gs4_endpoints}
\alias{gs4_endpoints}
\title{List Sheets endpoints}
\usage{
gs4_endpoints(i = NULL)
}
\arguments{
\item{i}{The name(s) or integer index(ices) of the endpoints to return.
Optional. By default, the entire list is returned.}
}
\value{
A list containing some or all of the subset of the Sheets API v4
endpoints that are used internally by googlesheets4.
}
\description{
Returns a list of selected Sheets API v4 endpoints, as stored inside the
googlesheets4 package. The names of this list (or the \code{id} sub-elements) are
the nicknames that can be used to specify an endpoint in
\code{\link[=request_generate]{request_generate()}}. For each endpoint, we store its nickname or \code{id}, the
associated HTTP \code{method}, the \code{path}, and details about the parameters. This
list is derived programmatically from the Sheets API v4 Discovery
Document (\verb{https://www.googleapis.com/discovery/v1/apis/sheets/v4/rest}).
}
\examples{
str(gs4_endpoints(), max.level = 2)
gs4_endpoints("sheets.spreadsheets.values.get")
gs4_endpoints(4)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_auth.R
\name{gs4_user}
\alias{gs4_user}
\title{Get info on current user}
\usage{
gs4_user()
}
\value{
An email address or, if no token has been loaded, \code{NULL}.
}
\description{
Reveals the email address of the user associated with the current token.
If no token has been loaded yet, this function does not initiate auth.
}
\examples{
gs4_user()
}
\seealso{
\code{\link[gargle:token-info]{gargle::token_userinfo()}}, \code{\link[gargle:token-info]{gargle::token_email()}},
\code{\link[gargle:token-info]{gargle::token_tokeninfo()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_example.R
\name{gs4_examples}
\alias{gs4_examples}
\alias{gs4_example}
\title{Example Sheets}
\usage{
gs4_examples(matches)

gs4_example(matches)
}
\arguments{
\item{matches}{A regular expression that matches the name of the desired
example Sheet(s). \code{matches} is optional for the plural \code{gs4_examples()}
and, if provided, it can match multiple Sheets. The singular
\code{gs4_example()} requires \code{matches} and it must match exactly one Sheet.}
}
\value{
\itemize{
\item \code{gs4_example()}: a \link{sheets_id}
\item \code{gs4_examples()}: a named vector of all built-in examples, with class
\code{\link[googledrive:drive_id]{drive_id}}
}
}
\description{
googlesheets4 makes a variety of world-readable example Sheets available for
use in documentation and reprexes. These functions help you access the
example Sheets. See \code{vignette("example-sheets", package = "googlesheets4")}
for more.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
gs4_examples()
gs4_examples("gap")

gs4_example("gapminder")
gs4_example("deaths")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sheet_properties.R
\name{sheet_properties}
\alias{sheet_properties}
\alias{sheet_names}
\title{Get data about (work)sheets}
\usage{
sheet_properties(ss)

sheet_names(ss)
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}
}
\value{
\itemize{
\item \code{sheet_properties()}: A tibble with one row per (work)sheet.
\item \code{sheet_names()}: A character vector of (work)sheet names.
}
}
\description{
Reveals full metadata or just the names for the (work)sheets inside a
(spread)Sheet.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
ss <- gs4_example("gapminder")
sheet_properties(ss)
sheet_names(ss)
\dontshow{\}) # examplesIf}
}
\seealso{
Other worksheet functions: 
\code{\link{sheet_add}()},
\code{\link{sheet_append}()},
\code{\link{sheet_copy}()},
\code{\link{sheet_delete}()},
\code{\link{sheet_relocate}()},
\code{\link{sheet_rename}()},
\code{\link{sheet_resize}()},
\code{\link{sheet_write}()}
}
\concept{worksheet functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_auth.R
\name{gs4_deauth}
\alias{gs4_deauth}
\title{Suspend authorization}
\usage{
gs4_deauth()
}
\description{
Put googlesheets4 into a de-authorized state. Instead of sending a token,
googlesheets4 will send an API key. This can be used to access public
resources for which no Google sign-in is required. This is handy for using
googlesheets4 in a non-interactive setting to make requests that do not
require a token. It will prevent the attempt to obtain a token
interactively in the browser. The user can configure their own API key
via \code{\link[=gs4_auth_configure]{gs4_auth_configure()}} and retrieve that key via
\code{\link[=gs4_api_key]{gs4_api_key()}}.
In the absence of a user-configured key, a built-in default key is used.
}
\examples{
\dontshow{if (rlang::is_interactive()) withAutoprint(\{ # examplesIf}
gs4_deauth()
gs4_user()

# get metadata on the public 'deaths' spreadsheet
gs4_example("deaths") \%>\%
  gs4_get()
\dontshow{\}) # examplesIf}
}
\seealso{
Other auth functions: 
\code{\link{gs4_auth}()},
\code{\link{gs4_auth_configure}()},
\code{\link{gs4_scopes}()}
}
\concept{auth functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/range_read_cells.R
\name{range_read_cells}
\alias{range_read_cells}
\title{Read cells from a Sheet}
\usage{
range_read_cells(
  ss,
  sheet = NULL,
  range = NULL,
  skip = 0,
  n_max = Inf,
  cell_data = c("default", "full"),
  discard_empty = TRUE
)
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}

\item{sheet}{Sheet to read, in the sense of "worksheet" or "tab". You can identify a sheet by name, with a string, or by position, with a number. Ignored if the sheet is specified via \code{range}. If neither argument specifies the sheet, defaults to the first visible sheet.}

\item{range}{A cell range to read from. If \code{NULL}, all non-empty cells are
read. Otherwise specify \code{range} as described in \href{https://developers.google.com/sheets/api/guides/concepts#a1_notation}{Sheets A1 notation}
or using the helpers documented in \link{cell-specification}. Sheets uses
fairly standard spreadsheet range notation, although a bit different from
Excel. Examples of valid ranges: \code{"Sheet1!A1:B2"}, \code{"Sheet1!A:A"},
\code{"Sheet1!1:2"}, \code{"Sheet1!A5:A"}, \code{"A1:B2"}, \code{"Sheet1"}. Interpreted
strictly, even if the range forces the inclusion of leading, trailing, or
embedded empty rows or columns. Takes precedence over \code{skip}, \code{n_max} and
\code{sheet}. Note \code{range} can be a named range, like \code{"sales_data"}, without
any cell reference.}

\item{skip}{Minimum number of rows to skip before reading anything, be it
column names or data. Leading empty rows are automatically skipped, so this
is a lower bound. Ignored if \code{range} is given.}

\item{n_max}{Maximum number of data rows to parse into the returned tibble.
Trailing empty rows are automatically skipped, so this is an upper bound on
the number of rows in the result. Ignored if \code{range} is given. \code{n_max} is
imposed locally, after reading all non-empty cells, so, if speed is an
issue, it is better to use \code{range}.}

\item{cell_data}{How much detail to get for each cell. \code{"default"} retrieves
the fields actually used when googlesheets4 guesses or imposes cell and
column types. \code{"full"} retrieves all fields in the \href{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#CellData}{\code{CellData} schema}.
The main differences relate to cell formatting.}

\item{discard_empty}{Whether to discard cells that have no data. Literally,
we check for an \code{effectiveValue}, which is one of the fields in the
\href{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#CellData}{\code{CellData} schema}.}
}
\value{
A tibble with one row per cell in the \code{range}.
}
\description{
This low-level function returns cell data in a tibble with one row per cell.
This tibble has integer variables \code{row} and \code{col} (referring to location
with the Google Sheet), an A1-style reference \code{loc}, and a \code{cell}
list-column. The flagship function \code{\link[=read_sheet]{read_sheet()}}, a.k.a. \code{\link[=range_read]{range_read()}}, is
what most users are looking for, rather than \code{range_read_cells()}.
\code{\link[=read_sheet]{read_sheet()}} is basically \code{range_read_cells()} (this function), followed by
\code{\link[=spread_sheet]{spread_sheet()}}, which looks after reshaping and column typing. But if you
really want raw cell data from the API, \code{range_read_cells()} is for you!
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
range_read_cells(gs4_example("deaths"), range = "arts_data")

# if you want detailed and exhaustive cell data, do this
range_read_cells(
  gs4_example("formulas-and-formats"),
  cell_data = "full",
  discard_empty = FALSE
)
\dontshow{\}) # examplesIf}
}
\seealso{
Wraps the \code{spreadsheets.get} endpoint:
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/get}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_auth.R
\name{gs4_scopes}
\alias{gs4_scopes}
\title{Produce scopes specific to the Sheets API}
\usage{
gs4_scopes(scopes = NULL)
}
\arguments{
\item{scopes}{One or more API scopes. Each scope can be specified in full or,
for Sheets API-specific scopes, in an abbreviated form that is recognized by
\code{\link[=gs4_scopes]{gs4_scopes()}}:
\itemize{
\item "spreadsheets" = "https://www.googleapis.com/auth/spreadsheets"
(the default)
\item "spreadsheets.readonly" =
"https://www.googleapis.com/auth/spreadsheets.readonly"
\item "drive" = "https://www.googleapis.com/auth/drive"
\item "drive.readonly" = "https://www.googleapis.com/auth/drive.readonly"
\item "drive.file" = "https://www.googleapis.com/auth/drive.file"
}

See
\url{https://developers.google.com/identity/protocols/oauth2/scopes#sheets} for
details on the permissions for each scope.}
}
\value{
A character vector of scopes.
}
\description{
When called with no arguments, \code{gs4_scopes()} returns a named character
vector of scopes associated with the Sheets API. If \code{gs4_scopes(scopes =)} is
given, an abbreviated entry such as \code{"sheets.readonly"} is expanded to a full
scope (\code{"https://www.googleapis.com/auth/sheets.readonly"} in this case).
Unrecognized scopes are passed through unchanged.
}
\examples{
gs4_scopes("spreadsheets")
gs4_scopes("spreadsheets.readonly")
gs4_scopes("drive")
gs4_scopes()
}
\seealso{
\url{https://developers.google.com/identity/protocols/oauth2/scopes#sheets} for
details on the permissions for each scope.

Other auth functions: 
\code{\link{gs4_auth}()},
\code{\link{gs4_auth_configure}()},
\code{\link{gs4_deauth}()}
}
\concept{auth functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/request_make.R
\name{request_make}
\alias{request_make}
\title{Make a Google Sheets API request}
\usage{
request_make(x, ..., encode = "json")
}
\arguments{
\item{x}{List. Holds the components for an HTTP request, presumably created
with \code{\link[=request_generate]{request_generate()}} or \code{\link[gargle:request_develop]{gargle::request_build()}}. Must contain a
\code{method} and \code{url}. If present, \code{body} and \code{token} are used.}

\item{...}{Optional arguments passed through to the HTTP method.}

\item{encode}{If the body is a named list, how should it be encoded? This has
the same meaning as \code{encode} in all the \code{\link[httr:VERB]{httr::VERB()}}s, such as
\code{\link[httr:POST]{httr::POST()}}. Note, however, that we default to \code{encode = "json"}, which
is what you want most of the time when calling the Sheets API. The httr
default is \code{"multipart"}. Other acceptable values are \code{"form"} and \code{"raw"}.}
}
\value{
Object of class \code{response} from \link[httr:httr-package]{httr::httr}.
}
\description{
Low-level function to execute a Sheets API request. Most users should,
instead, use higher-level wrappers that facilitate common tasks, such as
reading or writing worksheets or cell ranges. The functions here are intended
for internal use and for programming around the Sheets API.
}
\details{
\code{make_request()} is a very thin wrapper around \code{\link[gargle:request_retry]{gargle::request_retry()}},
only adding the googlesheets4 user agent. Typically the input has been
created with \code{\link[=request_generate]{request_generate()}} or \code{\link[gargle:request_develop]{gargle::request_build()}} and the output
is processed with \code{process_response()}.

\code{\link[gargle:request_retry]{gargle::request_retry()}} retries requests that error with \verb{429 RESOURCE_EXHAUSTED}. Its basic scheme is exponential backoff, with one tweak
that is very specific to the Sheets API, which has documented \href{https://developers.google.com/sheets/api/limits}{usage limits}:

"a limit of 500 requests per 100 seconds per project and 100 requests per 100
seconds per user"

Note that the "project" here means everyone using googlesheets4 who hasn't
configured their own OAuth client. This is potentially a lot of users, all
acting independently.

If you hit the "100 requests per 100 seconds per \strong{user}" limit (which
really does mean YOU), the first wait time is a bit more than 100 seconds,
then we revert to exponential backoff.

If you experience lots of retries, especially with 100 second delays, it
means your use of googlesheets4 is more than casual and \strong{it's time for you
to get your own OAuth client or use a service account token}. This is explained
in the gargle vignette \code{vignette("get-api-credentials", package = "gargle")}.
}
\seealso{
Other low-level API functions: 
\code{\link{gs4_has_token}()},
\code{\link{gs4_token}()},
\code{\link{request_generate}()}
}
\concept{low-level API functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/googlesheets4-package.R
\docType{package}
\name{googlesheets4-package}
\alias{googlesheets4}
\alias{googlesheets4-package}
\title{googlesheets4: Access Google Sheets using the Sheets API V4}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Interact with Google Sheets through the Sheets API v4 \url{https://developers.google.com/sheets/api}. "API" is an acronym for "application programming interface"; the Sheets API allows users to interact with Google Sheets programmatically, instead of via a web browser. The "v4" refers to the fact that the Sheets API is currently at version 4. This package can read and write both the metadata and the cell data in a Sheet.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://googlesheets4.tidyverse.org}
  \item \url{https://github.com/tidyverse/googlesheets4}
  \item Report bugs at \url{https://github.com/tidyverse/googlesheets4/issues}
}

}
\author{
\strong{Maintainer}: Jennifer Bryan \email{jenny@posit.co} (\href{https://orcid.org/0000-0002-6983-2759}{ORCID})

Other contributors:
\itemize{
  \item Posit Software, PBC [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/request_generate.R
\name{request_generate}
\alias{request_generate}
\title{Generate a Google Sheets API request}
\usage{
request_generate(
  endpoint = character(),
  params = list(),
  key = NULL,
  token = gs4_token()
)
}
\arguments{
\item{endpoint}{Character. Nickname for one of the selected Sheets API v4
endpoints built into googlesheets4. Learn more in \code{\link[=gs4_endpoints]{gs4_endpoints()}}.}

\item{params}{Named list. Parameters destined for endpoint URL substitution,
the query, or the body.}

\item{key}{API key. Needed for requests that don't contain a token. The need
for an API key in the absence of a token is explained in Google's document
"Credentials, access, security, and identity"
(\verb{https://support.google.com/googleapi/answer/6158857?hl=en&ref_topic=7013279}).
In order of precedence, these sources are consulted: the formal \code{key}
argument, a \code{key} parameter in \code{params}, a user-configured API key set up
with \code{\link[=gs4_auth_configure]{gs4_auth_configure()}} and retrieved with \code{\link[=gs4_api_key]{gs4_api_key()}}.}

\item{token}{Set this to \code{NULL} to suppress the inclusion of a token. Note
that, if auth has been de-activated via \code{\link[=gs4_deauth]{gs4_deauth()}},
\code{gs4_token()} will actually return \code{NULL}.}
}
\value{
\code{list()}\cr Components are \code{method}, \code{url}, \code{body}, and \code{token},
suitable as input for \code{\link[=request_make]{request_make()}}.
}
\description{
Generate a request, using knowledge of the \href{https://developers.google.com/sheets/api/}{Sheets API} from its Discovery
Document (\verb{https://www.googleapis.com/discovery/v1/apis/sheets/v4/rest}). Use
\code{\link[=request_make]{request_make()}} to execute the request. Most users should, instead, use
higher-level wrappers that facilitate common tasks, such as reading or
writing worksheets or cell ranges. The functions here are intended for
internal use and for programming around the Sheets API.

\code{request_generate()} lets you provide the bare minimum of input.
It takes a nickname for an endpoint and:
\itemize{
\item Uses the API spec to look up the \code{method}, \code{path}, and \code{base_url}.
\item Checks \code{params} for validity and completeness with respect to the
endpoint. Uses \code{params} for URL endpoint substitution and separates
remaining parameters into those destined for the body versus the query.
\item Adds an API key to the query if and only if \code{token = NULL}.
}
}
\examples{
req <- request_generate(
  "sheets.spreadsheets.get",
  list(spreadsheetId = gs4_example("deaths")),
  key = "PRETEND_I_AM_AN_API_KEY",
  token = NULL
)
req
}
\seealso{
\code{\link[gargle:request_develop]{gargle::request_develop()}}, \code{\link[gargle:request_develop]{gargle::request_build()}},
\code{\link[gargle:request_make]{gargle::request_make()}}

Other low-level API functions: 
\code{\link{gs4_has_token}()},
\code{\link{gs4_token}()},
\code{\link{request_make}()}
}
\concept{low-level API functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/googlesheets4-package.R, R/gs4_formula.R
\name{googlesheets4-vctrs}
\alias{googlesheets4-vctrs}
\alias{vec_ptype2.googlesheets4_formula}
\alias{vec_cast.googlesheets4_formula}
\title{Internal vctrs methods}
\usage{
\method{vec_ptype2}{googlesheets4_formula}(x, y, ...)

\method{vec_cast}{googlesheets4_formula}(x, to, ...)
}
\description{
Internal vctrs methods
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_browse.R
\name{gs4_browse}
\alias{gs4_browse}
\title{Visit a Sheet in a web browser}
\usage{
gs4_browse(ss)
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}
}
\value{
The Sheet's browser URL, invisibly.
}
\description{
Visits a Google Sheet in your default browser, if session is interactive.
}
\examples{
gs4_example("mini-gap") \%>\% gs4_browse()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sheet_resize.R
\name{sheet_resize}
\alias{sheet_resize}
\title{Change the size of a (work)sheet}
\usage{
sheet_resize(ss, sheet = NULL, nrow = NULL, ncol = NULL, exact = FALSE)
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}

\item{sheet}{Sheet to resize, in the sense of "worksheet" or "tab". You can identify a sheet by name, with a string, or by position, with a number.}

\item{nrow, ncol}{Desired number of rows or columns, respectively. The default
of \code{NULL} means to leave unchanged.}

\item{exact}{Logical, indicating whether to impose \code{nrow} and \code{ncol} exactly
or to treat them as lower bounds. If \code{exact = FALSE},
\code{sheet_resize()} can only add cells. If \code{exact = TRUE}, cells can be
deleted and their contents are lost.}
}
\value{
The input \code{ss}, as an instance of \code{\link{sheets_id}}
}
\description{
Changes the number of rows and/or columns in a (work)sheet.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
# create a Sheet with the default initial worksheet
(ss <- gs4_create("sheet-resize-demo"))

# see (work)sheet dims
sheet_properties(ss)

# no resize occurs
sheet_resize(ss, nrow = 2, ncol = 6)

# reduce sheet size
sheet_resize(ss, nrow = 5, ncol = 7, exact = TRUE)

# add rows
sheet_resize(ss, nrow = 7)

# add columns
sheet_resize(ss, ncol = 10)

# add rows and columns
sheet_resize(ss, nrow = 9, ncol = 12)

# re-inspect (work)sheet dims
sheet_properties(ss)

# clean up
gs4_find("sheet-resize-demo") \%>\%
  googledrive::drive_trash()
\dontshow{\}) # examplesIf}
}
\seealso{
Makes an \code{UpdateSheetPropertiesRequest}:
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#UpdateSheetPropertiesRequest}
}

Other worksheet functions: 
\code{\link{sheet_add}()},
\code{\link{sheet_append}()},
\code{\link{sheet_copy}()},
\code{\link{sheet_delete}()},
\code{\link{sheet_properties}()},
\code{\link{sheet_relocate}()},
\code{\link{sheet_rename}()},
\code{\link{sheet_write}()}
}
\concept{worksheet functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sheet_write.R
\name{sheet_write}
\alias{sheet_write}
\alias{write_sheet}
\title{(Over)write new data into a Sheet}
\usage{
sheet_write(data, ss = NULL, sheet = NULL)

write_sheet(data, ss = NULL, sheet = NULL)
}
\arguments{
\item{data}{A data frame. If it has zero rows, we send one empty pseudo-row
of data, so that we can apply the usual table styling. This empty row goes
away (gets filled, actually) the first time you send more data with
\code{\link[=sheet_append]{sheet_append()}}.}

\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}

\item{sheet}{Sheet to write into, in the sense of "worksheet" or "tab". You can identify a sheet by name, with a string, or by position, with a number.}
}
\value{
The input \code{ss}, as an instance of \code{\link{sheets_id}}
}
\description{
This is one of the main ways to write data with googlesheets4. This function
writes a data frame into a (work)sheet inside a (spread)Sheet. The target
sheet is styled as a table:
\itemize{
\item Special formatting is applied to the header row, which holds column
names.
\item The first row (header row) is frozen.
\item The sheet's dimensions are set to "shrink wrap" the \code{data}.
}

If no existing Sheet is specified via \code{ss}, this function delegates to
\code{\link[=gs4_create]{gs4_create()}} and the new Sheet's name is randomly generated. If that's
undesirable, call \code{\link[=gs4_create]{gs4_create()}} directly to get more control.

If no \code{sheet} is specified or if \code{sheet} doesn't identify an existing sheet,
a new sheet is added to receive the \code{data}. If \code{sheet} specifies an existing
sheet, it is effectively overwritten! All pre-existing values, formats, and
dimensions are cleared and the targeted sheet gets new values and dimensions
from \code{data}.

This function goes by two names, because we want it to make sense in two
contexts:
\itemize{
\item \code{write_sheet()} evokes other table-writing functions, like
\code{readr::write_csv()}. The \code{sheet} here technically refers to an individual
(work)sheet (but also sort of refers to the associated Google
(spread)Sheet).
\item \code{sheet_write()} is the right name according to the naming convention used
throughout the googlesheets4 package.
}

\code{write_sheet()} and \code{sheet_write()} are equivalent and you can use either one.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
df <- data.frame(
  x = 1:3,
  y = letters[1:3]
)

# specify only a data frame, get a new Sheet, with a random name
ss <- write_sheet(df)
read_sheet(ss)

# clean up
googledrive::drive_trash(ss)

# create a Sheet with some initial, placeholder data
ss <- gs4_create(
  "sheet-write-demo",
  sheets = list(alpha = data.frame(x = 1), omega = data.frame(x = 1))
)

# write df into its own, new sheet
sheet_write(df, ss = ss)

# write mtcars into the sheet named "omega"
sheet_write(mtcars, ss = ss, sheet = "omega")

# get an overview of the sheets
sheet_properties(ss)

# view your magnificent creation in the browser
gs4_browse(ss)

# clean up
gs4_find("sheet-write-demo") \%>\%
  googledrive::drive_trash()
\dontshow{\}) # examplesIf}
}
\seealso{
Other write functions: 
\code{\link{gs4_create}()},
\code{\link{gs4_formula}()},
\code{\link{range_delete}()},
\code{\link{range_flood}()},
\code{\link{range_write}()},
\code{\link{sheet_append}()}

Other worksheet functions: 
\code{\link{sheet_add}()},
\code{\link{sheet_append}()},
\code{\link{sheet_copy}()},
\code{\link{sheet_delete}()},
\code{\link{sheet_properties}()},
\code{\link{sheet_relocate}()},
\code{\link{sheet_rename}()},
\code{\link{sheet_resize}()}
}
\concept{worksheet functions}
\concept{write functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cell-specification.R
\name{cell-specification}
\alias{cell-specification}
\alias{cell_limits}
\alias{cell_rows}
\alias{cell_cols}
\alias{anchored}
\title{Specify cells}
\description{
Many functions in googlesheets4 use a \code{range} argument to target specific
cells. The Sheets v4 API expects user-specified ranges to be expressed via
\href{https://developers.google.com/sheets/api/guides/concepts#a1_notation}{its A1 notation},
but googlesheets4 accepts and converts a few alternative specifications
provided by the functions in the \link[cellranger:cellranger]{cellranger::cellranger} package. Of course,
you can always provide A1-style ranges directly to functions like
\code{\link[=read_sheet]{read_sheet()}} or \code{\link[=range_read_cells]{range_read_cells()}}. Why would you use the
\link[cellranger:cellranger]{cellranger::cellranger} helpers? Some ranges are practically impossible to
express in A1 notation, specifically when you want to describe rectangles
with some bounds that are specified and others determined by the data.
}
\examples{
\dontshow{if (gs4_has_token() && rlang::is_interactive()) withAutoprint(\{ # examplesIf}
ss <- gs4_example("mini-gap")

# Specify only the rows or only the columns
read_sheet(ss, range = cell_rows(1:3))
read_sheet(ss, range = cell_cols("C:D"))
read_sheet(ss, range = cell_cols(1))

# Specify upper or lower bound on row or column
read_sheet(ss, range = cell_rows(c(NA, 4)))
read_sheet(ss, range = cell_cols(c(NA, "D")))
read_sheet(ss, range = cell_rows(c(3, NA)))
read_sheet(ss, range = cell_cols(c(2, NA)))
read_sheet(ss, range = cell_cols(c("C", NA)))

# Specify a partially open rectangle
read_sheet(ss, range = cell_limits(c(2, 3), c(NA, NA)), col_names = FALSE)
read_sheet(ss, range = cell_limits(c(1, 2), c(NA, 4)))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_fodder.R
\name{gs4_fodder}
\alias{gs4_fodder}
\title{Create useful spreadsheet filler}
\usage{
gs4_fodder(n = 10, m = n)
}
\arguments{
\item{n}{Number of rows.}

\item{m}{Number of columns.}
}
\value{
A data frame of character vectors.
}
\description{
Creates a data frame that is useful for filling a spreadsheet, when you just
need a sheet to experiment with. The data frame has \code{n} rows and \code{m} columns
with these properties:
\itemize{
\item Column names match what Sheets displays: "A", "B", "C", and so on.
\item Inner cell values reflect the coordinates where each value will land in
the sheet, in A1-notation. So the first row is "B2", "C2", and so on.
Note that this \code{n}-row data frame will occupy \code{n + 1} rows in the sheet,
because the column names occupy the first row.
}
}
\examples{
gs4_fodder()
gs4_fodder(5, 3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sheets_id-class.R
\name{sheets_id}
\alias{sheets_id}
\alias{as_sheets_id}
\title{\code{sheets_id} class}
\usage{
as_sheets_id(x, ...)
}
\arguments{
\item{x}{Something that contains a Google Sheet id: an id string, a
\code{\link[googledrive:drive_id]{drive_id}}, a URL, a one-row
\code{\link[googledrive:dribble]{dribble}}, or a \code{googlesheets4_spreadsheet}.}

\item{...}{Other arguments passed down to methods. (Not used.)}
}
\description{
\code{sheets_id} is an S3 class that marks a string as a Google Sheet's id, which
the Sheets API docs refer to as \code{spreadsheetId}.

Any object of class \code{sheets_id} also has the \code{\link[googledrive:drive_id]{drive_id}}
class, which is used by \link[googledrive:googledrive-package]{googledrive::googledrive} for the same purpose. This means you
can provide a \code{sheets_id} to \link[googledrive:googledrive-package]{googledrive::googledrive} functions, in order to do anything
with your Sheet that has nothing to do with it being a spreadsheet. Examples:
change the Sheet's name, parent folder, or permissions. Read more about using
\link{googlesheets4} and \link[googledrive:googledrive-package]{googledrive::googledrive} together in \code{vignette("drive-and-sheets")}.
Note that a \code{sheets_id} object is intended to hold \strong{just one} id, while the
parent class \code{drive_id} can be used for multiple ids.

\code{as_sheets_id()} is a generic function that converts various inputs into an
instance of \code{sheets_id}. See more below.

When you print a \code{sheets_id}, we attempt to reveal the Sheet's current
metadata, via \code{\link[=gs4_get]{gs4_get()}}. This can fail for a variety of reasons (e.g. if
you're offline), but the input \code{sheets_id} is always revealed and returned,
invisibly.
}
\section{\code{as_sheets_id()}}{


These inputs can be converted to a \code{sheets_id}:
\itemize{
\item Spreadsheet id, "a string containing letters, numbers, and some special
characters", typically 44 characters long, in our experience. Example:
\verb{1qpyC0XzvTcKT6EISywvqESX3A0MwQoFDE8p-Bll4hps}.
\item A URL, from which we can excavate a spreadsheet or file id. Example:
\code{"https://docs.google.com/spreadsheets/d/1BzfL0kZUz1TsI5zxJF1WNF01IxvC67FbOJUiiGMZ_mQ/edit#gid=1150108545"}.
\item A one-row \code{\link[googledrive:dribble]{dribble}}, a "Drive tibble" used by the
\link[googledrive:googledrive-package]{googledrive::googledrive} package. In general, a \code{dribble} can represent several
files, one row per file. Since googlesheets4 is not vectorized over
spreadsheets, we are only prepared to accept a one-row \code{dribble}.
\itemize{
\item \code{\link[googledrive:drive_get]{googledrive::drive_get("YOUR_SHEET_NAME")}}
is a great way to look up a Sheet via its name.
\item \code{\link[=gs4_find]{gs4_find("YOUR_SHEET_NAME")}} is another good way
to get your hands on a Sheet.
}
\item Spreadsheet meta data, as returned by, e.g., \code{\link[=gs4_get]{gs4_get()}}. Literally,
this is an object of class \code{googlesheets4_spreadsheet}.
}
}

\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
mini_gap_id <- gs4_example("mini-gap")
class(mini_gap_id)
mini_gap_id

as_sheets_id("abc")
\dontshow{\}) # examplesIf}
}
\seealso{
\link[googledrive:drive_id]{googledrive::as_id}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/range_write.R
\name{range_write}
\alias{range_write}
\title{(Over)write new data into a range}
\usage{
range_write(
  ss,
  data,
  sheet = NULL,
  range = NULL,
  col_names = TRUE,
  reformat = TRUE
)
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}

\item{data}{A data frame.}

\item{sheet}{Sheet to write into, in the sense of "worksheet" or "tab". You can identify a sheet by name, with a string, or by position, with a number. Ignored if the sheet is specified via \code{range}. If neither argument specifies the sheet, defaults to the first visible sheet.}

\item{range}{Where to write. This \code{range} argument has important similarities
and differences to \code{range} elsewhere (e.g. \code{\link[=range_read]{range_read()}}):
\itemize{
\item Similarities: Can be a cell range, using A1 notation ("A1:D3") or using
the helpers in \code{\link{cell-specification}}. Can combine sheet name and cell
range ("Sheet1!A5:A") or refer to a sheet by name (\code{range = "Sheet1"},
although \code{sheet = "Sheet1"} is preferred for clarity).
\item Difference: Can NOT be a named range.
\item Difference: \code{range} can be interpreted as the \emph{start} of the target
rectangle (the upper left corner) or, more literally, as the actual
target rectangle. See the "Range specification" section for details.
}}

\item{col_names}{Logical, indicates whether to send the column names of
\code{data}.}

\item{reformat}{Logical, indicates whether to reformat the affected cells.
Currently googlesheets4 provides no real support for formatting, so
\code{reformat = TRUE} effectively means that edited cells become unformatted.}
}
\value{
The input \code{ss}, as an instance of \code{\link{sheets_id}}
}
\description{
Writes a data frame into a range of cells. Main differences from
\code{\link[=sheet_write]{sheet_write()}} (a.k.a. \code{\link[=write_sheet]{write_sheet()}}):
\itemize{
\item Narrower scope. \code{range_write()} literally targets some cells, not a whole
(work)sheet.
\item The edited rectangle is not explicitly styled as a table. Nothing special
is done re: formatting a header row or freezing rows.
\item Column names can be suppressed. This means that, although \code{data} must
be a data frame (at least for now), \code{range_write()} can actually be used
to write arbitrary data.
\item The target (spread)Sheet and (work)sheet must already exist. There is no
ability to create a Sheet or add a worksheet.
\item The target sheet dimensions are not "trimmed" to shrink-wrap the \code{data}.
However, the sheet might gain rows and/or columns, in order to write
\code{data} to the user-specified \code{range}.
}

If you just want to add rows to an existing table, the function you probably
want is \code{\link[=sheet_append]{sheet_append()}}.
}
\section{Range specification}{

The \code{range} argument of \code{range_write()} is special, because the Sheets API
can implement it in 2 different ways:
\itemize{
\item If \code{range} represents exactly 1 cell, like "B3", it is taken as the \emph{start}
(or upper left corner) of the targeted cell rectangle. The edited cells are
determined implicitly by the extent of the \code{data} we are writing. This
frees you from doing fiddly range computations based on the dimensions of
the \code{data}.
\item If \code{range} describes a rectangle with multiple cells, it is interpreted
as the \emph{actual} rectangle to edit. It is possible to describe a rectangle
that is unbounded on the right (e.g. "B2:4"), on the bottom (e.g. "A4:C"),
or on both the right and the bottom (e.g.
\code{cell_limits(c(2, 3), c(NA, NA))}. Note that \strong{all cells} inside the
rectangle receive updated data and format. Important implication: if the
\code{data} object isn't big enough to fill the target rectangle, the cells that
don't receive new data are effectively cleared, i.e. the existing value
and format are deleted.
}
}

\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
# create a Sheet with some initial, empty (work)sheets
(ss <- gs4_create("range-write-demo", sheets = c("alpha", "beta")))

df <- data.frame(
  x = 1:3,
  y = letters[1:3]
)

#  write df somewhere other than the "upper left corner"
range_write(ss, data = df, range = "D6")

# view your magnificent creation in the browser
gs4_browse(ss)

# send data of disparate types to a 1-row rectangle
dat <- tibble::tibble(
  string = "string",
  logical = TRUE,
  datetime = Sys.time()
)
range_write(ss, data = dat, sheet = "beta", col_names = FALSE)

# send data of disparate types to a 1-column rectangle
dat <- tibble::tibble(
  x = list(Sys.time(), FALSE, "string")
)
range_write(ss, data = dat, range = "beta!C5", col_names = FALSE)

# clean up
gs4_find("range-write-demo") \%>\%
  googledrive::drive_trash()
\dontshow{\}) # examplesIf}
}
\seealso{
If sheet size needs to change, makes an \code{UpdateSheetPropertiesRequest}:
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#UpdateSheetPropertiesRequest}
}

The main data write is done via an \code{UpdateCellsRequest}:
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#updatecellsrequest}
}

Other write functions: 
\code{\link{gs4_create}()},
\code{\link{gs4_formula}()},
\code{\link{range_delete}()},
\code{\link{range_flood}()},
\code{\link{sheet_append}()},
\code{\link{sheet_write}()}
}
\concept{write functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_formula.R
\name{gs4_formula}
\alias{gs4_formula}
\title{Class for Google Sheets formulas}
\usage{
gs4_formula(x = character())
}
\arguments{
\item{x}{Character.}
}
\value{
An S3 vector of class \code{googlesheets4_formula}.
}
\description{
In order to write a formula into Google Sheets, you need to store it as an
object of class \code{googlesheets4_formula}. This is how we distinguish a
"regular" character string from a string that should be interpreted as a
formula. \code{googlesheets4_formula} is an S3 class implemented using the \href{https://vctrs.r-lib.org/articles/s3-vector.html}{vctrs package}.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
dat <- data.frame(x = c(1, 5, 3, 2, 4, 6))

ss <- gs4_create("gs4-formula-demo", sheets = dat)
ss

summaries <- tibble::tribble(
  ~desc, ~summaries,
  "max", "=max(A:A)",
  "sum", "=sum(A:A)",
  "min", "=min(A:A)",
  "sparkline", "=SPARKLINE(A:A, {\"color\", \"blue\"})"
)

# explicitly declare a column as `googlesheets4_formula`
summaries$summaries <- gs4_formula(summaries$summaries)
summaries

range_write(ss, data = summaries, range = "C1", reformat = FALSE)

miscellany <- tibble::tribble(
  ~desc, ~example,
  "hyperlink", "=HYPERLINK(\"http://www.google.com/\",\"Google\")",
  "image", "=IMAGE(\"https://www.google.com/images/srpr/logo3w.png\")"
)
miscellany$example <- gs4_formula(miscellany$example)
miscellany

sheet_write(miscellany, ss = ss)

# clean up
gs4_find("gs4-formula-demo") \%>\%
  googledrive::drive_trash()
\dontshow{\}) # examplesIf}
}
\seealso{
Other write functions: 
\code{\link{gs4_create}()},
\code{\link{range_delete}()},
\code{\link{range_flood}()},
\code{\link{range_write}()},
\code{\link{sheet_append}()},
\code{\link{sheet_write}()}
}
\concept{write functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sheet_append.R
\name{sheet_append}
\alias{sheet_append}
\title{Append rows to a sheet}
\usage{
sheet_append(ss, data, sheet = 1)
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}

\item{data}{A data frame.}

\item{sheet}{Sheet to append to, in the sense of "worksheet" or "tab". You can identify a sheet by name, with a string, or by position, with a number.}
}
\value{
The input \code{ss}, as an instance of \code{\link{sheets_id}}
}
\description{
Adds one or more new rows after the last row with data in a (work)sheet,
increasing the row dimension of the sheet if necessary.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
# we will recreate the table of "other" deaths from this example Sheet
(deaths <- gs4_example("deaths") \%>\%
  range_read(range = "other_data", col_types = "????DD"))

# split the data into 3 pieces, which we will send separately
deaths_one <- deaths[1:5, ]
deaths_two <- deaths[6, ]
deaths_three <- deaths[7:10, ]

# create a Sheet and send the first chunk of data
ss <- gs4_create("sheet-append-demo", sheets = list(deaths = deaths_one))

# append a single row
ss \%>\% sheet_append(deaths_two)

# append remaining rows
ss \%>\% sheet_append(deaths_three)

# read and check against the original
deaths_replica <- range_read(ss, col_types = "????DD")
identical(deaths, deaths_replica)

# clean up
gs4_find("sheet-append-demo") \%>\%
  googledrive::drive_trash()
\dontshow{\}) # examplesIf}
}
\seealso{
Makes an \code{AppendCellsRequest}:
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#AppendCellsRequest}
}

Other write functions: 
\code{\link{gs4_create}()},
\code{\link{gs4_formula}()},
\code{\link{range_delete}()},
\code{\link{range_flood}()},
\code{\link{range_write}()},
\code{\link{sheet_write}()}

Other worksheet functions: 
\code{\link{sheet_add}()},
\code{\link{sheet_copy}()},
\code{\link{sheet_delete}()},
\code{\link{sheet_properties}()},
\code{\link{sheet_relocate}()},
\code{\link{sheet_rename}()},
\code{\link{sheet_resize}()},
\code{\link{sheet_write}()}
}
\concept{worksheet functions}
\concept{write functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sheet_relocate.R
\name{sheet_relocate}
\alias{sheet_relocate}
\title{Relocate one or more (work)sheets}
\usage{
sheet_relocate(ss, sheet, .before = if (is.null(.after)) 1, .after = NULL)
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}

\item{sheet}{Sheet to relocate, in the sense of "worksheet" or "tab". You can identify a sheet by name, with a string, or by position, with a number. You can pass a vector to move multiple sheets at once or even a list, if you need to mix names and positions.}

\item{.before, .after}{Specification of where to locate the sheets(s)
identified by \code{sheet}. Exactly one of \code{.before} and \code{.after} must be
specified. Refer to an existing sheet by name (via a string) or by position
(via a number).}
}
\value{
The input \code{ss}, as an instance of \code{\link{sheets_id}}
}
\description{
Move (work)sheets around within a (spread)Sheet. The outcome is most
predictable for these common and simple use cases:
\itemize{
\item Reorder and move one or more sheets to the front.
\item Move a single sheet to a specific (but arbitrary) location.
\item Move multiple sheets to the back with \code{.after = 100} (\code{.after} can be
any number greater than or equal to the number of sheets).
}

If your relocation task is more complicated and you are puzzled by the
result, break it into a sequence of simpler calls to
\code{sheet_relocate()}.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
sheet_names <- c("alfa", "bravo", "charlie", "delta", "echo", "foxtrot")
ss <- gs4_create("sheet-relocate-demo", sheets = sheet_names)
sheet_names(ss)

# move one sheet, forwards then backwards
ss \%>\%
  sheet_relocate("echo", .before = "bravo") \%>\%
  sheet_names()
ss \%>\%
  sheet_relocate("echo", .after = "delta") \%>\%
  sheet_names()

# reorder and move multiple sheets to the front
ss \%>\%
  sheet_relocate(list("foxtrot", 4)) \%>\%
  sheet_names()

# put the sheets back in the original order
ss \%>\%
  sheet_relocate(sheet_names) \%>\%
  sheet_names()

# reorder and move multiple sheets to the back
ss \%>\%
  sheet_relocate(c("bravo", "alfa", "echo"), .after = 10) \%>\%
  sheet_names()

# clean up
gs4_find("sheet-relocate-demo") \%>\%
  googledrive::drive_trash()
\dontshow{\}) # examplesIf}
}
\seealso{
Constructs a batch of \code{UpdateSheetPropertiesRequest}s (one per sheet):
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#UpdateSheetPropertiesRequest}
}

Other worksheet functions: 
\code{\link{sheet_add}()},
\code{\link{sheet_append}()},
\code{\link{sheet_copy}()},
\code{\link{sheet_delete}()},
\code{\link{sheet_properties}()},
\code{\link{sheet_rename}()},
\code{\link{sheet_resize}()},
\code{\link{sheet_write}()}
}
\concept{worksheet functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_create.R
\name{gs4_create}
\alias{gs4_create}
\title{Create a new Sheet}
\usage{
gs4_create(name = gs4_random(), ..., sheets = NULL)
}
\arguments{
\item{name}{The name of the new spreadsheet.}

\item{...}{Optional spreadsheet properties that can be set through this API
endpoint, such as locale and time zone.}

\item{sheets}{Optional input for initializing (work)sheets. If unspecified,
the Sheets API automatically creates an empty "Sheet1". You can provide a
vector of sheet names, a data frame, or a (possibly named) list of data
frames. See the examples.}
}
\value{
The input \code{ss}, as an instance of \code{\link{sheets_id}}
}
\description{
Creates an entirely new (spread)Sheet (or, in Excel-speak, workbook).
Optionally, you can also provide names and/or data for the initial set of
(work)sheets. Any initial data provided via \code{sheets} is styled as a table,
as described in \code{\link[=sheet_write]{sheet_write()}}.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
gs4_create("gs4-create-demo-1")

gs4_create("gs4-create-demo-2", locale = "en_CA")

gs4_create(
  "gs4-create-demo-3",
  locale = "fr_FR",
  timeZone = "Europe/Paris"
)

gs4_create(
  "gs4-create-demo-4",
  sheets = c("alpha", "beta")
)

my_data <- data.frame(x = 1)
gs4_create(
  "gs4-create-demo-5",
  sheets = my_data
)

gs4_create(
  "gs4-create-demo-6",
  sheets = list(chickwts = head(chickwts), mtcars = head(mtcars))
)

# Clean up
gs4_find("gs4-create-demo") \%>\%
  googledrive::drive_trash()
\dontshow{\}) # examplesIf}
}
\seealso{
Wraps the \code{spreadsheets.create} endpoint:
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/create}
}

There is an article on writing Sheets:
\itemize{
\item \url{https://googlesheets4.tidyverse.org/articles/articles/write-sheets.html}
}

Other write functions: 
\code{\link{gs4_formula}()},
\code{\link{range_delete}()},
\code{\link{range_flood}()},
\code{\link{range_write}()},
\code{\link{sheet_append}()},
\code{\link{sheet_write}()}
}
\concept{write functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sheet_delete.R
\name{sheet_delete}
\alias{sheet_delete}
\title{Delete one or more (work)sheets}
\usage{
sheet_delete(ss, sheet)
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}

\item{sheet}{Sheet to delete, in the sense of "worksheet" or "tab". You can identify a sheet by name, with a string, or by position, with a number. You can pass a vector to delete multiple sheets at once or even a list, if you need to mix names and positions.}
}
\value{
The input \code{ss}, as an instance of \code{\link{sheets_id}}
}
\description{
Deletes one or more (work)sheets from a (spread)Sheet.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
ss <- gs4_create("delete-sheets-from-me")
sheet_add(ss, c("alpha", "beta", "gamma", "delta"))

# get an overview of the sheets
sheet_properties(ss)

# delete sheets
sheet_delete(ss, 1)
sheet_delete(ss, "gamma")
sheet_delete(ss, list("alpha", 2))

# get an overview of the sheets
sheet_properties(ss)

# clean up
gs4_find("delete-sheets-from-me") \%>\%
  googledrive::drive_trash()
\dontshow{\}) # examplesIf}
}
\seealso{
Makes an \code{DeleteSheetsRequest}:
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#DeleteSheetRequest}
}

Other worksheet functions: 
\code{\link{sheet_add}()},
\code{\link{sheet_append}()},
\code{\link{sheet_copy}()},
\code{\link{sheet_properties}()},
\code{\link{sheet_relocate}()},
\code{\link{sheet_rename}()},
\code{\link{sheet_resize}()},
\code{\link{sheet_write}()}
}
\concept{worksheet functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/range_read.R
\name{range_read}
\alias{range_read}
\alias{read_sheet}
\title{Read a Sheet into a data frame}
\usage{
range_read(
  ss,
  sheet = NULL,
  range = NULL,
  col_names = TRUE,
  col_types = NULL,
  na = "",
  trim_ws = TRUE,
  skip = 0,
  n_max = Inf,
  guess_max = min(1000, n_max),
  .name_repair = "unique"
)

read_sheet(
  ss,
  sheet = NULL,
  range = NULL,
  col_names = TRUE,
  col_types = NULL,
  na = "",
  trim_ws = TRUE,
  skip = 0,
  n_max = Inf,
  guess_max = min(1000, n_max),
  .name_repair = "unique"
)
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}

\item{sheet}{Sheet to read, in the sense of "worksheet" or "tab". You can identify a sheet by name, with a string, or by position, with a number. Ignored if the sheet is specified via \code{range}. If neither argument specifies the sheet, defaults to the first visible sheet.}

\item{range}{A cell range to read from. If \code{NULL}, all non-empty cells are
read. Otherwise specify \code{range} as described in \href{https://developers.google.com/sheets/api/guides/concepts#a1_notation}{Sheets A1 notation}
or using the helpers documented in \link{cell-specification}. Sheets uses
fairly standard spreadsheet range notation, although a bit different from
Excel. Examples of valid ranges: \code{"Sheet1!A1:B2"}, \code{"Sheet1!A:A"},
\code{"Sheet1!1:2"}, \code{"Sheet1!A5:A"}, \code{"A1:B2"}, \code{"Sheet1"}. Interpreted
strictly, even if the range forces the inclusion of leading, trailing, or
embedded empty rows or columns. Takes precedence over \code{skip}, \code{n_max} and
\code{sheet}. Note \code{range} can be a named range, like \code{"sales_data"}, without
any cell reference.}

\item{col_names}{\code{TRUE} to use the first row as column names, \code{FALSE} to get
default names, or a character vector to provide column names directly. If
user provides \code{col_types}, \code{col_names} can have one entry per column or one
entry per unskipped column.}

\item{col_types}{Column types. Either \code{NULL} to guess all from the
spreadsheet or a string of readr-style shortcodes, with one character or
code per column. If exactly one \code{col_type} is specified, it is recycled.
See Column Specification for more.}

\item{na}{Character vector of strings to interpret as missing values. By
default, blank cells are treated as missing data.}

\item{trim_ws}{Logical. Should leading and trailing whitespace be trimmed
from cell contents?}

\item{skip}{Minimum number of rows to skip before reading anything, be it
column names or data. Leading empty rows are automatically skipped, so this
is a lower bound. Ignored if \code{range} is given.}

\item{n_max}{Maximum number of data rows to parse into the returned tibble.
Trailing empty rows are automatically skipped, so this is an upper bound on
the number of rows in the result. Ignored if \code{range} is given. \code{n_max} is
imposed locally, after reading all non-empty cells, so, if speed is an
issue, it is better to use \code{range}.}

\item{guess_max}{Maximum number of data rows to use for guessing column
types.}

\item{.name_repair}{Handling of column names. By default, googlesheets4
ensures column names are not empty and are unique. There is full support
for \code{.name_repair} as documented in \code{\link[tibble:tibble]{tibble::tibble()}}.}
}
\value{
A \link[tibble:tibble-package]{tibble}
}
\description{
This is the main "read" function of the googlesheets4 package. It goes by two
names, because we want it to make sense in two contexts:
\itemize{
\item \code{read_sheet()} evokes other table-reading functions, like
\code{readr::read_csv()} and \code{readxl::read_excel()}. The \code{sheet} in this case
refers to a Google (spread)Sheet.
\item \code{range_read()} is the right name according to the naming convention used
throughout the googlesheets4 package.
}

\code{read_sheet()} and \code{range_read()} are synonyms and you can use either one.
}
\section{Column Specification}{


Column types must be specified in a single string of readr-style short
codes, e.g. "cci?l" means "character, character, integer, guess, logical".
This is not where googlesheets4's col spec will end up, but it gets the
ball rolling in a way that is consistent with readr and doesn't reinvent
any wheels.

Shortcodes for column types:
\itemize{
\item \verb{_} or \code{-}: Skip. Data in a skipped column is still requested from the
API (the high-level functions in this package are rectangle-oriented), but
is not parsed into the data frame output.
\item \verb{?}: Guess. A type is guessed for each cell and then a consensus type is
selected for the column. If no atomic type is suitable for all cells, a
list-column is created, in which each cell is converted to an R object of
"best" type. If no column types are specified, i.e. \code{col_types = NULL},
all types are guessed.
\item \code{l}: Logical.
\item \code{i}: Integer. This type is never guessed from the data, because Sheets
have no formal cell type for integers.
\item \code{d} or \code{n}: Numeric, in the sense of "double".
\item \code{D}: Date. This type is never guessed from the data, because date cells
are just serial datetimes that bear a "date" format.
\item \code{t}: Time of day. This type is never guessed from the data, because time
cells are just serial datetimes that bear a "time" format. \emph{Not implemented
yet; returns POSIXct.}
\item \code{T}: Datetime, specifically POSIXct.
\item \code{c}: Character.
\item \code{C}: Cell. This type is unique to googlesheets4. This returns raw cell
data, as an R list, which consists of everything sent by the Sheets API for
that cell. Has S3 type of \code{"CELL_SOMETHING"} and \code{"SHEETS_CELL"}. Mostly
useful internally, but exposed for those who want direct access to, e.g.,
formulas and formats.
\item \code{L}: List, as in "list-column". Each cell is a length-1 atomic vector of
its discovered type.
\item \emph{Still to come}: duration (code will be \code{:}) and factor (code will be
\code{f}).
}
}

\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
ss <- gs4_example("deaths")
read_sheet(ss, range = "A5:F15")
read_sheet(ss, range = "other!A5:F15", col_types = "ccilDD")
read_sheet(ss, range = "arts_data", col_types = "ccilDD")

read_sheet(gs4_example("mini-gap"))
read_sheet(
  gs4_example("mini-gap"),
  sheet = "Europe",
  range = "A:D",
  col_types = "ccid"
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/range_speedread.R
\name{range_speedread}
\alias{range_speedread}
\title{Read Sheet as CSV}
\usage{
range_speedread(ss, sheet = NULL, range = NULL, skip = 0, ...)
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}

\item{sheet}{Sheet to read, in the sense of "worksheet" or "tab". You can identify a sheet by name, with a string, or by position, with a number. Ignored if the sheet is specified via \code{range}. If neither argument specifies the sheet, defaults to the first visible sheet.}

\item{range}{A cell range to read from. If \code{NULL}, all non-empty cells are
read. Otherwise specify \code{range} as described in \href{https://developers.google.com/sheets/api/guides/concepts#a1_notation}{Sheets A1 notation}
or using the helpers documented in \link{cell-specification}. Sheets uses
fairly standard spreadsheet range notation, although a bit different from
Excel. Examples of valid ranges: \code{"Sheet1!A1:B2"}, \code{"Sheet1!A:A"},
\code{"Sheet1!1:2"}, \code{"Sheet1!A5:A"}, \code{"A1:B2"}, \code{"Sheet1"}. Interpreted
strictly, even if the range forces the inclusion of leading, trailing, or
embedded empty rows or columns. Takes precedence over \code{skip}, \code{n_max} and
\code{sheet}. Note \code{range} can be a named range, like \code{"sales_data"}, without
any cell reference.}

\item{skip}{Minimum number of rows to skip before reading anything, be it
column names or data. Leading empty rows are automatically skipped, so this
is a lower bound. Ignored if \code{range} is given.}

\item{...}{Passed along to the CSV parsing function (currently
\code{readr::read_csv()}).}
}
\value{
A \link[tibble:tibble-package]{tibble}
}
\description{
This function uses a quick-and-dirty method to read a Sheet that bypasses the
Sheets API and, instead, parses a CSV representation of the data. This can be
much faster than \code{\link[=range_read]{range_read()}} -- noticeably so for "large" spreadsheets.
There are real downsides, though, so we recommend this approach only when the
speed difference justifies it. Here are the limitations we must accept to get
faster reading:
\itemize{
\item Only formatted cell values are available, not underlying values or details
on the formats.
\item We can't target a named range as the \code{range}.
\item We have no access to the data type of a cell, i.e. we don't know that it's
logical, numeric, or datetime. That must be re-discovered based on the
CSV data (or specified by the user).
\item Auth and error handling have to be handled a bit differently internally,
which may lead to behaviour that differs from other functions in
googlesheets4.
}

Note that the Sheets API is still used to retrieve metadata on the target
Sheet, in order to support range specification. \code{range_speedread()} also
sends an auth token with the request, unless a previous call to
\code{\link[=gs4_deauth]{gs4_deauth()}} has put googlesheets4 into a de-authorized state.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
if (require("readr")) {
  # since cell type is not available, use readr's col type specification
  range_speedread(
    gs4_example("deaths"),
    sheet = "other",
    range = "A5:F15",
    col_types = cols(
      Age = col_integer(),
      `Date of birth` = col_date("\%m/\%d/\%Y"),
      `Date of death` = col_date("\%m/\%d/\%Y")
    )
  )
}

# write a Sheet that, by default, is NOT world-readable
(ss <- sheet_write(chickwts))

# demo that range_speedread() sends a token, which is why we can read this
range_speedread(ss)

# clean up
googledrive::drive_trash(ss)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/range_read.R
\name{spread_sheet}
\alias{spread_sheet}
\title{Spread a data frame of cells into spreadsheet shape}
\usage{
spread_sheet(
  df,
  col_names = TRUE,
  col_types = NULL,
  na = "",
  trim_ws = TRUE,
  guess_max = min(1000, max(df$row)),
  .name_repair = "unique"
)
}
\arguments{
\item{df}{A data frame with one row per (nonempty) cell, integer variables
\code{row} and \code{column} (probably referring to location within the spreadsheet),
and a list-column \code{cell} of \code{SHEET_CELL} objects.}

\item{col_names}{\code{TRUE} to use the first row as column names, \code{FALSE} to get
default names, or a character vector to provide column names directly. If
user provides \code{col_types}, \code{col_names} can have one entry per column or one
entry per unskipped column.}

\item{col_types}{Column types. Either \code{NULL} to guess all from the
spreadsheet or a string of readr-style shortcodes, with one character or
code per column. If exactly one \code{col_type} is specified, it is recycled.
See Column Specification for more.}

\item{na}{Character vector of strings to interpret as missing values. By
default, blank cells are treated as missing data.}

\item{trim_ws}{Logical. Should leading and trailing whitespace be trimmed
from cell contents?}

\item{guess_max}{Maximum number of data rows to use for guessing column
types.}

\item{.name_repair}{Handling of column names. By default, googlesheets4
ensures column names are not empty and are unique. There is full support
for \code{.name_repair} as documented in \code{\link[tibble:tibble]{tibble::tibble()}}.}
}
\value{
A tibble in the shape of the original spreadsheet, but enforcing
user's wishes regarding column names, column types, \code{NA} strings, and
whitespace trimming.
}
\description{
Reshapes a data frame of cells (presumably the output of
\code{\link[=range_read_cells]{range_read_cells()}}) into another data frame, i.e., puts it back into the
shape of the source spreadsheet. This function exists primarily for internal
use and for testing. The flagship function \code{\link[=range_read]{range_read()}}, a.k.a.
\code{\link[=read_sheet]{read_sheet()}}, is what most users are looking for. It is basically
\code{\link[=range_read_cells]{range_read_cells()}} + \code{spread_sheet()}.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
df <- gs4_example("mini-gap") \%>\%
  range_read_cells()
spread_sheet(df)

# ^^ gets same result as ...
read_sheet(gs4_example("mini-gap"))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/range_autofit.R
\name{range_autofit}
\alias{range_autofit}
\title{Auto-fit columns or rows to the data}
\usage{
range_autofit(ss, sheet = NULL, range = NULL, dimension = c("columns", "rows"))
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}

\item{sheet}{Sheet to modify, in the sense of "worksheet" or "tab". You can identify a sheet by name, with a string, or by position, with a number. Ignored if the sheet is specified via \code{range}. If neither argument specifies the sheet, defaults to the first visible sheet.}

\item{range}{Which columns or rows to resize. Optional. If you want to resize
all columns or all rows, use \code{dimension} instead. All the usual \code{range}
specifications are accepted, but the targeted range must specify only
columns (e.g. "B:F") or only rows (e.g. "2:7").}

\item{dimension}{Ignored if \code{range} is given. If consulted, \code{dimension} must
be either \code{"columns"} (the default) or \code{"rows"}. This is the simplest way
to request auto-resize for all columns or all rows.}
}
\value{
The input \code{ss}, as an instance of \code{\link{sheets_id}}
}
\description{
Applies automatic resizing to either columns or rows of a (work)sheet. The
width or height of targeted columns or rows, respectively, is determined
from the current cell contents. This only affects the appearance of a sheet
in the browser and doesn't affect its values or dimensions in any way.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
dat <- tibble::tibble(
  fruit = c("date", "lime", "pear", "plum")
)

ss <- gs4_create("range-autofit-demo", sheets = dat)
ss

# open in the browser
gs4_browse(ss)

# shrink column A to fit the short fruit names
range_autofit(ss)
# in the browser, notice how the column width shrank

# send some longer fruit names
dat2 <- tibble::tibble(
  fruit = c("cucumber", "honeydew")
)
ss \%>\% sheet_append(dat2)
# in the browser, see that column A is now too narrow to show the data

range_autofit(ss)
# in the browser, see the column A reveals all the data now

# clean up
gs4_find("range-autofit-demo") \%>\%
  googledrive::drive_trash()
\dontshow{\}) # examplesIf}
}
\seealso{
Makes an \code{AutoResizeDimensionsRequest}:
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#autoresizedimensionsrequest}
}
}
\concept{formatting functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_create.R
\name{gs4_random}
\alias{gs4_random}
\title{Generate a random Sheet name}
\usage{
gs4_random(n = 1)
}
\arguments{
\item{n}{Number of names to generate.}
}
\value{
A character vector.
}
\description{
Generates a random name, suitable for a newly created Sheet, using
\code{\link[ids:adjective_animal]{ids::adjective_animal()}}.
}
\examples{
gs4_random()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sheet_add.R
\name{sheet_add}
\alias{sheet_add}
\title{Add one or more (work)sheets}
\usage{
sheet_add(ss, sheet = NULL, ..., .before = NULL, .after = NULL)
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}

\item{sheet}{One or more new sheet names. If unspecified, one new sheet is
added and Sheets autogenerates a name of the form "SheetN".}

\item{...}{Optional parameters to specify additional properties, common to
all of the new sheet(s). Not relevant to most users. Specify fields of the
\href{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#SheetProperties}{\code{SheetProperties} schema}
in \code{name = value} form.}

\item{.before, .after}{Optional specification of where to put the new sheet(s). Specify, at most, one of \code{.before} and \code{.after}. Refer to an existing sheet by name (via a string) or by position (via a number). If unspecified, Sheets puts the new sheet(s) at the end.}
}
\value{
The input \code{ss}, as an instance of \code{\link{sheets_id}}
}
\description{
Adds one or more (work)sheets to an existing (spread)Sheet. Note that sheet
names must be unique.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
ss <- gs4_create("add-sheets-to-me")

# the only required argument is the target spreadsheet
ss \%>\% sheet_add()

# but you CAN specify sheet name and/or position
ss \%>\% sheet_add("apple", .after = 1)
ss \%>\% sheet_add("banana", .after = "apple")

# add multiple sheets at once
ss \%>\% sheet_add(c("coconut", "dragonfruit"))

# keeners can even specify additional sheet properties
ss \%>\%
  sheet_add(
    sheet = "eggplant",
    .before = 1,
    gridProperties = list(
      rowCount = 3, columnCount = 6, frozenRowCount = 1
    )
  )

# get an overview of the sheets
sheet_properties(ss)

# clean up
gs4_find("add-sheets-to-me") \%>\%
  googledrive::drive_trash()
\dontshow{\}) # examplesIf}
}
\seealso{
Makes a batch of \code{AddSheetRequest}s (one per sheet):
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#addsheetrequest}
}

Other worksheet functions: 
\code{\link{sheet_append}()},
\code{\link{sheet_copy}()},
\code{\link{sheet_delete}()},
\code{\link{sheet_properties}()},
\code{\link{sheet_relocate}()},
\code{\link{sheet_rename}()},
\code{\link{sheet_resize}()},
\code{\link{sheet_write}()}
}
\concept{worksheet functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sheet_copy.R
\name{sheet_copy}
\alias{sheet_copy}
\title{Copy a (work)sheet}
\usage{
sheet_copy(
  from_ss,
  from_sheet = NULL,
  to_ss = from_ss,
  to_sheet = NULL,
  .before = NULL,
  .after = NULL
)
}
\arguments{
\item{from_ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}

\item{from_sheet}{Sheet to copy, in the sense of "worksheet" or "tab". You can identify a sheet by name, with a string, or by position, with a number. Defaults to the first visible sheet.}

\item{to_ss}{The Sheet to copy \emph{to}. Accepts all the same types of input as
\code{from_ss}, which is also what this defaults to, if unspecified.}

\item{to_sheet}{Optional. Name of the new sheet, as a string. If you don't
specify this, Google generates a name, along the lines of "Copy of blah".
Note that sheet names must be unique within a Sheet, so if the automatic
name would violate this, Google also de-duplicates it for you, meaning you
could conceivably end up with "Copy of blah 2". If you have better ideas
about sheet names, specify \code{to_sheet}.}

\item{.before, .after}{Optional specification of where to put the new sheet. Specify, at most, one of \code{.before} and \code{.after}. Refer to an existing sheet by name (via a string) or by position (via a number). If unspecified, Sheets puts the new sheet at the end.}
}
\value{
The receiving Sheet, \verb{to_ ss}, as an instance of \code{\link{sheets_id}}.
}
\description{
Copies a (work)sheet, within its current (spread)Sheet or to another Sheet.
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
ss_aaa <- gs4_create(
  "sheet-copy-demo-aaa",
  sheets = list(mtcars = head(mtcars), chickwts = head(chickwts))
)

# copy 'mtcars' sheet within existing Sheet, accept autogenerated name
ss_aaa \%>\%
  sheet_copy()

# copy 'mtcars' sheet within existing Sheet
# specify new sheet's name and location
ss_aaa \%>\%
  sheet_copy(to_sheet = "mtcars-the-sequel", .after = 1)

# make a second Sheet
ss_bbb <- gs4_create("sheet-copy-demo-bbb")

# copy 'chickwts' sheet from first Sheet to second
# accept auto-generated name and default location
ss_aaa \%>\%
  sheet_copy("chickwts", to_ss = ss_bbb)

# copy 'chickwts' sheet from first Sheet to second,
# WITH a specific name and into a specific location
ss_aaa \%>\%
  sheet_copy(
    "chickwts",
    to_ss = ss_bbb, to_sheet = "chicks-two", .before = 1
  )

# clean up
gs4_find("sheet-copy-demo") \%>\%
  googledrive::drive_trash()
\dontshow{\}) # examplesIf}
}
\seealso{
If the copy happens within one Sheet, makes a \code{DuplicateSheetRequest}:
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#duplicatesheetrequest}
}

If the copy is from one Sheet to another, wraps the
\code{spreadsheets.sheets/copyTo} endpoint:
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets.sheets/copyTo}
}

and possibly makes a subsequent \code{UpdateSheetPropertiesRequest}:
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#UpdateSheetPropertiesRequest}
}

Other worksheet functions: 
\code{\link{sheet_add}()},
\code{\link{sheet_append}()},
\code{\link{sheet_delete}()},
\code{\link{sheet_properties}()},
\code{\link{sheet_relocate}()},
\code{\link{sheet_rename}()},
\code{\link{sheet_resize}()},
\code{\link{sheet_write}()}
}
\concept{worksheet functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/range_delete.R
\name{range_delete}
\alias{range_delete}
\title{Delete cells}
\usage{
range_delete(ss, sheet = NULL, range, shift = NULL)
}
\arguments{
\item{ss}{Something that identifies a Google Sheet:
\itemize{
\item its file id as a string or \code{\link[googledrive:drive_id]{drive_id}}
\item a URL from which we can recover the id
\item a one-row \code{\link[googledrive:dribble]{dribble}}, which is how googledrive
represents Drive files
\item an instance of \code{googlesheets4_spreadsheet}, which is what \code{\link[=gs4_get]{gs4_get()}}
returns
}

Processed through \code{\link[=as_sheets_id]{as_sheets_id()}}.}

\item{sheet}{Sheet to delete, in the sense of "worksheet" or "tab". You can identify a sheet by name, with a string, or by position, with a number. Ignored if the sheet is specified via \code{range}. If neither argument specifies the sheet, defaults to the first visible sheet.}

\item{range}{Cells to delete. There are a couple differences between \code{range}
here and how it works in other functions (e.g. \code{\link[=range_read]{range_read()}}):
\itemize{
\item \code{range} must be specified.
\item \code{range} must not be a named range.
\item \code{range} must not be the name of a (work) sheet. Instead, use
\code{\link[=sheet_delete]{sheet_delete()}} to delete an entire sheet.
Row-only and column-only ranges are especially relevant, such as "2:6" or
"D". Remember you can also use the helpers in \code{\link{cell-specification}},
such as \code{cell_cols(4:6)}, or \code{cell_rows(5)}.
}}

\item{shift}{Must be one of "up" or "left", if specified. Required if \code{range}
is NOT a rows-only or column-only range (in which case, we can figure it
out for you). Determines whether the deleted area is filled by shifting
surrounding cells up or to the left.}
}
\value{
The input \code{ss}, as an instance of \code{\link{sheets_id}}
}
\description{
Deletes a range of cells and shifts other cells into the deleted area. There
are several related tasks that are implemented by other functions:
\itemize{
\item To clear cells of their value and/or format, use \code{\link[=range_clear]{range_clear()}}.
\item To delete an entire (work)sheet, use \code{\link[=sheet_delete]{sheet_delete()}}.
\item To change the dimensions of a (work)sheet, use \code{\link[=sheet_resize]{sheet_resize()}}.
}
}
\examples{
\dontshow{if (gs4_has_token()) withAutoprint(\{ # examplesIf}
# create a data frame to use as initial data
df <- gs4_fodder(10)

# create Sheet
ss <- gs4_create("range-delete-example", sheets = list(df))

# delete some rows
range_delete(ss, range = "2:4")

# delete a column
range_delete(ss, range = "C")

# delete a rectangle and specify how to shift remaining cells
range_delete(ss, range = "B3:F4", shift = "left")

# clean up
gs4_find("range-delete-example") \%>\%
  googledrive::drive_trash()
\dontshow{\}) # examplesIf}
}
\seealso{
Makes a \code{DeleteRangeRequest}:
\itemize{
\item \url{https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/request#DeleteRangeRequest}
}

Other write functions: 
\code{\link{gs4_create}()},
\code{\link{gs4_formula}()},
\code{\link{range_flood}()},
\code{\link{range_write}()},
\code{\link{sheet_append}()},
\code{\link{sheet_write}()}
}
\concept{write functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_auth.R
\name{gs4_oauth_app}
\alias{gs4_oauth_app}
\title{Get currently configured OAuth app (deprecated)}
\usage{
gs4_oauth_app()
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

In light of the new \code{\link[gargle:gargle_oauth_client_from_json]{gargle::gargle_oauth_client()}} constructor and class of
the same name, \code{gs4_oauth_app()} is being replaced by
\code{\link[=gs4_oauth_client]{gs4_oauth_client()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_auth.R
\name{gs4_auth_configure}
\alias{gs4_auth_configure}
\alias{gs4_api_key}
\alias{gs4_oauth_client}
\title{Edit and view auth configuration}
\usage{
gs4_auth_configure(client, path, api_key, app = deprecated())

gs4_api_key()

gs4_oauth_client()
}
\arguments{
\item{client}{A Google OAuth client, presumably constructed via
\code{\link[gargle:gargle_oauth_client_from_json]{gargle::gargle_oauth_client_from_json()}}. Note, however, that it is
preferred to specify the client with JSON, using the \code{path} argument.}

\item{path}{JSON downloaded from \href{https://console.cloud.google.com}{Google Cloud Console}, containing a client id and
secret, in one of the forms supported for the \code{txt} argument of
\code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}} (typically, a file path or JSON string).}

\item{api_key}{API key.}

\item{app}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Replaced by the \code{client}
argument.}
}
\value{
\itemize{
\item \code{gs4_auth_configure()}: An object of R6 class
\link[gargle:AuthState-class]{gargle::AuthState}, invisibly.
\item \code{gs4_oauth_client()}: the current user-configured OAuth client.
\item \code{gs4_api_key()}: the current user-configured API key.
}
}
\description{
These functions give more control over and visibility into the auth
configuration than \code{\link[=gs4_auth]{gs4_auth()}} does. \code{gs4_auth_configure()}
lets the user specify their own:
\itemize{
\item OAuth client, which is used when obtaining a user token.
\item API key. If googlesheets4 is de-authorized via \code{\link[=gs4_deauth]{gs4_deauth()}}, all
requests are sent with an API key in lieu of a token.
}

See the \code{vignette("get-api-credentials", package = "gargle")}
for more.
If the user does not configure these settings, internal defaults
are used.

\code{gs4_oauth_client()} and \code{gs4_api_key()} retrieve the
currently configured OAuth client and API key, respectively.
}
\examples{
# see and store the current user-configured OAuth client (probably `NULL`)
(original_client <- gs4_oauth_client())

# see and store the current user-configured API key (probably `NULL`)
(original_api_key <- gs4_api_key())

# the preferred way to configure your own client is via a JSON file
# downloaded from Google Developers Console
# this example JSON is indicative, but fake
path_to_json <- system.file(
  "extdata", "data", "client_secret_123.googleusercontent.com.json",
  package = "googledrive"
)
gs4_auth_configure(path = path_to_json)

# this is also obviously a fake API key
gs4_auth_configure(api_key = "the_key_I_got_for_a_google_API")

# confirm the changes
gs4_oauth_client()
gs4_api_key()

# restore original auth config
gs4_auth_configure(client = original_client, api_key = original_api_key)
}
\seealso{
Other auth functions: 
\code{\link{gs4_auth}()},
\code{\link{gs4_deauth}()}
}
\concept{auth functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gs4_auth.R
\name{gs4_token}
\alias{gs4_token}
\title{Produce configured token}
\usage{
gs4_token()
}
\value{
A \code{request} object (an S3 class provided by \link[httr:httr-package]{httr}).
}
\description{
For internal use or for those programming around the Sheets API.
Returns a token pre-processed with \code{\link[httr:config]{httr::config()}}. Most users
do not need to handle tokens "by hand" or, even if they need some
control, \code{\link[=gs4_auth]{gs4_auth()}} is what they need. If there is no current
token, \code{\link[=gs4_auth]{gs4_auth()}} is called to either load from cache or
initiate OAuth2.0 flow.
If auth has been deactivated via \code{\link[=gs4_deauth]{gs4_deauth()}}, \code{gs4_token()}
returns \code{NULL}.
}
\examples{
\dontshow{if (gs4_has_token()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
req <- request_generate(
  "sheets.spreadsheets.get",
  list(spreadsheetId = "abc"),
  token = gs4_token()
)
req
\dontshow{\}) # examplesIf}
}
\seealso{
Other low-level API functions: 
\code{\link{gs4_has_token}()},
\code{\link{request_generate}()},
\code{\link{request_make}()}
}
\concept{low-level API functions}
